
#line 1 "./zscanner/scanner.rl"
/*  Copyright (C) 2011 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "zscanner/scanner.h"

#include <stdint.h>			// uint32_t
#include <stdlib.h>			// calloc
#include <stdio.h>			// sprintf
#include <libgen.h>			// dirname
#include <stdbool.h>			// bool
#include <math.h>			// pow
#include <string.h>			// strdup
#include <sys/types.h>			// (OpenBSD)
#include <sys/socket.h>			// AF_INET (BSD)
#include <netinet/in.h>			// in_addr (BSD)
#include <arpa/inet.h>			// inet_pton

#include "common/errcode.h"		// error codes
#include "common/descriptor.h"		// KNOT_RRTYPE_A
#include "zscanner/file_loader.h"	// file_loader
#include "zscanner/scanner_functions.h"	// Base64

/*! \brief Shorthand for setting warning data. */
#define SCANNER_WARNING(code) { s->error_code = code; }
/*! \brief Shorthand for setting error data. */
#define SCANNER_ERROR(code)   { s->error_code = code; s->stop = true; }

/*!
 * \brief Writes record type number to r_data.
 *
 * \param type		Type number.
 * \param rdata_tail	Position where to write type number to.
 */
static inline void type_num(const uint16_t type, uint8_t **rdata_tail)
{
	*((uint16_t *)*rdata_tail) = htons(type);
	*rdata_tail += 2;
}

/*!
 * \brief Sets bit to bitmap window.
 *
 * \param type		Type number.
 * \param s		Scanner context.
 */
static inline void window_add_bit(const uint16_t type, scanner_t *s) {
	uint8_t win      = type / 256;
	uint8_t bit_pos  = type % 256;
	uint8_t byte_pos = bit_pos / 8;

	((s->windows[win]).bitmap)[byte_pos] |= 128 >> (bit_pos % 8);

	if ((s->windows[win]).length < byte_pos + 1) {
		(s->windows[win]).length = byte_pos + 1;
	}

	if (s->last_window < win) {
		s->last_window = win;
	}
}

// Include scanner file (in Ragel).

#line 80 "zscanner/scanner.c"
static const int zone_scanner_start = 1060;
static const int zone_scanner_first_final = 1060;
static const int zone_scanner_error = 0;

static const int zone_scanner_en_err_line = 247;
static const int zone_scanner_en_dname_ = 249;
static const int zone_scanner_en_text_ = 259;
static const int zone_scanner_en_default_ttl_ = 271;
static const int zone_scanner_en_zone_origin_ = 280;
static const int zone_scanner_en_include_file_ = 293;
static const int zone_scanner_en_base64_ = 311;
static const int zone_scanner_en_bitmap_ = 318;
static const int zone_scanner_en_nonempty_hex_r_data = 438;
static const int zone_scanner_en_hex_r_data = 447;
static const int zone_scanner_en_dns_alg_ = 457;
static const int zone_scanner_en_cert_type_ = 552;
static const int zone_scanner_en_r_data_a = 593;
static const int zone_scanner_en_r_data_ns = 595;
static const int zone_scanner_en_r_data_soa = 597;
static const int zone_scanner_en_r_data_hinfo = 629;
static const int zone_scanner_en_r_data_minfo = 634;
static const int zone_scanner_en_r_data_mx = 639;
static const int zone_scanner_en_r_data_txt = 644;
static const int zone_scanner_en_r_data_aaaa = 648;
static const int zone_scanner_en_r_data_loc = 650;
static const int zone_scanner_en_r_data_srv = 705;
static const int zone_scanner_en_r_data_naptr = 716;
static const int zone_scanner_en_r_data_cert = 733;
static const int zone_scanner_en_r_data_apl = 744;
static const int zone_scanner_en_r_data_ds = 755;
static const int zone_scanner_en_r_data_sshfp = 768;
static const int zone_scanner_en_r_data_ipseckey = 778;
static const int zone_scanner_en_r_data_rrsig = 817;
static const int zone_scanner_en_r_data_nsec = 959;
static const int zone_scanner_en_r_data_dnskey = 962;
static const int zone_scanner_en_r_data_dhcid = 973;
static const int zone_scanner_en_r_data_nsec3 = 975;
static const int zone_scanner_en_r_data_nsec3param = 1004;
static const int zone_scanner_en_r_data_tlsa = 1017;
static const int zone_scanner_en_r_data_l32 = 1030;
static const int zone_scanner_en_r_data_l64 = 1035;
static const int zone_scanner_en_r_data_eui48 = 1048;
static const int zone_scanner_en_r_data_eui64 = 1054;
static const int zone_scanner_en_main = 1060;


#line 82 "./zscanner/scanner.rl"


scanner_t* scanner_create(const char *file_name)
{
	scanner_t *s = calloc(1, sizeof(scanner_t));
	if (s == NULL) {
		return NULL;
	}

	s->file_name = strdup(file_name);
	s->line_counter = 1;

	// Nonzero initial scanner state.
	s->cs = zone_scanner_start;

	return s;
}

void scanner_free(scanner_t *s)
{
	if (s != NULL) {
		free(s->file_name);
		free(s);
	}
}

int scanner_process(const char *start,
		    const char *end,
		    const bool is_last_block,
		    scanner_t  *s)
{
	// Necessary scanner variables.
	const char *p = start;
	const char *pe = end;
	char       *eof = NULL;
	int        stack[RAGEL_STACK_SIZE];

	// Auxiliary variables which are used in scanner body.
	struct in_addr  addr4;
	struct in6_addr addr6;
	uint32_t timestamp;
	int16_t  window;
	int	 ret;

	// Next 2 variables are for better performance.
	// Restoring r_data pointer to next free space.
	uint8_t *rdata_tail = s->r_data + s->r_data_tail;
	// Initialization of the last r_data byte.
	uint8_t *rdata_stop = s->r_data + MAX_RDATA_LENGTH - 1;

	// Restoring scanner states.
	int cs  = s->cs;
	int top = s->top;
	memcpy(stack, s->stack, sizeof(stack));

	// End of file check.
	if (is_last_block == true) {
		eof = (char *)pe;
	}

	// Writing scanner body (in C).
	
#line 190 "zscanner/scanner.c"
	{
	short _widec;
	if ( p == pe )
		goto _test_eof;
	goto _resume;

_again:
	switch ( cs ) {
		case 1060: goto st1060;
		case 0: goto st0;
		case 1: goto st1;
		case 2: goto st2;
		case 3: goto st3;
		case 4: goto st4;
		case 5: goto st5;
		case 6: goto st6;
		case 7: goto st7;
		case 8: goto st8;
		case 9: goto st9;
		case 10: goto st10;
		case 11: goto st11;
		case 12: goto st12;
		case 13: goto st13;
		case 1061: goto st1061;
		case 14: goto st14;
		case 15: goto st15;
		case 16: goto st16;
		case 17: goto st17;
		case 18: goto st18;
		case 19: goto st19;
		case 20: goto st20;
		case 21: goto st21;
		case 22: goto st22;
		case 23: goto st23;
		case 24: goto st24;
		case 25: goto st25;
		case 26: goto st26;
		case 27: goto st27;
		case 28: goto st28;
		case 29: goto st29;
		case 30: goto st30;
		case 31: goto st31;
		case 32: goto st32;
		case 33: goto st33;
		case 34: goto st34;
		case 35: goto st35;
		case 36: goto st36;
		case 37: goto st37;
		case 38: goto st38;
		case 39: goto st39;
		case 40: goto st40;
		case 41: goto st41;
		case 42: goto st42;
		case 43: goto st43;
		case 44: goto st44;
		case 45: goto st45;
		case 46: goto st46;
		case 47: goto st47;
		case 48: goto st48;
		case 49: goto st49;
		case 50: goto st50;
		case 51: goto st51;
		case 52: goto st52;
		case 53: goto st53;
		case 54: goto st54;
		case 55: goto st55;
		case 56: goto st56;
		case 57: goto st57;
		case 58: goto st58;
		case 59: goto st59;
		case 60: goto st60;
		case 61: goto st61;
		case 62: goto st62;
		case 63: goto st63;
		case 64: goto st64;
		case 65: goto st65;
		case 66: goto st66;
		case 67: goto st67;
		case 68: goto st68;
		case 69: goto st69;
		case 70: goto st70;
		case 71: goto st71;
		case 72: goto st72;
		case 73: goto st73;
		case 74: goto st74;
		case 75: goto st75;
		case 76: goto st76;
		case 77: goto st77;
		case 78: goto st78;
		case 79: goto st79;
		case 80: goto st80;
		case 81: goto st81;
		case 82: goto st82;
		case 83: goto st83;
		case 84: goto st84;
		case 85: goto st85;
		case 86: goto st86;
		case 87: goto st87;
		case 88: goto st88;
		case 89: goto st89;
		case 90: goto st90;
		case 91: goto st91;
		case 92: goto st92;
		case 93: goto st93;
		case 94: goto st94;
		case 95: goto st95;
		case 96: goto st96;
		case 97: goto st97;
		case 98: goto st98;
		case 99: goto st99;
		case 100: goto st100;
		case 101: goto st101;
		case 102: goto st102;
		case 103: goto st103;
		case 104: goto st104;
		case 105: goto st105;
		case 106: goto st106;
		case 107: goto st107;
		case 108: goto st108;
		case 109: goto st109;
		case 110: goto st110;
		case 111: goto st111;
		case 112: goto st112;
		case 113: goto st113;
		case 114: goto st114;
		case 115: goto st115;
		case 116: goto st116;
		case 117: goto st117;
		case 118: goto st118;
		case 119: goto st119;
		case 120: goto st120;
		case 121: goto st121;
		case 122: goto st122;
		case 123: goto st123;
		case 124: goto st124;
		case 125: goto st125;
		case 126: goto st126;
		case 127: goto st127;
		case 128: goto st128;
		case 1062: goto st1062;
		case 129: goto st129;
		case 130: goto st130;
		case 131: goto st131;
		case 132: goto st132;
		case 133: goto st133;
		case 134: goto st134;
		case 135: goto st135;
		case 136: goto st136;
		case 137: goto st137;
		case 138: goto st138;
		case 1063: goto st1063;
		case 139: goto st139;
		case 140: goto st140;
		case 141: goto st141;
		case 142: goto st142;
		case 143: goto st143;
		case 144: goto st144;
		case 145: goto st145;
		case 146: goto st146;
		case 1064: goto st1064;
		case 147: goto st147;
		case 148: goto st148;
		case 149: goto st149;
		case 1065: goto st1065;
		case 150: goto st150;
		case 151: goto st151;
		case 152: goto st152;
		case 153: goto st153;
		case 154: goto st154;
		case 155: goto st155;
		case 156: goto st156;
		case 157: goto st157;
		case 158: goto st158;
		case 159: goto st159;
		case 160: goto st160;
		case 1066: goto st1066;
		case 161: goto st161;
		case 162: goto st162;
		case 163: goto st163;
		case 164: goto st164;
		case 1067: goto st1067;
		case 165: goto st165;
		case 166: goto st166;
		case 167: goto st167;
		case 168: goto st168;
		case 169: goto st169;
		case 170: goto st170;
		case 171: goto st171;
		case 172: goto st172;
		case 173: goto st173;
		case 174: goto st174;
		case 175: goto st175;
		case 176: goto st176;
		case 177: goto st177;
		case 178: goto st178;
		case 179: goto st179;
		case 1068: goto st1068;
		case 180: goto st180;
		case 181: goto st181;
		case 182: goto st182;
		case 183: goto st183;
		case 184: goto st184;
		case 185: goto st185;
		case 186: goto st186;
		case 187: goto st187;
		case 188: goto st188;
		case 189: goto st189;
		case 190: goto st190;
		case 191: goto st191;
		case 192: goto st192;
		case 193: goto st193;
		case 194: goto st194;
		case 195: goto st195;
		case 1069: goto st1069;
		case 196: goto st196;
		case 197: goto st197;
		case 198: goto st198;
		case 199: goto st199;
		case 200: goto st200;
		case 201: goto st201;
		case 202: goto st202;
		case 203: goto st203;
		case 204: goto st204;
		case 205: goto st205;
		case 206: goto st206;
		case 207: goto st207;
		case 208: goto st208;
		case 209: goto st209;
		case 210: goto st210;
		case 211: goto st211;
		case 212: goto st212;
		case 213: goto st213;
		case 214: goto st214;
		case 215: goto st215;
		case 216: goto st216;
		case 217: goto st217;
		case 218: goto st218;
		case 219: goto st219;
		case 220: goto st220;
		case 221: goto st221;
		case 222: goto st222;
		case 223: goto st223;
		case 224: goto st224;
		case 225: goto st225;
		case 226: goto st226;
		case 227: goto st227;
		case 228: goto st228;
		case 229: goto st229;
		case 230: goto st230;
		case 231: goto st231;
		case 232: goto st232;
		case 233: goto st233;
		case 234: goto st234;
		case 235: goto st235;
		case 236: goto st236;
		case 237: goto st237;
		case 238: goto st238;
		case 239: goto st239;
		case 240: goto st240;
		case 241: goto st241;
		case 242: goto st242;
		case 243: goto st243;
		case 244: goto st244;
		case 245: goto st245;
		case 246: goto st246;
		case 247: goto st247;
		case 248: goto st248;
		case 1070: goto st1070;
		case 249: goto st249;
		case 250: goto st250;
		case 1071: goto st1071;
		case 251: goto st251;
		case 252: goto st252;
		case 253: goto st253;
		case 254: goto st254;
		case 255: goto st255;
		case 256: goto st256;
		case 257: goto st257;
		case 258: goto st258;
		case 259: goto st259;
		case 260: goto st260;
		case 1072: goto st1072;
		case 1073: goto st1073;
		case 261: goto st261;
		case 262: goto st262;
		case 263: goto st263;
		case 264: goto st264;
		case 265: goto st265;
		case 266: goto st266;
		case 267: goto st267;
		case 268: goto st268;
		case 269: goto st269;
		case 270: goto st270;
		case 271: goto st271;
		case 272: goto st272;
		case 273: goto st273;
		case 274: goto st274;
		case 1074: goto st1074;
		case 275: goto st275;
		case 276: goto st276;
		case 277: goto st277;
		case 278: goto st278;
		case 279: goto st279;
		case 280: goto st280;
		case 281: goto st281;
		case 282: goto st282;
		case 283: goto st283;
		case 284: goto st284;
		case 1075: goto st1075;
		case 285: goto st285;
		case 286: goto st286;
		case 287: goto st287;
		case 288: goto st288;
		case 289: goto st289;
		case 290: goto st290;
		case 291: goto st291;
		case 292: goto st292;
		case 293: goto st293;
		case 294: goto st294;
		case 295: goto st295;
		case 296: goto st296;
		case 297: goto st297;
		case 298: goto st298;
		case 299: goto st299;
		case 1076: goto st1076;
		case 300: goto st300;
		case 301: goto st301;
		case 302: goto st302;
		case 303: goto st303;
		case 304: goto st304;
		case 305: goto st305;
		case 306: goto st306;
		case 307: goto st307;
		case 1077: goto st1077;
		case 308: goto st308;
		case 309: goto st309;
		case 310: goto st310;
		case 311: goto st311;
		case 312: goto st312;
		case 313: goto st313;
		case 314: goto st314;
		case 315: goto st315;
		case 316: goto st316;
		case 1078: goto st1078;
		case 1079: goto st1079;
		case 1080: goto st1080;
		case 317: goto st317;
		case 318: goto st318;
		case 319: goto st319;
		case 320: goto st320;
		case 321: goto st321;
		case 322: goto st322;
		case 323: goto st323;
		case 324: goto st324;
		case 1081: goto st1081;
		case 1082: goto st1082;
		case 325: goto st325;
		case 326: goto st326;
		case 327: goto st327;
		case 328: goto st328;
		case 1083: goto st1083;
		case 329: goto st329;
		case 330: goto st330;
		case 331: goto st331;
		case 332: goto st332;
		case 333: goto st333;
		case 334: goto st334;
		case 335: goto st335;
		case 336: goto st336;
		case 337: goto st337;
		case 338: goto st338;
		case 339: goto st339;
		case 340: goto st340;
		case 341: goto st341;
		case 342: goto st342;
		case 343: goto st343;
		case 344: goto st344;
		case 345: goto st345;
		case 346: goto st346;
		case 347: goto st347;
		case 348: goto st348;
		case 349: goto st349;
		case 350: goto st350;
		case 351: goto st351;
		case 352: goto st352;
		case 353: goto st353;
		case 354: goto st354;
		case 355: goto st355;
		case 356: goto st356;
		case 357: goto st357;
		case 358: goto st358;
		case 359: goto st359;
		case 360: goto st360;
		case 361: goto st361;
		case 362: goto st362;
		case 363: goto st363;
		case 364: goto st364;
		case 365: goto st365;
		case 366: goto st366;
		case 367: goto st367;
		case 368: goto st368;
		case 369: goto st369;
		case 370: goto st370;
		case 371: goto st371;
		case 372: goto st372;
		case 373: goto st373;
		case 374: goto st374;
		case 375: goto st375;
		case 376: goto st376;
		case 377: goto st377;
		case 378: goto st378;
		case 379: goto st379;
		case 380: goto st380;
		case 381: goto st381;
		case 382: goto st382;
		case 383: goto st383;
		case 384: goto st384;
		case 385: goto st385;
		case 386: goto st386;
		case 387: goto st387;
		case 388: goto st388;
		case 389: goto st389;
		case 390: goto st390;
		case 391: goto st391;
		case 392: goto st392;
		case 393: goto st393;
		case 394: goto st394;
		case 395: goto st395;
		case 396: goto st396;
		case 397: goto st397;
		case 398: goto st398;
		case 399: goto st399;
		case 400: goto st400;
		case 401: goto st401;
		case 402: goto st402;
		case 403: goto st403;
		case 404: goto st404;
		case 405: goto st405;
		case 406: goto st406;
		case 407: goto st407;
		case 408: goto st408;
		case 409: goto st409;
		case 410: goto st410;
		case 411: goto st411;
		case 412: goto st412;
		case 413: goto st413;
		case 414: goto st414;
		case 415: goto st415;
		case 416: goto st416;
		case 417: goto st417;
		case 418: goto st418;
		case 419: goto st419;
		case 420: goto st420;
		case 421: goto st421;
		case 422: goto st422;
		case 423: goto st423;
		case 424: goto st424;
		case 425: goto st425;
		case 426: goto st426;
		case 427: goto st427;
		case 428: goto st428;
		case 429: goto st429;
		case 430: goto st430;
		case 431: goto st431;
		case 432: goto st432;
		case 433: goto st433;
		case 434: goto st434;
		case 435: goto st435;
		case 436: goto st436;
		case 437: goto st437;
		case 438: goto st438;
		case 439: goto st439;
		case 440: goto st440;
		case 441: goto st441;
		case 442: goto st442;
		case 443: goto st443;
		case 444: goto st444;
		case 1084: goto st1084;
		case 1085: goto st1085;
		case 1086: goto st1086;
		case 445: goto st445;
		case 446: goto st446;
		case 447: goto st447;
		case 448: goto st448;
		case 449: goto st449;
		case 1087: goto st1087;
		case 450: goto st450;
		case 451: goto st451;
		case 452: goto st452;
		case 453: goto st453;
		case 1088: goto st1088;
		case 1089: goto st1089;
		case 1090: goto st1090;
		case 454: goto st454;
		case 455: goto st455;
		case 1091: goto st1091;
		case 456: goto st456;
		case 457: goto st457;
		case 458: goto st458;
		case 1092: goto st1092;
		case 459: goto st459;
		case 460: goto st460;
		case 461: goto st461;
		case 462: goto st462;
		case 463: goto st463;
		case 464: goto st464;
		case 465: goto st465;
		case 466: goto st466;
		case 467: goto st467;
		case 468: goto st468;
		case 469: goto st469;
		case 470: goto st470;
		case 471: goto st471;
		case 472: goto st472;
		case 473: goto st473;
		case 474: goto st474;
		case 475: goto st475;
		case 476: goto st476;
		case 477: goto st477;
		case 478: goto st478;
		case 479: goto st479;
		case 480: goto st480;
		case 481: goto st481;
		case 482: goto st482;
		case 483: goto st483;
		case 484: goto st484;
		case 485: goto st485;
		case 486: goto st486;
		case 487: goto st487;
		case 488: goto st488;
		case 489: goto st489;
		case 490: goto st490;
		case 491: goto st491;
		case 492: goto st492;
		case 493: goto st493;
		case 494: goto st494;
		case 495: goto st495;
		case 496: goto st496;
		case 497: goto st497;
		case 498: goto st498;
		case 499: goto st499;
		case 500: goto st500;
		case 501: goto st501;
		case 502: goto st502;
		case 503: goto st503;
		case 504: goto st504;
		case 505: goto st505;
		case 506: goto st506;
		case 507: goto st507;
		case 508: goto st508;
		case 509: goto st509;
		case 510: goto st510;
		case 511: goto st511;
		case 512: goto st512;
		case 513: goto st513;
		case 514: goto st514;
		case 515: goto st515;
		case 516: goto st516;
		case 517: goto st517;
		case 518: goto st518;
		case 519: goto st519;
		case 520: goto st520;
		case 521: goto st521;
		case 522: goto st522;
		case 523: goto st523;
		case 524: goto st524;
		case 525: goto st525;
		case 526: goto st526;
		case 527: goto st527;
		case 528: goto st528;
		case 529: goto st529;
		case 530: goto st530;
		case 531: goto st531;
		case 532: goto st532;
		case 533: goto st533;
		case 534: goto st534;
		case 535: goto st535;
		case 536: goto st536;
		case 537: goto st537;
		case 538: goto st538;
		case 539: goto st539;
		case 540: goto st540;
		case 541: goto st541;
		case 542: goto st542;
		case 543: goto st543;
		case 544: goto st544;
		case 545: goto st545;
		case 546: goto st546;
		case 547: goto st547;
		case 548: goto st548;
		case 549: goto st549;
		case 550: goto st550;
		case 551: goto st551;
		case 552: goto st552;
		case 553: goto st553;
		case 1093: goto st1093;
		case 554: goto st554;
		case 555: goto st555;
		case 556: goto st556;
		case 557: goto st557;
		case 558: goto st558;
		case 559: goto st559;
		case 560: goto st560;
		case 561: goto st561;
		case 562: goto st562;
		case 563: goto st563;
		case 564: goto st564;
		case 565: goto st565;
		case 566: goto st566;
		case 567: goto st567;
		case 568: goto st568;
		case 569: goto st569;
		case 570: goto st570;
		case 571: goto st571;
		case 572: goto st572;
		case 573: goto st573;
		case 574: goto st574;
		case 575: goto st575;
		case 576: goto st576;
		case 577: goto st577;
		case 578: goto st578;
		case 579: goto st579;
		case 580: goto st580;
		case 581: goto st581;
		case 582: goto st582;
		case 583: goto st583;
		case 584: goto st584;
		case 585: goto st585;
		case 586: goto st586;
		case 587: goto st587;
		case 588: goto st588;
		case 589: goto st589;
		case 590: goto st590;
		case 591: goto st591;
		case 592: goto st592;
		case 593: goto st593;
		case 594: goto st594;
		case 1094: goto st1094;
		case 595: goto st595;
		case 596: goto st596;
		case 1095: goto st1095;
		case 597: goto st597;
		case 598: goto st598;
		case 599: goto st599;
		case 600: goto st600;
		case 601: goto st601;
		case 602: goto st602;
		case 603: goto st603;
		case 604: goto st604;
		case 605: goto st605;
		case 606: goto st606;
		case 607: goto st607;
		case 608: goto st608;
		case 609: goto st609;
		case 610: goto st610;
		case 1096: goto st1096;
		case 611: goto st611;
		case 612: goto st612;
		case 613: goto st613;
		case 614: goto st614;
		case 615: goto st615;
		case 616: goto st616;
		case 617: goto st617;
		case 618: goto st618;
		case 619: goto st619;
		case 620: goto st620;
		case 621: goto st621;
		case 622: goto st622;
		case 623: goto st623;
		case 624: goto st624;
		case 625: goto st625;
		case 626: goto st626;
		case 627: goto st627;
		case 628: goto st628;
		case 629: goto st629;
		case 630: goto st630;
		case 631: goto st631;
		case 632: goto st632;
		case 1097: goto st1097;
		case 633: goto st633;
		case 634: goto st634;
		case 635: goto st635;
		case 636: goto st636;
		case 637: goto st637;
		case 1098: goto st1098;
		case 638: goto st638;
		case 639: goto st639;
		case 640: goto st640;
		case 641: goto st641;
		case 642: goto st642;
		case 1099: goto st1099;
		case 643: goto st643;
		case 644: goto st644;
		case 645: goto st645;
		case 646: goto st646;
		case 647: goto st647;
		case 1100: goto st1100;
		case 1101: goto st1101;
		case 1102: goto st1102;
		case 648: goto st648;
		case 649: goto st649;
		case 1103: goto st1103;
		case 650: goto st650;
		case 651: goto st651;
		case 652: goto st652;
		case 653: goto st653;
		case 654: goto st654;
		case 655: goto st655;
		case 656: goto st656;
		case 657: goto st657;
		case 658: goto st658;
		case 659: goto st659;
		case 660: goto st660;
		case 661: goto st661;
		case 662: goto st662;
		case 663: goto st663;
		case 664: goto st664;
		case 665: goto st665;
		case 666: goto st666;
		case 667: goto st667;
		case 668: goto st668;
		case 669: goto st669;
		case 670: goto st670;
		case 671: goto st671;
		case 672: goto st672;
		case 673: goto st673;
		case 674: goto st674;
		case 675: goto st675;
		case 676: goto st676;
		case 1104: goto st1104;
		case 1105: goto st1105;
		case 1106: goto st1106;
		case 677: goto st677;
		case 678: goto st678;
		case 679: goto st679;
		case 1107: goto st1107;
		case 1108: goto st1108;
		case 680: goto st680;
		case 681: goto st681;
		case 682: goto st682;
		case 683: goto st683;
		case 1109: goto st1109;
		case 1110: goto st1110;
		case 684: goto st684;
		case 685: goto st685;
		case 686: goto st686;
		case 687: goto st687;
		case 1111: goto st1111;
		case 1112: goto st1112;
		case 688: goto st688;
		case 689: goto st689;
		case 690: goto st690;
		case 691: goto st691;
		case 692: goto st692;
		case 693: goto st693;
		case 694: goto st694;
		case 695: goto st695;
		case 696: goto st696;
		case 697: goto st697;
		case 698: goto st698;
		case 699: goto st699;
		case 700: goto st700;
		case 701: goto st701;
		case 702: goto st702;
		case 703: goto st703;
		case 704: goto st704;
		case 705: goto st705;
		case 706: goto st706;
		case 707: goto st707;
		case 708: goto st708;
		case 709: goto st709;
		case 710: goto st710;
		case 711: goto st711;
		case 712: goto st712;
		case 1113: goto st1113;
		case 713: goto st713;
		case 714: goto st714;
		case 715: goto st715;
		case 716: goto st716;
		case 717: goto st717;
		case 718: goto st718;
		case 719: goto st719;
		case 720: goto st720;
		case 721: goto st721;
		case 722: goto st722;
		case 723: goto st723;
		case 724: goto st724;
		case 725: goto st725;
		case 726: goto st726;
		case 727: goto st727;
		case 1114: goto st1114;
		case 728: goto st728;
		case 729: goto st729;
		case 730: goto st730;
		case 731: goto st731;
		case 732: goto st732;
		case 733: goto st733;
		case 734: goto st734;
		case 735: goto st735;
		case 736: goto st736;
		case 737: goto st737;
		case 738: goto st738;
		case 739: goto st739;
		case 740: goto st740;
		case 1115: goto st1115;
		case 741: goto st741;
		case 742: goto st742;
		case 743: goto st743;
		case 744: goto st744;
		case 745: goto st745;
		case 746: goto st746;
		case 747: goto st747;
		case 748: goto st748;
		case 749: goto st749;
		case 750: goto st750;
		case 751: goto st751;
		case 1116: goto st1116;
		case 1117: goto st1117;
		case 752: goto st752;
		case 753: goto st753;
		case 754: goto st754;
		case 1118: goto st1118;
		case 755: goto st755;
		case 756: goto st756;
		case 757: goto st757;
		case 758: goto st758;
		case 759: goto st759;
		case 760: goto st760;
		case 761: goto st761;
		case 762: goto st762;
		case 763: goto st763;
		case 764: goto st764;
		case 1119: goto st1119;
		case 1120: goto st1120;
		case 1121: goto st1121;
		case 765: goto st765;
		case 766: goto st766;
		case 767: goto st767;
		case 768: goto st768;
		case 769: goto st769;
		case 770: goto st770;
		case 771: goto st771;
		case 772: goto st772;
		case 773: goto st773;
		case 774: goto st774;
		case 775: goto st775;
		case 1122: goto st1122;
		case 1123: goto st1123;
		case 1124: goto st1124;
		case 776: goto st776;
		case 777: goto st777;
		case 778: goto st778;
		case 779: goto st779;
		case 780: goto st780;
		case 781: goto st781;
		case 782: goto st782;
		case 783: goto st783;
		case 784: goto st784;
		case 785: goto st785;
		case 786: goto st786;
		case 787: goto st787;
		case 1125: goto st1125;
		case 788: goto st788;
		case 789: goto st789;
		case 790: goto st790;
		case 1126: goto st1126;
		case 1127: goto st1127;
		case 791: goto st791;
		case 1128: goto st1128;
		case 1129: goto st1129;
		case 792: goto st792;
		case 1130: goto st1130;
		case 1131: goto st1131;
		case 793: goto st793;
		case 794: goto st794;
		case 795: goto st795;
		case 796: goto st796;
		case 797: goto st797;
		case 798: goto st798;
		case 799: goto st799;
		case 800: goto st800;
		case 801: goto st801;
		case 802: goto st802;
		case 803: goto st803;
		case 804: goto st804;
		case 805: goto st805;
		case 806: goto st806;
		case 807: goto st807;
		case 808: goto st808;
		case 809: goto st809;
		case 810: goto st810;
		case 811: goto st811;
		case 812: goto st812;
		case 813: goto st813;
		case 814: goto st814;
		case 815: goto st815;
		case 816: goto st816;
		case 817: goto st817;
		case 818: goto st818;
		case 819: goto st819;
		case 820: goto st820;
		case 821: goto st821;
		case 822: goto st822;
		case 823: goto st823;
		case 824: goto st824;
		case 825: goto st825;
		case 826: goto st826;
		case 827: goto st827;
		case 828: goto st828;
		case 829: goto st829;
		case 830: goto st830;
		case 831: goto st831;
		case 832: goto st832;
		case 833: goto st833;
		case 834: goto st834;
		case 1132: goto st1132;
		case 835: goto st835;
		case 836: goto st836;
		case 837: goto st837;
		case 838: goto st838;
		case 839: goto st839;
		case 840: goto st840;
		case 841: goto st841;
		case 842: goto st842;
		case 843: goto st843;
		case 844: goto st844;
		case 845: goto st845;
		case 846: goto st846;
		case 847: goto st847;
		case 848: goto st848;
		case 849: goto st849;
		case 850: goto st850;
		case 851: goto st851;
		case 852: goto st852;
		case 853: goto st853;
		case 854: goto st854;
		case 855: goto st855;
		case 856: goto st856;
		case 857: goto st857;
		case 858: goto st858;
		case 859: goto st859;
		case 860: goto st860;
		case 861: goto st861;
		case 862: goto st862;
		case 863: goto st863;
		case 864: goto st864;
		case 865: goto st865;
		case 866: goto st866;
		case 867: goto st867;
		case 868: goto st868;
		case 869: goto st869;
		case 870: goto st870;
		case 871: goto st871;
		case 872: goto st872;
		case 873: goto st873;
		case 874: goto st874;
		case 875: goto st875;
		case 876: goto st876;
		case 877: goto st877;
		case 878: goto st878;
		case 879: goto st879;
		case 880: goto st880;
		case 881: goto st881;
		case 882: goto st882;
		case 883: goto st883;
		case 884: goto st884;
		case 885: goto st885;
		case 886: goto st886;
		case 887: goto st887;
		case 888: goto st888;
		case 889: goto st889;
		case 890: goto st890;
		case 891: goto st891;
		case 892: goto st892;
		case 893: goto st893;
		case 894: goto st894;
		case 895: goto st895;
		case 896: goto st896;
		case 897: goto st897;
		case 898: goto st898;
		case 899: goto st899;
		case 900: goto st900;
		case 901: goto st901;
		case 902: goto st902;
		case 903: goto st903;
		case 904: goto st904;
		case 905: goto st905;
		case 906: goto st906;
		case 907: goto st907;
		case 908: goto st908;
		case 909: goto st909;
		case 910: goto st910;
		case 911: goto st911;
		case 912: goto st912;
		case 913: goto st913;
		case 914: goto st914;
		case 915: goto st915;
		case 916: goto st916;
		case 917: goto st917;
		case 918: goto st918;
		case 919: goto st919;
		case 920: goto st920;
		case 921: goto st921;
		case 922: goto st922;
		case 923: goto st923;
		case 924: goto st924;
		case 925: goto st925;
		case 926: goto st926;
		case 927: goto st927;
		case 928: goto st928;
		case 929: goto st929;
		case 930: goto st930;
		case 931: goto st931;
		case 932: goto st932;
		case 933: goto st933;
		case 934: goto st934;
		case 935: goto st935;
		case 936: goto st936;
		case 937: goto st937;
		case 938: goto st938;
		case 939: goto st939;
		case 940: goto st940;
		case 941: goto st941;
		case 942: goto st942;
		case 943: goto st943;
		case 944: goto st944;
		case 945: goto st945;
		case 946: goto st946;
		case 947: goto st947;
		case 948: goto st948;
		case 949: goto st949;
		case 950: goto st950;
		case 951: goto st951;
		case 952: goto st952;
		case 953: goto st953;
		case 954: goto st954;
		case 955: goto st955;
		case 956: goto st956;
		case 957: goto st957;
		case 958: goto st958;
		case 959: goto st959;
		case 960: goto st960;
		case 961: goto st961;
		case 1133: goto st1133;
		case 962: goto st962;
		case 963: goto st963;
		case 964: goto st964;
		case 965: goto st965;
		case 966: goto st966;
		case 967: goto st967;
		case 968: goto st968;
		case 969: goto st969;
		case 1134: goto st1134;
		case 970: goto st970;
		case 971: goto st971;
		case 972: goto st972;
		case 973: goto st973;
		case 974: goto st974;
		case 1135: goto st1135;
		case 975: goto st975;
		case 976: goto st976;
		case 977: goto st977;
		case 978: goto st978;
		case 979: goto st979;
		case 980: goto st980;
		case 981: goto st981;
		case 982: goto st982;
		case 983: goto st983;
		case 984: goto st984;
		case 985: goto st985;
		case 986: goto st986;
		case 987: goto st987;
		case 988: goto st988;
		case 989: goto st989;
		case 990: goto st990;
		case 991: goto st991;
		case 992: goto st992;
		case 1136: goto st1136;
		case 993: goto st993;
		case 994: goto st994;
		case 995: goto st995;
		case 996: goto st996;
		case 997: goto st997;
		case 998: goto st998;
		case 999: goto st999;
		case 1000: goto st1000;
		case 1001: goto st1001;
		case 1002: goto st1002;
		case 1003: goto st1003;
		case 1004: goto st1004;
		case 1005: goto st1005;
		case 1006: goto st1006;
		case 1007: goto st1007;
		case 1008: goto st1008;
		case 1009: goto st1009;
		case 1010: goto st1010;
		case 1011: goto st1011;
		case 1137: goto st1137;
		case 1012: goto st1012;
		case 1013: goto st1013;
		case 1014: goto st1014;
		case 1015: goto st1015;
		case 1016: goto st1016;
		case 1017: goto st1017;
		case 1018: goto st1018;
		case 1019: goto st1019;
		case 1020: goto st1020;
		case 1021: goto st1021;
		case 1022: goto st1022;
		case 1023: goto st1023;
		case 1024: goto st1024;
		case 1025: goto st1025;
		case 1026: goto st1026;
		case 1138: goto st1138;
		case 1139: goto st1139;
		case 1140: goto st1140;
		case 1027: goto st1027;
		case 1028: goto st1028;
		case 1029: goto st1029;
		case 1030: goto st1030;
		case 1031: goto st1031;
		case 1032: goto st1032;
		case 1033: goto st1033;
		case 1141: goto st1141;
		case 1034: goto st1034;
		case 1035: goto st1035;
		case 1036: goto st1036;
		case 1037: goto st1037;
		case 1038: goto st1038;
		case 1039: goto st1039;
		case 1040: goto st1040;
		case 1041: goto st1041;
		case 1042: goto st1042;
		case 1043: goto st1043;
		case 1044: goto st1044;
		case 1045: goto st1045;
		case 1046: goto st1046;
		case 1142: goto st1142;
		case 1047: goto st1047;
		case 1048: goto st1048;
		case 1049: goto st1049;
		case 1050: goto st1050;
		case 1051: goto st1051;
		case 1052: goto st1052;
		case 1053: goto st1053;
		case 1143: goto st1143;
		case 1054: goto st1054;
		case 1055: goto st1055;
		case 1056: goto st1056;
		case 1057: goto st1057;
		case 1058: goto st1058;
		case 1059: goto st1059;
		case 1144: goto st1144;
	default: break;
	}

	if ( ++p == pe )
		goto _test_eof;
_resume:
	switch ( cs )
	{
tr19:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1060;
tr82:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1060;
tr88:
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1060;
tr3214:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1060;
st1060:
	if ( ++p == pe )
		goto _test_eof1060;
case 1060:
#line 1409 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3155;
		case 32: goto tr3155;
		case 36: goto st139;
		case 40: goto tr3157;
		case 41: goto tr3158;
		case 42: goto tr3159;
		case 92: goto tr3159;
		case 95: goto tr3159;
		case 778: goto tr19;
		case 827: goto st138;
		case 1034: goto tr3160;
		case 1083: goto tr3161;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3159;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3159;
	} else
		goto tr3159;
	goto tr3154;
tr0:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr23:
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	goto st0;
tr35:
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	goto st0;
tr55:
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr69:
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr78:
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr84:
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr101:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr125:
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr131:
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr530:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	goto st0;
tr532:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr597:
#line 724 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_DIRECTIVE);
		p--; {goto st247;}
	}
	goto st0;
tr608:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
	goto st0;
tr620:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	goto st0;
tr644:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr659:
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr687:
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 724 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_DIRECTIVE);
		p--; {goto st247;}
	}
	goto st0;
tr697:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr703:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr714:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr727:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr788:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	goto st0;
tr793:
#line 189 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_DNAME_CHAR);
		p--; {goto st247;}
	}
	goto st0;
tr808:
#line 143 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 189 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_DNAME_CHAR);
		p--; {goto st247;}
	}
	goto st0;
tr816:
#line 509 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT_CHAR);
		p--; {goto st247;}
	}
#line 513 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT);
		p--; {goto st247;}
	}
	goto st0;
tr823:
#line 543 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 509 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT_CHAR);
		p--; {goto st247;}
	}
#line 513 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT);
		p--; {goto st247;}
	}
	goto st0;
tr833:
#line 513 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT);
		p--; {goto st247;}
	}
	goto st0;
tr840:
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	goto st0;
tr846:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	goto st0;
tr848:
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	goto st0;
tr860:
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	goto st0;
tr862:
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	goto st0;
tr875:
#line 600 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_ORIGIN);
		p--; {goto st247;}
	}
	goto st0;
tr886:
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 600 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_ORIGIN);
		p--; {goto st247;}
	}
	goto st0;
tr903:
#line 143 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 600 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_ORIGIN);
		p--; {goto st247;}
	}
	goto st0;
tr915:
#line 629 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
		p--; {goto st247;}
	}
	goto st0;
tr916:
#line 629 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr923:
#line 640 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_ORIGIN);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr934:
#line 640 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_ORIGIN);
		p--; {goto st247;}
	}
	goto st0;
tr950:
#line 143 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 640 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_ORIGIN);
		p--; {goto st247;}
	}
	goto st0;
tr962:
#line 974 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_BASE64_CHAR);
		p--; {goto st247;}
	}
	goto st0;
tr976:
#line 1334 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_BITMAP);
		p--; {goto st247;}
	}
	goto st0;
tr1423:
#line 1500 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr1429:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1500 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr1437:
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 932 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
#line 1500 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr1487:
#line 1576 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ALGORITHM);
		p--; {goto st247;}
	}
	goto st0;
tr1598:
#line 1580 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CERT_TYPE);
		p--; {goto st247;}
	}
	goto st0;
tr1651:
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr1674:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr1693:
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr1860:
#line 1487 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_LOC_DATA);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2217:
#line 875 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_APL);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2232:
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 875 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_APL);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2275:
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2332:
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2347:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2361:
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1186 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY_KEY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2384:
#line 1186 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY_KEY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2398:
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2405:
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1186 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY_KEY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2510:
#line 1204 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2563:
#line 491 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2902:
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1204 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr2990:
#line 1058 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_BASE32HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr3133:
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1563 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CHAR_COLON);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr3139:
#line 1563 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CHAR_COLON);
		p--; {goto st247;}
	}
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr3143:
#line 1536 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CHAR_DASH);                       
		p--; {goto st247;}
	}
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	goto st0;
tr3154:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr3172:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr3192:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr3208:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 724 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_DIRECTIVE);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
tr3231:
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	goto st0;
#line 2471 "zscanner/scanner.c"
st0:
cs = 0;
	goto _out;
tr2:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1;
tr3:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1;
tr3155:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1;
tr3157:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1;
tr3158:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1;
tr3218:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1;
tr3219:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1;
tr3209:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1;
tr3211:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1;
tr3212:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1;
st1:
	if ( ++p == pe )
		goto _test_eof1;
case 1:
#line 2631 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1;
		case 32: goto st1;
		case 40: goto tr2;
		case 41: goto tr3;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr19;
		case 827: goto st138;
		case 1034: goto tr21;
		case 1083: goto st243;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr4:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st2;
tr27:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st2;
st2:
	if ( ++p == pe )
		goto _test_eof2;
case 2:
#line 2729 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr24;
		case 32: goto tr24;
		case 40: goto tr25;
		case 41: goto tr26;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st153;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st153;
		case 119: goto tr32;
		case 1034: goto tr33;
		case 1083: goto tr34;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr23;
tr37:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st3;
tr38:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st3;
tr53:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st3;
tr24:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st3;
tr25:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st3;
tr26:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st3;
tr33:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st3;
tr638:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st3;
tr639:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st3;
tr640:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st3;
tr642:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st3;
st3:
	if ( ++p == pe )
		goto _test_eof3;
case 3:
#line 2959 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st3;
		case 32: goto st3;
		case 40: goto tr37;
		case 41: goto tr38;
		case 65: goto tr39;
		case 67: goto tr40;
		case 68: goto tr41;
		case 69: goto tr42;
		case 72: goto tr43;
		case 73: goto tr44;
		case 75: goto tr45;
		case 76: goto tr46;
		case 77: goto tr47;
		case 78: goto tr48;
		case 80: goto tr49;
		case 82: goto tr50;
		case 83: goto tr51;
		case 84: goto tr52;
		case 97: goto tr39;
		case 99: goto tr40;
		case 100: goto tr41;
		case 101: goto tr42;
		case 104: goto tr43;
		case 105: goto tr44;
		case 107: goto tr45;
		case 108: goto tr46;
		case 109: goto tr47;
		case 110: goto tr48;
		case 112: goto tr49;
		case 114: goto tr50;
		case 115: goto tr51;
		case 116: goto tr52;
		case 1034: goto tr53;
		case 1083: goto st156;
	}
	goto tr35;
tr5:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
tr39:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st4;
tr557:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
st4:
	if ( ++p == pe )
		goto _test_eof4;
case 4:
#line 3037 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr56;
		case 32: goto tr56;
		case 40: goto tr57;
		case 41: goto tr58;
		case 65: goto st214;
		case 70: goto st217;
		case 80: goto st221;
		case 97: goto st214;
		case 102: goto st217;
		case 112: goto st221;
		case 2058: goto tr62;
		case 2107: goto tr63;
		case 2314: goto tr64;
		case 2363: goto tr64;
		case 2570: goto tr65;
		case 2619: goto tr66;
	}
	goto tr55;
tr70:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr71:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr73:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr56:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr57:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr58:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr62:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr141:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr142:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr143:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr144:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr155:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr156:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr157:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr158:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr169:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr170:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr171:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr172:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr181:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr182:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr183:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr184:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr192:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr193:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr194:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr195:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr200:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr201:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr202:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr203:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr213:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr214:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr215:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr216:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr222:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr223:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr224:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr225:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr234:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr235:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr236:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr237:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr274:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr275:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr276:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr277:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr285:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr286:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr287:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr288:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr293:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr294:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr295:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr296:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr306:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr307:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr308:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr309:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr315:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr316:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr317:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr318:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr324:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr325:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr326:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr327:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr332:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr333:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr334:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr335:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr345:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr346:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr347:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr348:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr353:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr354:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr355:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr356:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr367:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr368:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr369:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr370:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr376:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr377:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr378:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr379:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr384:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr385:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr386:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr388:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr394:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr395:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr396:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr398:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr403:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr404:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr405:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr407:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr416:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr417:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr418:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr419:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr426:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr427:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr428:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr429:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr437:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr438:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr439:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr440:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr448:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr449:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr450:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr451:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr456:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr457:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr458:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr459:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr469:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr470:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr471:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr472:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr478:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr479:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr480:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr481:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr487:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr488:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr489:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr490:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr498:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr499:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr500:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr501:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr511:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr512:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr513:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr514:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr520:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr521:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr522:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr523:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr533:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr534:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr535:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr537:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr748:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr749:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr750:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr751:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr759:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr760:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr761:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr762:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
tr768:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st5;
tr769:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st5;
tr770:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st5;
tr771:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st5;
st5:
	if ( ++p == pe )
		goto _test_eof5;
case 5:
#line 5475 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st5;
		case 32: goto st5;
		case 40: goto tr70;
		case 41: goto tr71;
		case 92: goto st9;
		case 2058: goto tr73;
		case 2107: goto st11;
		case 2314: goto tr75;
		case 2363: goto tr75;
		case 2570: goto tr76;
		case 2619: goto tr77;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr67;
	} else
		goto tr67;
	goto tr69;
tr67:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 6; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st6;
tr90:
#line 1873 "./zscanner/scanner_body.rl"
	{ p--; }
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 6; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 6; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st6;
st6:
	if ( ++p == pe )
		goto _test_eof6;
case 6:
#line 5676 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr78;
tr86:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st7;
tr87:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st7;
tr79:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st7;
tr80:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st7;
tr81:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st7;
st7:
	if ( ++p == pe )
		goto _test_eof7;
case 7:
#line 5763 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr86;
		case 41: goto tr87;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr88;
		case 1083: goto st8;
	}
	goto tr84;
tr83:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st8;
st8:
	if ( ++p == pe )
		goto _test_eof8;
case 8:
#line 5800 "zscanner/scanner.c"
	if ( (*p) == 10 )
		goto tr88;
	goto st8;
st9:
	if ( ++p == pe )
		goto _test_eof9;
case 9:
	if ( (*p) == 35 )
		goto tr91;
	goto tr90;
tr64:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr75:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr91:
#line 1817 "./zscanner/scanner_body.rl"
	{
		switch (s->r_type) {
		// Next types must not have empty rdata.
		case KNOT_RRTYPE_A:
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
		case KNOT_RRTYPE_SOA:
		case KNOT_RRTYPE_HINFO:
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
		case KNOT_RRTYPE_RP:
		case KNOT_RRTYPE_AAAA:
		case KNOT_RRTYPE_LOC:
		case KNOT_RRTYPE_SRV:
		case KNOT_RRTYPE_NAPTR:
		case KNOT_RRTYPE_CERT:
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_SSHFP:
		case KNOT_RRTYPE_IPSECKEY:
		case KNOT_RRTYPE_RRSIG:
		case KNOT_RRTYPE_NSEC:
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_NSEC3:
		case KNOT_RRTYPE_NSEC3PARAM:
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L32:
		case KNOT_RRTYPE_L64:
		case KNOT_RRTYPE_LP:
		case KNOT_RRTYPE_EUI48:
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st438;}
		// Next types can have empty rdata.
		case KNOT_RRTYPE_APL:
		default:
			{stack[top++] = 10; goto st447;}
		}
	}
	goto st10;
tr146:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr160:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr174:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr186:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr197:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr205:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr218:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr227:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr239:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr279:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr290:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr298:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr311:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr320:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr329:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr337:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr350:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr358:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr372:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr381:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr390:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr400:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr409:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr421:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr431:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr442:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr453:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr461:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr474:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr483:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr492:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr503:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr516:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr525:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr539:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr753:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr764:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
tr773:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 10; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 10; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 10; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 10; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 10; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 10; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 10; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 10; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 10; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 10; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 10; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 10; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 10; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 10; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 10; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 10; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 10; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 10; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 10; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 10; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 10; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 10; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 10; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 10; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 10; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 10; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 10; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st10;
st10:
	if ( ++p == pe )
		goto _test_eof10;
case 10:
#line 9224 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr86;
		case 41: goto tr87;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr88;
		case 1083: goto st8;
	}
	goto tr78;
tr63:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr145:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr159:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr173:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr185:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr196:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr204:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr217:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr226:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr238:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr278:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr289:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr297:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr310:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr319:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr328:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr336:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr349:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr357:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr371:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr380:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr389:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr399:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr408:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr420:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr430:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr441:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr452:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr460:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr473:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr482:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr491:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr502:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr515:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr524:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr538:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr752:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr763:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
tr772:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st11;
st11:
	if ( ++p == pe )
		goto _test_eof11;
case 11:
#line 9651 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr73;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st11;
	goto tr69;
tr93:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st12;
tr94:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st12;
tr65:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr76:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr147:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr161:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr175:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr187:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr198:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr206:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr219:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr228:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr240:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr280:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr291:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr299:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr312:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr321:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr330:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr338:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr351:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr359:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr373:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr382:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr391:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr401:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr410:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr422:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr432:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr443:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr454:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr462:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr475:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr484:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr493:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr504:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr517:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr526:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr540:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr754:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr765:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
tr774:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 12; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 12; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 12; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 12; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 12; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 12; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 12; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 12; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 12; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 12; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 12; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 12; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 12; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 12; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 12; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 12; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 12; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 12; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 12; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 12; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 12; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 12; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 12; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 12; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 12; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 12; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 12; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st12;
st12:
	if ( ++p == pe )
		goto _test_eof12;
case 12:
#line 13221 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st12;
		case 32: goto st12;
		case 40: goto tr93;
		case 41: goto tr94;
		case 92: goto st9;
		case 1802: goto tr88;
		case 1851: goto st8;
		case 2058: goto tr88;
		case 2107: goto st13;
		case 2314: goto tr96;
		case 2363: goto tr97;
		case 2570: goto tr88;
		case 2619: goto tr98;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr67;
	} else
		goto tr67;
	goto tr78;
tr664:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st13;
st13:
	if ( ++p == pe )
		goto _test_eof13;
case 13:
#line 13285 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr88;
		case 1034: goto tr99;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st13;
	} else if ( _widec >= 768 )
		goto st8;
	goto tr78;
tr99:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1061;
tr663:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1061;
st1061:
	if ( ++p == pe )
		goto _test_eof1061;
case 1061:
#line 13366 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3162;
		case 32: goto tr3162;
		case 36: goto tr3163;
		case 40: goto tr3164;
		case 41: goto tr3165;
		case 42: goto tr3166;
		case 58: goto tr67;
		case 92: goto tr3167;
		case 95: goto tr3166;
		case 1802: goto tr19;
		case 1851: goto st138;
		case 2058: goto tr3168;
		case 2107: goto tr3169;
		case 2314: goto tr121;
		case 2363: goto tr122;
		case 2570: goto tr3170;
		case 2619: goto tr3171;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3166;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3166;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 97 )
				goto tr3166;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr697;
tr102:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st14;
tr103:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st14;
tr3162:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st14;
tr3164:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st14;
tr3165:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st14;
st14:
	if ( ++p == pe )
		goto _test_eof14;
case 14:
#line 13497 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st14;
		case 32: goto st14;
		case 40: goto tr102;
		case 41: goto tr103;
		case 58: goto tr67;
		case 65: goto tr105;
		case 67: goto tr106;
		case 68: goto tr107;
		case 69: goto tr108;
		case 72: goto tr109;
		case 73: goto tr110;
		case 75: goto tr111;
		case 76: goto tr112;
		case 77: goto tr113;
		case 78: goto tr114;
		case 80: goto tr115;
		case 82: goto tr116;
		case 83: goto tr117;
		case 84: goto tr118;
		case 92: goto st9;
		case 97: goto tr105;
		case 99: goto tr106;
		case 100: goto tr107;
		case 101: goto tr108;
		case 104: goto tr109;
		case 105: goto tr110;
		case 107: goto tr111;
		case 108: goto tr112;
		case 109: goto tr113;
		case 110: goto tr114;
		case 112: goto tr115;
		case 114: goto tr116;
		case 115: goto tr117;
		case 116: goto tr118;
		case 1802: goto tr19;
		case 1851: goto st138;
		case 2058: goto tr119;
		case 2107: goto st223;
		case 2314: goto tr121;
		case 2363: goto tr122;
		case 2570: goto tr123;
		case 2619: goto tr124;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr104;
	} else
		goto tr67;
	goto tr101;
tr104:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 15; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 15; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 15; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 15; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 15; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 15; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 15; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 15; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 15; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 15; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 15; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 15; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 15; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 15; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 15; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 15; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 15; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 15; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 15; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 15; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 15; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 15; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 15; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 15; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 15; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 15; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 15; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st15;
st15:
	if ( ++p == pe )
		goto _test_eof15;
case 15:
#line 13673 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr126;
		case 32: goto tr126;
		case 40: goto tr127;
		case 41: goto tr128;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st153;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st153;
		case 119: goto tr32;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr129;
		case 1083: goto tr130;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr125;
tr133:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st16;
tr134:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st16;
tr126:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st16;
tr127:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st16;
tr128:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st16;
st16:
	if ( ++p == pe )
		goto _test_eof16;
case 16:
#line 13799 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st16;
		case 32: goto st16;
		case 40: goto tr133;
		case 41: goto tr134;
		case 65: goto tr39;
		case 67: goto tr40;
		case 68: goto tr41;
		case 69: goto tr42;
		case 72: goto tr43;
		case 73: goto tr44;
		case 75: goto tr45;
		case 76: goto tr46;
		case 77: goto tr47;
		case 78: goto tr48;
		case 80: goto tr49;
		case 82: goto tr50;
		case 83: goto tr51;
		case 84: goto tr52;
		case 97: goto tr39;
		case 99: goto tr40;
		case 100: goto tr41;
		case 101: goto tr42;
		case 104: goto tr43;
		case 105: goto tr44;
		case 107: goto tr45;
		case 108: goto tr46;
		case 109: goto tr47;
		case 110: goto tr48;
		case 112: goto tr49;
		case 114: goto tr50;
		case 115: goto tr51;
		case 116: goto tr52;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr135;
		case 1083: goto st246;
	}
	goto tr131;
tr6:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st17;
tr40:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st17;
tr558:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st17;
st17:
	if ( ++p == pe )
		goto _test_eof17;
case 17:
#line 13879 "zscanner/scanner.c"
	switch( (*p) ) {
		case 69: goto st18;
		case 78: goto st22;
		case 101: goto st18;
		case 110: goto st22;
	}
	goto tr35;
st18:
	if ( ++p == pe )
		goto _test_eof18;
case 18:
	switch( (*p) ) {
		case 82: goto st19;
		case 114: goto st19;
	}
	goto tr35;
st19:
	if ( ++p == pe )
		goto _test_eof19;
case 19:
	switch( (*p) ) {
		case 84: goto st20;
		case 116: goto st20;
	}
	goto tr35;
st20:
	if ( ++p == pe )
		goto _test_eof20;
case 20:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr141;
		case 32: goto tr141;
		case 40: goto tr142;
		case 41: goto tr143;
		case 2058: goto tr144;
		case 2107: goto tr145;
		case 2314: goto tr146;
		case 2363: goto tr146;
		case 2570: goto tr147;
		case 2619: goto tr148;
	}
	goto tr55;
tr150:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st21;
tr151:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st21;
tr66:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr77:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr148:
#line 1904 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CERT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr162:
#line 1888 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_CNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr176:
#line 1913 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DHCID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr188:
#line 1905 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNAME; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr199:
#line 1912 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr207:
#line 1907 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_DS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr220:
#line 1922 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI48; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr229:
#line 1923 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_EUI64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr241:
#line 1891 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_HINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr281:
#line 1909 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr292:
#line 1898 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KEY; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr300:
#line 1903 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_KX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr313:
#line 1919 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L32; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr322:
#line 1920 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_L64; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr331:
#line 1900 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LOC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr339:
#line 1921 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_LP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr352:
#line 1892 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MINFO; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr360:
#line 1893 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_MX; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr374:
#line 1902 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr383:
#line 1918 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NID; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr392:
#line 1887 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NS; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr402:
#line 1911 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr411:
#line 1914 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr423:
#line 1915 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr433:
#line 1890 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_PTR; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr444:
#line 1895 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr455:
#line 1910 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr463:
#line 1897 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_RT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr476:
#line 1889 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SOA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr485:
#line 1917 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SPF; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr494:
#line 1901 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SRV; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr505:
#line 1908 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr518:
#line 1916 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TLSA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr527:
#line 1894 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_TXT; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr541:
#line 360 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr755:
#line 1899 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AAAA; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr766:
#line 1896 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
tr775:
#line 1906 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_APL; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 21; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 21; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 21; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 21; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 21; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 21; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 21; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 21; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 21; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 21; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 21; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 21; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 21; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 21; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 21; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 21; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 21; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 21; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 21; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 21; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 21; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 21; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 21; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 21; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 21; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 21; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 21; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st21;
st21:
	if ( ++p == pe )
		goto _test_eof21;
case 21:
#line 17325 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr88;
		case 800: goto st7;
		case 808: goto tr86;
		case 809: goto tr87;
		case 827: goto st8;
		case 1033: goto st21;
		case 1034: goto tr88;
		case 1056: goto st21;
		case 1064: goto tr150;
		case 1065: goto tr151;
		case 1083: goto st13;
	}
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st11;
	goto tr78;
st22:
	if ( ++p == pe )
		goto _test_eof22;
case 22:
	switch( (*p) ) {
		case 65: goto st23;
		case 97: goto st23;
	}
	goto tr35;
st23:
	if ( ++p == pe )
		goto _test_eof23;
case 23:
	switch( (*p) ) {
		case 77: goto st24;
		case 109: goto st24;
	}
	goto tr35;
st24:
	if ( ++p == pe )
		goto _test_eof24;
case 24:
	switch( (*p) ) {
		case 69: goto st25;
		case 101: goto st25;
	}
	goto tr35;
st25:
	if ( ++p == pe )
		goto _test_eof25;
case 25:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr155;
		case 32: goto tr155;
		case 40: goto tr156;
		case 41: goto tr157;
		case 2058: goto tr158;
		case 2107: goto tr159;
		case 2314: goto tr160;
		case 2363: goto tr160;
		case 2570: goto tr161;
		case 2619: goto tr162;
	}
	goto tr55;
tr7:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st26;
tr41:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st26;
tr559:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st26;
st26:
	if ( ++p == pe )
		goto _test_eof26;
case 26:
#line 17468 "zscanner/scanner.c"
	switch( (*p) ) {
		case 72: goto st27;
		case 78: goto st31;
		case 83: goto st39;
		case 104: goto st27;
		case 110: goto st31;
		case 115: goto st39;
	}
	goto tr35;
st27:
	if ( ++p == pe )
		goto _test_eof27;
case 27:
	switch( (*p) ) {
		case 67: goto st28;
		case 99: goto st28;
	}
	goto tr35;
st28:
	if ( ++p == pe )
		goto _test_eof28;
case 28:
	switch( (*p) ) {
		case 73: goto st29;
		case 105: goto st29;
	}
	goto tr35;
st29:
	if ( ++p == pe )
		goto _test_eof29;
case 29:
	switch( (*p) ) {
		case 68: goto st30;
		case 100: goto st30;
	}
	goto tr35;
st30:
	if ( ++p == pe )
		goto _test_eof30;
case 30:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr169;
		case 32: goto tr169;
		case 40: goto tr170;
		case 41: goto tr171;
		case 2058: goto tr172;
		case 2107: goto tr173;
		case 2314: goto tr174;
		case 2363: goto tr174;
		case 2570: goto tr175;
		case 2619: goto tr176;
	}
	goto tr55;
st31:
	if ( ++p == pe )
		goto _test_eof31;
case 31:
	switch( (*p) ) {
		case 65: goto st32;
		case 83: goto st35;
		case 97: goto st32;
		case 115: goto st35;
	}
	goto tr35;
st32:
	if ( ++p == pe )
		goto _test_eof32;
case 32:
	switch( (*p) ) {
		case 77: goto st33;
		case 109: goto st33;
	}
	goto tr35;
st33:
	if ( ++p == pe )
		goto _test_eof33;
case 33:
	switch( (*p) ) {
		case 69: goto st34;
		case 101: goto st34;
	}
	goto tr35;
st34:
	if ( ++p == pe )
		goto _test_eof34;
case 34:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr181;
		case 32: goto tr181;
		case 40: goto tr182;
		case 41: goto tr183;
		case 2058: goto tr184;
		case 2107: goto tr185;
		case 2314: goto tr186;
		case 2363: goto tr186;
		case 2570: goto tr187;
		case 2619: goto tr188;
	}
	goto tr55;
st35:
	if ( ++p == pe )
		goto _test_eof35;
case 35:
	switch( (*p) ) {
		case 75: goto st36;
		case 107: goto st36;
	}
	goto tr35;
st36:
	if ( ++p == pe )
		goto _test_eof36;
case 36:
	switch( (*p) ) {
		case 69: goto st37;
		case 101: goto st37;
	}
	goto tr35;
st37:
	if ( ++p == pe )
		goto _test_eof37;
case 37:
	switch( (*p) ) {
		case 89: goto st38;
		case 121: goto st38;
	}
	goto tr35;
st38:
	if ( ++p == pe )
		goto _test_eof38;
case 38:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr192;
		case 32: goto tr192;
		case 40: goto tr193;
		case 41: goto tr194;
		case 2058: goto tr195;
		case 2107: goto tr196;
		case 2314: goto tr197;
		case 2363: goto tr197;
		case 2570: goto tr198;
		case 2619: goto tr199;
	}
	goto tr55;
st39:
	if ( ++p == pe )
		goto _test_eof39;
case 39:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr200;
		case 32: goto tr200;
		case 40: goto tr201;
		case 41: goto tr202;
		case 2058: goto tr203;
		case 2107: goto tr204;
		case 2314: goto tr205;
		case 2363: goto tr205;
		case 2570: goto tr206;
		case 2619: goto tr207;
	}
	goto tr55;
tr8:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st40;
tr42:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st40;
tr560:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st40;
st40:
	if ( ++p == pe )
		goto _test_eof40;
case 40:
#line 17735 "zscanner/scanner.c"
	switch( (*p) ) {
		case 85: goto st41;
		case 117: goto st41;
	}
	goto tr35;
st41:
	if ( ++p == pe )
		goto _test_eof41;
case 41:
	switch( (*p) ) {
		case 73: goto st42;
		case 105: goto st42;
	}
	goto tr35;
st42:
	if ( ++p == pe )
		goto _test_eof42;
case 42:
	switch( (*p) ) {
		case 52: goto st43;
		case 54: goto st45;
	}
	goto tr35;
st43:
	if ( ++p == pe )
		goto _test_eof43;
case 43:
	if ( (*p) == 56 )
		goto st44;
	goto tr35;
st44:
	if ( ++p == pe )
		goto _test_eof44;
case 44:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr213;
		case 32: goto tr213;
		case 40: goto tr214;
		case 41: goto tr215;
		case 2058: goto tr216;
		case 2107: goto tr217;
		case 2314: goto tr218;
		case 2363: goto tr218;
		case 2570: goto tr219;
		case 2619: goto tr220;
	}
	goto tr55;
st45:
	if ( ++p == pe )
		goto _test_eof45;
case 45:
	if ( (*p) == 52 )
		goto st46;
	goto tr35;
st46:
	if ( ++p == pe )
		goto _test_eof46;
case 46:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr222;
		case 32: goto tr222;
		case 40: goto tr223;
		case 41: goto tr224;
		case 2058: goto tr225;
		case 2107: goto tr226;
		case 2314: goto tr227;
		case 2363: goto tr227;
		case 2570: goto tr228;
		case 2619: goto tr229;
	}
	goto tr55;
tr9:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st47;
tr43:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st47;
tr561:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st47;
st47:
	if ( ++p == pe )
		goto _test_eof47;
case 47:
#line 17873 "zscanner/scanner.c"
	switch( (*p) ) {
		case 73: goto st48;
		case 105: goto st48;
	}
	goto tr35;
st48:
	if ( ++p == pe )
		goto _test_eof48;
case 48:
	switch( (*p) ) {
		case 78: goto st49;
		case 110: goto st49;
	}
	goto tr35;
st49:
	if ( ++p == pe )
		goto _test_eof49;
case 49:
	switch( (*p) ) {
		case 70: goto st50;
		case 102: goto st50;
	}
	goto tr35;
st50:
	if ( ++p == pe )
		goto _test_eof50;
case 50:
	switch( (*p) ) {
		case 79: goto st51;
		case 111: goto st51;
	}
	goto tr35;
st51:
	if ( ++p == pe )
		goto _test_eof51;
case 51:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr234;
		case 32: goto tr234;
		case 40: goto tr235;
		case 41: goto tr236;
		case 2058: goto tr237;
		case 2107: goto tr238;
		case 2314: goto tr239;
		case 2363: goto tr239;
		case 2570: goto tr240;
		case 2619: goto tr241;
	}
	goto tr55;
tr44:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st52;
st52:
	if ( ++p == pe )
		goto _test_eof52;
case 52:
#line 17953 "zscanner/scanner.c"
	switch( (*p) ) {
		case 78: goto st53;
		case 80: goto st56;
		case 110: goto st53;
		case 112: goto st56;
	}
	goto tr35;
st53:
	if ( ++p == pe )
		goto _test_eof53;
case 53:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr244;
		case 32: goto tr244;
		case 40: goto tr245;
		case 41: goto tr246;
		case 1034: goto tr247;
		case 1083: goto tr248;
	}
	goto tr35;
tr250:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st54;
tr251:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st54;
tr266:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st54;
tr573:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st54;
tr574:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st54;
tr575:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st54;
tr582:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st54;
tr244:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st54;
tr245:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st54;
tr246:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st54;
tr247:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st54;
tr591:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st54;
tr592:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st54;
tr593:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st54;
tr595:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st54;
st54:
	if ( ++p == pe )
		goto _test_eof54;
case 54:
#line 18226 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st54;
		case 32: goto st54;
		case 40: goto tr250;
		case 41: goto tr251;
		case 65: goto st4;
		case 67: goto st17;
		case 68: goto st26;
		case 69: goto st40;
		case 72: goto st47;
		case 73: goto st55;
		case 75: goto st63;
		case 76: goto st67;
		case 77: goto st75;
		case 78: goto st81;
		case 80: goto st97;
		case 82: goto st100;
		case 83: goto st107;
		case 84: goto st118;
		case 97: goto st4;
		case 99: goto st17;
		case 100: goto st26;
		case 101: goto st40;
		case 104: goto st47;
		case 105: goto st55;
		case 107: goto st63;
		case 108: goto st67;
		case 109: goto st75;
		case 110: goto st81;
		case 112: goto st97;
		case 114: goto st100;
		case 115: goto st107;
		case 116: goto st118;
		case 1034: goto tr266;
		case 1083: goto st128;
	}
	goto tr35;
tr562:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st55;
st55:
	if ( ++p == pe )
		goto _test_eof55;
case 55:
#line 18288 "zscanner/scanner.c"
	switch( (*p) ) {
		case 80: goto st56;
		case 112: goto st56;
	}
	goto tr35;
st56:
	if ( ++p == pe )
		goto _test_eof56;
case 56:
	switch( (*p) ) {
		case 83: goto st57;
		case 115: goto st57;
	}
	goto tr35;
st57:
	if ( ++p == pe )
		goto _test_eof57;
case 57:
	switch( (*p) ) {
		case 69: goto st58;
		case 101: goto st58;
	}
	goto tr35;
st58:
	if ( ++p == pe )
		goto _test_eof58;
case 58:
	switch( (*p) ) {
		case 67: goto st59;
		case 99: goto st59;
	}
	goto tr35;
st59:
	if ( ++p == pe )
		goto _test_eof59;
case 59:
	switch( (*p) ) {
		case 75: goto st60;
		case 107: goto st60;
	}
	goto tr35;
st60:
	if ( ++p == pe )
		goto _test_eof60;
case 60:
	switch( (*p) ) {
		case 69: goto st61;
		case 101: goto st61;
	}
	goto tr35;
st61:
	if ( ++p == pe )
		goto _test_eof61;
case 61:
	switch( (*p) ) {
		case 89: goto st62;
		case 121: goto st62;
	}
	goto tr35;
st62:
	if ( ++p == pe )
		goto _test_eof62;
case 62:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr274;
		case 32: goto tr274;
		case 40: goto tr275;
		case 41: goto tr276;
		case 2058: goto tr277;
		case 2107: goto tr278;
		case 2314: goto tr279;
		case 2363: goto tr279;
		case 2570: goto tr280;
		case 2619: goto tr281;
	}
	goto tr55;
tr11:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st63;
tr45:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st63;
tr563:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st63;
st63:
	if ( ++p == pe )
		goto _test_eof63;
case 63:
#line 18411 "zscanner/scanner.c"
	switch( (*p) ) {
		case 69: goto st64;
		case 88: goto st66;
		case 101: goto st64;
		case 120: goto st66;
	}
	goto tr35;
st64:
	if ( ++p == pe )
		goto _test_eof64;
case 64:
	switch( (*p) ) {
		case 89: goto st65;
		case 121: goto st65;
	}
	goto tr35;
st65:
	if ( ++p == pe )
		goto _test_eof65;
case 65:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr285;
		case 32: goto tr285;
		case 40: goto tr286;
		case 41: goto tr287;
		case 2058: goto tr288;
		case 2107: goto tr289;
		case 2314: goto tr290;
		case 2363: goto tr290;
		case 2570: goto tr291;
		case 2619: goto tr292;
	}
	goto tr55;
st66:
	if ( ++p == pe )
		goto _test_eof66;
case 66:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr293;
		case 32: goto tr293;
		case 40: goto tr294;
		case 41: goto tr295;
		case 2058: goto tr296;
		case 2107: goto tr297;
		case 2314: goto tr298;
		case 2363: goto tr298;
		case 2570: goto tr299;
		case 2619: goto tr300;
	}
	goto tr55;
tr12:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st67;
tr46:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st67;
tr564:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st67;
st67:
	if ( ++p == pe )
		goto _test_eof67;
case 67:
#line 18528 "zscanner/scanner.c"
	switch( (*p) ) {
		case 51: goto st68;
		case 54: goto st70;
		case 79: goto st72;
		case 80: goto st74;
		case 111: goto st72;
		case 112: goto st74;
	}
	goto tr35;
st68:
	if ( ++p == pe )
		goto _test_eof68;
case 68:
	if ( (*p) == 50 )
		goto st69;
	goto tr35;
st69:
	if ( ++p == pe )
		goto _test_eof69;
case 69:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr306;
		case 32: goto tr306;
		case 40: goto tr307;
		case 41: goto tr308;
		case 2058: goto tr309;
		case 2107: goto tr310;
		case 2314: goto tr311;
		case 2363: goto tr311;
		case 2570: goto tr312;
		case 2619: goto tr313;
	}
	goto tr55;
st70:
	if ( ++p == pe )
		goto _test_eof70;
case 70:
	if ( (*p) == 52 )
		goto st71;
	goto tr35;
st71:
	if ( ++p == pe )
		goto _test_eof71;
case 71:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr315;
		case 32: goto tr315;
		case 40: goto tr316;
		case 41: goto tr317;
		case 2058: goto tr318;
		case 2107: goto tr319;
		case 2314: goto tr320;
		case 2363: goto tr320;
		case 2570: goto tr321;
		case 2619: goto tr322;
	}
	goto tr55;
st72:
	if ( ++p == pe )
		goto _test_eof72;
case 72:
	switch( (*p) ) {
		case 67: goto st73;
		case 99: goto st73;
	}
	goto tr35;
st73:
	if ( ++p == pe )
		goto _test_eof73;
case 73:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr324;
		case 32: goto tr324;
		case 40: goto tr325;
		case 41: goto tr326;
		case 2058: goto tr327;
		case 2107: goto tr328;
		case 2314: goto tr329;
		case 2363: goto tr329;
		case 2570: goto tr330;
		case 2619: goto tr331;
	}
	goto tr55;
st74:
	if ( ++p == pe )
		goto _test_eof74;
case 74:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr332;
		case 32: goto tr332;
		case 40: goto tr333;
		case 41: goto tr334;
		case 2058: goto tr335;
		case 2107: goto tr336;
		case 2314: goto tr337;
		case 2363: goto tr337;
		case 2570: goto tr338;
		case 2619: goto tr339;
	}
	goto tr55;
tr13:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st75;
tr47:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st75;
tr565:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st75;
st75:
	if ( ++p == pe )
		goto _test_eof75;
case 75:
#line 18735 "zscanner/scanner.c"
	switch( (*p) ) {
		case 73: goto st76;
		case 88: goto st80;
		case 105: goto st76;
		case 120: goto st80;
	}
	goto tr35;
st76:
	if ( ++p == pe )
		goto _test_eof76;
case 76:
	switch( (*p) ) {
		case 78: goto st77;
		case 110: goto st77;
	}
	goto tr35;
st77:
	if ( ++p == pe )
		goto _test_eof77;
case 77:
	switch( (*p) ) {
		case 70: goto st78;
		case 102: goto st78;
	}
	goto tr35;
st78:
	if ( ++p == pe )
		goto _test_eof78;
case 78:
	switch( (*p) ) {
		case 79: goto st79;
		case 111: goto st79;
	}
	goto tr35;
st79:
	if ( ++p == pe )
		goto _test_eof79;
case 79:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr345;
		case 32: goto tr345;
		case 40: goto tr346;
		case 41: goto tr347;
		case 2058: goto tr348;
		case 2107: goto tr349;
		case 2314: goto tr350;
		case 2363: goto tr350;
		case 2570: goto tr351;
		case 2619: goto tr352;
	}
	goto tr55;
st80:
	if ( ++p == pe )
		goto _test_eof80;
case 80:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr353;
		case 32: goto tr353;
		case 40: goto tr354;
		case 41: goto tr355;
		case 2058: goto tr356;
		case 2107: goto tr357;
		case 2314: goto tr358;
		case 2363: goto tr358;
		case 2570: goto tr359;
		case 2619: goto tr360;
	}
	goto tr55;
tr14:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st81;
tr48:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st81;
tr566:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st81;
st81:
	if ( ++p == pe )
		goto _test_eof81;
case 81:
#line 18870 "zscanner/scanner.c"
	switch( (*p) ) {
		case 65: goto st82;
		case 73: goto st86;
		case 83: goto st88;
		case 97: goto st82;
		case 105: goto st86;
		case 115: goto st88;
	}
	goto tr35;
st82:
	if ( ++p == pe )
		goto _test_eof82;
case 82:
	switch( (*p) ) {
		case 80: goto st83;
		case 112: goto st83;
	}
	goto tr35;
st83:
	if ( ++p == pe )
		goto _test_eof83;
case 83:
	switch( (*p) ) {
		case 84: goto st84;
		case 116: goto st84;
	}
	goto tr35;
st84:
	if ( ++p == pe )
		goto _test_eof84;
case 84:
	switch( (*p) ) {
		case 82: goto st85;
		case 114: goto st85;
	}
	goto tr35;
st85:
	if ( ++p == pe )
		goto _test_eof85;
case 85:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr367;
		case 32: goto tr367;
		case 40: goto tr368;
		case 41: goto tr369;
		case 2058: goto tr370;
		case 2107: goto tr371;
		case 2314: goto tr372;
		case 2363: goto tr372;
		case 2570: goto tr373;
		case 2619: goto tr374;
	}
	goto tr55;
st86:
	if ( ++p == pe )
		goto _test_eof86;
case 86:
	switch( (*p) ) {
		case 68: goto st87;
		case 100: goto st87;
	}
	goto tr35;
st87:
	if ( ++p == pe )
		goto _test_eof87;
case 87:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr376;
		case 32: goto tr376;
		case 40: goto tr377;
		case 41: goto tr378;
		case 2058: goto tr379;
		case 2107: goto tr380;
		case 2314: goto tr381;
		case 2363: goto tr381;
		case 2570: goto tr382;
		case 2619: goto tr383;
	}
	goto tr55;
st88:
	if ( ++p == pe )
		goto _test_eof88;
case 88:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr384;
		case 32: goto tr384;
		case 40: goto tr385;
		case 41: goto tr386;
		case 69: goto st89;
		case 101: goto st89;
		case 2058: goto tr388;
		case 2107: goto tr389;
		case 2314: goto tr390;
		case 2363: goto tr390;
		case 2570: goto tr391;
		case 2619: goto tr392;
	}
	goto tr55;
st89:
	if ( ++p == pe )
		goto _test_eof89;
case 89:
	switch( (*p) ) {
		case 67: goto st90;
		case 99: goto st90;
	}
	goto tr35;
st90:
	if ( ++p == pe )
		goto _test_eof90;
case 90:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr394;
		case 32: goto tr394;
		case 40: goto tr395;
		case 41: goto tr396;
		case 51: goto st91;
		case 2058: goto tr398;
		case 2107: goto tr399;
		case 2314: goto tr400;
		case 2363: goto tr400;
		case 2570: goto tr401;
		case 2619: goto tr402;
	}
	goto tr55;
st91:
	if ( ++p == pe )
		goto _test_eof91;
case 91:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr403;
		case 32: goto tr403;
		case 40: goto tr404;
		case 41: goto tr405;
		case 80: goto st92;
		case 112: goto st92;
		case 2058: goto tr407;
		case 2107: goto tr408;
		case 2314: goto tr409;
		case 2363: goto tr409;
		case 2570: goto tr410;
		case 2619: goto tr411;
	}
	goto tr55;
st92:
	if ( ++p == pe )
		goto _test_eof92;
case 92:
	switch( (*p) ) {
		case 65: goto st93;
		case 97: goto st93;
	}
	goto tr35;
st93:
	if ( ++p == pe )
		goto _test_eof93;
case 93:
	switch( (*p) ) {
		case 82: goto st94;
		case 114: goto st94;
	}
	goto tr35;
st94:
	if ( ++p == pe )
		goto _test_eof94;
case 94:
	switch( (*p) ) {
		case 65: goto st95;
		case 97: goto st95;
	}
	goto tr35;
st95:
	if ( ++p == pe )
		goto _test_eof95;
case 95:
	switch( (*p) ) {
		case 77: goto st96;
		case 109: goto st96;
	}
	goto tr35;
st96:
	if ( ++p == pe )
		goto _test_eof96;
case 96:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr416;
		case 32: goto tr416;
		case 40: goto tr417;
		case 41: goto tr418;
		case 2058: goto tr419;
		case 2107: goto tr420;
		case 2314: goto tr421;
		case 2363: goto tr421;
		case 2570: goto tr422;
		case 2619: goto tr423;
	}
	goto tr55;
tr15:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st97;
tr49:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st97;
tr567:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st97;
st97:
	if ( ++p == pe )
		goto _test_eof97;
case 97:
#line 19214 "zscanner/scanner.c"
	switch( (*p) ) {
		case 84: goto st98;
		case 116: goto st98;
	}
	goto tr35;
st98:
	if ( ++p == pe )
		goto _test_eof98;
case 98:
	switch( (*p) ) {
		case 82: goto st99;
		case 114: goto st99;
	}
	goto tr35;
st99:
	if ( ++p == pe )
		goto _test_eof99;
case 99:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr426;
		case 32: goto tr426;
		case 40: goto tr427;
		case 41: goto tr428;
		case 2058: goto tr429;
		case 2107: goto tr430;
		case 2314: goto tr431;
		case 2363: goto tr431;
		case 2570: goto tr432;
		case 2619: goto tr433;
	}
	goto tr55;
tr16:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st100;
tr50:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st100;
tr568:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st100;
st100:
	if ( ++p == pe )
		goto _test_eof100;
case 100:
#line 19292 "zscanner/scanner.c"
	switch( (*p) ) {
		case 80: goto st101;
		case 82: goto st102;
		case 84: goto st106;
		case 112: goto st101;
		case 114: goto st102;
		case 116: goto st106;
	}
	goto tr35;
st101:
	if ( ++p == pe )
		goto _test_eof101;
case 101:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr437;
		case 32: goto tr437;
		case 40: goto tr438;
		case 41: goto tr439;
		case 2058: goto tr440;
		case 2107: goto tr441;
		case 2314: goto tr442;
		case 2363: goto tr442;
		case 2570: goto tr443;
		case 2619: goto tr444;
	}
	goto tr55;
st102:
	if ( ++p == pe )
		goto _test_eof102;
case 102:
	switch( (*p) ) {
		case 83: goto st103;
		case 115: goto st103;
	}
	goto tr35;
st103:
	if ( ++p == pe )
		goto _test_eof103;
case 103:
	switch( (*p) ) {
		case 73: goto st104;
		case 105: goto st104;
	}
	goto tr35;
st104:
	if ( ++p == pe )
		goto _test_eof104;
case 104:
	switch( (*p) ) {
		case 71: goto st105;
		case 103: goto st105;
	}
	goto tr35;
st105:
	if ( ++p == pe )
		goto _test_eof105;
case 105:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr448;
		case 32: goto tr448;
		case 40: goto tr449;
		case 41: goto tr450;
		case 2058: goto tr451;
		case 2107: goto tr452;
		case 2314: goto tr453;
		case 2363: goto tr453;
		case 2570: goto tr454;
		case 2619: goto tr455;
	}
	goto tr55;
st106:
	if ( ++p == pe )
		goto _test_eof106;
case 106:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr456;
		case 32: goto tr456;
		case 40: goto tr457;
		case 41: goto tr458;
		case 2058: goto tr459;
		case 2107: goto tr460;
		case 2314: goto tr461;
		case 2363: goto tr461;
		case 2570: goto tr462;
		case 2619: goto tr463;
	}
	goto tr55;
tr17:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st107;
tr51:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st107;
tr569:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st107;
st107:
	if ( ++p == pe )
		goto _test_eof107;
case 107:
#line 19466 "zscanner/scanner.c"
	switch( (*p) ) {
		case 79: goto st108;
		case 80: goto st110;
		case 82: goto st112;
		case 83: goto st114;
		case 111: goto st108;
		case 112: goto st110;
		case 114: goto st112;
		case 115: goto st114;
	}
	goto tr35;
st108:
	if ( ++p == pe )
		goto _test_eof108;
case 108:
	switch( (*p) ) {
		case 65: goto st109;
		case 97: goto st109;
	}
	goto tr35;
st109:
	if ( ++p == pe )
		goto _test_eof109;
case 109:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr469;
		case 32: goto tr469;
		case 40: goto tr470;
		case 41: goto tr471;
		case 2058: goto tr472;
		case 2107: goto tr473;
		case 2314: goto tr474;
		case 2363: goto tr474;
		case 2570: goto tr475;
		case 2619: goto tr476;
	}
	goto tr55;
st110:
	if ( ++p == pe )
		goto _test_eof110;
case 110:
	switch( (*p) ) {
		case 70: goto st111;
		case 102: goto st111;
	}
	goto tr35;
st111:
	if ( ++p == pe )
		goto _test_eof111;
case 111:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr478;
		case 32: goto tr478;
		case 40: goto tr479;
		case 41: goto tr480;
		case 2058: goto tr481;
		case 2107: goto tr482;
		case 2314: goto tr483;
		case 2363: goto tr483;
		case 2570: goto tr484;
		case 2619: goto tr485;
	}
	goto tr55;
st112:
	if ( ++p == pe )
		goto _test_eof112;
case 112:
	switch( (*p) ) {
		case 86: goto st113;
		case 118: goto st113;
	}
	goto tr35;
st113:
	if ( ++p == pe )
		goto _test_eof113;
case 113:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr487;
		case 32: goto tr487;
		case 40: goto tr488;
		case 41: goto tr489;
		case 2058: goto tr490;
		case 2107: goto tr491;
		case 2314: goto tr492;
		case 2363: goto tr492;
		case 2570: goto tr493;
		case 2619: goto tr494;
	}
	goto tr55;
st114:
	if ( ++p == pe )
		goto _test_eof114;
case 114:
	switch( (*p) ) {
		case 72: goto st115;
		case 104: goto st115;
	}
	goto tr35;
st115:
	if ( ++p == pe )
		goto _test_eof115;
case 115:
	switch( (*p) ) {
		case 70: goto st116;
		case 102: goto st116;
	}
	goto tr35;
st116:
	if ( ++p == pe )
		goto _test_eof116;
case 116:
	switch( (*p) ) {
		case 80: goto st117;
		case 112: goto st117;
	}
	goto tr35;
st117:
	if ( ++p == pe )
		goto _test_eof117;
case 117:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr498;
		case 32: goto tr498;
		case 40: goto tr499;
		case 41: goto tr500;
		case 2058: goto tr501;
		case 2107: goto tr502;
		case 2314: goto tr503;
		case 2363: goto tr503;
		case 2570: goto tr504;
		case 2619: goto tr505;
	}
	goto tr55;
tr18:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st118;
tr52:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
	goto st118;
tr570:
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st118;
st118:
	if ( ++p == pe )
		goto _test_eof118;
case 118:
#line 19706 "zscanner/scanner.c"
	switch( (*p) ) {
		case 76: goto st119;
		case 88: goto st122;
		case 89: goto st124;
		case 108: goto st119;
		case 120: goto st122;
		case 121: goto st124;
	}
	goto tr35;
st119:
	if ( ++p == pe )
		goto _test_eof119;
case 119:
	switch( (*p) ) {
		case 83: goto st120;
		case 115: goto st120;
	}
	goto tr35;
st120:
	if ( ++p == pe )
		goto _test_eof120;
case 120:
	switch( (*p) ) {
		case 65: goto st121;
		case 97: goto st121;
	}
	goto tr35;
st121:
	if ( ++p == pe )
		goto _test_eof121;
case 121:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr511;
		case 32: goto tr511;
		case 40: goto tr512;
		case 41: goto tr513;
		case 2058: goto tr514;
		case 2107: goto tr515;
		case 2314: goto tr516;
		case 2363: goto tr516;
		case 2570: goto tr517;
		case 2619: goto tr518;
	}
	goto tr55;
st122:
	if ( ++p == pe )
		goto _test_eof122;
case 122:
	switch( (*p) ) {
		case 84: goto st123;
		case 116: goto st123;
	}
	goto tr35;
st123:
	if ( ++p == pe )
		goto _test_eof123;
case 123:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr520;
		case 32: goto tr520;
		case 40: goto tr521;
		case 41: goto tr522;
		case 2058: goto tr523;
		case 2107: goto tr524;
		case 2314: goto tr525;
		case 2363: goto tr525;
		case 2570: goto tr526;
		case 2619: goto tr527;
	}
	goto tr55;
st124:
	if ( ++p == pe )
		goto _test_eof124;
case 124:
	switch( (*p) ) {
		case 80: goto st125;
		case 112: goto st125;
	}
	goto tr35;
st125:
	if ( ++p == pe )
		goto _test_eof125;
case 125:
	switch( (*p) ) {
		case 69: goto st126;
		case 101: goto st126;
	}
	goto tr35;
st126:
	if ( ++p == pe )
		goto _test_eof126;
case 126:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr531;
	goto tr530;
tr531:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st127;
tr536:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st127;
st127:
	if ( ++p == pe )
		goto _test_eof127;
case 127:
#line 19884 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr533;
		case 32: goto tr533;
		case 40: goto tr534;
		case 41: goto tr535;
		case 2058: goto tr537;
		case 2107: goto tr538;
		case 2314: goto tr539;
		case 2363: goto tr539;
		case 2570: goto tr540;
		case 2619: goto tr541;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr536;
	goto tr532;
tr583:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st128;
tr248:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st128;
tr596:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st128;
st128:
	if ( ++p == pe )
		goto _test_eof128;
case 128:
#line 19961 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr266;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st128;
	goto tr35;
tr135:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1062;
tr129:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1062;
st1062:
	if ( ++p == pe )
		goto _test_eof1062;
case 1062:
#line 20038 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3173;
		case 32: goto tr3173;
		case 36: goto st139;
		case 40: goto tr3174;
		case 41: goto tr3175;
		case 42: goto tr3159;
		case 65: goto tr3176;
		case 67: goto tr3177;
		case 68: goto tr3178;
		case 69: goto tr3179;
		case 72: goto tr3180;
		case 73: goto tr3181;
		case 75: goto tr3182;
		case 76: goto tr3183;
		case 77: goto tr3184;
		case 78: goto tr3185;
		case 80: goto tr3186;
		case 82: goto tr3187;
		case 83: goto tr3188;
		case 84: goto tr3189;
		case 92: goto tr3159;
		case 95: goto tr3159;
		case 97: goto tr3176;
		case 99: goto tr3177;
		case 100: goto tr3178;
		case 101: goto tr3179;
		case 104: goto tr3180;
		case 105: goto tr3181;
		case 107: goto tr3182;
		case 108: goto tr3183;
		case 109: goto tr3184;
		case 110: goto tr3185;
		case 112: goto tr3186;
		case 114: goto tr3187;
		case 115: goto tr3188;
		case 116: goto tr3189;
		case 778: goto tr19;
		case 827: goto st138;
		case 1034: goto tr3190;
		case 1083: goto tr3191;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3159;
	} else if ( _widec > 90 ) {
		if ( 98 <= _widec && _widec <= 122 )
			goto tr3159;
	} else
		goto tr3159;
	goto tr3172;
tr543:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st129;
tr544:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st129;
tr3173:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st129;
tr3174:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st129;
tr3175:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st129;
st129:
	if ( ++p == pe )
		goto _test_eof129;
case 129:
#line 20171 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st129;
		case 32: goto st129;
		case 40: goto tr543;
		case 41: goto tr544;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr19;
		case 827: goto st138;
		case 1034: goto tr545;
		case 1083: goto st244;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr10:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
	goto st130;
st130:
	if ( ++p == pe )
		goto _test_eof130;
case 130:
#line 20241 "zscanner/scanner.c"
	switch( (*p) ) {
		case 78: goto st131;
		case 80: goto st56;
		case 110: goto st131;
		case 112: goto st56;
	}
	goto tr35;
st131:
	if ( ++p == pe )
		goto _test_eof131;
case 131:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr548;
		case 32: goto tr548;
		case 40: goto tr549;
		case 41: goto tr550;
		case 1034: goto tr551;
		case 1083: goto tr552;
	}
	goto tr35;
tr554:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st132;
tr555:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st132;
tr571:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st132;
tr548:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st132;
tr549:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st132;
tr550:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st132;
tr551:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st132;
st132:
	if ( ++p == pe )
		goto _test_eof132;
case 132:
#line 20350 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st132;
		case 32: goto st132;
		case 40: goto tr554;
		case 41: goto tr555;
		case 65: goto tr557;
		case 67: goto tr558;
		case 68: goto tr559;
		case 69: goto tr560;
		case 72: goto tr561;
		case 73: goto tr562;
		case 75: goto tr563;
		case 76: goto tr564;
		case 77: goto tr565;
		case 78: goto tr566;
		case 80: goto tr567;
		case 82: goto tr568;
		case 83: goto tr569;
		case 84: goto tr570;
		case 97: goto tr557;
		case 99: goto tr558;
		case 100: goto tr559;
		case 101: goto tr560;
		case 104: goto tr561;
		case 105: goto tr562;
		case 107: goto tr563;
		case 108: goto tr564;
		case 109: goto tr565;
		case 110: goto tr566;
		case 112: goto tr567;
		case 114: goto tr568;
		case 115: goto tr569;
		case 116: goto tr570;
		case 1034: goto tr571;
		case 1083: goto st137;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr556;
	goto tr530;
tr556:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st133;
tr576:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st133;
st133:
	if ( ++p == pe )
		goto _test_eof133;
case 133:
#line 20446 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr573;
		case 32: goto tr573;
		case 40: goto tr574;
		case 41: goto tr575;
		case 68: goto tr577;
		case 72: goto tr578;
		case 77: goto tr579;
		case 83: goto st134;
		case 87: goto tr581;
		case 100: goto tr577;
		case 104: goto tr578;
		case 109: goto tr579;
		case 115: goto st134;
		case 119: goto tr581;
		case 1034: goto tr582;
		case 1083: goto tr583;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr576;
	goto tr23;
tr577:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st134;
tr578:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st134;
tr579:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st134;
tr581:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st134;
st134:
	if ( ++p == pe )
		goto _test_eof134;
case 134:
#line 20526 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr573;
		case 32: goto tr573;
		case 40: goto tr574;
		case 41: goto tr575;
		case 1034: goto tr582;
		case 1083: goto tr583;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr584;
	goto tr23;
tr585:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st135;
tr584:
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st135;
tr594:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st135;
st135:
	if ( ++p == pe )
		goto _test_eof135;
case 135:
#line 20632 "zscanner/scanner.c"
	switch( (*p) ) {
		case 68: goto tr586;
		case 72: goto tr587;
		case 77: goto tr588;
		case 83: goto st136;
		case 87: goto tr590;
		case 100: goto tr586;
		case 104: goto tr587;
		case 109: goto tr588;
		case 115: goto st136;
		case 119: goto tr590;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr585;
	goto tr23;
tr586:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st136;
tr587:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st136;
tr588:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st136;
tr590:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st136;
st136:
	if ( ++p == pe )
		goto _test_eof136;
case 136:
#line 20692 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr591;
		case 32: goto tr591;
		case 40: goto tr592;
		case 41: goto tr593;
		case 1034: goto tr595;
		case 1083: goto tr596;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr594;
	goto tr23;
tr552:
#line 744 "./zscanner/scanner_body.rl"
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st137;
st137:
	if ( ++p == pe )
		goto _test_eof137;
case 137:
#line 20728 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr571;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st137;
	goto tr35;
tr3215:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
	goto st138;
st138:
	if ( ++p == pe )
		goto _test_eof138;
case 138:
#line 20765 "zscanner/scanner.c"
	if ( (*p) == 10 )
		goto tr19;
	goto st138;
tr545:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1063;
tr712:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1063;
tr707:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1063;
tr3190:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1063;
st1063:
	if ( ++p == pe )
		goto _test_eof1063;
case 1063:
#line 20842 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3173;
		case 32: goto tr3173;
		case 36: goto st139;
		case 40: goto tr3174;
		case 41: goto tr3175;
		case 42: goto tr3159;
		case 65: goto tr3194;
		case 67: goto tr3195;
		case 68: goto tr3196;
		case 69: goto tr3197;
		case 72: goto tr3198;
		case 73: goto tr3199;
		case 75: goto tr3200;
		case 76: goto tr3201;
		case 77: goto tr3202;
		case 78: goto tr3203;
		case 80: goto tr3204;
		case 82: goto tr3205;
		case 83: goto tr3206;
		case 84: goto tr3207;
		case 92: goto tr3159;
		case 95: goto tr3159;
		case 97: goto tr3194;
		case 99: goto tr3195;
		case 100: goto tr3196;
		case 101: goto tr3197;
		case 104: goto tr3198;
		case 105: goto tr3199;
		case 107: goto tr3200;
		case 108: goto tr3201;
		case 109: goto tr3202;
		case 110: goto tr3203;
		case 112: goto tr3204;
		case 114: goto tr3205;
		case 115: goto tr3206;
		case 116: goto tr3207;
		case 778: goto tr19;
		case 827: goto st138;
		case 1034: goto tr3190;
		case 1083: goto tr3191;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3159;
	} else if ( _widec > 57 ) {
		if ( _widec > 90 ) {
			if ( 98 <= _widec && _widec <= 122 )
				goto tr3159;
		} else if ( _widec >= 64 )
			goto tr3159;
	} else
		goto tr3193;
	goto tr3192;
tr3210:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
	goto st139;
st139:
	if ( ++p == pe )
		goto _test_eof139;
case 139:
#line 20921 "zscanner/scanner.c"
	switch( (*p) ) {
		case 73: goto tr598;
		case 79: goto tr599;
		case 84: goto tr600;
		case 105: goto tr598;
		case 111: goto tr599;
		case 116: goto tr600;
	}
	goto tr597;
tr598:
#line 717 "./zscanner/scanner_body.rl"
	{
		s->stop = true;
	}
	goto st140;
st140:
	if ( ++p == pe )
		goto _test_eof140;
case 140:
#line 20941 "zscanner/scanner.c"
	switch( (*p) ) {
		case 78: goto st141;
		case 110: goto st141;
	}
	goto tr597;
st141:
	if ( ++p == pe )
		goto _test_eof141;
case 141:
	switch( (*p) ) {
		case 67: goto st142;
		case 99: goto st142;
	}
	goto tr597;
st142:
	if ( ++p == pe )
		goto _test_eof142;
case 142:
	switch( (*p) ) {
		case 76: goto st143;
		case 108: goto st143;
	}
	goto tr597;
st143:
	if ( ++p == pe )
		goto _test_eof143;
case 143:
	switch( (*p) ) {
		case 85: goto st144;
		case 117: goto st144;
	}
	goto tr597;
st144:
	if ( ++p == pe )
		goto _test_eof144;
case 144:
	switch( (*p) ) {
		case 68: goto st145;
		case 100: goto st145;
	}
	goto tr597;
st145:
	if ( ++p == pe )
		goto _test_eof145;
case 145:
	switch( (*p) ) {
		case 69: goto st146;
		case 101: goto st146;
	}
	goto tr597;
st146:
	if ( ++p == pe )
		goto _test_eof146;
case 146:
	switch( (*p) ) {
		case 32: goto tr607;
		case 59: goto tr607;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr607;
	} else if ( (*p) >= 9 )
		goto tr607;
	goto tr597;
tr607:
#line 711 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 1064; goto st293;} }
	goto st1064;
tr693:
#line 607 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 1064; goto st280;} }
	goto st1064;
tr696:
#line 590 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 1064; goto st271;} }
	goto st1064;
st1064:
	if ( ++p == pe )
		goto _test_eof1064;
case 1064:
#line 21022 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3209;
		case 32: goto tr3209;
		case 36: goto tr3210;
		case 40: goto tr3211;
		case 41: goto tr3212;
		case 42: goto tr3213;
		case 92: goto tr3213;
		case 95: goto tr3213;
		case 778: goto tr3214;
		case 827: goto tr3215;
		case 1034: goto tr3216;
		case 1083: goto tr3217;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3213;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3213;
	} else
		goto tr3213;
	goto tr3208;
tr3159:
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 147; goto st249;} }
	goto st147;
tr3213:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 147; goto st249;} }
	goto st147;
st147:
	if ( ++p == pe )
		goto _test_eof147;
case 147:
#line 21086 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr608;
tr615:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st148;
tr616:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st148;
tr617:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st148;
tr609:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st148;
tr610:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st148;
tr611:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st148;
tr612:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st148;
st148:
	if ( ++p == pe )
		goto _test_eof148;
case 148:
#line 21184 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st148;
		case 32: goto st148;
		case 40: goto tr615;
		case 41: goto tr616;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 1034: goto tr617;
		case 1083: goto st149;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr530;
tr613:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st149;
st149:
	if ( ++p == pe )
		goto _test_eof149;
case 149:
#line 21248 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr617;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st149;
	goto st0;
tr21:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1065;
tr676:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1065;
tr701:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1065;
tr735:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1065;
tr3160:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1065;
tr3220:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1065;
tr3216:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1065;
tr3224:
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1065;
tr3247:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1065;
st1065:
	if ( ++p == pe )
		goto _test_eof1065;
case 1065:
#line 21436 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3155;
		case 32: goto tr3155;
		case 36: goto st139;
		case 40: goto tr3218;
		case 41: goto tr3219;
		case 42: goto tr3159;
		case 65: goto tr3194;
		case 67: goto tr3195;
		case 68: goto tr3196;
		case 69: goto tr3197;
		case 72: goto tr3198;
		case 73: goto tr3199;
		case 75: goto tr3200;
		case 76: goto tr3201;
		case 77: goto tr3202;
		case 78: goto tr3203;
		case 80: goto tr3204;
		case 82: goto tr3205;
		case 83: goto tr3206;
		case 84: goto tr3207;
		case 92: goto tr3159;
		case 95: goto tr3159;
		case 97: goto tr3194;
		case 99: goto tr3195;
		case 100: goto tr3196;
		case 101: goto tr3197;
		case 104: goto tr3198;
		case 105: goto tr3199;
		case 107: goto tr3200;
		case 108: goto tr3201;
		case 109: goto tr3202;
		case 110: goto tr3203;
		case 112: goto tr3204;
		case 114: goto tr3205;
		case 115: goto tr3206;
		case 116: goto tr3207;
		case 778: goto tr19;
		case 827: goto st138;
		case 1034: goto tr3220;
		case 1083: goto tr3161;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3159;
	} else if ( _widec > 57 ) {
		if ( _widec > 90 ) {
			if ( 98 <= _widec && _widec <= 122 )
				goto tr3159;
		} else if ( _widec >= 64 )
			goto tr3159;
	} else
		goto tr3193;
	goto tr3192;
tr3193:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 150; goto st249;} }
	goto st150;
st150:
	if ( ++p == pe )
		goto _test_eof150;
case 150:
#line 21537 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr621;
		case 32: goto tr621;
		case 40: goto tr622;
		case 41: goto tr623;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st153;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st153;
		case 119: goto tr32;
		case 1034: goto tr624;
		case 1083: goto tr625;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr620;
tr627:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st151;
tr628:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st151;
tr629:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st151;
tr621:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st151;
tr622:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st151;
tr623:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st151;
tr624:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st151;
st151:
	if ( ++p == pe )
		goto _test_eof151;
case 151:
#line 21683 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st151;
		case 32: goto st151;
		case 40: goto tr627;
		case 41: goto tr628;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 1034: goto tr629;
		case 1083: goto st152;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr530;
tr625:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st152;
st152:
	if ( ++p == pe )
		goto _test_eof152;
case 152:
#line 21756 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr629;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st152;
	goto tr35;
tr28:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st153;
tr29:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st153;
tr30:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st153;
tr32:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st153;
st153:
	if ( ++p == pe )
		goto _test_eof153;
case 153:
#line 21827 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr24;
		case 32: goto tr24;
		case 40: goto tr25;
		case 41: goto tr26;
		case 1034: goto tr33;
		case 1083: goto tr34;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr631;
	goto tr23;
tr632:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st154;
tr631:
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st154;
tr641:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st154;
st154:
	if ( ++p == pe )
		goto _test_eof154;
case 154:
#line 21933 "zscanner/scanner.c"
	switch( (*p) ) {
		case 68: goto tr633;
		case 72: goto tr634;
		case 77: goto tr635;
		case 83: goto st155;
		case 87: goto tr637;
		case 100: goto tr633;
		case 104: goto tr634;
		case 109: goto tr635;
		case 115: goto st155;
		case 119: goto tr637;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr632;
	goto tr23;
tr633:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st155;
tr634:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st155;
tr635:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st155;
tr637:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st155;
st155:
	if ( ++p == pe )
		goto _test_eof155;
case 155:
#line 21993 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr638;
		case 32: goto tr638;
		case 40: goto tr639;
		case 41: goto tr640;
		case 1034: goto tr642;
		case 1083: goto tr643;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr641;
	goto tr23;
tr34:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st156;
tr643:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st156;
st156:
	if ( ++p == pe )
		goto _test_eof156;
case 156:
#line 22054 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr53;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st156;
	goto tr35;
tr3176:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 157; goto st249;} }
	goto st157;
tr3194:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 157; goto st249;} }
	goto st157;
st157:
	if ( ++p == pe )
		goto _test_eof157;
case 157:
#line 22115 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr645;
		case 32: goto tr645;
		case 40: goto tr646;
		case 41: goto tr647;
		case 65: goto st214;
		case 70: goto st217;
		case 80: goto st221;
		case 97: goto st214;
		case 102: goto st217;
		case 112: goto st221;
		case 2058: goto tr648;
		case 2107: goto tr649;
		case 2314: goto tr64;
		case 2363: goto tr64;
		case 2570: goto tr650;
		case 2619: goto tr651;
	}
	goto tr644;
tr653:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st158;
tr654:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st158;
tr655:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st158;
tr645:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st158;
tr646:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st158;
tr647:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st158;
tr648:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st158;
st158:
	if ( ++p == pe )
		goto _test_eof158;
case 158:
#line 22261 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st158;
		case 32: goto st158;
		case 40: goto tr653;
		case 41: goto tr654;
		case 58: goto tr67;
		case 65: goto tr105;
		case 67: goto tr106;
		case 68: goto tr107;
		case 69: goto tr108;
		case 72: goto tr109;
		case 73: goto tr110;
		case 75: goto tr111;
		case 76: goto tr112;
		case 77: goto tr113;
		case 78: goto tr114;
		case 80: goto tr115;
		case 82: goto tr116;
		case 83: goto tr117;
		case 84: goto tr118;
		case 92: goto st9;
		case 97: goto tr105;
		case 99: goto tr106;
		case 100: goto tr107;
		case 101: goto tr108;
		case 104: goto tr109;
		case 105: goto tr110;
		case 107: goto tr111;
		case 108: goto tr112;
		case 109: goto tr113;
		case 110: goto tr114;
		case 112: goto tr115;
		case 114: goto tr116;
		case 115: goto tr117;
		case 116: goto tr118;
		case 2058: goto tr655;
		case 2107: goto st228;
		case 2314: goto tr75;
		case 2363: goto tr75;
		case 2570: goto tr657;
		case 2619: goto tr658;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr104;
	} else
		goto tr67;
	goto tr532;
tr105:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 159; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 159; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 159; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 159; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 159; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 159; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 159; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 159; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 159; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 159; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 159; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 159; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 159; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 159; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 159; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 159; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 159; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 159; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 159; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 159; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 159; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 159; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 159; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 159; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 159; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 159; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 159; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st159;
st159:
	if ( ++p == pe )
		goto _test_eof159;
case 159:
#line 22424 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr660;
		case 32: goto tr660;
		case 40: goto tr661;
		case 41: goto tr662;
		case 65: goto st214;
		case 70: goto st217;
		case 80: goto st221;
		case 97: goto st214;
		case 102: goto st217;
		case 112: goto st221;
		case 1802: goto tr82;
		case 1851: goto tr83;
		case 2058: goto tr663;
		case 2107: goto tr664;
		case 2314: goto tr665;
		case 2363: goto tr666;
		case 2570: goto tr667;
		case 2619: goto tr668;
	}
	goto tr659;
tr670:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st160;
tr671:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st160;
tr660:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st160;
tr661:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st160;
tr662:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st160;
st160:
	if ( ++p == pe )
		goto _test_eof160;
case 160:
#line 22551 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st160;
		case 32: goto st160;
		case 40: goto tr670;
		case 41: goto tr671;
		case 92: goto st9;
		case 1802: goto tr88;
		case 1851: goto st8;
		case 2058: goto tr99;
		case 2107: goto st13;
		case 2314: goto tr96;
		case 2363: goto tr97;
		case 2570: goto tr672;
		case 2619: goto tr98;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr67;
	} else
		goto tr67;
	goto tr78;
tr96:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1066; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1066; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1066; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1066; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1066; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1066; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1066; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1066; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1066; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1066; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1066; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1066; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1066; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1066; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1066; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1066; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1066; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1066; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1066; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1066; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1066; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1066; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1066; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1066; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1066; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1066; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1066; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1066;
tr121:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1066; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1066; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1066; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1066; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1066; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1066; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1066; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1066; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1066; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1066; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1066; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1066; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1066; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1066; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1066; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1066; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1066; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1066; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1066; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1066; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1066; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1066; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1066; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1066; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1066; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1066; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1066; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1066;
tr665:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1066; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1066; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1066; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1066; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1066; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1066; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1066; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1066; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1066; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1066; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1066; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1066; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1066; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1066; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1066; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1066; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1066; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1066; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1066; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1066; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1066; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1066; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1066; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1066; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1066; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1066; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1066; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1066;
st1066:
	if ( ++p == pe )
		goto _test_eof1066;
case 1066:
#line 22873 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3221;
		case 32: goto tr3221;
		case 36: goto st139;
		case 40: goto tr3222;
		case 41: goto tr3223;
		case 42: goto tr3159;
		case 92: goto tr3159;
		case 95: goto tr3159;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr3224;
		case 1083: goto tr3225;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3159;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3159;
	} else
		goto tr3159;
	goto tr697;
tr674:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st161;
tr675:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st161;
tr732:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st161;
tr698:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st161;
tr699:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st161;
tr700:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st161;
tr733:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st161;
tr734:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st161;
tr3221:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st161;
tr3222:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st161;
tr3223:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st161;
st161:
	if ( ++p == pe )
		goto _test_eof161;
case 161:
#line 23064 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st161;
		case 32: goto st161;
		case 40: goto tr674;
		case 41: goto tr675;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr676;
		case 1083: goto st162;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr736:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st162;
tr702:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st162;
tr3225:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st162;
st162:
	if ( ++p == pe )
		goto _test_eof162;
case 162:
#line 23151 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr88;
		case 1034: goto tr676;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st162;
	} else if ( _widec >= 768 )
		goto st8;
	goto tr84;
tr97:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 163; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 163; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 163; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 163; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 163; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 163; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 163; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 163; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 163; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 163; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 163; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 163; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 163; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 163; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 163; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 163; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 163; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 163; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 163; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 163; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 163; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 163; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 163; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 163; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 163; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 163; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 163; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st163;
tr666:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 163; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 163; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 163; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 163; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 163; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 163; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 163; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 163; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 163; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 163; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 163; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 163; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 163; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 163; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 163; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 163; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 163; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 163; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 163; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 163; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 163; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 163; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 163; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 163; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 163; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 163; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 163; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st163;
st163:
	if ( ++p == pe )
		goto _test_eof163;
case 163:
#line 23352 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st164;
		case 32: goto st164;
		case 40: goto tr679;
		case 41: goto tr680;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr88;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st8;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st8;
	} else
		goto st8;
	goto tr78;
tr679:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st164;
tr680:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st164;
st164:
	if ( ++p == pe )
		goto _test_eof164;
case 164:
#line 23410 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st164;
		case 32: goto st164;
		case 40: goto tr679;
		case 41: goto tr680;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr88;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st8;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st8;
	} else
		goto st8;
	goto tr84;
tr667:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1067; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1067; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1067; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1067; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1067; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1067; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1067; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1067; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1067; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1067; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1067; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1067; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1067; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1067; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1067; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1067; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1067; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1067; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1067; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1067; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1067; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1067; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1067; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1067; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1067; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1067; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1067; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1067;
tr672:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1067; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1067; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1067; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1067; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1067; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1067; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1067; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1067; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1067; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1067; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1067; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1067; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1067; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1067; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1067; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1067; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1067; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1067; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1067; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1067; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1067; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1067; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1067; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1067; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1067; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1067; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1067; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1067;
st1067:
	if ( ++p == pe )
		goto _test_eof1067;
case 1067:
#line 23641 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3226;
		case 32: goto tr3226;
		case 36: goto tr3163;
		case 40: goto tr3227;
		case 41: goto tr3228;
		case 42: goto tr3166;
		case 58: goto tr67;
		case 92: goto tr3167;
		case 95: goto tr3166;
		case 1802: goto tr88;
		case 1851: goto st8;
		case 2058: goto tr3224;
		case 2107: goto tr3229;
		case 2314: goto tr96;
		case 2363: goto tr97;
		case 2570: goto tr3224;
		case 2619: goto tr3230;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3166;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3166;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 97 )
				goto tr3166;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr697;
tr682:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st165;
tr683:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st165;
tr657:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 165; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 165; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 165; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 165; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 165; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 165; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 165; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 165; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 165; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 165; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 165; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 165; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 165; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 165; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 165; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 165; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 165; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 165; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 165; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 165; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 165; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 165; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 165; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 165; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 165; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 165; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 165; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st165;
tr650:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 165; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 165; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 165; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 165; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 165; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 165; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 165; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 165; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 165; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 165; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 165; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 165; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 165; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 165; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 165; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 165; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 165; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 165; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 165; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 165; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 165; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 165; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 165; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 165; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 165; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 165; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 165; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st165;
tr3226:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st165;
tr3227:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st165;
tr3228:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st165;
st165:
	if ( ++p == pe )
		goto _test_eof165;
case 165:
#line 23944 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st165;
		case 32: goto st165;
		case 40: goto tr682;
		case 41: goto tr683;
		case 58: goto tr67;
		case 65: goto tr105;
		case 67: goto tr106;
		case 68: goto tr107;
		case 69: goto tr108;
		case 72: goto tr109;
		case 73: goto tr110;
		case 75: goto tr111;
		case 76: goto tr112;
		case 77: goto tr113;
		case 78: goto tr114;
		case 80: goto tr115;
		case 82: goto tr116;
		case 83: goto tr117;
		case 84: goto tr118;
		case 92: goto st9;
		case 97: goto tr105;
		case 99: goto tr106;
		case 100: goto tr107;
		case 101: goto tr108;
		case 104: goto tr109;
		case 105: goto tr110;
		case 107: goto tr111;
		case 108: goto tr112;
		case 109: goto tr113;
		case 110: goto tr114;
		case 112: goto tr115;
		case 114: goto tr116;
		case 115: goto tr117;
		case 116: goto tr118;
		case 1802: goto tr88;
		case 1851: goto st8;
		case 2058: goto tr676;
		case 2107: goto st179;
		case 2314: goto tr96;
		case 2363: goto tr97;
		case 2570: goto tr676;
		case 2619: goto tr685;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr104;
	} else
		goto tr67;
	goto tr101;
tr106:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 166; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 166; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 166; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 166; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 166; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 166; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 166; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 166; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 166; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 166; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 166; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 166; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 166; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 166; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 166; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 166; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 166; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 166; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 166; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 166; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 166; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 166; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 166; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 166; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 166; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 166; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 166; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st166;
st166:
	if ( ++p == pe )
		goto _test_eof166;
case 166:
#line 24109 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 69: goto st18;
		case 78: goto st22;
		case 101: goto st18;
		case 110: goto st22;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr107:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 167; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 167; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 167; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 167; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 167; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 167; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 167; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 167; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 167; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 167; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 167; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 167; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 167; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 167; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 167; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 167; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 167; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 167; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 167; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 167; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 167; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 167; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 167; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 167; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 167; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 167; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 167; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st167;
st167:
	if ( ++p == pe )
		goto _test_eof167;
case 167:
#line 24227 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 72: goto st27;
		case 78: goto st31;
		case 83: goto st39;
		case 104: goto st27;
		case 110: goto st31;
		case 115: goto st39;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr108:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 168; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 168; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 168; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 168; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 168; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 168; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 168; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 168; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 168; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 168; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 168; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 168; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 168; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 168; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 168; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 168; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 168; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 168; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 168; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 168; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 168; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 168; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 168; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 168; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 168; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 168; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 168; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st168;
st168:
	if ( ++p == pe )
		goto _test_eof168;
case 168:
#line 24347 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 85: goto st41;
		case 117: goto st41;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr109:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 169; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 169; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 169; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 169; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 169; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 169; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 169; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 169; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 169; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 169; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 169; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 169; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 169; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 169; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 169; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 169; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 169; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 169; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 169; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 169; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 169; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 169; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 169; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 169; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 169; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 169; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 169; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st169;
st169:
	if ( ++p == pe )
		goto _test_eof169;
case 169:
#line 24463 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 73: goto st48;
		case 105: goto st48;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr110:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 170; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 170; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 170; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 170; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 170; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 170; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 170; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 170; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 170; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 170; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 170; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 170; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 170; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 170; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 170; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 170; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 170; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 170; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 170; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 170; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 170; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 170; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 170; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 170; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 170; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 170; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 170; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st170;
st170:
	if ( ++p == pe )
		goto _test_eof170;
case 170:
#line 24579 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 78: goto st131;
		case 80: goto st56;
		case 110: goto st131;
		case 112: goto st56;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr111:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 171; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 171; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 171; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 171; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 171; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 171; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 171; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 171; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 171; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 171; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 171; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 171; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 171; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 171; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 171; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 171; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 171; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 171; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 171; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 171; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 171; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 171; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 171; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 171; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 171; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 171; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 171; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st171;
st171:
	if ( ++p == pe )
		goto _test_eof171;
case 171:
#line 24697 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 69: goto st64;
		case 88: goto st66;
		case 101: goto st64;
		case 120: goto st66;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr112:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 172; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 172; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 172; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 172; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 172; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 172; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 172; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 172; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 172; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 172; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 172; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 172; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 172; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 172; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 172; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 172; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 172; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 172; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 172; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 172; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 172; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 172; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 172; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 172; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 172; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 172; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 172; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st172;
st172:
	if ( ++p == pe )
		goto _test_eof172;
case 172:
#line 24815 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 51: goto st68;
		case 54: goto st70;
		case 79: goto st72;
		case 80: goto st74;
		case 111: goto st72;
		case 112: goto st74;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr113:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 173; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 173; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 173; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 173; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 173; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 173; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 173; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 173; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 173; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 173; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 173; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 173; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 173; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 173; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 173; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 173; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 173; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 173; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 173; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 173; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 173; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 173; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 173; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 173; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 173; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 173; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 173; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st173;
st173:
	if ( ++p == pe )
		goto _test_eof173;
case 173:
#line 24935 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 73: goto st76;
		case 88: goto st80;
		case 105: goto st76;
		case 120: goto st80;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr114:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 174; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 174; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 174; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 174; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 174; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 174; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 174; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 174; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 174; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 174; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 174; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 174; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 174; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 174; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 174; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 174; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 174; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 174; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 174; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 174; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 174; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 174; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 174; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 174; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 174; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 174; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 174; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st174;
st174:
	if ( ++p == pe )
		goto _test_eof174;
case 174:
#line 25053 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 65: goto st82;
		case 73: goto st86;
		case 83: goto st88;
		case 97: goto st82;
		case 105: goto st86;
		case 115: goto st88;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr115:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 175; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 175; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 175; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 175; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 175; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 175; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 175; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 175; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 175; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 175; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 175; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 175; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 175; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 175; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 175; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 175; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 175; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 175; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 175; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 175; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 175; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 175; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 175; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 175; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 175; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 175; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 175; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st175;
st175:
	if ( ++p == pe )
		goto _test_eof175;
case 175:
#line 25173 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 84: goto st98;
		case 116: goto st98;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr116:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 176; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 176; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 176; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 176; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 176; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 176; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 176; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 176; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 176; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 176; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 176; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 176; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 176; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 176; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 176; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 176; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 176; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 176; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 176; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 176; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 176; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 176; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 176; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 176; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 176; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 176; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 176; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st176;
st176:
	if ( ++p == pe )
		goto _test_eof176;
case 176:
#line 25289 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 80: goto st101;
		case 82: goto st102;
		case 84: goto st106;
		case 112: goto st101;
		case 114: goto st102;
		case 116: goto st106;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr117:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 177; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 177; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 177; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 177; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 177; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 177; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 177; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 177; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 177; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 177; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 177; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 177; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 177; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 177; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 177; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 177; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 177; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 177; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 177; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 177; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 177; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 177; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 177; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 177; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 177; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 177; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 177; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st177;
st177:
	if ( ++p == pe )
		goto _test_eof177;
case 177:
#line 25409 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 79: goto st108;
		case 80: goto st110;
		case 82: goto st112;
		case 83: goto st114;
		case 111: goto st108;
		case 112: goto st110;
		case 114: goto st112;
		case 115: goto st114;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr118:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 178; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 178; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 178; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 178; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 178; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 178; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 178; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 178; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 178; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 178; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 178; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 178; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 178; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 178; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 178; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 178; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 178; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 178; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 178; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 178; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 178; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 178; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 178; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 178; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 178; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 178; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 178; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st178;
st178:
	if ( ++p == pe )
		goto _test_eof178;
case 178:
#line 25531 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 76: goto st119;
		case 88: goto st122;
		case 89: goto st124;
		case 108: goto st119;
		case 120: goto st122;
		case 121: goto st124;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr659;
tr719:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st179;
tr3229:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st179;
st179:
	if ( ++p == pe )
		goto _test_eof179;
case 179:
#line 25595 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr88;
		case 1034: goto tr686;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st179;
	} else if ( _widec >= 768 )
		goto st8;
	goto tr78;
tr119:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1068;
tr686:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1068;
tr718:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1068;
tr3168:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1068;
st1068:
	if ( ++p == pe )
		goto _test_eof1068;
case 1068:
#line 25699 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3162;
		case 32: goto tr3162;
		case 36: goto tr3163;
		case 40: goto tr3164;
		case 41: goto tr3165;
		case 42: goto tr3166;
		case 58: goto tr67;
		case 65: goto tr3233;
		case 67: goto tr3234;
		case 68: goto tr3235;
		case 69: goto tr3236;
		case 72: goto tr3237;
		case 73: goto tr3238;
		case 75: goto tr3239;
		case 76: goto tr3240;
		case 77: goto tr3241;
		case 78: goto tr3242;
		case 80: goto tr3243;
		case 82: goto tr3244;
		case 83: goto tr3245;
		case 84: goto tr3246;
		case 92: goto tr3167;
		case 95: goto tr3166;
		case 97: goto tr3233;
		case 99: goto tr3234;
		case 100: goto tr3235;
		case 101: goto tr3236;
		case 104: goto tr3237;
		case 105: goto tr3238;
		case 107: goto tr3239;
		case 108: goto tr3240;
		case 109: goto tr3241;
		case 110: goto tr3242;
		case 112: goto tr3243;
		case 114: goto tr3244;
		case 115: goto tr3245;
		case 116: goto tr3246;
		case 1802: goto tr19;
		case 1851: goto st138;
		case 2058: goto tr3168;
		case 2107: goto tr3169;
		case 2314: goto tr121;
		case 2363: goto tr122;
		case 2570: goto tr3170;
		case 2619: goto tr3171;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3232;
			} else if ( _widec >= 45 )
				goto tr3166;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3166;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 98 )
				goto tr3166;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr3231;
tr3163:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 180; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 180; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 180; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 180; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 180; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 180; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 180; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 180; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 180; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 180; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 180; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 180; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 180; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 180; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 180; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 180; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 180; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 180; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 180; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 180; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 180; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 180; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 180; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 180; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 180; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 180; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 180; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st180;
st180:
	if ( ++p == pe )
		goto _test_eof180;
case 180:
#line 25874 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr79;
		case 32: goto tr79;
		case 40: goto tr80;
		case 41: goto tr81;
		case 73: goto tr598;
		case 79: goto tr599;
		case 84: goto tr600;
		case 105: goto tr598;
		case 111: goto tr599;
		case 116: goto tr600;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr82;
		case 1083: goto tr83;
	}
	goto tr687;
tr599:
#line 717 "./zscanner/scanner_body.rl"
	{
		s->stop = true;
	}
	goto st181;
st181:
	if ( ++p == pe )
		goto _test_eof181;
case 181:
#line 25916 "zscanner/scanner.c"
	switch( (*p) ) {
		case 82: goto st182;
		case 114: goto st182;
	}
	goto tr597;
st182:
	if ( ++p == pe )
		goto _test_eof182;
case 182:
	switch( (*p) ) {
		case 73: goto st183;
		case 105: goto st183;
	}
	goto tr597;
st183:
	if ( ++p == pe )
		goto _test_eof183;
case 183:
	switch( (*p) ) {
		case 71: goto st184;
		case 103: goto st184;
	}
	goto tr597;
st184:
	if ( ++p == pe )
		goto _test_eof184;
case 184:
	switch( (*p) ) {
		case 73: goto st185;
		case 105: goto st185;
	}
	goto tr597;
st185:
	if ( ++p == pe )
		goto _test_eof185;
case 185:
	switch( (*p) ) {
		case 78: goto st186;
		case 110: goto st186;
	}
	goto tr597;
st186:
	if ( ++p == pe )
		goto _test_eof186;
case 186:
	switch( (*p) ) {
		case 32: goto tr693;
		case 59: goto tr693;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr693;
	} else if ( (*p) >= 9 )
		goto tr693;
	goto tr597;
tr600:
#line 717 "./zscanner/scanner_body.rl"
	{
		s->stop = true;
	}
	goto st187;
st187:
	if ( ++p == pe )
		goto _test_eof187;
case 187:
#line 25982 "zscanner/scanner.c"
	switch( (*p) ) {
		case 84: goto st188;
		case 116: goto st188;
	}
	goto tr597;
st188:
	if ( ++p == pe )
		goto _test_eof188;
case 188:
	switch( (*p) ) {
		case 76: goto st189;
		case 108: goto st189;
	}
	goto tr597;
st189:
	if ( ++p == pe )
		goto _test_eof189;
case 189:
	switch( (*p) ) {
		case 32: goto tr696;
		case 59: goto tr696;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr696;
	} else if ( (*p) >= 9 )
		goto tr696;
	goto tr597;
tr3166:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 190; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 190; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 190; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 190; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 190; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 190; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 190; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 190; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 190; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 190; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 190; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 190; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 190; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 190; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 190; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 190; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 190; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 190; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 190; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 190; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 190; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 190; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 190; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 190; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 190; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 190; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 190; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 190; goto st249;} }
	goto st190;
st190:
	if ( ++p == pe )
		goto _test_eof190;
case 190:
#line 26098 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr697;
tr3232:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 191; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 191; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 191; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 191; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 191; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 191; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 191; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 191; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 191; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 191; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 191; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 191; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 191; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 191; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 191; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 191; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 191; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 191; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 191; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 191; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 191; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 191; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 191; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 191; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 191; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 191; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 191; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 191; goto st249;} }
	goto st191;
st191:
	if ( ++p == pe )
		goto _test_eof191;
case 191:
#line 26230 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr704;
		case 32: goto tr704;
		case 40: goto tr705;
		case 41: goto tr706;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st153;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st153;
		case 119: goto tr32;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr707;
		case 1083: goto tr708;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr703;
tr710:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st192;
tr711:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st192;
tr704:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st192;
tr705:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st192;
tr706:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st192;
st192:
	if ( ++p == pe )
		goto _test_eof192;
case 192:
#line 26368 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st192;
		case 32: goto st192;
		case 40: goto tr710;
		case 41: goto tr711;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr712;
		case 1083: goto st193;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr708:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st193;
st193:
	if ( ++p == pe )
		goto _test_eof193;
case 193:
#line 26448 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr88;
		case 1034: goto tr712;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st193;
	} else if ( _widec >= 768 )
		goto st8;
	goto tr131;
tr3233:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 194; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 194; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 194; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 194; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 194; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 194; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 194; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 194; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 194; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 194; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 194; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 194; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 194; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 194; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 194; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 194; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 194; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 194; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 194; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 194; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 194; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 194; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 194; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 194; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 194; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 194; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 194; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 194; goto st249;} }
	goto st194;
st194:
	if ( ++p == pe )
		goto _test_eof194;
case 194:
#line 26575 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 65: goto st214;
		case 70: goto st217;
		case 80: goto st221;
		case 97: goto st214;
		case 102: goto st217;
		case 112: goto st221;
		case 1802: goto tr82;
		case 1851: goto tr83;
		case 2058: goto tr718;
		case 2107: goto tr719;
		case 2314: goto tr665;
		case 2363: goto tr666;
		case 2570: goto tr720;
		case 2619: goto tr721;
	}
	goto tr714;
tr723:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st195;
tr724:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st195;
tr715:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st195;
tr716:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st195;
tr717:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st195;
st195:
	if ( ++p == pe )
		goto _test_eof195;
case 195:
#line 26714 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st195;
		case 32: goto st195;
		case 40: goto tr723;
		case 41: goto tr724;
		case 58: goto tr67;
		case 65: goto tr105;
		case 67: goto tr106;
		case 68: goto tr107;
		case 69: goto tr108;
		case 72: goto tr109;
		case 73: goto tr110;
		case 75: goto tr111;
		case 76: goto tr112;
		case 77: goto tr113;
		case 78: goto tr114;
		case 80: goto tr115;
		case 82: goto tr116;
		case 83: goto tr117;
		case 84: goto tr118;
		case 92: goto st9;
		case 97: goto tr105;
		case 99: goto tr106;
		case 100: goto tr107;
		case 101: goto tr108;
		case 104: goto tr109;
		case 105: goto tr110;
		case 107: goto tr111;
		case 108: goto tr112;
		case 109: goto tr113;
		case 110: goto tr114;
		case 112: goto tr115;
		case 114: goto tr116;
		case 115: goto tr117;
		case 116: goto tr118;
		case 1802: goto tr88;
		case 1851: goto st8;
		case 2058: goto tr686;
		case 2107: goto st179;
		case 2314: goto tr96;
		case 2363: goto tr97;
		case 2570: goto tr725;
		case 2619: goto tr685;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr104;
	} else
		goto tr67;
	goto tr101;
tr123:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1069; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1069; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1069; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1069; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1069; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1069; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1069; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1069; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1069; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1069; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1069; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1069; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1069; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1069; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1069; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1069; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1069; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1069; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1069; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1069; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1069; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1069; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1069; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1069; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1069; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1069; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1069; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st1069;
tr725:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1069; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1069; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1069; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1069; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1069; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1069; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1069; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1069; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1069; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1069; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1069; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1069; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1069; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1069; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1069; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1069; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1069; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1069; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1069; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1069; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1069; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1069; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1069; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1069; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1069; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1069; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1069; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1069;
tr720:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1069; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1069; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1069; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1069; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1069; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1069; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1069; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1069; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1069; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1069; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1069; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1069; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1069; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1069; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1069; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1069; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1069; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1069; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1069; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1069; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1069; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1069; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1069; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1069; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1069; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1069; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1069; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 1929 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);
	}
	goto st1069;
tr3170:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1069; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1069; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1069; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1069; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1069; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1069; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1069; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1069; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1069; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1069; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1069; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1069; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1069; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 1069; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1069; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1069; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1069; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1069; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 1069; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1069; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1069; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1069; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1069; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1069; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1069; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1069; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1069; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st1069;
st1069:
	if ( ++p == pe )
		goto _test_eof1069;
case 1069:
#line 27159 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3226;
		case 32: goto tr3226;
		case 36: goto tr3163;
		case 40: goto tr3227;
		case 41: goto tr3228;
		case 42: goto tr3166;
		case 58: goto tr67;
		case 65: goto tr3233;
		case 67: goto tr3234;
		case 68: goto tr3235;
		case 69: goto tr3236;
		case 72: goto tr3237;
		case 73: goto tr3238;
		case 75: goto tr3239;
		case 76: goto tr3240;
		case 77: goto tr3241;
		case 78: goto tr3242;
		case 80: goto tr3243;
		case 82: goto tr3244;
		case 83: goto tr3245;
		case 84: goto tr3246;
		case 92: goto tr3167;
		case 95: goto tr3166;
		case 97: goto tr3233;
		case 99: goto tr3234;
		case 100: goto tr3235;
		case 101: goto tr3236;
		case 104: goto tr3237;
		case 105: goto tr3238;
		case 107: goto tr3239;
		case 108: goto tr3240;
		case 109: goto tr3241;
		case 110: goto tr3242;
		case 112: goto tr3243;
		case 114: goto tr3244;
		case 115: goto tr3245;
		case 116: goto tr3246;
		case 1802: goto tr88;
		case 1851: goto st8;
		case 2058: goto tr3247;
		case 2107: goto tr3229;
		case 2314: goto tr96;
		case 2363: goto tr97;
		case 2570: goto tr3247;
		case 2619: goto tr3230;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3232;
			} else if ( _widec >= 45 )
				goto tr3166;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3166;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 98 )
				goto tr3166;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr3231;
tr3234:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 196; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 196; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 196; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 196; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 196; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 196; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 196; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 196; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 196; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 196; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 196; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 196; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 196; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 196; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 196; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 196; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 196; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 196; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 196; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 196; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 196; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 196; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 196; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 196; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 196; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 196; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 196; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 196; goto st249;} }
	goto st196;
st196:
	if ( ++p == pe )
		goto _test_eof196;
case 196:
#line 27349 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 69: goto st18;
		case 78: goto st22;
		case 101: goto st18;
		case 110: goto st22;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3235:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 197; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 197; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 197; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 197; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 197; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 197; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 197; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 197; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 197; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 197; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 197; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 197; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 197; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 197; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 197; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 197; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 197; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 197; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 197; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 197; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 197; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 197; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 197; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 197; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 197; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 197; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 197; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 197; goto st249;} }
	goto st197;
st197:
	if ( ++p == pe )
		goto _test_eof197;
case 197:
#line 27474 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 72: goto st27;
		case 78: goto st31;
		case 83: goto st39;
		case 104: goto st27;
		case 110: goto st31;
		case 115: goto st39;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3236:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 198; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 198; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 198; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 198; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 198; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 198; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 198; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 198; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 198; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 198; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 198; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 198; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 198; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 198; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 198; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 198; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 198; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 198; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 198; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 198; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 198; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 198; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 198; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 198; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 198; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 198; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 198; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 198; goto st249;} }
	goto st198;
st198:
	if ( ++p == pe )
		goto _test_eof198;
case 198:
#line 27601 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 85: goto st41;
		case 117: goto st41;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3237:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 199; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 199; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 199; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 199; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 199; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 199; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 199; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 199; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 199; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 199; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 199; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 199; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 199; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 199; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 199; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 199; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 199; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 199; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 199; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 199; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 199; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 199; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 199; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 199; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 199; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 199; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 199; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 199; goto st249;} }
	goto st199;
st199:
	if ( ++p == pe )
		goto _test_eof199;
case 199:
#line 27724 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 73: goto st48;
		case 105: goto st48;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3238:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 200; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 200; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 200; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 200; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 200; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 200; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 200; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 200; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 200; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 200; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 200; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 200; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 200; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 200; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 200; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 200; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 200; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 200; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 200; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 200; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 200; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 200; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 200; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 200; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 200; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 200; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 200; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 200; goto st249;} }
	goto st200;
st200:
	if ( ++p == pe )
		goto _test_eof200;
case 200:
#line 27847 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 78: goto st131;
		case 80: goto st56;
		case 110: goto st131;
		case 112: goto st56;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3239:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 201; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 201; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 201; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 201; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 201; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 201; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 201; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 201; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 201; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 201; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 201; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 201; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 201; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 201; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 201; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 201; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 201; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 201; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 201; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 201; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 201; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 201; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 201; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 201; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 201; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 201; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 201; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 201; goto st249;} }
	goto st201;
st201:
	if ( ++p == pe )
		goto _test_eof201;
case 201:
#line 27972 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 69: goto st64;
		case 88: goto st66;
		case 101: goto st64;
		case 120: goto st66;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3240:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 202; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 202; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 202; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 202; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 202; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 202; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 202; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 202; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 202; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 202; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 202; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 202; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 202; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 202; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 202; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 202; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 202; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 202; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 202; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 202; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 202; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 202; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 202; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 202; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 202; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 202; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 202; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 202; goto st249;} }
	goto st202;
st202:
	if ( ++p == pe )
		goto _test_eof202;
case 202:
#line 28097 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 51: goto st68;
		case 54: goto st70;
		case 79: goto st72;
		case 80: goto st74;
		case 111: goto st72;
		case 112: goto st74;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3241:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 203; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 203; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 203; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 203; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 203; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 203; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 203; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 203; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 203; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 203; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 203; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 203; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 203; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 203; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 203; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 203; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 203; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 203; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 203; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 203; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 203; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 203; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 203; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 203; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 203; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 203; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 203; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 203; goto st249;} }
	goto st203;
st203:
	if ( ++p == pe )
		goto _test_eof203;
case 203:
#line 28224 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 73: goto st76;
		case 88: goto st80;
		case 105: goto st76;
		case 120: goto st80;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3242:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 204; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 204; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 204; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 204; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 204; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 204; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 204; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 204; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 204; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 204; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 204; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 204; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 204; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 204; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 204; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 204; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 204; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 204; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 204; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 204; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 204; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 204; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 204; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 204; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 204; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 204; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 204; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 204; goto st249;} }
	goto st204;
st204:
	if ( ++p == pe )
		goto _test_eof204;
case 204:
#line 28349 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 65: goto st82;
		case 73: goto st86;
		case 83: goto st88;
		case 97: goto st82;
		case 105: goto st86;
		case 115: goto st88;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3243:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 205; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 205; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 205; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 205; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 205; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 205; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 205; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 205; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 205; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 205; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 205; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 205; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 205; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 205; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 205; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 205; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 205; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 205; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 205; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 205; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 205; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 205; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 205; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 205; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 205; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 205; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 205; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 205; goto st249;} }
	goto st205;
st205:
	if ( ++p == pe )
		goto _test_eof205;
case 205:
#line 28476 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 84: goto st98;
		case 116: goto st98;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3244:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 206; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 206; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 206; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 206; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 206; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 206; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 206; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 206; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 206; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 206; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 206; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 206; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 206; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 206; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 206; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 206; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 206; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 206; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 206; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 206; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 206; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 206; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 206; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 206; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 206; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 206; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 206; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 206; goto st249;} }
	goto st206;
st206:
	if ( ++p == pe )
		goto _test_eof206;
case 206:
#line 28599 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 80: goto st101;
		case 82: goto st102;
		case 84: goto st106;
		case 112: goto st101;
		case 114: goto st102;
		case 116: goto st106;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3245:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 207; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 207; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 207; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 207; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 207; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 207; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 207; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 207; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 207; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 207; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 207; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 207; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 207; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 207; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 207; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 207; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 207; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 207; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 207; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 207; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 207; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 207; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 207; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 207; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 207; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 207; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 207; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 207; goto st249;} }
	goto st207;
st207:
	if ( ++p == pe )
		goto _test_eof207;
case 207:
#line 28726 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 79: goto st108;
		case 80: goto st110;
		case 82: goto st112;
		case 83: goto st114;
		case 111: goto st108;
		case 112: goto st110;
		case 114: goto st112;
		case 115: goto st114;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3246:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 208; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 208; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 208; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 208; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 208; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 208; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 208; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 208; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 208; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 208; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 208; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 208; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 208; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 208; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 208; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 208; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 208; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 208; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 208; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 208; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 208; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 208; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 208; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 208; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 208; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 208; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 208; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 208; goto st249;} }
	goto st208;
st208:
	if ( ++p == pe )
		goto _test_eof208;
case 208:
#line 28855 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr698;
		case 32: goto tr698;
		case 40: goto tr699;
		case 41: goto tr700;
		case 76: goto st119;
		case 88: goto st122;
		case 89: goto st124;
		case 108: goto st119;
		case 120: goto st122;
		case 121: goto st124;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr701;
		case 1083: goto tr702;
	}
	goto tr714;
tr3167:
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 209; goto st249;} }
	goto st209;
st209:
	if ( ++p == pe )
		goto _test_eof209;
case 209:
#line 28900 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr726;
		case 32: goto tr726;
		case 35: goto tr91;
		case 40: goto tr728;
		case 41: goto tr729;
		case 778: goto tr90;
		case 827: goto tr90;
		case 1034: goto tr730;
		case 1083: goto tr731;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr90;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr90;
	} else
		goto tr90;
	goto tr727;
tr726:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1873 "./zscanner/scanner_body.rl"
	{ p--; }
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 210; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 210; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 210; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 210; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 210; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 210; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 210; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 210; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 210; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 210; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 210; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 210; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 210; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 210; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 210; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 210; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 210; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 210; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 210; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 210; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 210; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 210; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 210; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 210; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 210; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 210; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 210; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st210;
tr728:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
#line 1873 "./zscanner/scanner_body.rl"
	{ p--; }
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 210; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 210; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 210; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 210; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 210; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 210; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 210; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 210; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 210; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 210; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 210; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 210; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 210; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 210; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 210; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 210; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 210; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 210; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 210; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 210; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 210; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 210; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 210; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 210; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 210; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 210; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 210; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st210;
tr729:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
#line 1873 "./zscanner/scanner_body.rl"
	{ p--; }
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 210; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 210; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 210; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 210; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 210; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 210; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 210; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 210; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 210; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 210; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 210; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 210; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 210; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 210; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 210; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 210; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 210; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 210; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 210; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 210; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 210; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 210; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 210; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 210; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 210; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 210; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 210; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st210;
tr730:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1873 "./zscanner/scanner_body.rl"
	{ p--; }
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 210; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 210; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 210; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 210; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 210; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 210; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 210; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 210; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 210; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 210; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 210; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 210; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 210; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 210; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 210; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 210; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 210; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 210; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 210; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 210; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 210; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 210; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 210; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 210; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 210; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 210; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 210; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st210;
st210:
	if ( ++p == pe )
		goto _test_eof210;
case 210:
#line 29287 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr732;
		case 32: goto tr732;
		case 40: goto tr733;
		case 41: goto tr734;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr82;
		case 827: goto tr83;
		case 1034: goto tr735;
		case 1083: goto tr736;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr101;
tr731:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1873 "./zscanner/scanner_body.rl"
	{ p--; }
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 211; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 211; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 211; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 211; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 211; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 211; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 211; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 211; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 211; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 211; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 211; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 211; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 211; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 211; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 211; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 211; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 211; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 211; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 211; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 211; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 211; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 211; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 211; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 211; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 211; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 211; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 211; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st211;
st211:
	if ( ++p == pe )
		goto _test_eof211;
case 211:
#line 29429 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto tr79;
		case 778: goto tr82;
		case 800: goto tr79;
		case 808: goto tr80;
		case 809: goto tr81;
		case 827: goto tr83;
		case 1033: goto tr737;
		case 1034: goto tr735;
		case 1056: goto tr737;
		case 1064: goto tr738;
		case 1065: goto tr739;
		case 1083: goto tr736;
	}
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st149;
	goto tr78;
tr741:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st212;
tr742:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st212;
tr737:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st212;
tr738:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st212;
tr739:
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st212;
st212:
	if ( ++p == pe )
		goto _test_eof212;
case 212:
#line 29543 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr88;
		case 800: goto st7;
		case 808: goto tr86;
		case 809: goto tr87;
		case 827: goto st8;
		case 1033: goto st212;
		case 1034: goto tr676;
		case 1056: goto st212;
		case 1064: goto tr741;
		case 1065: goto tr742;
		case 1083: goto st162;
	}
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st149;
	goto tr84;
tr744:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st213;
tr745:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st213;
tr685:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 213; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 213; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 213; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 213; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 213; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 213; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 213; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 213; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 213; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 213; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 213; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 213; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 213; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 213; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 213; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 213; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 213; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 213; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 213; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 213; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 213; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 213; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 213; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 213; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 213; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 213; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 213; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st213;
tr721:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 213; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 213; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 213; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 213; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 213; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 213; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 213; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 213; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 213; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 213; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 213; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 213; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 213; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 213; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 213; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 213; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 213; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 213; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 213; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 213; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 213; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 213; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 213; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 213; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 213; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 213; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 213; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st213;
tr3230:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 213; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 213; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 213; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 213; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 213; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 213; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 213; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 213; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 213; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 213; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 213; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 213; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 213; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 213; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 213; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 213; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 213; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 213; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 213; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 213; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 213; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 213; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 213; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 213; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 213; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 213; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 213; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st213;
st213:
	if ( ++p == pe )
		goto _test_eof213;
case 213:
#line 29872 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st164;
		case 778: goto tr88;
		case 800: goto st164;
		case 808: goto tr679;
		case 809: goto tr680;
		case 1033: goto st213;
		case 1034: goto tr676;
		case 1056: goto st213;
		case 1064: goto tr744;
		case 1065: goto tr745;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st179;
	} else if ( _widec >= 768 )
		goto st8;
	goto tr78;
st214:
	if ( ++p == pe )
		goto _test_eof214;
case 214:
	switch( (*p) ) {
		case 65: goto st215;
		case 97: goto st215;
	}
	goto tr35;
st215:
	if ( ++p == pe )
		goto _test_eof215;
case 215:
	switch( (*p) ) {
		case 65: goto st216;
		case 97: goto st216;
	}
	goto tr35;
st216:
	if ( ++p == pe )
		goto _test_eof216;
case 216:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr748;
		case 32: goto tr748;
		case 40: goto tr749;
		case 41: goto tr750;
		case 2058: goto tr751;
		case 2107: goto tr752;
		case 2314: goto tr753;
		case 2363: goto tr753;
		case 2570: goto tr754;
		case 2619: goto tr755;
	}
	goto tr55;
st217:
	if ( ++p == pe )
		goto _test_eof217;
case 217:
	switch( (*p) ) {
		case 83: goto st218;
		case 115: goto st218;
	}
	goto tr35;
st218:
	if ( ++p == pe )
		goto _test_eof218;
case 218:
	switch( (*p) ) {
		case 68: goto st219;
		case 100: goto st219;
	}
	goto tr35;
st219:
	if ( ++p == pe )
		goto _test_eof219;
case 219:
	switch( (*p) ) {
		case 66: goto st220;
		case 98: goto st220;
	}
	goto tr35;
st220:
	if ( ++p == pe )
		goto _test_eof220;
case 220:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr759;
		case 32: goto tr759;
		case 40: goto tr760;
		case 41: goto tr761;
		case 2058: goto tr762;
		case 2107: goto tr763;
		case 2314: goto tr764;
		case 2363: goto tr764;
		case 2570: goto tr765;
		case 2619: goto tr766;
	}
	goto tr55;
st221:
	if ( ++p == pe )
		goto _test_eof221;
case 221:
	switch( (*p) ) {
		case 76: goto st222;
		case 108: goto st222;
	}
	goto tr35;
st222:
	if ( ++p == pe )
		goto _test_eof222;
case 222:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr768;
		case 32: goto tr768;
		case 40: goto tr769;
		case 41: goto tr770;
		case 2058: goto tr771;
		case 2107: goto tr772;
		case 2314: goto tr773;
		case 2363: goto tr773;
		case 2570: goto tr774;
		case 2619: goto tr775;
	}
	goto tr55;
tr3169:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st223;
st223:
	if ( ++p == pe )
		goto _test_eof223;
case 223:
#line 30104 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr19;
		case 1034: goto tr119;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st223;
	} else if ( _widec >= 768 )
		goto st138;
	goto tr78;
tr122:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 224; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 224; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 224; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 224; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 224; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 224; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 224; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 224; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 224; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 224; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 224; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 224; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 224; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 224; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 224; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 224; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 224; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 224; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 224; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 224; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 224; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 224; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 224; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 224; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 224; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 224; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 224; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st224;
st224:
	if ( ++p == pe )
		goto _test_eof224;
case 224:
#line 30216 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st225;
		case 32: goto st225;
		case 40: goto tr777;
		case 41: goto tr778;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr88;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st138;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st138;
	} else
		goto st138;
	goto tr78;
tr777:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st225;
tr778:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st225;
st225:
	if ( ++p == pe )
		goto _test_eof225;
case 225:
#line 30274 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st225;
		case 32: goto st225;
		case 40: goto tr777;
		case 41: goto tr778;
		case 778: goto tr88;
		case 827: goto st8;
		case 1034: goto tr88;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st138;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st138;
	} else
		goto st138;
	goto tr84;
tr780:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st226;
tr781:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st226;
tr124:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 226; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 226; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 226; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 226; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 226; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 226; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 226; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 226; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 226; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 226; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 226; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 226; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 226; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 226; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 226; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 226; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 226; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 226; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 226; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 226; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 226; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 226; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 226; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 226; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 226; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 226; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 226; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st226;
tr3171:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 226; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 226; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 226; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 226; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 226; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 226; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 226; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 226; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 226; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 226; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 226; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 226; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 226; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 226; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 226; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 226; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 226; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 226; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 226; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 226; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 226; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 226; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 226; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 226; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 226; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 226; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 226; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st226;
st226:
	if ( ++p == pe )
		goto _test_eof226;
case 226:
#line 30491 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st225;
		case 778: goto tr88;
		case 800: goto st225;
		case 808: goto tr777;
		case 809: goto tr778;
		case 827: goto st8;
		case 1033: goto st226;
		case 1034: goto tr676;
		case 1056: goto st226;
		case 1064: goto tr780;
		case 1065: goto tr781;
		case 1083: goto st179;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st223;
	} else if ( _widec >= 768 )
		goto st138;
	goto tr78;
tr783:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st227;
tr784:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st227;
tr98:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 227; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 227; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 227; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 227; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 227; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 227; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 227; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 227; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 227; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 227; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 227; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 227; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 227; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 227; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 227; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 227; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 227; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 227; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 227; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 227; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 227; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 227; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 227; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 227; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 227; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 227; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 227; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st227;
tr668:
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 227; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 227; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 227; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 227; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 227; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 227; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 227; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 227; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 227; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 227; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 227; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 227; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 227; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 227; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 227; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 227; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 227; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 227; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 227; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 227; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 227; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 227; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 227; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 227; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 227; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 227; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 227; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st227;
st227:
	if ( ++p == pe )
		goto _test_eof227;
case 227:
#line 30736 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st164;
		case 778: goto tr88;
		case 800: goto st164;
		case 808: goto tr679;
		case 809: goto tr680;
		case 1033: goto st227;
		case 1034: goto tr88;
		case 1056: goto st227;
		case 1064: goto tr783;
		case 1065: goto tr784;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st13;
	} else if ( _widec >= 768 )
		goto st8;
	goto tr78;
tr649:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
	goto st228;
st228:
	if ( ++p == pe )
		goto _test_eof228;
case 228:
#line 30808 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr655;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st228;
	goto tr69;
tr786:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st229;
tr787:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st229;
tr658:
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 229; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 229; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 229; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 229; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 229; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 229; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 229; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 229; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 229; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 229; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 229; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 229; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 229; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 229; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 229; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 229; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 229; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 229; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 229; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 229; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 229; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 229; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 229; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 229; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 229; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 229; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 229; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st229;
tr651:
#line 233 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = s->dname_tmp_length;
	}
#line 1886 "./zscanner/scanner_body.rl"
	{ s->r_type = KNOT_RRTYPE_A; }
#line 1621 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[0] = 0;
		s->r_data_blocks_count = 0;
		rdata_tail = s->r_data;
	}
#line 1744 "./zscanner/scanner_body.rl"
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 229; goto st593;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 229; goto st595;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 229; goto st597;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 229; goto st629;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 229; goto st634;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 229; goto st639;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 229; goto st644;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 229; goto st648;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 229; goto st650;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 229; goto st705;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 229; goto st716;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 229; goto st733;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 229; goto st744;}
		case KNOT_RRTYPE_DS:
			{stack[top++] = 229; goto st755;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 229; goto st768;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 229; goto st778;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 229; goto st817;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 229; goto st959;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
			{stack[top++] = 229; goto st962;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 229; goto st973;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 229; goto st975;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 229; goto st1004;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 229; goto st1017;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 229; goto st1035;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 229; goto st1030;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 229; goto st1048;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 229; goto st1054;}
		default:
			SCANNER_WARNING(ZSCANNER_ECANNOT_TEXT_DATA);
			{goto st247;}
		}
	}
	goto st229;
st229:
	if ( ++p == pe )
		goto _test_eof229;
case 229:
#line 31023 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(768 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr88;
		case 800: goto st7;
		case 808: goto tr86;
		case 809: goto tr87;
		case 827: goto st8;
		case 1033: goto st229;
		case 1034: goto tr676;
		case 1056: goto st229;
		case 1064: goto tr786;
		case 1065: goto tr787;
		case 1083: goto st179;
	}
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st228;
	goto tr78;
tr3177:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 230; goto st249;} }
	goto st230;
tr3195:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 230; goto st249;} }
	goto st230;
st230:
	if ( ++p == pe )
		goto _test_eof230;
case 230:
#line 31110 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 69: goto st18;
		case 78: goto st22;
		case 101: goto st18;
		case 110: goto st22;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3178:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 231; goto st249;} }
	goto st231;
tr3196:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 231; goto st249;} }
	goto st231;
st231:
	if ( ++p == pe )
		goto _test_eof231;
case 231:
#line 31172 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 72: goto st27;
		case 78: goto st31;
		case 83: goto st39;
		case 104: goto st27;
		case 110: goto st31;
		case 115: goto st39;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3179:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 232; goto st249;} }
	goto st232;
tr3197:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 232; goto st249;} }
	goto st232;
st232:
	if ( ++p == pe )
		goto _test_eof232;
case 232:
#line 31236 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 85: goto st41;
		case 117: goto st41;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3180:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 233; goto st249;} }
	goto st233;
tr3198:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 233; goto st249;} }
	goto st233;
st233:
	if ( ++p == pe )
		goto _test_eof233;
case 233:
#line 31296 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 73: goto st48;
		case 105: goto st48;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3199:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 234; goto st249;} }
	goto st234;
st234:
	if ( ++p == pe )
		goto _test_eof234;
case 234:
#line 31343 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 78: goto st131;
		case 80: goto st56;
		case 110: goto st131;
		case 112: goto st56;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3182:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 235; goto st249;} }
	goto st235;
tr3200:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 235; goto st249;} }
	goto st235;
st235:
	if ( ++p == pe )
		goto _test_eof235;
case 235:
#line 31405 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 69: goto st64;
		case 88: goto st66;
		case 101: goto st64;
		case 120: goto st66;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3183:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 236; goto st249;} }
	goto st236;
tr3201:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 236; goto st249;} }
	goto st236;
st236:
	if ( ++p == pe )
		goto _test_eof236;
case 236:
#line 31467 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 51: goto st68;
		case 54: goto st70;
		case 79: goto st72;
		case 80: goto st74;
		case 111: goto st72;
		case 112: goto st74;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3184:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 237; goto st249;} }
	goto st237;
tr3202:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 237; goto st249;} }
	goto st237;
st237:
	if ( ++p == pe )
		goto _test_eof237;
case 237:
#line 31531 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 73: goto st76;
		case 88: goto st80;
		case 105: goto st76;
		case 120: goto st80;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3185:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 238; goto st249;} }
	goto st238;
tr3203:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 238; goto st249;} }
	goto st238;
st238:
	if ( ++p == pe )
		goto _test_eof238;
case 238:
#line 31593 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 65: goto st82;
		case 73: goto st86;
		case 83: goto st88;
		case 97: goto st82;
		case 105: goto st86;
		case 115: goto st88;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3186:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 239; goto st249;} }
	goto st239;
tr3204:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 239; goto st249;} }
	goto st239;
st239:
	if ( ++p == pe )
		goto _test_eof239;
case 239:
#line 31657 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 84: goto st98;
		case 116: goto st98;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3187:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 240; goto st249;} }
	goto st240;
tr3205:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 240; goto st249;} }
	goto st240;
st240:
	if ( ++p == pe )
		goto _test_eof240;
case 240:
#line 31717 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 80: goto st101;
		case 82: goto st102;
		case 84: goto st106;
		case 112: goto st101;
		case 114: goto st102;
		case 116: goto st106;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3188:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 241; goto st249;} }
	goto st241;
tr3206:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 241; goto st249;} }
	goto st241;
st241:
	if ( ++p == pe )
		goto _test_eof241;
case 241:
#line 31781 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 79: goto st108;
		case 80: goto st110;
		case 82: goto st112;
		case 83: goto st114;
		case 111: goto st108;
		case 112: goto st110;
		case 114: goto st112;
		case 115: goto st114;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3189:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 242; goto st249;} }
	goto st242;
tr3207:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 740 "./zscanner/scanner_body.rl"
	{
		s->r_ttl = s->default_ttl;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 242; goto st249;} }
	goto st242;
st242:
	if ( ++p == pe )
		goto _test_eof242;
case 242:
#line 31847 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 76: goto st119;
		case 88: goto st122;
		case 89: goto st124;
		case 108: goto st119;
		case 120: goto st122;
		case 121: goto st124;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr3161:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st243;
tr3217:
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st243;
st243:
	if ( ++p == pe )
		goto _test_eof243;
case 243:
#line 31903 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr19;
		case 1034: goto tr21;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st243;
	} else if ( _widec >= 768 )
		goto st138;
	goto tr84;
tr3191:
#line 236 "./zscanner/scanner_body.rl"
	{
		if (s->r_owner_length == 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_PREVIOUS_OWNER);
			p--; {goto st247;}
		}
	}
	goto st244;
st244:
	if ( ++p == pe )
		goto _test_eof244;
case 244:
#line 31948 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr19;
		case 1034: goto tr545;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st244;
	} else if ( _widec >= 768 )
		goto st138;
	goto tr131;
tr3181:
#line 736 "./zscanner/scanner_body.rl"
	{
		s->r_class = s->default_class;
	}
#line 229 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 245; goto st249;} }
	goto st245;
st245:
	if ( ++p == pe )
		goto _test_eof245;
case 245:
#line 31997 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr609;
		case 32: goto tr609;
		case 40: goto tr610;
		case 41: goto tr611;
		case 78: goto st53;
		case 80: goto st56;
		case 110: goto st53;
		case 112: goto st56;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	goto tr788;
tr130:
#line 748 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1865 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			(uint16_t)(rdata_tail - s->r_data);
	}
	goto st246;
st246:
	if ( ++p == pe )
		goto _test_eof246;
case 246:
#line 32045 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr88;
		case 1034: goto tr135;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st246;
	} else if ( _widec >= 768 )
		goto st8;
	goto tr131;
st247:
	if ( ++p == pe )
		goto _test_eof247;
case 247:
	if ( (*p) == 10 )
		goto tr790;
	goto tr789;
tr789:
#line 67 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 70 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st248;
tr791:
#line 70 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st248;
st248:
	if ( ++p == pe )
		goto _test_eof248;
case 248:
#line 32108 "zscanner/scanner.c"
	if ( (*p) == 10 )
		goto tr792;
	goto tr791;
tr790:
#line 67 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 75 "./zscanner/scanner_body.rl"
	{
		// Ending string in buffer.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error_counter++;

		// Initialization of fcall stack.
		top = 0;

		// Process error message.
		s->process_error(s);

		// Reset.
		s->error_code = KNOT_EOK;
		s->multiline = false;

		// In case of serious error, stop scanner.
		if (s->stop == true) {
			return -1;
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 100 "./zscanner/scanner_body.rl"
	{ {goto st1060;} }
	goto st1070;
tr792:
#line 75 "./zscanner/scanner_body.rl"
	{
		// Ending string in buffer.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error_counter++;

		// Initialization of fcall stack.
		top = 0;

		// Process error message.
		s->process_error(s);

		// Reset.
		s->error_code = KNOT_EOK;
		s->multiline = false;

		// In case of serious error, stop scanner.
		if (s->stop == true) {
			return -1;
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 100 "./zscanner/scanner_body.rl"
	{ {goto st1060;} }
	goto st1070;
st1070:
	if ( ++p == pe )
		goto _test_eof1070;
case 1070:
#line 32182 "zscanner/scanner.c"
	goto st0;
st249:
	if ( ++p == pe )
		goto _test_eof249;
case 249:
	switch( (*p) ) {
		case 42: goto tr794;
		case 46: goto tr796;
		case 64: goto st258;
		case 92: goto tr798;
		case 95: goto tr795;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr795;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr795;
	} else
		goto tr795;
	goto tr793;
tr794:
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st250;
tr802:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st250;
st250:
	if ( ++p == pe )
		goto _test_eof250;
case 250:
#line 32247 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr799;
		case 46: goto tr800;
		case 59: goto tr799;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr799;
	} else if ( (*p) >= 9 )
		goto tr799;
	goto tr793;
tr799:
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 165 "./zscanner/scanner_body.rl"
	{
		memcpy(s->dname + s->dname_tmp_length,
		       s->zone_origin,
		       s->zone_origin_length);

		s->dname_tmp_length += s->zone_origin_length;

		if (s->dname_tmp_length > MAX_DNAME_LENGTH) {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1071;
tr801:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1071;
tr811:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 165 "./zscanner/scanner_body.rl"
	{
		memcpy(s->dname + s->dname_tmp_length,
		       s->zone_origin,
		       s->zone_origin_length);

		s->dname_tmp_length += s->zone_origin_length;

		if (s->dname_tmp_length > MAX_DNAME_LENGTH) {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1071;
tr815:
#line 177 "./zscanner/scanner_body.rl"
	{
		memcpy(s->dname,
		       s->zone_origin,
		       s->zone_origin_length);

		s->dname_tmp_length = s->zone_origin_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1071;
st1071:
	if ( ++p == pe )
		goto _test_eof1071;
case 1071:
#line 32349 "zscanner/scanner.c"
	goto st0;
tr800:
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st251;
tr813:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st251;
st251:
	if ( ++p == pe )
		goto _test_eof251;
case 251:
#line 32383 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr801;
		case 42: goto tr802;
		case 45: goto tr803;
		case 59: goto tr801;
		case 92: goto tr804;
		case 95: goto tr803;
	}
	if ( (*p) < 47 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr801;
		} else if ( (*p) >= 9 )
			goto tr801;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr803;
		} else if ( (*p) >= 65 )
			goto tr803;
	} else
		goto tr803;
	goto tr793;
tr795:
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st252;
tr803:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st252;
tr805:
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st252;
tr812:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st252;
st252:
	if ( ++p == pe )
		goto _test_eof252;
case 252:
#line 32478 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr799;
		case 46: goto tr800;
		case 59: goto tr799;
		case 92: goto st253;
		case 95: goto tr805;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr799;
		} else if ( (*p) >= 9 )
			goto tr799;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr805;
		} else if ( (*p) >= 65 )
			goto tr805;
	} else
		goto tr805;
	goto tr793;
tr798:
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st253;
tr804:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st253;
tr814:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
	goto st253;
st253:
	if ( ++p == pe )
		goto _test_eof253;
case 253:
#line 32530 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr807;
	goto tr805;
tr807:
#line 127 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st254;
st254:
	if ( ++p == pe )
		goto _test_eof254;
case 254:
#line 32555 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr809;
	goto tr808;
tr809:
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st255;
st255:
	if ( ++p == pe )
		goto _test_eof255;
case 255:
#line 32570 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr810;
	goto tr808;
tr810:
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st256;
st256:
	if ( ++p == pe )
		goto _test_eof256;
case 256:
#line 32585 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr811;
		case 46: goto tr813;
		case 59: goto tr811;
		case 92: goto tr814;
		case 95: goto tr812;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr811;
		} else if ( (*p) >= 9 )
			goto tr811;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr812;
		} else if ( (*p) >= 65 )
			goto tr812;
	} else
		goto tr812;
	goto tr808;
tr796:
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st257;
st257:
	if ( ++p == pe )
		goto _test_eof257;
case 257:
#line 32619 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr801;
		case 59: goto tr801;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr801;
	} else if ( (*p) >= 9 )
		goto tr801;
	goto tr793;
st258:
	if ( ++p == pe )
		goto _test_eof258;
case 258:
	switch( (*p) ) {
		case 32: goto tr815;
		case 59: goto tr815;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr815;
	} else if ( (*p) >= 9 )
		goto tr815;
	goto tr793;
st259:
	if ( ++p == pe )
		goto _test_eof259;
case 259:
	switch( (*p) ) {
		case 34: goto st265;
		case 92: goto st261;
	}
	if ( (*p) > 58 ) {
		if ( 60 <= (*p) && (*p) <= 126 )
			goto tr817;
	} else if ( (*p) >= 33 )
		goto tr817;
	goto tr816;
tr817:
#line 501 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*(rdata_tail++) = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st260;
tr827:
#line 540 "./zscanner/scanner_body.rl"
	{
		rdata_tail++;
	}
#line 501 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*(rdata_tail++) = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st260;
st260:
	if ( ++p == pe )
		goto _test_eof260;
case 260:
#line 32688 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr820;
		case 33: goto tr817;
		case 59: goto tr820;
		case 92: goto st261;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr820;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr817;
		} else if ( (*p) >= 40 )
			goto tr821;
	} else
		goto tr817;
	goto tr816;
tr820:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1072;
tr826:
#line 540 "./zscanner/scanner_body.rl"
	{
		rdata_tail++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1072;
st1072:
	if ( ++p == pe )
		goto _test_eof1072;
case 1072:
#line 32727 "zscanner/scanner.c"
	goto st0;
tr821:
#line 501 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*(rdata_tail++) = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1073;
tr828:
#line 540 "./zscanner/scanner_body.rl"
	{
		rdata_tail++;
	}
#line 501 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*(rdata_tail++) = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1073;
st1073:
	if ( ++p == pe )
		goto _test_eof1073;
case 1073:
#line 32767 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr820;
		case 33: goto tr817;
		case 59: goto tr820;
		case 92: goto st261;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr820;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr817;
		} else if ( (*p) >= 40 )
			goto tr821;
	} else
		goto tr817;
	goto tr816;
tr829:
#line 540 "./zscanner/scanner_body.rl"
	{
		rdata_tail++;
	}
	goto st261;
st261:
	if ( ++p == pe )
		goto _test_eof261;
case 261:
#line 32796 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr822;
	goto tr817;
tr822:
#line 518 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = 0;
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 527 "./zscanner/scanner_body.rl"
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + ASCII_0)
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st262;
st262:
	if ( ++p == pe )
		goto _test_eof262;
case 262:
#line 32830 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr824;
	goto tr823;
tr824:
#line 527 "./zscanner/scanner_body.rl"
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + ASCII_0)
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st263;
st263:
	if ( ++p == pe )
		goto _test_eof263;
case 263:
#line 32854 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr825;
	goto tr823;
tr825:
#line 527 "./zscanner/scanner_body.rl"
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + ASCII_0)
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st264;
st264:
	if ( ++p == pe )
		goto _test_eof264;
case 264:
#line 32878 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr826;
		case 33: goto tr827;
		case 59: goto tr826;
		case 92: goto tr829;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr826;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr827;
		} else if ( (*p) >= 40 )
			goto tr828;
	} else
		goto tr827;
	goto tr823;
tr830:
#line 501 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*(rdata_tail++) = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st265;
tr837:
#line 540 "./zscanner/scanner_body.rl"
	{
		rdata_tail++;
	}
#line 501 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*(rdata_tail++) = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st265;
st265:
	if ( ++p == pe )
		goto _test_eof265;
case 265:
#line 32927 "zscanner/scanner.c"
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(256 + ((*p) - 0));
		if ( 
#line 559 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr830;
		case 34: goto st266;
		case 92: goto st267;
		case 522: goto tr830;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr830;
	goto tr816;
tr838:
#line 540 "./zscanner/scanner_body.rl"
	{
		rdata_tail++;
	}
	goto st266;
st266:
	if ( ++p == pe )
		goto _test_eof266;
case 266:
#line 32954 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr820;
		case 59: goto tr820;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr820;
	} else if ( (*p) >= 9 )
		goto tr820;
	goto tr833;
tr839:
#line 540 "./zscanner/scanner_body.rl"
	{
		rdata_tail++;
	}
	goto st267;
st267:
	if ( ++p == pe )
		goto _test_eof267;
case 267:
#line 32975 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr834;
	goto tr830;
tr834:
#line 518 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = 0;
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ETEXT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 527 "./zscanner/scanner_body.rl"
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + ASCII_0)
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st268;
st268:
	if ( ++p == pe )
		goto _test_eof268;
case 268:
#line 33009 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr835;
	goto tr823;
tr835:
#line 527 "./zscanner/scanner_body.rl"
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + ASCII_0)
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st269;
st269:
	if ( ++p == pe )
		goto _test_eof269;
case 269:
#line 33033 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr836;
	goto tr823;
tr836:
#line 527 "./zscanner/scanner_body.rl"
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + ASCII_0)
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st270;
st270:
	if ( ++p == pe )
		goto _test_eof270;
case 270:
#line 33057 "zscanner/scanner.c"
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(256 + ((*p) - 0));
		if ( 
#line 559 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr837;
		case 34: goto tr838;
		case 92: goto tr839;
		case 522: goto tr837;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr837;
	goto tr823;
st271:
	if ( ++p == pe )
		goto _test_eof271;
case 271:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st272;
		case 32: goto st272;
		case 40: goto tr842;
		case 41: goto tr843;
		case 1034: goto tr844;
		case 1083: goto st279;
	}
	goto tr840;
tr842:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st272;
tr843:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st272;
tr844:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st272;
st272:
	if ( ++p == pe )
		goto _test_eof272;
case 272:
#line 33131 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st272;
		case 32: goto st272;
		case 40: goto tr842;
		case 41: goto tr843;
		case 1034: goto tr844;
		case 1083: goto st279;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr847;
	goto tr846;
tr847:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st273;
tr852:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st273;
st273:
	if ( ++p == pe )
		goto _test_eof273;
case 273:
#line 33199 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st274;
		case 32: goto st274;
		case 40: goto tr850;
		case 41: goto tr851;
		case 68: goto tr853;
		case 72: goto tr854;
		case 77: goto tr855;
		case 83: goto st276;
		case 87: goto tr857;
		case 100: goto tr853;
		case 104: goto tr854;
		case 109: goto tr855;
		case 115: goto st276;
		case 119: goto tr857;
		case 778: goto tr858;
		case 827: goto st275;
		case 1034: goto tr858;
		case 1083: goto st275;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr852;
	goto tr848;
tr850:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st274;
tr851:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st274;
tr869:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st274;
tr870:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st274;
tr871:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st274;
st274:
	if ( ++p == pe )
		goto _test_eof274;
case 274:
#line 33310 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st274;
		case 32: goto st274;
		case 40: goto tr850;
		case 41: goto tr851;
		case 778: goto tr858;
		case 827: goto st275;
		case 1034: goto tr858;
		case 1083: goto st275;
	}
	goto tr860;
tr858:
#line 575 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_ERROR(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1074;
tr873:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 575 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			SCANNER_ERROR(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1074;
st1074:
	if ( ++p == pe )
		goto _test_eof1074;
case 1074:
#line 33387 "zscanner/scanner.c"
	goto st0;
tr874:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st275;
st275:
	if ( ++p == pe )
		goto _test_eof275;
case 275:
#line 33404 "zscanner/scanner.c"
	if ( (*p) == 10 )
		goto tr858;
	goto st275;
tr853:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st276;
tr854:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st276;
tr855:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st276;
tr857:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st276;
st276:
	if ( ++p == pe )
		goto _test_eof276;
case 276:
#line 33452 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st274;
		case 32: goto st274;
		case 40: goto tr850;
		case 41: goto tr851;
		case 778: goto tr858;
		case 827: goto st275;
		case 1034: goto tr858;
		case 1083: goto st275;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr861;
	goto tr848;
tr863:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st277;
tr861:
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st277;
tr872:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st277;
st277:
	if ( ++p == pe )
		goto _test_eof277;
case 277:
#line 33560 "zscanner/scanner.c"
	switch( (*p) ) {
		case 68: goto tr864;
		case 72: goto tr865;
		case 77: goto tr866;
		case 83: goto st278;
		case 87: goto tr868;
		case 100: goto tr864;
		case 104: goto tr865;
		case 109: goto tr866;
		case 115: goto st278;
		case 119: goto tr868;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr863;
	goto tr862;
tr864:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st278;
tr865:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st278;
tr866:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st278;
tr868:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st278;
st278:
	if ( ++p == pe )
		goto _test_eof278;
case 278:
#line 33620 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr869;
		case 32: goto tr869;
		case 40: goto tr870;
		case 41: goto tr871;
		case 778: goto tr873;
		case 827: goto tr874;
		case 1034: goto tr873;
		case 1083: goto tr874;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr872;
	goto tr848;
st279:
	if ( ++p == pe )
		goto _test_eof279;
case 279:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr844;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st279;
	goto tr840;
st280:
	if ( ++p == pe )
		goto _test_eof280;
case 280:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st281;
		case 32: goto st281;
		case 40: goto tr877;
		case 41: goto tr878;
		case 1034: goto tr879;
		case 1083: goto st292;
	}
	goto tr875;
tr877:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st281;
tr878:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st281;
tr879:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st281;
st281:
	if ( ++p == pe )
		goto _test_eof281;
case 281:
#line 33735 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st281;
		case 32: goto st281;
		case 40: goto tr877;
		case 41: goto tr878;
		case 42: goto tr881;
		case 46: goto tr883;
		case 92: goto tr884;
		case 95: goto tr882;
		case 1034: goto tr879;
		case 1083: goto st292;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr882;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr882;
	} else
		goto tr882;
	goto tr875;
tr890:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st282;
tr881:
#line 594 "./zscanner/scanner_body.rl"
	{
		s->dname = s->zone_origin;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st282;
st282:
	if ( ++p == pe )
		goto _test_eof282;
case 282:
#line 33818 "zscanner/scanner.c"
	if ( (*p) == 46 )
		goto tr885;
	goto tr875;
tr885:
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st283;
tr907:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st283;
st283:
	if ( ++p == pe )
		goto _test_eof283;
case 283:
#line 33854 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr887;
		case 32: goto tr887;
		case 40: goto tr888;
		case 41: goto tr889;
		case 42: goto tr890;
		case 45: goto tr891;
		case 92: goto tr892;
		case 95: goto tr891;
		case 778: goto tr893;
		case 827: goto tr894;
		case 1034: goto tr893;
		case 1083: goto tr894;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr891;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr891;
	} else
		goto tr891;
	goto tr886;
tr896:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st284;
tr897:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st284;
tr887:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	goto st284;
tr888:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st284;
tr889:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st284;
st284:
	if ( ++p == pe )
		goto _test_eof284;
case 284:
#line 33950 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st284;
		case 32: goto st284;
		case 40: goto tr896;
		case 41: goto tr897;
		case 778: goto tr898;
		case 827: goto st285;
		case 1034: goto tr898;
		case 1083: goto st285;
	}
	goto tr886;
tr893:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 597 "./zscanner/scanner_body.rl"
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1075;
tr898:
#line 597 "./zscanner/scanner_body.rl"
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1075;
st1075:
	if ( ++p == pe )
		goto _test_eof1075;
case 1075:
#line 34012 "zscanner/scanner.c"
	goto st0;
tr894:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	goto st285;
st285:
	if ( ++p == pe )
		goto _test_eof285;
case 285:
#line 34024 "zscanner/scanner.c"
	if ( (*p) == 10 )
		goto tr898;
	goto st285;
tr891:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st286;
tr900:
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st286;
tr906:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st286;
tr882:
#line 594 "./zscanner/scanner_body.rl"
	{
		s->dname = s->zone_origin;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st286;
st286:
	if ( ++p == pe )
		goto _test_eof286;
case 286:
#line 34103 "zscanner/scanner.c"
	switch( (*p) ) {
		case 46: goto tr885;
		case 92: goto st287;
		case 95: goto tr900;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr900;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr900;
	} else
		goto tr900;
	goto tr875;
tr892:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st287;
tr908:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
	goto st287;
tr884:
#line 594 "./zscanner/scanner_body.rl"
	{
		s->dname = s->zone_origin;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st287;
st287:
	if ( ++p == pe )
		goto _test_eof287;
case 287:
#line 34151 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr902;
	goto tr900;
tr902:
#line 127 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st288;
st288:
	if ( ++p == pe )
		goto _test_eof288;
case 288:
#line 34176 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr904;
	goto tr903;
tr904:
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st289;
st289:
	if ( ++p == pe )
		goto _test_eof289;
case 289:
#line 34191 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr905;
	goto tr903;
tr905:
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st290;
st290:
	if ( ++p == pe )
		goto _test_eof290;
case 290:
#line 34206 "zscanner/scanner.c"
	switch( (*p) ) {
		case 46: goto tr907;
		case 92: goto tr908;
		case 95: goto tr906;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr906;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr906;
	} else
		goto tr906;
	goto tr903;
tr883:
#line 594 "./zscanner/scanner_body.rl"
	{
		s->dname = s->zone_origin;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st291;
st291:
	if ( ++p == pe )
		goto _test_eof291;
case 291:
#line 34236 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr887;
		case 32: goto tr887;
		case 40: goto tr888;
		case 41: goto tr889;
		case 778: goto tr893;
		case 827: goto tr894;
		case 1034: goto tr893;
		case 1083: goto tr894;
	}
	goto tr886;
st292:
	if ( ++p == pe )
		goto _test_eof292;
case 292:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr879;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st292;
	goto tr875;
st293:
	if ( ++p == pe )
		goto _test_eof293;
case 293:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st294;
		case 32: goto st294;
		case 40: goto tr910;
		case 41: goto tr911;
		case 1034: goto tr912;
		case 1083: goto st310;
	}
	goto st0;
tr910:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st294;
tr911:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st294;
tr912:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st294;
st294:
	if ( ++p == pe )
		goto _test_eof294;
case 294:
#line 34349 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st294;
		case 32: goto st294;
		case 40: goto tr910;
		case 41: goto tr911;
		case 1034: goto tr912;
		case 1083: goto st310;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr914;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr914;
	} else
		goto tr914;
	goto tr915;
tr914:
#line 611 "./zscanner/scanner_body.rl"
	{
		rdata_tail = s->r_data;
	}
#line 565 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 295; goto st259;} }
	goto st295;
st295:
	if ( ++p == pe )
		goto _test_eof295;
case 295:
#line 34393 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr917;
		case 32: goto tr917;
		case 40: goto tr918;
		case 41: goto tr919;
		case 778: goto tr920;
		case 827: goto tr921;
		case 1034: goto tr920;
		case 1083: goto tr922;
	}
	goto tr916;
tr925:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st296;
tr926:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st296;
tr917:
#line 614 "./zscanner/scanner_body.rl"
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
			p--; {goto st247;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	goto st296;
tr918:
#line 614 "./zscanner/scanner_body.rl"
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
			p--; {goto st247;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st296;
tr919:
#line 614 "./zscanner/scanner_body.rl"
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
			p--; {goto st247;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st296;
st296:
	if ( ++p == pe )
		goto _test_eof296;
case 296:
#line 34513 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st296;
		case 32: goto st296;
		case 40: goto tr925;
		case 41: goto tr926;
		case 42: goto tr927;
		case 46: goto tr929;
		case 92: goto tr930;
		case 95: goto tr928;
		case 778: goto tr931;
		case 827: goto st300;
		case 1034: goto tr931;
		case 1083: goto st307;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr928;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr928;
	} else
		goto tr928;
	goto tr923;
tr939:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st297;
tr927:
#line 634 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_data;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st297;
st297:
	if ( ++p == pe )
		goto _test_eof297;
case 297:
#line 34598 "zscanner/scanner.c"
	if ( (*p) == 46 )
		goto tr935;
	goto tr934;
tr935:
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st298;
tr954:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
#line 117 "./zscanner/scanner_body.rl"
	{
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EDNAME_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st298;
st298:
	if ( ++p == pe )
		goto _test_eof298;
case 298:
#line 34634 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr936;
		case 32: goto tr936;
		case 40: goto tr937;
		case 41: goto tr938;
		case 42: goto tr939;
		case 45: goto tr940;
		case 92: goto tr941;
		case 95: goto tr940;
		case 778: goto tr942;
		case 827: goto tr943;
		case 1034: goto tr942;
		case 1083: goto tr943;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr940;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr940;
	} else
		goto tr940;
	goto tr923;
tr945:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st299;
tr946:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st299;
tr936:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 637 "./zscanner/scanner_body.rl"
	{
		s->r_data_length = s->dname_tmp_length;
	}
	goto st299;
tr937:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 637 "./zscanner/scanner_body.rl"
	{
		s->r_data_length = s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st299;
tr938:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 637 "./zscanner/scanner_body.rl"
	{
		s->r_data_length = s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st299;
st299:
	if ( ++p == pe )
		goto _test_eof299;
case 299:
#line 34742 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st299;
		case 32: goto st299;
		case 40: goto tr945;
		case 41: goto tr946;
		case 778: goto tr931;
		case 827: goto st300;
		case 1034: goto tr931;
		case 1083: goto st300;
	}
	goto tr84;
tr920:
#line 614 "./zscanner/scanner_body.rl"
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
			p--; {goto st247;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
#line 645 "./zscanner/scanner_body.rl"
	{
		char text_origin[MAX_DNAME_LENGTH];

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
					  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
					  text_origin);
		}

		if (s->include_filename[0] != '/') { // Relative file path.
			// Get absolute path of the current zone file.
			if (realpath(s->file_name, (char*)(s->buffer)) != NULL) {
				char *full_current_zone_file_name =
					strdup((char*)(s->buffer));

				// Creating full include file name.
				snprintf((char*)(s->buffer), sizeof(s->buffer),
				        "%s/%s",
				        dirname(full_current_zone_file_name),
				        s->include_filename);

				free(full_current_zone_file_name);
			} else {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new file loader for included zone file.
		file_loader_t *fl = file_loader_create((char*)(s->buffer),
		                                       text_origin,
		                                       DEFAULT_CLASS,
		                                       DEFAULT_TTL,
		                                       s->process_record,
		                                       s->process_error,
		                                       s->data);
		if (fl != NULL) {
			// Process included zone file.
			ret = file_loader_process(fl);
			file_loader_free(fl);

			if (ret != 0) {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			SCANNER_ERROR(ZSCANNER_EUNOPENED_INCLUDE);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1076;
tr931:
#line 645 "./zscanner/scanner_body.rl"
	{
		char text_origin[MAX_DNAME_LENGTH];

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
					  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
					  text_origin);
		}

		if (s->include_filename[0] != '/') { // Relative file path.
			// Get absolute path of the current zone file.
			if (realpath(s->file_name, (char*)(s->buffer)) != NULL) {
				char *full_current_zone_file_name =
					strdup((char*)(s->buffer));

				// Creating full include file name.
				snprintf((char*)(s->buffer), sizeof(s->buffer),
				        "%s/%s",
				        dirname(full_current_zone_file_name),
				        s->include_filename);

				free(full_current_zone_file_name);
			} else {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new file loader for included zone file.
		file_loader_t *fl = file_loader_create((char*)(s->buffer),
		                                       text_origin,
		                                       DEFAULT_CLASS,
		                                       DEFAULT_TTL,
		                                       s->process_record,
		                                       s->process_error,
		                                       s->data);
		if (fl != NULL) {
			// Process included zone file.
			ret = file_loader_process(fl);
			file_loader_free(fl);

			if (ret != 0) {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			SCANNER_ERROR(ZSCANNER_EUNOPENED_INCLUDE);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1076;
tr942:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 637 "./zscanner/scanner_body.rl"
	{
		s->r_data_length = s->dname_tmp_length;
	}
#line 645 "./zscanner/scanner_body.rl"
	{
		char text_origin[MAX_DNAME_LENGTH];

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
					  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
					  text_origin);
		}

		if (s->include_filename[0] != '/') { // Relative file path.
			// Get absolute path of the current zone file.
			if (realpath(s->file_name, (char*)(s->buffer)) != NULL) {
				char *full_current_zone_file_name =
					strdup((char*)(s->buffer));

				// Creating full include file name.
				snprintf((char*)(s->buffer), sizeof(s->buffer),
				        "%s/%s",
				        dirname(full_current_zone_file_name),
				        s->include_filename);

				free(full_current_zone_file_name);
			} else {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new file loader for included zone file.
		file_loader_t *fl = file_loader_create((char*)(s->buffer),
		                                       text_origin,
		                                       DEFAULT_CLASS,
		                                       DEFAULT_TTL,
		                                       s->process_record,
		                                       s->process_error,
		                                       s->data);
		if (fl != NULL) {
			// Process included zone file.
			ret = file_loader_process(fl);
			file_loader_free(fl);

			if (ret != 0) {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			SCANNER_ERROR(ZSCANNER_EUNOPENED_INCLUDE);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1076;
st1076:
	if ( ++p == pe )
		goto _test_eof1076;
case 1076:
#line 35003 "zscanner/scanner.c"
	goto st0;
tr921:
#line 614 "./zscanner/scanner_body.rl"
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
			p--; {goto st247;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	goto st300;
tr943:
#line 162 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length++] = 0;
	}
#line 637 "./zscanner/scanner_body.rl"
	{
		s->r_data_length = s->dname_tmp_length;
	}
	goto st300;
st300:
	if ( ++p == pe )
		goto _test_eof300;
case 300:
#line 35037 "zscanner/scanner.c"
	if ( (*p) == 10 )
		goto tr931;
	goto st300;
tr940:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st301;
tr947:
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st301;
tr953:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st301;
tr928:
#line 634 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_data;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
#line 108 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st301;
st301:
	if ( ++p == pe )
		goto _test_eof301;
case 301:
#line 35116 "zscanner/scanner.c"
	switch( (*p) ) {
		case 46: goto tr935;
		case 92: goto st302;
		case 95: goto tr947;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr947;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr947;
	} else
		goto tr947;
	goto tr934;
tr941:
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st302;
tr955:
#line 140 "./zscanner/scanner_body.rl"
	{
		s->dname_tmp_length++;
	}
	goto st302;
tr930:
#line 634 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_data;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
#line 104 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st302;
st302:
	if ( ++p == pe )
		goto _test_eof302;
case 302:
#line 35164 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr949;
	goto tr947;
tr949:
#line 127 "./zscanner/scanner_body.rl"
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			SCANNER_WARNING(ZSCANNER_ELABEL_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st303;
st303:
	if ( ++p == pe )
		goto _test_eof303;
case 303:
#line 35189 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr951;
	goto tr950;
tr951:
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st304;
st304:
	if ( ++p == pe )
		goto _test_eof304;
case 304:
#line 35204 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr952;
	goto tr950;
tr952:
#line 136 "./zscanner/scanner_body.rl"
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st305;
st305:
	if ( ++p == pe )
		goto _test_eof305;
case 305:
#line 35219 "zscanner/scanner.c"
	switch( (*p) ) {
		case 46: goto tr954;
		case 92: goto tr955;
		case 95: goto tr953;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr953;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr953;
	} else
		goto tr953;
	goto tr950;
tr929:
#line 634 "./zscanner/scanner_body.rl"
	{
		s->dname = s->r_data;
	}
#line 185 "./zscanner/scanner_body.rl"
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st306;
st306:
	if ( ++p == pe )
		goto _test_eof306;
case 306:
#line 35249 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr936;
		case 32: goto tr936;
		case 40: goto tr937;
		case 41: goto tr938;
		case 778: goto tr942;
		case 827: goto tr943;
		case 1034: goto tr942;
		case 1083: goto tr943;
	}
	goto tr923;
tr922:
#line 614 "./zscanner/scanner_body.rl"
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
			p--; {goto st247;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	goto st307;
st307:
	if ( ++p == pe )
		goto _test_eof307;
case 307:
#line 35297 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr931;
		case 1034: goto tr956;
	}
	if ( _widec > 1023 ) {
		if ( 1024 <= _widec && _widec <= 1279 )
			goto st307;
	} else if ( _widec >= 768 )
		goto st300;
	goto tr84;
tr956:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 645 "./zscanner/scanner_body.rl"
	{
		char text_origin[MAX_DNAME_LENGTH];

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
					  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
					  text_origin);
		}

		if (s->include_filename[0] != '/') { // Relative file path.
			// Get absolute path of the current zone file.
			if (realpath(s->file_name, (char*)(s->buffer)) != NULL) {
				char *full_current_zone_file_name =
					strdup((char*)(s->buffer));

				// Creating full include file name.
				snprintf((char*)(s->buffer), sizeof(s->buffer),
				        "%s/%s",
				        dirname(full_current_zone_file_name),
				        s->include_filename);

				free(full_current_zone_file_name);
			} else {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new file loader for included zone file.
		file_loader_t *fl = file_loader_create((char*)(s->buffer),
		                                       text_origin,
		                                       DEFAULT_CLASS,
		                                       DEFAULT_TTL,
		                                       s->process_record,
		                                       s->process_error,
		                                       s->data);
		if (fl != NULL) {
			// Process included zone file.
			ret = file_loader_process(fl);
			file_loader_free(fl);

			if (ret != 0) {
				SCANNER_ERROR(ZSCANNER_EUNPROCESSED_INCLUDE);
				p--; {goto st247;}
			}
		} else {
			SCANNER_ERROR(ZSCANNER_EUNOPENED_INCLUDE);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1077;
st1077:
	if ( ++p == pe )
		goto _test_eof1077;
case 1077:
#line 35402 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st308;
		case 32: goto st308;
		case 40: goto tr958;
		case 41: goto tr959;
		case 42: goto tr927;
		case 46: goto tr929;
		case 92: goto tr930;
		case 95: goto tr928;
		case 1034: goto tr960;
		case 1083: goto st309;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr928;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr928;
	} else
		goto tr928;
	goto tr934;
tr958:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st308;
tr959:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st308;
tr960:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st308;
st308:
	if ( ++p == pe )
		goto _test_eof308;
case 308:
#line 35468 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st308;
		case 32: goto st308;
		case 40: goto tr958;
		case 41: goto tr959;
		case 42: goto tr927;
		case 46: goto tr929;
		case 92: goto tr930;
		case 95: goto tr928;
		case 1034: goto tr960;
		case 1083: goto st309;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr928;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr928;
	} else
		goto tr928;
	goto tr934;
st309:
	if ( ++p == pe )
		goto _test_eof309;
case 309:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr960;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st309;
	goto st0;
st310:
	if ( ++p == pe )
		goto _test_eof310;
case 310:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr912;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st310;
	goto st0;
st311:
	if ( ++p == pe )
		goto _test_eof311;
case 311:
	if ( (*p) == 43 )
		goto tr963;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr963;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr963;
	} else
		goto tr963;
	goto tr962;
tr963:
#line 942 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base64_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st312;
st312:
	if ( ++p == pe )
		goto _test_eof312;
case 312:
#line 35594 "zscanner/scanner.c"
	if ( (*p) == 43 )
		goto tr964;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr964;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr964;
	} else
		goto tr964;
	goto tr962;
tr964:
#line 950 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += second_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base64_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st313;
st313:
	if ( ++p == pe )
		goto _test_eof313;
case 313:
#line 35623 "zscanner/scanner.c"
	switch( (*p) ) {
		case 43: goto tr965;
		case 61: goto st317;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr965;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr965;
	} else
		goto tr965;
	goto tr962;
tr965:
#line 960 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += third_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = third_right_base64_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st314;
st314:
	if ( ++p == pe )
		goto _test_eof314;
case 314:
#line 35654 "zscanner/scanner.c"
	switch( (*p) ) {
		case 43: goto tr967;
		case 61: goto st315;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr967;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr967;
	} else
		goto tr967;
	goto tr962;
tr969:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st315;
tr970:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st315;
tr971:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st315;
tr967:
#line 970 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += fourth_base64_to_num[(uint8_t)(*p)];
	}
	goto st315;
st315:
	if ( ++p == pe )
		goto _test_eof315;
case 315:
#line 35704 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st315;
		case 32: goto st315;
		case 40: goto tr969;
		case 41: goto tr970;
		case 43: goto tr963;
		case 2058: goto tr971;
		case 2107: goto st316;
		case 2314: goto tr973;
		case 2363: goto tr973;
		case 2570: goto tr974;
		case 2619: goto tr975;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr963;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr963;
	} else
		goto tr963;
	goto tr962;
st316:
	if ( ++p == pe )
		goto _test_eof316;
case 316:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr971;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st316;
	goto tr962;
tr973:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1078;
st1078:
	if ( ++p == pe )
		goto _test_eof1078;
case 1078:
#line 35787 "zscanner/scanner.c"
	goto st0;
tr974:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1079;
st1079:
	if ( ++p == pe )
		goto _test_eof1079;
case 1079:
#line 35803 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st315;
		case 32: goto st315;
		case 40: goto tr969;
		case 41: goto tr970;
		case 43: goto tr963;
		case 2058: goto tr971;
		case 2107: goto st316;
		case 2314: goto tr973;
		case 2363: goto tr973;
		case 2570: goto tr974;
		case 2619: goto tr975;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr963;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr963;
	} else
		goto tr963;
	goto tr962;
tr975:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1080;
st1080:
	if ( ++p == pe )
		goto _test_eof1080;
case 1080:
#line 35856 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr971;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st316;
	goto tr962;
st317:
	if ( ++p == pe )
		goto _test_eof317;
case 317:
	if ( (*p) == 61 )
		goto st315;
	goto tr962;
st318:
	if ( ++p == pe )
		goto _test_eof318;
case 318:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr977;
		case 32: goto tr977;
		case 40: goto tr978;
		case 41: goto tr979;
		case 2058: goto tr980;
		case 2107: goto tr981;
		case 2314: goto tr982;
		case 2363: goto tr982;
		case 2570: goto tr983;
		case 2619: goto tr984;
	}
	goto tr976;
tr986:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr987:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1002:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr977:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	goto st319;
tr978:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr979:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr980:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1007:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	goto st319;
tr1008:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1009:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1013:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1020:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	goto st319;
tr1021:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1022:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1023:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1032:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	goto st319;
tr1033:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1034:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1035:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1043:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	goto st319;
tr1044:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1045:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1046:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1057:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	goto st319;
tr1058:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1059:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1060:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1069:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	goto st319;
tr1070:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1071:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1072:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1080:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	goto st319;
tr1081:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1082:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1083:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1088:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	goto st319;
tr1089:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1090:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1091:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1101:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	goto st319;
tr1102:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1103:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1104:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1110:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	goto st319;
tr1111:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1112:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1113:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1122:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	goto st319;
tr1123:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1124:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1125:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1137:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	goto st319;
tr1138:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1139:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1140:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1148:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	goto st319;
tr1149:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1150:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1151:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1156:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	goto st319;
tr1157:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1158:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1159:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1169:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	goto st319;
tr1170:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1171:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1172:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1178:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	goto st319;
tr1179:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1180:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1181:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1187:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	goto st319;
tr1188:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1189:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1190:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1195:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	goto st319;
tr1196:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1197:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1198:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1208:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	goto st319;
tr1209:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1210:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1211:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1216:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	goto st319;
tr1217:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1218:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1219:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1230:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	goto st319;
tr1231:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1232:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1233:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1239:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	goto st319;
tr1240:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1241:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1242:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1247:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	goto st319;
tr1248:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1249:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1251:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1257:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	goto st319;
tr1258:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1259:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1261:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1266:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	goto st319;
tr1267:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1268:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1270:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1279:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	goto st319;
tr1280:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1281:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1282:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1289:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	goto st319;
tr1290:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1291:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1292:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1300:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	goto st319;
tr1301:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1302:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1303:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1311:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	goto st319;
tr1312:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1313:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1314:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1319:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	goto st319;
tr1320:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1321:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1322:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1332:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	goto st319;
tr1333:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1334:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1335:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1341:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	goto st319;
tr1342:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1343:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1344:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1350:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	goto st319;
tr1351:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1352:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1353:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1361:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	goto st319;
tr1362:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1363:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1364:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1374:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	goto st319;
tr1375:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1376:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1377:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1383:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	goto st319;
tr1384:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1385:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1386:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1394:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st319;
tr1395:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1396:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1398:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1406:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	goto st319;
tr1407:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1408:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1409:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
tr1415:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	goto st319;
tr1416:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st319;
tr1417:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st319;
tr1418:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st319;
st319:
	if ( ++p == pe )
		goto _test_eof319;
case 319:
#line 37437 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st319;
		case 32: goto st319;
		case 40: goto tr986;
		case 41: goto tr987;
		case 65: goto st320;
		case 67: goto st325;
		case 68: goto st333;
		case 69: goto st347;
		case 72: goto st354;
		case 73: goto st359;
		case 75: goto st367;
		case 76: goto st371;
		case 77: goto st379;
		case 78: goto st385;
		case 80: goto st401;
		case 82: goto st404;
		case 83: goto st411;
		case 84: goto st422;
		case 97: goto st320;
		case 99: goto st325;
		case 100: goto st333;
		case 101: goto st347;
		case 104: goto st354;
		case 105: goto st359;
		case 107: goto st367;
		case 108: goto st371;
		case 109: goto st379;
		case 110: goto st385;
		case 112: goto st401;
		case 114: goto st404;
		case 115: goto st411;
		case 116: goto st422;
		case 2058: goto tr1002;
		case 2107: goto st324;
		case 2314: goto tr1004;
		case 2363: goto tr1004;
		case 2570: goto tr1005;
		case 2619: goto tr1006;
	}
	goto tr976;
st320:
	if ( ++p == pe )
		goto _test_eof320;
case 320:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1007;
		case 32: goto tr1007;
		case 40: goto tr1008;
		case 41: goto tr1009;
		case 65: goto st321;
		case 70: goto st432;
		case 80: goto st436;
		case 97: goto st321;
		case 102: goto st432;
		case 112: goto st436;
		case 2058: goto tr1013;
		case 2107: goto tr1014;
		case 2314: goto tr1015;
		case 2363: goto tr1015;
		case 2570: goto tr1016;
		case 2619: goto tr1017;
	}
	goto tr976;
st321:
	if ( ++p == pe )
		goto _test_eof321;
case 321:
	switch( (*p) ) {
		case 65: goto st322;
		case 97: goto st322;
	}
	goto tr976;
st322:
	if ( ++p == pe )
		goto _test_eof322;
case 322:
	switch( (*p) ) {
		case 65: goto st323;
		case 97: goto st323;
	}
	goto tr976;
st323:
	if ( ++p == pe )
		goto _test_eof323;
case 323:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1020;
		case 32: goto tr1020;
		case 40: goto tr1021;
		case 41: goto tr1022;
		case 2058: goto tr1023;
		case 2107: goto tr1024;
		case 2314: goto tr1025;
		case 2363: goto tr1025;
		case 2570: goto tr1026;
		case 2619: goto tr1027;
	}
	goto tr976;
tr981:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	goto st324;
tr1014:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	goto st324;
tr1024:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	goto st324;
tr1036:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	goto st324;
tr1047:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	goto st324;
tr1061:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	goto st324;
tr1073:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	goto st324;
tr1084:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	goto st324;
tr1092:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	goto st324;
tr1105:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	goto st324;
tr1114:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	goto st324;
tr1126:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	goto st324;
tr1141:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	goto st324;
tr1152:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	goto st324;
tr1160:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	goto st324;
tr1173:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	goto st324;
tr1182:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	goto st324;
tr1191:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	goto st324;
tr1199:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	goto st324;
tr1212:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	goto st324;
tr1220:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	goto st324;
tr1234:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	goto st324;
tr1243:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	goto st324;
tr1252:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	goto st324;
tr1262:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	goto st324;
tr1271:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	goto st324;
tr1283:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	goto st324;
tr1293:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	goto st324;
tr1304:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	goto st324;
tr1315:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	goto st324;
tr1323:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	goto st324;
tr1336:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	goto st324;
tr1345:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	goto st324;
tr1354:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	goto st324;
tr1365:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	goto st324;
tr1378:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	goto st324;
tr1387:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	goto st324;
tr1399:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st324;
tr1410:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	goto st324;
tr1419:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	goto st324;
st324:
	if ( ++p == pe )
		goto _test_eof324;
case 324:
#line 37771 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1002;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st324;
	goto tr976;
tr982:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1004:
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1015:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1025:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1037:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1048:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1062:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1074:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1085:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1093:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1106:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1115:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1127:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1142:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1153:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1161:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1174:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1183:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1192:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1200:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1213:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1221:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1235:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1244:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1253:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1263:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1272:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1284:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1294:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1305:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1316:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1324:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1337:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1346:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1355:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1366:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1379:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1388:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1400:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1411:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
tr1420:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1081;
st1081:
	if ( ++p == pe )
		goto _test_eof1081;
case 1081:
#line 39122 "zscanner/scanner.c"
	goto st0;
tr983:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1005:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1016:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1026:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1038:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1049:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1063:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1075:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1086:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1094:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1107:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1116:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1128:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1143:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1154:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1162:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1175:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1184:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1193:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1201:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1214:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1222:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1236:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1245:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1254:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1264:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1273:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1285:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1295:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1306:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1317:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1325:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1338:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1347:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1356:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1367:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1380:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1389:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1401:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1412:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
tr1421:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1082;
st1082:
	if ( ++p == pe )
		goto _test_eof1082;
case 1082:
#line 40612 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st319;
		case 32: goto st319;
		case 40: goto tr986;
		case 41: goto tr987;
		case 65: goto st320;
		case 67: goto st325;
		case 68: goto st333;
		case 69: goto st347;
		case 72: goto st354;
		case 73: goto st359;
		case 75: goto st367;
		case 76: goto st371;
		case 77: goto st379;
		case 78: goto st385;
		case 80: goto st401;
		case 82: goto st404;
		case 83: goto st411;
		case 84: goto st422;
		case 97: goto st320;
		case 99: goto st325;
		case 100: goto st333;
		case 101: goto st347;
		case 104: goto st354;
		case 105: goto st359;
		case 107: goto st367;
		case 108: goto st371;
		case 109: goto st379;
		case 110: goto st385;
		case 112: goto st401;
		case 114: goto st404;
		case 115: goto st411;
		case 116: goto st422;
		case 2058: goto tr1002;
		case 2107: goto st324;
		case 2314: goto tr1004;
		case 2363: goto tr1004;
		case 2570: goto tr1005;
		case 2619: goto tr1006;
	}
	goto tr976;
st325:
	if ( ++p == pe )
		goto _test_eof325;
case 325:
	switch( (*p) ) {
		case 69: goto st326;
		case 78: goto st329;
		case 101: goto st326;
		case 110: goto st329;
	}
	goto tr976;
st326:
	if ( ++p == pe )
		goto _test_eof326;
case 326:
	switch( (*p) ) {
		case 82: goto st327;
		case 114: goto st327;
	}
	goto tr976;
st327:
	if ( ++p == pe )
		goto _test_eof327;
case 327:
	switch( (*p) ) {
		case 84: goto st328;
		case 116: goto st328;
	}
	goto tr976;
st328:
	if ( ++p == pe )
		goto _test_eof328;
case 328:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1032;
		case 32: goto tr1032;
		case 40: goto tr1033;
		case 41: goto tr1034;
		case 2058: goto tr1035;
		case 2107: goto tr1036;
		case 2314: goto tr1037;
		case 2363: goto tr1037;
		case 2570: goto tr1038;
		case 2619: goto tr1039;
	}
	goto tr976;
tr984:
#line 1307 "./zscanner/scanner_body.rl"
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1006:
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1017:
#line 1266 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_A, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1027:
#line 1279 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1039:
#line 1284 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1050:
#line 1268 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1064:
#line 1293 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1076:
#line 1285 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1087:
#line 1292 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1095:
#line 1287 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1108:
#line 1302 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1117:
#line 1303 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1129:
#line 1271 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1144:
#line 1289 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1155:
#line 1278 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1163:
#line 1283 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1176:
#line 1299 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1185:
#line 1300 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1194:
#line 1280 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1202:
#line 1301 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1215:
#line 1272 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1223:
#line 1273 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1237:
#line 1282 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1246:
#line 1298 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1255:
#line 1267 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1265:
#line 1291 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1274:
#line 1294 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1286:
#line 1295 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1296:
#line 1270 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1307:
#line 1275 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1318:
#line 1290 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1326:
#line 1277 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1339:
#line 1269 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1348:
#line 1297 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1357:
#line 1281 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1368:
#line 1288 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1381:
#line 1296 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1390:
#line 1274 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1402:
#line 1253 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1413:
#line 1276 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
tr1422:
#line 1286 "./zscanner/scanner_body.rl"
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
#line 1311 "./zscanner/scanner_body.rl"
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
					p--; {goto st247;}
				}
			}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1083;
st1083:
	if ( ++p == pe )
		goto _test_eof1083;
case 1083:
#line 42064 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1002;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st324;
	goto tr976;
st329:
	if ( ++p == pe )
		goto _test_eof329;
case 329:
	switch( (*p) ) {
		case 65: goto st330;
		case 97: goto st330;
	}
	goto tr976;
st330:
	if ( ++p == pe )
		goto _test_eof330;
case 330:
	switch( (*p) ) {
		case 77: goto st331;
		case 109: goto st331;
	}
	goto tr976;
st331:
	if ( ++p == pe )
		goto _test_eof331;
case 331:
	switch( (*p) ) {
		case 69: goto st332;
		case 101: goto st332;
	}
	goto tr976;
st332:
	if ( ++p == pe )
		goto _test_eof332;
case 332:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1043;
		case 32: goto tr1043;
		case 40: goto tr1044;
		case 41: goto tr1045;
		case 2058: goto tr1046;
		case 2107: goto tr1047;
		case 2314: goto tr1048;
		case 2363: goto tr1048;
		case 2570: goto tr1049;
		case 2619: goto tr1050;
	}
	goto tr976;
st333:
	if ( ++p == pe )
		goto _test_eof333;
case 333:
	switch( (*p) ) {
		case 72: goto st334;
		case 78: goto st338;
		case 83: goto st346;
		case 104: goto st334;
		case 110: goto st338;
		case 115: goto st346;
	}
	goto tr976;
st334:
	if ( ++p == pe )
		goto _test_eof334;
case 334:
	switch( (*p) ) {
		case 67: goto st335;
		case 99: goto st335;
	}
	goto tr976;
st335:
	if ( ++p == pe )
		goto _test_eof335;
case 335:
	switch( (*p) ) {
		case 73: goto st336;
		case 105: goto st336;
	}
	goto tr976;
st336:
	if ( ++p == pe )
		goto _test_eof336;
case 336:
	switch( (*p) ) {
		case 68: goto st337;
		case 100: goto st337;
	}
	goto tr976;
st337:
	if ( ++p == pe )
		goto _test_eof337;
case 337:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1057;
		case 32: goto tr1057;
		case 40: goto tr1058;
		case 41: goto tr1059;
		case 2058: goto tr1060;
		case 2107: goto tr1061;
		case 2314: goto tr1062;
		case 2363: goto tr1062;
		case 2570: goto tr1063;
		case 2619: goto tr1064;
	}
	goto tr976;
st338:
	if ( ++p == pe )
		goto _test_eof338;
case 338:
	switch( (*p) ) {
		case 65: goto st339;
		case 83: goto st342;
		case 97: goto st339;
		case 115: goto st342;
	}
	goto tr976;
st339:
	if ( ++p == pe )
		goto _test_eof339;
case 339:
	switch( (*p) ) {
		case 77: goto st340;
		case 109: goto st340;
	}
	goto tr976;
st340:
	if ( ++p == pe )
		goto _test_eof340;
case 340:
	switch( (*p) ) {
		case 69: goto st341;
		case 101: goto st341;
	}
	goto tr976;
st341:
	if ( ++p == pe )
		goto _test_eof341;
case 341:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1069;
		case 32: goto tr1069;
		case 40: goto tr1070;
		case 41: goto tr1071;
		case 2058: goto tr1072;
		case 2107: goto tr1073;
		case 2314: goto tr1074;
		case 2363: goto tr1074;
		case 2570: goto tr1075;
		case 2619: goto tr1076;
	}
	goto tr976;
st342:
	if ( ++p == pe )
		goto _test_eof342;
case 342:
	switch( (*p) ) {
		case 75: goto st343;
		case 107: goto st343;
	}
	goto tr976;
st343:
	if ( ++p == pe )
		goto _test_eof343;
case 343:
	switch( (*p) ) {
		case 69: goto st344;
		case 101: goto st344;
	}
	goto tr976;
st344:
	if ( ++p == pe )
		goto _test_eof344;
case 344:
	switch( (*p) ) {
		case 89: goto st345;
		case 121: goto st345;
	}
	goto tr976;
st345:
	if ( ++p == pe )
		goto _test_eof345;
case 345:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1080;
		case 32: goto tr1080;
		case 40: goto tr1081;
		case 41: goto tr1082;
		case 2058: goto tr1083;
		case 2107: goto tr1084;
		case 2314: goto tr1085;
		case 2363: goto tr1085;
		case 2570: goto tr1086;
		case 2619: goto tr1087;
	}
	goto tr976;
st346:
	if ( ++p == pe )
		goto _test_eof346;
case 346:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1088;
		case 32: goto tr1088;
		case 40: goto tr1089;
		case 41: goto tr1090;
		case 2058: goto tr1091;
		case 2107: goto tr1092;
		case 2314: goto tr1093;
		case 2363: goto tr1093;
		case 2570: goto tr1094;
		case 2619: goto tr1095;
	}
	goto tr976;
st347:
	if ( ++p == pe )
		goto _test_eof347;
case 347:
	switch( (*p) ) {
		case 85: goto st348;
		case 117: goto st348;
	}
	goto tr976;
st348:
	if ( ++p == pe )
		goto _test_eof348;
case 348:
	switch( (*p) ) {
		case 73: goto st349;
		case 105: goto st349;
	}
	goto tr976;
st349:
	if ( ++p == pe )
		goto _test_eof349;
case 349:
	switch( (*p) ) {
		case 52: goto st350;
		case 54: goto st352;
	}
	goto tr976;
st350:
	if ( ++p == pe )
		goto _test_eof350;
case 350:
	if ( (*p) == 56 )
		goto st351;
	goto tr976;
st351:
	if ( ++p == pe )
		goto _test_eof351;
case 351:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1101;
		case 32: goto tr1101;
		case 40: goto tr1102;
		case 41: goto tr1103;
		case 2058: goto tr1104;
		case 2107: goto tr1105;
		case 2314: goto tr1106;
		case 2363: goto tr1106;
		case 2570: goto tr1107;
		case 2619: goto tr1108;
	}
	goto tr976;
st352:
	if ( ++p == pe )
		goto _test_eof352;
case 352:
	if ( (*p) == 52 )
		goto st353;
	goto tr976;
st353:
	if ( ++p == pe )
		goto _test_eof353;
case 353:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1110;
		case 32: goto tr1110;
		case 40: goto tr1111;
		case 41: goto tr1112;
		case 2058: goto tr1113;
		case 2107: goto tr1114;
		case 2314: goto tr1115;
		case 2363: goto tr1115;
		case 2570: goto tr1116;
		case 2619: goto tr1117;
	}
	goto tr976;
st354:
	if ( ++p == pe )
		goto _test_eof354;
case 354:
	switch( (*p) ) {
		case 73: goto st355;
		case 105: goto st355;
	}
	goto tr976;
st355:
	if ( ++p == pe )
		goto _test_eof355;
case 355:
	switch( (*p) ) {
		case 78: goto st356;
		case 110: goto st356;
	}
	goto tr976;
st356:
	if ( ++p == pe )
		goto _test_eof356;
case 356:
	switch( (*p) ) {
		case 70: goto st357;
		case 102: goto st357;
	}
	goto tr976;
st357:
	if ( ++p == pe )
		goto _test_eof357;
case 357:
	switch( (*p) ) {
		case 79: goto st358;
		case 111: goto st358;
	}
	goto tr976;
st358:
	if ( ++p == pe )
		goto _test_eof358;
case 358:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1122;
		case 32: goto tr1122;
		case 40: goto tr1123;
		case 41: goto tr1124;
		case 2058: goto tr1125;
		case 2107: goto tr1126;
		case 2314: goto tr1127;
		case 2363: goto tr1127;
		case 2570: goto tr1128;
		case 2619: goto tr1129;
	}
	goto tr976;
st359:
	if ( ++p == pe )
		goto _test_eof359;
case 359:
	switch( (*p) ) {
		case 80: goto st360;
		case 112: goto st360;
	}
	goto tr976;
st360:
	if ( ++p == pe )
		goto _test_eof360;
case 360:
	switch( (*p) ) {
		case 83: goto st361;
		case 115: goto st361;
	}
	goto tr976;
st361:
	if ( ++p == pe )
		goto _test_eof361;
case 361:
	switch( (*p) ) {
		case 69: goto st362;
		case 101: goto st362;
	}
	goto tr976;
st362:
	if ( ++p == pe )
		goto _test_eof362;
case 362:
	switch( (*p) ) {
		case 67: goto st363;
		case 99: goto st363;
	}
	goto tr976;
st363:
	if ( ++p == pe )
		goto _test_eof363;
case 363:
	switch( (*p) ) {
		case 75: goto st364;
		case 107: goto st364;
	}
	goto tr976;
st364:
	if ( ++p == pe )
		goto _test_eof364;
case 364:
	switch( (*p) ) {
		case 69: goto st365;
		case 101: goto st365;
	}
	goto tr976;
st365:
	if ( ++p == pe )
		goto _test_eof365;
case 365:
	switch( (*p) ) {
		case 89: goto st366;
		case 121: goto st366;
	}
	goto tr976;
st366:
	if ( ++p == pe )
		goto _test_eof366;
case 366:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1137;
		case 32: goto tr1137;
		case 40: goto tr1138;
		case 41: goto tr1139;
		case 2058: goto tr1140;
		case 2107: goto tr1141;
		case 2314: goto tr1142;
		case 2363: goto tr1142;
		case 2570: goto tr1143;
		case 2619: goto tr1144;
	}
	goto tr976;
st367:
	if ( ++p == pe )
		goto _test_eof367;
case 367:
	switch( (*p) ) {
		case 69: goto st368;
		case 88: goto st370;
		case 101: goto st368;
		case 120: goto st370;
	}
	goto tr976;
st368:
	if ( ++p == pe )
		goto _test_eof368;
case 368:
	switch( (*p) ) {
		case 89: goto st369;
		case 121: goto st369;
	}
	goto tr976;
st369:
	if ( ++p == pe )
		goto _test_eof369;
case 369:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1148;
		case 32: goto tr1148;
		case 40: goto tr1149;
		case 41: goto tr1150;
		case 2058: goto tr1151;
		case 2107: goto tr1152;
		case 2314: goto tr1153;
		case 2363: goto tr1153;
		case 2570: goto tr1154;
		case 2619: goto tr1155;
	}
	goto tr976;
st370:
	if ( ++p == pe )
		goto _test_eof370;
case 370:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1156;
		case 32: goto tr1156;
		case 40: goto tr1157;
		case 41: goto tr1158;
		case 2058: goto tr1159;
		case 2107: goto tr1160;
		case 2314: goto tr1161;
		case 2363: goto tr1161;
		case 2570: goto tr1162;
		case 2619: goto tr1163;
	}
	goto tr976;
st371:
	if ( ++p == pe )
		goto _test_eof371;
case 371:
	switch( (*p) ) {
		case 51: goto st372;
		case 54: goto st374;
		case 79: goto st376;
		case 80: goto st378;
		case 111: goto st376;
		case 112: goto st378;
	}
	goto tr976;
st372:
	if ( ++p == pe )
		goto _test_eof372;
case 372:
	if ( (*p) == 50 )
		goto st373;
	goto tr976;
st373:
	if ( ++p == pe )
		goto _test_eof373;
case 373:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1169;
		case 32: goto tr1169;
		case 40: goto tr1170;
		case 41: goto tr1171;
		case 2058: goto tr1172;
		case 2107: goto tr1173;
		case 2314: goto tr1174;
		case 2363: goto tr1174;
		case 2570: goto tr1175;
		case 2619: goto tr1176;
	}
	goto tr976;
st374:
	if ( ++p == pe )
		goto _test_eof374;
case 374:
	if ( (*p) == 52 )
		goto st375;
	goto tr976;
st375:
	if ( ++p == pe )
		goto _test_eof375;
case 375:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1178;
		case 32: goto tr1178;
		case 40: goto tr1179;
		case 41: goto tr1180;
		case 2058: goto tr1181;
		case 2107: goto tr1182;
		case 2314: goto tr1183;
		case 2363: goto tr1183;
		case 2570: goto tr1184;
		case 2619: goto tr1185;
	}
	goto tr976;
st376:
	if ( ++p == pe )
		goto _test_eof376;
case 376:
	switch( (*p) ) {
		case 67: goto st377;
		case 99: goto st377;
	}
	goto tr976;
st377:
	if ( ++p == pe )
		goto _test_eof377;
case 377:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1187;
		case 32: goto tr1187;
		case 40: goto tr1188;
		case 41: goto tr1189;
		case 2058: goto tr1190;
		case 2107: goto tr1191;
		case 2314: goto tr1192;
		case 2363: goto tr1192;
		case 2570: goto tr1193;
		case 2619: goto tr1194;
	}
	goto tr976;
st378:
	if ( ++p == pe )
		goto _test_eof378;
case 378:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1195;
		case 32: goto tr1195;
		case 40: goto tr1196;
		case 41: goto tr1197;
		case 2058: goto tr1198;
		case 2107: goto tr1199;
		case 2314: goto tr1200;
		case 2363: goto tr1200;
		case 2570: goto tr1201;
		case 2619: goto tr1202;
	}
	goto tr976;
st379:
	if ( ++p == pe )
		goto _test_eof379;
case 379:
	switch( (*p) ) {
		case 73: goto st380;
		case 88: goto st384;
		case 105: goto st380;
		case 120: goto st384;
	}
	goto tr976;
st380:
	if ( ++p == pe )
		goto _test_eof380;
case 380:
	switch( (*p) ) {
		case 78: goto st381;
		case 110: goto st381;
	}
	goto tr976;
st381:
	if ( ++p == pe )
		goto _test_eof381;
case 381:
	switch( (*p) ) {
		case 70: goto st382;
		case 102: goto st382;
	}
	goto tr976;
st382:
	if ( ++p == pe )
		goto _test_eof382;
case 382:
	switch( (*p) ) {
		case 79: goto st383;
		case 111: goto st383;
	}
	goto tr976;
st383:
	if ( ++p == pe )
		goto _test_eof383;
case 383:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1208;
		case 32: goto tr1208;
		case 40: goto tr1209;
		case 41: goto tr1210;
		case 2058: goto tr1211;
		case 2107: goto tr1212;
		case 2314: goto tr1213;
		case 2363: goto tr1213;
		case 2570: goto tr1214;
		case 2619: goto tr1215;
	}
	goto tr976;
st384:
	if ( ++p == pe )
		goto _test_eof384;
case 384:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1216;
		case 32: goto tr1216;
		case 40: goto tr1217;
		case 41: goto tr1218;
		case 2058: goto tr1219;
		case 2107: goto tr1220;
		case 2314: goto tr1221;
		case 2363: goto tr1221;
		case 2570: goto tr1222;
		case 2619: goto tr1223;
	}
	goto tr976;
st385:
	if ( ++p == pe )
		goto _test_eof385;
case 385:
	switch( (*p) ) {
		case 65: goto st386;
		case 73: goto st390;
		case 83: goto st392;
		case 97: goto st386;
		case 105: goto st390;
		case 115: goto st392;
	}
	goto tr976;
st386:
	if ( ++p == pe )
		goto _test_eof386;
case 386:
	switch( (*p) ) {
		case 80: goto st387;
		case 112: goto st387;
	}
	goto tr976;
st387:
	if ( ++p == pe )
		goto _test_eof387;
case 387:
	switch( (*p) ) {
		case 84: goto st388;
		case 116: goto st388;
	}
	goto tr976;
st388:
	if ( ++p == pe )
		goto _test_eof388;
case 388:
	switch( (*p) ) {
		case 82: goto st389;
		case 114: goto st389;
	}
	goto tr976;
st389:
	if ( ++p == pe )
		goto _test_eof389;
case 389:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1230;
		case 32: goto tr1230;
		case 40: goto tr1231;
		case 41: goto tr1232;
		case 2058: goto tr1233;
		case 2107: goto tr1234;
		case 2314: goto tr1235;
		case 2363: goto tr1235;
		case 2570: goto tr1236;
		case 2619: goto tr1237;
	}
	goto tr976;
st390:
	if ( ++p == pe )
		goto _test_eof390;
case 390:
	switch( (*p) ) {
		case 68: goto st391;
		case 100: goto st391;
	}
	goto tr976;
st391:
	if ( ++p == pe )
		goto _test_eof391;
case 391:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1239;
		case 32: goto tr1239;
		case 40: goto tr1240;
		case 41: goto tr1241;
		case 2058: goto tr1242;
		case 2107: goto tr1243;
		case 2314: goto tr1244;
		case 2363: goto tr1244;
		case 2570: goto tr1245;
		case 2619: goto tr1246;
	}
	goto tr976;
st392:
	if ( ++p == pe )
		goto _test_eof392;
case 392:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1247;
		case 32: goto tr1247;
		case 40: goto tr1248;
		case 41: goto tr1249;
		case 69: goto st393;
		case 101: goto st393;
		case 2058: goto tr1251;
		case 2107: goto tr1252;
		case 2314: goto tr1253;
		case 2363: goto tr1253;
		case 2570: goto tr1254;
		case 2619: goto tr1255;
	}
	goto tr976;
st393:
	if ( ++p == pe )
		goto _test_eof393;
case 393:
	switch( (*p) ) {
		case 67: goto st394;
		case 99: goto st394;
	}
	goto tr976;
st394:
	if ( ++p == pe )
		goto _test_eof394;
case 394:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1257;
		case 32: goto tr1257;
		case 40: goto tr1258;
		case 41: goto tr1259;
		case 51: goto st395;
		case 2058: goto tr1261;
		case 2107: goto tr1262;
		case 2314: goto tr1263;
		case 2363: goto tr1263;
		case 2570: goto tr1264;
		case 2619: goto tr1265;
	}
	goto tr976;
st395:
	if ( ++p == pe )
		goto _test_eof395;
case 395:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1266;
		case 32: goto tr1266;
		case 40: goto tr1267;
		case 41: goto tr1268;
		case 80: goto st396;
		case 112: goto st396;
		case 2058: goto tr1270;
		case 2107: goto tr1271;
		case 2314: goto tr1272;
		case 2363: goto tr1272;
		case 2570: goto tr1273;
		case 2619: goto tr1274;
	}
	goto tr976;
st396:
	if ( ++p == pe )
		goto _test_eof396;
case 396:
	switch( (*p) ) {
		case 65: goto st397;
		case 97: goto st397;
	}
	goto tr976;
st397:
	if ( ++p == pe )
		goto _test_eof397;
case 397:
	switch( (*p) ) {
		case 82: goto st398;
		case 114: goto st398;
	}
	goto tr976;
st398:
	if ( ++p == pe )
		goto _test_eof398;
case 398:
	switch( (*p) ) {
		case 65: goto st399;
		case 97: goto st399;
	}
	goto tr976;
st399:
	if ( ++p == pe )
		goto _test_eof399;
case 399:
	switch( (*p) ) {
		case 77: goto st400;
		case 109: goto st400;
	}
	goto tr976;
st400:
	if ( ++p == pe )
		goto _test_eof400;
case 400:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1279;
		case 32: goto tr1279;
		case 40: goto tr1280;
		case 41: goto tr1281;
		case 2058: goto tr1282;
		case 2107: goto tr1283;
		case 2314: goto tr1284;
		case 2363: goto tr1284;
		case 2570: goto tr1285;
		case 2619: goto tr1286;
	}
	goto tr976;
st401:
	if ( ++p == pe )
		goto _test_eof401;
case 401:
	switch( (*p) ) {
		case 84: goto st402;
		case 116: goto st402;
	}
	goto tr976;
st402:
	if ( ++p == pe )
		goto _test_eof402;
case 402:
	switch( (*p) ) {
		case 82: goto st403;
		case 114: goto st403;
	}
	goto tr976;
st403:
	if ( ++p == pe )
		goto _test_eof403;
case 403:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1289;
		case 32: goto tr1289;
		case 40: goto tr1290;
		case 41: goto tr1291;
		case 2058: goto tr1292;
		case 2107: goto tr1293;
		case 2314: goto tr1294;
		case 2363: goto tr1294;
		case 2570: goto tr1295;
		case 2619: goto tr1296;
	}
	goto tr976;
st404:
	if ( ++p == pe )
		goto _test_eof404;
case 404:
	switch( (*p) ) {
		case 80: goto st405;
		case 82: goto st406;
		case 84: goto st410;
		case 112: goto st405;
		case 114: goto st406;
		case 116: goto st410;
	}
	goto tr976;
st405:
	if ( ++p == pe )
		goto _test_eof405;
case 405:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1300;
		case 32: goto tr1300;
		case 40: goto tr1301;
		case 41: goto tr1302;
		case 2058: goto tr1303;
		case 2107: goto tr1304;
		case 2314: goto tr1305;
		case 2363: goto tr1305;
		case 2570: goto tr1306;
		case 2619: goto tr1307;
	}
	goto tr976;
st406:
	if ( ++p == pe )
		goto _test_eof406;
case 406:
	switch( (*p) ) {
		case 83: goto st407;
		case 115: goto st407;
	}
	goto tr976;
st407:
	if ( ++p == pe )
		goto _test_eof407;
case 407:
	switch( (*p) ) {
		case 73: goto st408;
		case 105: goto st408;
	}
	goto tr976;
st408:
	if ( ++p == pe )
		goto _test_eof408;
case 408:
	switch( (*p) ) {
		case 71: goto st409;
		case 103: goto st409;
	}
	goto tr976;
st409:
	if ( ++p == pe )
		goto _test_eof409;
case 409:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1311;
		case 32: goto tr1311;
		case 40: goto tr1312;
		case 41: goto tr1313;
		case 2058: goto tr1314;
		case 2107: goto tr1315;
		case 2314: goto tr1316;
		case 2363: goto tr1316;
		case 2570: goto tr1317;
		case 2619: goto tr1318;
	}
	goto tr976;
st410:
	if ( ++p == pe )
		goto _test_eof410;
case 410:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1319;
		case 32: goto tr1319;
		case 40: goto tr1320;
		case 41: goto tr1321;
		case 2058: goto tr1322;
		case 2107: goto tr1323;
		case 2314: goto tr1324;
		case 2363: goto tr1324;
		case 2570: goto tr1325;
		case 2619: goto tr1326;
	}
	goto tr976;
st411:
	if ( ++p == pe )
		goto _test_eof411;
case 411:
	switch( (*p) ) {
		case 79: goto st412;
		case 80: goto st414;
		case 82: goto st416;
		case 83: goto st418;
		case 111: goto st412;
		case 112: goto st414;
		case 114: goto st416;
		case 115: goto st418;
	}
	goto tr976;
st412:
	if ( ++p == pe )
		goto _test_eof412;
case 412:
	switch( (*p) ) {
		case 65: goto st413;
		case 97: goto st413;
	}
	goto tr976;
st413:
	if ( ++p == pe )
		goto _test_eof413;
case 413:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1332;
		case 32: goto tr1332;
		case 40: goto tr1333;
		case 41: goto tr1334;
		case 2058: goto tr1335;
		case 2107: goto tr1336;
		case 2314: goto tr1337;
		case 2363: goto tr1337;
		case 2570: goto tr1338;
		case 2619: goto tr1339;
	}
	goto tr976;
st414:
	if ( ++p == pe )
		goto _test_eof414;
case 414:
	switch( (*p) ) {
		case 70: goto st415;
		case 102: goto st415;
	}
	goto tr976;
st415:
	if ( ++p == pe )
		goto _test_eof415;
case 415:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1341;
		case 32: goto tr1341;
		case 40: goto tr1342;
		case 41: goto tr1343;
		case 2058: goto tr1344;
		case 2107: goto tr1345;
		case 2314: goto tr1346;
		case 2363: goto tr1346;
		case 2570: goto tr1347;
		case 2619: goto tr1348;
	}
	goto tr976;
st416:
	if ( ++p == pe )
		goto _test_eof416;
case 416:
	switch( (*p) ) {
		case 86: goto st417;
		case 118: goto st417;
	}
	goto tr976;
st417:
	if ( ++p == pe )
		goto _test_eof417;
case 417:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1350;
		case 32: goto tr1350;
		case 40: goto tr1351;
		case 41: goto tr1352;
		case 2058: goto tr1353;
		case 2107: goto tr1354;
		case 2314: goto tr1355;
		case 2363: goto tr1355;
		case 2570: goto tr1356;
		case 2619: goto tr1357;
	}
	goto tr976;
st418:
	if ( ++p == pe )
		goto _test_eof418;
case 418:
	switch( (*p) ) {
		case 72: goto st419;
		case 104: goto st419;
	}
	goto tr976;
st419:
	if ( ++p == pe )
		goto _test_eof419;
case 419:
	switch( (*p) ) {
		case 70: goto st420;
		case 102: goto st420;
	}
	goto tr976;
st420:
	if ( ++p == pe )
		goto _test_eof420;
case 420:
	switch( (*p) ) {
		case 80: goto st421;
		case 112: goto st421;
	}
	goto tr976;
st421:
	if ( ++p == pe )
		goto _test_eof421;
case 421:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1361;
		case 32: goto tr1361;
		case 40: goto tr1362;
		case 41: goto tr1363;
		case 2058: goto tr1364;
		case 2107: goto tr1365;
		case 2314: goto tr1366;
		case 2363: goto tr1366;
		case 2570: goto tr1367;
		case 2619: goto tr1368;
	}
	goto tr976;
st422:
	if ( ++p == pe )
		goto _test_eof422;
case 422:
	switch( (*p) ) {
		case 76: goto st423;
		case 88: goto st426;
		case 89: goto st428;
		case 108: goto st423;
		case 120: goto st426;
		case 121: goto st428;
	}
	goto tr976;
st423:
	if ( ++p == pe )
		goto _test_eof423;
case 423:
	switch( (*p) ) {
		case 83: goto st424;
		case 115: goto st424;
	}
	goto tr976;
st424:
	if ( ++p == pe )
		goto _test_eof424;
case 424:
	switch( (*p) ) {
		case 65: goto st425;
		case 97: goto st425;
	}
	goto tr976;
st425:
	if ( ++p == pe )
		goto _test_eof425;
case 425:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1374;
		case 32: goto tr1374;
		case 40: goto tr1375;
		case 41: goto tr1376;
		case 2058: goto tr1377;
		case 2107: goto tr1378;
		case 2314: goto tr1379;
		case 2363: goto tr1379;
		case 2570: goto tr1380;
		case 2619: goto tr1381;
	}
	goto tr976;
st426:
	if ( ++p == pe )
		goto _test_eof426;
case 426:
	switch( (*p) ) {
		case 84: goto st427;
		case 116: goto st427;
	}
	goto tr976;
st427:
	if ( ++p == pe )
		goto _test_eof427;
case 427:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1383;
		case 32: goto tr1383;
		case 40: goto tr1384;
		case 41: goto tr1385;
		case 2058: goto tr1386;
		case 2107: goto tr1387;
		case 2314: goto tr1388;
		case 2363: goto tr1388;
		case 2570: goto tr1389;
		case 2619: goto tr1390;
	}
	goto tr976;
st428:
	if ( ++p == pe )
		goto _test_eof428;
case 428:
	switch( (*p) ) {
		case 80: goto st429;
		case 112: goto st429;
	}
	goto tr976;
st429:
	if ( ++p == pe )
		goto _test_eof429;
case 429:
	switch( (*p) ) {
		case 69: goto st430;
		case 101: goto st430;
	}
	goto tr976;
st430:
	if ( ++p == pe )
		goto _test_eof430;
case 430:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1393;
	goto tr976;
tr1393:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st431;
tr1397:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st431;
st431:
	if ( ++p == pe )
		goto _test_eof431;
case 431:
#line 44002 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1394;
		case 32: goto tr1394;
		case 40: goto tr1395;
		case 41: goto tr1396;
		case 2058: goto tr1398;
		case 2107: goto tr1399;
		case 2314: goto tr1400;
		case 2363: goto tr1400;
		case 2570: goto tr1401;
		case 2619: goto tr1402;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1397;
	goto tr976;
st432:
	if ( ++p == pe )
		goto _test_eof432;
case 432:
	switch( (*p) ) {
		case 83: goto st433;
		case 115: goto st433;
	}
	goto tr976;
st433:
	if ( ++p == pe )
		goto _test_eof433;
case 433:
	switch( (*p) ) {
		case 68: goto st434;
		case 100: goto st434;
	}
	goto tr976;
st434:
	if ( ++p == pe )
		goto _test_eof434;
case 434:
	switch( (*p) ) {
		case 66: goto st435;
		case 98: goto st435;
	}
	goto tr976;
st435:
	if ( ++p == pe )
		goto _test_eof435;
case 435:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1406;
		case 32: goto tr1406;
		case 40: goto tr1407;
		case 41: goto tr1408;
		case 2058: goto tr1409;
		case 2107: goto tr1410;
		case 2314: goto tr1411;
		case 2363: goto tr1411;
		case 2570: goto tr1412;
		case 2619: goto tr1413;
	}
	goto tr976;
st436:
	if ( ++p == pe )
		goto _test_eof436;
case 436:
	switch( (*p) ) {
		case 76: goto st437;
		case 108: goto st437;
	}
	goto tr976;
st437:
	if ( ++p == pe )
		goto _test_eof437;
case 437:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1415;
		case 32: goto tr1415;
		case 40: goto tr1416;
		case 41: goto tr1417;
		case 2058: goto tr1418;
		case 2107: goto tr1419;
		case 2314: goto tr1420;
		case 2363: goto tr1420;
		case 2570: goto tr1421;
		case 2619: goto tr1422;
	}
	goto tr976;
st438:
	if ( ++p == pe )
		goto _test_eof438;
case 438:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st439;
		case 32: goto st439;
		case 40: goto tr1425;
		case 41: goto tr1426;
		case 1034: goto tr1427;
		case 1083: goto st446;
	}
	goto tr1423;
tr1425:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st439;
tr1426:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st439;
tr1427:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st439;
st439:
	if ( ++p == pe )
		goto _test_eof439;
case 439:
#line 44205 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st439;
		case 32: goto st439;
		case 40: goto tr1425;
		case 41: goto tr1426;
		case 1034: goto tr1427;
		case 1083: goto st446;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1430;
	goto tr1429;
tr1430:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st440;
tr1434:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st440;
st440:
	if ( ++p == pe )
		goto _test_eof440;
case 440:
#line 44273 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1431;
		case 32: goto tr1431;
		case 40: goto tr1432;
		case 41: goto tr1433;
		case 1034: goto tr1435;
		case 1083: goto tr1436;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1434;
	goto tr1429;
tr1439:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st441;
tr1440:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st441;
tr1442:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st441;
tr1431:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st441;
tr1432:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st441;
tr1433:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st441;
tr1435:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st441;
st441:
	if ( ++p == pe )
		goto _test_eof441;
case 441:
#line 44393 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st441;
		case 32: goto st441;
		case 40: goto tr1439;
		case 41: goto tr1440;
		case 1034: goto tr1442;
		case 1083: goto st445;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1441;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1441;
	} else
		goto tr1441;
	goto tr1437;
tr1441:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st442;
st442:
	if ( ++p == pe )
		goto _test_eof442;
case 442:
#line 44440 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1444;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1444;
	} else
		goto tr1444;
	goto tr1437;
tr1446:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st443;
tr1447:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st443;
tr1448:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st443;
tr1444:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st443;
st443:
	if ( ++p == pe )
		goto _test_eof443;
case 443:
#line 44487 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st443;
		case 32: goto st443;
		case 40: goto tr1446;
		case 41: goto tr1447;
		case 2058: goto tr1448;
		case 2107: goto st444;
		case 2314: goto tr1450;
		case 2363: goto tr1450;
		case 2570: goto tr1451;
		case 2619: goto tr1452;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1441;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1441;
	} else
		goto tr1441;
	goto tr1437;
st444:
	if ( ++p == pe )
		goto _test_eof444;
case 444:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1448;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st444;
	goto tr1437;
tr1450:
#line 919 "./zscanner/scanner_body.rl"
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			SCANNER_WARNING(ZSCANNER_EBAD_RDATA_LENGTH);
			p--; {goto st247;}
		}

		ret = find_rdata_blocks(s);
		if (ret != KNOT_EOK) {
			SCANNER_WARNING(ret);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1084;
st1084:
	if ( ++p == pe )
		goto _test_eof1084;
case 1084:
#line 44582 "zscanner/scanner.c"
	goto st0;
tr1451:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 919 "./zscanner/scanner_body.rl"
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			SCANNER_WARNING(ZSCANNER_EBAD_RDATA_LENGTH);
			p--; {goto st247;}
		}

		ret = find_rdata_blocks(s);
		if (ret != KNOT_EOK) {
			SCANNER_WARNING(ret);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1085;
st1085:
	if ( ++p == pe )
		goto _test_eof1085;
case 1085:
#line 44611 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st443;
		case 32: goto st443;
		case 40: goto tr1446;
		case 41: goto tr1447;
		case 2058: goto tr1448;
		case 2107: goto st444;
		case 2314: goto tr1450;
		case 2363: goto tr1450;
		case 2570: goto tr1451;
		case 2619: goto tr1452;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1441;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1441;
	} else
		goto tr1441;
	goto tr1437;
tr1452:
#line 919 "./zscanner/scanner_body.rl"
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			SCANNER_WARNING(ZSCANNER_EBAD_RDATA_LENGTH);
			p--; {goto st247;}
		}

		ret = find_rdata_blocks(s);
		if (ret != KNOT_EOK) {
			SCANNER_WARNING(ret);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1086;
st1086:
	if ( ++p == pe )
		goto _test_eof1086;
case 1086:
#line 44676 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1448;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st444;
	goto tr1437;
tr1436:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st445;
st445:
	if ( ++p == pe )
		goto _test_eof445;
case 445:
#line 44718 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1442;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st445;
	goto tr1423;
st446:
	if ( ++p == pe )
		goto _test_eof446;
case 446:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1427;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st446;
	goto tr1423;
st447:
	if ( ++p == pe )
		goto _test_eof447;
case 447:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st448;
		case 32: goto st448;
		case 40: goto tr1454;
		case 41: goto tr1455;
		case 1034: goto tr1456;
		case 1083: goto st456;
	}
	goto tr1423;
tr1454:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st448;
tr1455:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st448;
tr1456:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st448;
st448:
	if ( ++p == pe )
		goto _test_eof448;
case 448:
#line 44832 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st448;
		case 32: goto st448;
		case 40: goto tr1454;
		case 41: goto tr1455;
		case 48: goto tr1458;
		case 1034: goto tr1456;
		case 1083: goto st456;
	}
	if ( 49 <= _widec && _widec <= 57 )
		goto tr1459;
	goto tr1429;
tr1458:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st449;
st449:
	if ( ++p == pe )
		goto _test_eof449;
case 449:
#line 44884 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1460;
		case 32: goto tr1460;
		case 40: goto tr1461;
		case 41: goto tr1462;
		case 778: goto tr1464;
		case 827: goto tr1464;
		case 1034: goto tr1465;
		case 1083: goto tr1466;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1463;
	goto tr1429;
tr1460:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1087;
tr1461:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1087;
tr1462:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1087;
tr1465:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1087;
st1087:
	if ( ++p == pe )
		goto _test_eof1087;
case 1087:
#line 44996 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st450;
		case 32: goto st450;
		case 40: goto tr1468;
		case 41: goto tr1469;
		case 1034: goto tr1471;
		case 1083: goto st454;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1470;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1470;
	} else
		goto tr1470;
	goto tr1437;
tr1468:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st450;
tr1469:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st450;
tr1471:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st450;
tr1482:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st450;
tr1483:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st450;
tr1484:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st450;
tr1485:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st450;
st450:
	if ( ++p == pe )
		goto _test_eof450;
case 450:
#line 45122 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st450;
		case 32: goto st450;
		case 40: goto tr1468;
		case 41: goto tr1469;
		case 1034: goto tr1471;
		case 1083: goto st454;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1470;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1470;
	} else
		goto tr1470;
	goto tr1437;
tr1470:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st451;
st451:
	if ( ++p == pe )
		goto _test_eof451;
case 451:
#line 45169 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1473;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1473;
	} else
		goto tr1473;
	goto tr1437;
tr1475:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st452;
tr1476:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st452;
tr1477:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st452;
tr1473:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st452;
st452:
	if ( ++p == pe )
		goto _test_eof452;
case 452:
#line 45216 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st452;
		case 32: goto st452;
		case 40: goto tr1475;
		case 41: goto tr1476;
		case 2058: goto tr1477;
		case 2107: goto st453;
		case 2314: goto tr1479;
		case 2363: goto tr1479;
		case 2570: goto tr1480;
		case 2619: goto tr1481;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1470;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1470;
	} else
		goto tr1470;
	goto tr1437;
st453:
	if ( ++p == pe )
		goto _test_eof453;
case 453:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1477;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st453;
	goto tr1437;
tr1464:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1088;
tr1479:
#line 919 "./zscanner/scanner_body.rl"
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			SCANNER_WARNING(ZSCANNER_EBAD_RDATA_LENGTH);
			p--; {goto st247;}
		}

		ret = find_rdata_blocks(s);
		if (ret != KNOT_EOK) {
			SCANNER_WARNING(ret);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1088;
st1088:
	if ( ++p == pe )
		goto _test_eof1088;
case 1088:
#line 45317 "zscanner/scanner.c"
	goto tr1423;
tr1480:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 919 "./zscanner/scanner_body.rl"
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			SCANNER_WARNING(ZSCANNER_EBAD_RDATA_LENGTH);
			p--; {goto st247;}
		}

		ret = find_rdata_blocks(s);
		if (ret != KNOT_EOK) {
			SCANNER_WARNING(ret);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1089;
st1089:
	if ( ++p == pe )
		goto _test_eof1089;
case 1089:
#line 45346 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st452;
		case 32: goto st452;
		case 40: goto tr1475;
		case 41: goto tr1476;
		case 2058: goto tr1477;
		case 2107: goto st453;
		case 2314: goto tr1479;
		case 2363: goto tr1479;
		case 2570: goto tr1480;
		case 2619: goto tr1481;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1470;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1470;
	} else
		goto tr1470;
	goto tr1437;
tr1481:
#line 919 "./zscanner/scanner_body.rl"
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			SCANNER_WARNING(ZSCANNER_EBAD_RDATA_LENGTH);
			p--; {goto st247;}
		}

		ret = find_rdata_blocks(s);
		if (ret != KNOT_EOK) {
			SCANNER_WARNING(ret);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1090;
st1090:
	if ( ++p == pe )
		goto _test_eof1090;
case 1090:
#line 45411 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1477;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st453;
	goto tr1437;
tr1486:
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st454;
st454:
	if ( ++p == pe )
		goto _test_eof454;
case 454:
#line 45453 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1471;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st454;
	goto tr1423;
tr1459:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st455;
tr1463:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st455;
st455:
	if ( ++p == pe )
		goto _test_eof455;
case 455:
#line 45522 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1482;
		case 32: goto tr1482;
		case 40: goto tr1483;
		case 41: goto tr1484;
		case 1034: goto tr1485;
		case 1083: goto tr1486;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1463;
	goto tr1429;
tr1466:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
#line 369 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1091;
st1091:
	if ( ++p == pe )
		goto _test_eof1091;
case 1091:
#line 45567 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1471;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st454;
	goto tr1423;
st456:
	if ( ++p == pe )
		goto _test_eof456;
case 456:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1456;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st456;
	goto tr1423;
st457:
	if ( ++p == pe )
		goto _test_eof457;
case 457:
	switch( (*p) ) {
		case 68: goto st459;
		case 69: goto st474;
		case 73: goto st504;
		case 80: goto st512;
		case 82: goto st525;
		case 100: goto st459;
		case 101: goto st474;
		case 105: goto st504;
		case 112: goto st512;
		case 114: goto st525;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1488;
	goto tr1487;
tr1488:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st458;
tr1495:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st458;
st458:
	if ( ++p == pe )
		goto _test_eof458;
case 458:
#line 45685 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1494;
		case 59: goto tr1494;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1494;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1495;
	} else
		goto tr1494;
	goto tr1487;
tr1494:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1498:
#line 1099 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1500:
#line 1102 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 3;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1512:
#line 1108 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 6;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1521:
#line 1120 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 12;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1535:
#line 1123 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 13;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1544:
#line 1126 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 14;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1552:
#line 1129 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 252;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1563:
#line 1132 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 253;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1566:
#line 1135 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 254;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1573:
#line 1096 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 1;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1579:
#line 1105 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 5;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1591:
#line 1111 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 7;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1594:
#line 1114 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 8;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
tr1597:
#line 1117 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 10;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1092;
st1092:
	if ( ++p == pe )
		goto _test_eof1092;
case 1092:
#line 45859 "zscanner/scanner.c"
	goto st0;
st459:
	if ( ++p == pe )
		goto _test_eof459;
case 459:
	switch( (*p) ) {
		case 72: goto st460;
		case 83: goto st461;
		case 104: goto st460;
		case 115: goto st461;
	}
	goto tr1487;
st460:
	if ( ++p == pe )
		goto _test_eof460;
case 460:
	switch( (*p) ) {
		case 32: goto tr1498;
		case 59: goto tr1498;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1498;
	} else if ( (*p) >= 9 )
		goto tr1498;
	goto tr1487;
st461:
	if ( ++p == pe )
		goto _test_eof461;
case 461:
	switch( (*p) ) {
		case 65: goto st462;
		case 97: goto st462;
	}
	goto tr1487;
st462:
	if ( ++p == pe )
		goto _test_eof462;
case 462:
	switch( (*p) ) {
		case 32: goto tr1500;
		case 45: goto st463;
		case 59: goto tr1500;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1500;
	} else if ( (*p) >= 9 )
		goto tr1500;
	goto tr1487;
st463:
	if ( ++p == pe )
		goto _test_eof463;
case 463:
	switch( (*p) ) {
		case 78: goto st464;
		case 110: goto st464;
	}
	goto tr1487;
st464:
	if ( ++p == pe )
		goto _test_eof464;
case 464:
	switch( (*p) ) {
		case 83: goto st465;
		case 115: goto st465;
	}
	goto tr1487;
st465:
	if ( ++p == pe )
		goto _test_eof465;
case 465:
	switch( (*p) ) {
		case 69: goto st466;
		case 101: goto st466;
	}
	goto tr1487;
st466:
	if ( ++p == pe )
		goto _test_eof466;
case 466:
	switch( (*p) ) {
		case 67: goto st467;
		case 99: goto st467;
	}
	goto tr1487;
st467:
	if ( ++p == pe )
		goto _test_eof467;
case 467:
	if ( (*p) == 51 )
		goto st468;
	goto tr1487;
st468:
	if ( ++p == pe )
		goto _test_eof468;
case 468:
	if ( (*p) == 45 )
		goto st469;
	goto tr1487;
st469:
	if ( ++p == pe )
		goto _test_eof469;
case 469:
	switch( (*p) ) {
		case 83: goto st470;
		case 115: goto st470;
	}
	goto tr1487;
st470:
	if ( ++p == pe )
		goto _test_eof470;
case 470:
	switch( (*p) ) {
		case 72: goto st471;
		case 104: goto st471;
	}
	goto tr1487;
st471:
	if ( ++p == pe )
		goto _test_eof471;
case 471:
	switch( (*p) ) {
		case 65: goto st472;
		case 97: goto st472;
	}
	goto tr1487;
st472:
	if ( ++p == pe )
		goto _test_eof472;
case 472:
	if ( (*p) == 49 )
		goto st473;
	goto tr1487;
st473:
	if ( ++p == pe )
		goto _test_eof473;
case 473:
	switch( (*p) ) {
		case 32: goto tr1512;
		case 59: goto tr1512;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1512;
	} else if ( (*p) >= 9 )
		goto tr1512;
	goto tr1487;
st474:
	if ( ++p == pe )
		goto _test_eof474;
case 474:
	switch( (*p) ) {
		case 67: goto st475;
		case 99: goto st475;
	}
	goto tr1487;
st475:
	if ( ++p == pe )
		goto _test_eof475;
case 475:
	switch( (*p) ) {
		case 67: goto st476;
		case 68: goto st482;
		case 99: goto st476;
		case 100: goto st482;
	}
	goto tr1487;
st476:
	if ( ++p == pe )
		goto _test_eof476;
case 476:
	if ( (*p) == 45 )
		goto st477;
	goto tr1487;
st477:
	if ( ++p == pe )
		goto _test_eof477;
case 477:
	switch( (*p) ) {
		case 71: goto st478;
		case 103: goto st478;
	}
	goto tr1487;
st478:
	if ( ++p == pe )
		goto _test_eof478;
case 478:
	switch( (*p) ) {
		case 79: goto st479;
		case 111: goto st479;
	}
	goto tr1487;
st479:
	if ( ++p == pe )
		goto _test_eof479;
case 479:
	switch( (*p) ) {
		case 83: goto st480;
		case 115: goto st480;
	}
	goto tr1487;
st480:
	if ( ++p == pe )
		goto _test_eof480;
case 480:
	switch( (*p) ) {
		case 84: goto st481;
		case 116: goto st481;
	}
	goto tr1487;
st481:
	if ( ++p == pe )
		goto _test_eof481;
case 481:
	switch( (*p) ) {
		case 32: goto tr1521;
		case 59: goto tr1521;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1521;
	} else if ( (*p) >= 9 )
		goto tr1521;
	goto tr1487;
st482:
	if ( ++p == pe )
		goto _test_eof482;
case 482:
	switch( (*p) ) {
		case 83: goto st483;
		case 115: goto st483;
	}
	goto tr1487;
st483:
	if ( ++p == pe )
		goto _test_eof483;
case 483:
	switch( (*p) ) {
		case 65: goto st484;
		case 97: goto st484;
	}
	goto tr1487;
st484:
	if ( ++p == pe )
		goto _test_eof484;
case 484:
	switch( (*p) ) {
		case 80: goto st485;
		case 112: goto st485;
	}
	goto tr1487;
st485:
	if ( ++p == pe )
		goto _test_eof485;
case 485:
	switch( (*p) ) {
		case 50: goto st486;
		case 51: goto st495;
	}
	goto tr1487;
st486:
	if ( ++p == pe )
		goto _test_eof486;
case 486:
	if ( (*p) == 53 )
		goto st487;
	goto tr1487;
st487:
	if ( ++p == pe )
		goto _test_eof487;
case 487:
	if ( (*p) == 54 )
		goto st488;
	goto tr1487;
st488:
	if ( ++p == pe )
		goto _test_eof488;
case 488:
	switch( (*p) ) {
		case 83: goto st489;
		case 115: goto st489;
	}
	goto tr1487;
st489:
	if ( ++p == pe )
		goto _test_eof489;
case 489:
	switch( (*p) ) {
		case 72: goto st490;
		case 104: goto st490;
	}
	goto tr1487;
st490:
	if ( ++p == pe )
		goto _test_eof490;
case 490:
	switch( (*p) ) {
		case 65: goto st491;
		case 97: goto st491;
	}
	goto tr1487;
st491:
	if ( ++p == pe )
		goto _test_eof491;
case 491:
	if ( (*p) == 50 )
		goto st492;
	goto tr1487;
st492:
	if ( ++p == pe )
		goto _test_eof492;
case 492:
	if ( (*p) == 53 )
		goto st493;
	goto tr1487;
st493:
	if ( ++p == pe )
		goto _test_eof493;
case 493:
	if ( (*p) == 54 )
		goto st494;
	goto tr1487;
st494:
	if ( ++p == pe )
		goto _test_eof494;
case 494:
	switch( (*p) ) {
		case 32: goto tr1535;
		case 59: goto tr1535;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1535;
	} else if ( (*p) >= 9 )
		goto tr1535;
	goto tr1487;
st495:
	if ( ++p == pe )
		goto _test_eof495;
case 495:
	if ( (*p) == 56 )
		goto st496;
	goto tr1487;
st496:
	if ( ++p == pe )
		goto _test_eof496;
case 496:
	if ( (*p) == 52 )
		goto st497;
	goto tr1487;
st497:
	if ( ++p == pe )
		goto _test_eof497;
case 497:
	switch( (*p) ) {
		case 83: goto st498;
		case 115: goto st498;
	}
	goto tr1487;
st498:
	if ( ++p == pe )
		goto _test_eof498;
case 498:
	switch( (*p) ) {
		case 72: goto st499;
		case 104: goto st499;
	}
	goto tr1487;
st499:
	if ( ++p == pe )
		goto _test_eof499;
case 499:
	switch( (*p) ) {
		case 65: goto st500;
		case 97: goto st500;
	}
	goto tr1487;
st500:
	if ( ++p == pe )
		goto _test_eof500;
case 500:
	if ( (*p) == 51 )
		goto st501;
	goto tr1487;
st501:
	if ( ++p == pe )
		goto _test_eof501;
case 501:
	if ( (*p) == 56 )
		goto st502;
	goto tr1487;
st502:
	if ( ++p == pe )
		goto _test_eof502;
case 502:
	if ( (*p) == 52 )
		goto st503;
	goto tr1487;
st503:
	if ( ++p == pe )
		goto _test_eof503;
case 503:
	switch( (*p) ) {
		case 32: goto tr1544;
		case 59: goto tr1544;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1544;
	} else if ( (*p) >= 9 )
		goto tr1544;
	goto tr1487;
st504:
	if ( ++p == pe )
		goto _test_eof504;
case 504:
	switch( (*p) ) {
		case 78: goto st505;
		case 110: goto st505;
	}
	goto tr1487;
st505:
	if ( ++p == pe )
		goto _test_eof505;
case 505:
	switch( (*p) ) {
		case 68: goto st506;
		case 100: goto st506;
	}
	goto tr1487;
st506:
	if ( ++p == pe )
		goto _test_eof506;
case 506:
	switch( (*p) ) {
		case 73: goto st507;
		case 105: goto st507;
	}
	goto tr1487;
st507:
	if ( ++p == pe )
		goto _test_eof507;
case 507:
	switch( (*p) ) {
		case 82: goto st508;
		case 114: goto st508;
	}
	goto tr1487;
st508:
	if ( ++p == pe )
		goto _test_eof508;
case 508:
	switch( (*p) ) {
		case 69: goto st509;
		case 101: goto st509;
	}
	goto tr1487;
st509:
	if ( ++p == pe )
		goto _test_eof509;
case 509:
	switch( (*p) ) {
		case 67: goto st510;
		case 99: goto st510;
	}
	goto tr1487;
st510:
	if ( ++p == pe )
		goto _test_eof510;
case 510:
	switch( (*p) ) {
		case 84: goto st511;
		case 116: goto st511;
	}
	goto tr1487;
st511:
	if ( ++p == pe )
		goto _test_eof511;
case 511:
	switch( (*p) ) {
		case 32: goto tr1552;
		case 59: goto tr1552;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1552;
	} else if ( (*p) >= 9 )
		goto tr1552;
	goto tr1487;
st512:
	if ( ++p == pe )
		goto _test_eof512;
case 512:
	switch( (*p) ) {
		case 82: goto st513;
		case 114: goto st513;
	}
	goto tr1487;
st513:
	if ( ++p == pe )
		goto _test_eof513;
case 513:
	switch( (*p) ) {
		case 73: goto st514;
		case 105: goto st514;
	}
	goto tr1487;
st514:
	if ( ++p == pe )
		goto _test_eof514;
case 514:
	switch( (*p) ) {
		case 86: goto st515;
		case 118: goto st515;
	}
	goto tr1487;
st515:
	if ( ++p == pe )
		goto _test_eof515;
case 515:
	switch( (*p) ) {
		case 65: goto st516;
		case 97: goto st516;
	}
	goto tr1487;
st516:
	if ( ++p == pe )
		goto _test_eof516;
case 516:
	switch( (*p) ) {
		case 84: goto st517;
		case 116: goto st517;
	}
	goto tr1487;
st517:
	if ( ++p == pe )
		goto _test_eof517;
case 517:
	switch( (*p) ) {
		case 69: goto st518;
		case 101: goto st518;
	}
	goto tr1487;
st518:
	if ( ++p == pe )
		goto _test_eof518;
case 518:
	switch( (*p) ) {
		case 68: goto st519;
		case 79: goto st522;
		case 100: goto st519;
		case 111: goto st522;
	}
	goto tr1487;
st519:
	if ( ++p == pe )
		goto _test_eof519;
case 519:
	switch( (*p) ) {
		case 78: goto st520;
		case 110: goto st520;
	}
	goto tr1487;
st520:
	if ( ++p == pe )
		goto _test_eof520;
case 520:
	switch( (*p) ) {
		case 83: goto st521;
		case 115: goto st521;
	}
	goto tr1487;
st521:
	if ( ++p == pe )
		goto _test_eof521;
case 521:
	switch( (*p) ) {
		case 32: goto tr1563;
		case 59: goto tr1563;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1563;
	} else if ( (*p) >= 9 )
		goto tr1563;
	goto tr1487;
st522:
	if ( ++p == pe )
		goto _test_eof522;
case 522:
	switch( (*p) ) {
		case 73: goto st523;
		case 105: goto st523;
	}
	goto tr1487;
st523:
	if ( ++p == pe )
		goto _test_eof523;
case 523:
	switch( (*p) ) {
		case 68: goto st524;
		case 100: goto st524;
	}
	goto tr1487;
st524:
	if ( ++p == pe )
		goto _test_eof524;
case 524:
	switch( (*p) ) {
		case 32: goto tr1566;
		case 59: goto tr1566;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1566;
	} else if ( (*p) >= 9 )
		goto tr1566;
	goto tr1487;
st525:
	if ( ++p == pe )
		goto _test_eof525;
case 525:
	switch( (*p) ) {
		case 83: goto st526;
		case 115: goto st526;
	}
	goto tr1487;
st526:
	if ( ++p == pe )
		goto _test_eof526;
case 526:
	switch( (*p) ) {
		case 65: goto st527;
		case 97: goto st527;
	}
	goto tr1487;
st527:
	if ( ++p == pe )
		goto _test_eof527;
case 527:
	switch( (*p) ) {
		case 77: goto st528;
		case 83: goto st531;
		case 109: goto st528;
		case 115: goto st531;
	}
	goto tr1487;
st528:
	if ( ++p == pe )
		goto _test_eof528;
case 528:
	switch( (*p) ) {
		case 68: goto st529;
		case 100: goto st529;
	}
	goto tr1487;
st529:
	if ( ++p == pe )
		goto _test_eof529;
case 529:
	if ( (*p) == 53 )
		goto st530;
	goto tr1487;
st530:
	if ( ++p == pe )
		goto _test_eof530;
case 530:
	switch( (*p) ) {
		case 32: goto tr1573;
		case 59: goto tr1573;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1573;
	} else if ( (*p) >= 9 )
		goto tr1573;
	goto tr1487;
st531:
	if ( ++p == pe )
		goto _test_eof531;
case 531:
	switch( (*p) ) {
		case 72: goto st532;
		case 104: goto st532;
	}
	goto tr1487;
st532:
	if ( ++p == pe )
		goto _test_eof532;
case 532:
	switch( (*p) ) {
		case 65: goto st533;
		case 97: goto st533;
	}
	goto tr1487;
st533:
	if ( ++p == pe )
		goto _test_eof533;
case 533:
	switch( (*p) ) {
		case 49: goto st534;
		case 50: goto st546;
		case 53: goto st549;
	}
	goto tr1487;
st534:
	if ( ++p == pe )
		goto _test_eof534;
case 534:
	switch( (*p) ) {
		case 32: goto tr1579;
		case 45: goto st535;
		case 59: goto tr1579;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1579;
	} else if ( (*p) >= 9 )
		goto tr1579;
	goto tr1487;
st535:
	if ( ++p == pe )
		goto _test_eof535;
case 535:
	switch( (*p) ) {
		case 78: goto st536;
		case 110: goto st536;
	}
	goto tr1487;
st536:
	if ( ++p == pe )
		goto _test_eof536;
case 536:
	switch( (*p) ) {
		case 83: goto st537;
		case 115: goto st537;
	}
	goto tr1487;
st537:
	if ( ++p == pe )
		goto _test_eof537;
case 537:
	switch( (*p) ) {
		case 69: goto st538;
		case 101: goto st538;
	}
	goto tr1487;
st538:
	if ( ++p == pe )
		goto _test_eof538;
case 538:
	switch( (*p) ) {
		case 67: goto st539;
		case 99: goto st539;
	}
	goto tr1487;
st539:
	if ( ++p == pe )
		goto _test_eof539;
case 539:
	if ( (*p) == 51 )
		goto st540;
	goto tr1487;
st540:
	if ( ++p == pe )
		goto _test_eof540;
case 540:
	if ( (*p) == 45 )
		goto st541;
	goto tr1487;
st541:
	if ( ++p == pe )
		goto _test_eof541;
case 541:
	switch( (*p) ) {
		case 83: goto st542;
		case 115: goto st542;
	}
	goto tr1487;
st542:
	if ( ++p == pe )
		goto _test_eof542;
case 542:
	switch( (*p) ) {
		case 72: goto st543;
		case 104: goto st543;
	}
	goto tr1487;
st543:
	if ( ++p == pe )
		goto _test_eof543;
case 543:
	switch( (*p) ) {
		case 65: goto st544;
		case 97: goto st544;
	}
	goto tr1487;
st544:
	if ( ++p == pe )
		goto _test_eof544;
case 544:
	if ( (*p) == 49 )
		goto st545;
	goto tr1487;
st545:
	if ( ++p == pe )
		goto _test_eof545;
case 545:
	switch( (*p) ) {
		case 32: goto tr1591;
		case 59: goto tr1591;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1591;
	} else if ( (*p) >= 9 )
		goto tr1591;
	goto tr1487;
st546:
	if ( ++p == pe )
		goto _test_eof546;
case 546:
	if ( (*p) == 53 )
		goto st547;
	goto tr1487;
st547:
	if ( ++p == pe )
		goto _test_eof547;
case 547:
	if ( (*p) == 54 )
		goto st548;
	goto tr1487;
st548:
	if ( ++p == pe )
		goto _test_eof548;
case 548:
	switch( (*p) ) {
		case 32: goto tr1594;
		case 59: goto tr1594;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1594;
	} else if ( (*p) >= 9 )
		goto tr1594;
	goto tr1487;
st549:
	if ( ++p == pe )
		goto _test_eof549;
case 549:
	if ( (*p) == 49 )
		goto st550;
	goto tr1487;
st550:
	if ( ++p == pe )
		goto _test_eof550;
case 550:
	if ( (*p) == 50 )
		goto st551;
	goto tr1487;
st551:
	if ( ++p == pe )
		goto _test_eof551;
case 551:
	switch( (*p) ) {
		case 32: goto tr1597;
		case 59: goto tr1597;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1597;
	} else if ( (*p) >= 9 )
		goto tr1597;
	goto tr1487;
st552:
	if ( ++p == pe )
		goto _test_eof552;
case 552:
	switch( (*p) ) {
		case 65: goto st554;
		case 73: goto st560;
		case 79: goto st577;
		case 80: goto st580;
		case 83: goto st586;
		case 85: goto st590;
		case 97: goto st554;
		case 105: goto st560;
		case 111: goto st577;
		case 112: goto st580;
		case 115: goto st586;
		case 117: goto st590;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1599;
	goto tr1598;
tr1599:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st553;
tr1607:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st553;
st553:
	if ( ++p == pe )
		goto _test_eof553;
case 553:
#line 46798 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1606;
		case 59: goto tr1606;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1606;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1607;
	} else
		goto tr1606;
	goto tr1598;
tr1606:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1613:
#line 1163 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(7);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1622:
#line 1167 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(8);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1626:
#line 1159 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(6);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1629:
#line 1151 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(4);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1633:
#line 1155 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(5);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1636:
#line 1175 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(254);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1640:
#line 1147 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(3);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1643:
#line 1139 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(1);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1647:
#line 1143 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(2);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
tr1650:
#line 1171 "./zscanner/scanner_body.rl"
	{
		*((uint16_t *)rdata_tail) = htons(253);
		rdata_tail += 2;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1093;
st1093:
	if ( ++p == pe )
		goto _test_eof1093;
case 1093:
#line 46942 "zscanner/scanner.c"
	goto st0;
st554:
	if ( ++p == pe )
		goto _test_eof554;
case 554:
	switch( (*p) ) {
		case 67: goto st555;
		case 99: goto st555;
	}
	goto tr1598;
st555:
	if ( ++p == pe )
		goto _test_eof555;
case 555:
	switch( (*p) ) {
		case 80: goto st556;
		case 112: goto st556;
	}
	goto tr1598;
st556:
	if ( ++p == pe )
		goto _test_eof556;
case 556:
	switch( (*p) ) {
		case 75: goto st557;
		case 107: goto st557;
	}
	goto tr1598;
st557:
	if ( ++p == pe )
		goto _test_eof557;
case 557:
	switch( (*p) ) {
		case 73: goto st558;
		case 105: goto st558;
	}
	goto tr1598;
st558:
	if ( ++p == pe )
		goto _test_eof558;
case 558:
	switch( (*p) ) {
		case 88: goto st559;
		case 120: goto st559;
	}
	goto tr1598;
st559:
	if ( ++p == pe )
		goto _test_eof559;
case 559:
	switch( (*p) ) {
		case 32: goto tr1613;
		case 59: goto tr1613;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1613;
	} else if ( (*p) >= 9 )
		goto tr1613;
	goto tr1598;
st560:
	if ( ++p == pe )
		goto _test_eof560;
case 560:
	switch( (*p) ) {
		case 65: goto st561;
		case 80: goto st567;
		case 83: goto st573;
		case 97: goto st561;
		case 112: goto st567;
		case 115: goto st573;
	}
	goto tr1598;
st561:
	if ( ++p == pe )
		goto _test_eof561;
case 561:
	switch( (*p) ) {
		case 67: goto st562;
		case 99: goto st562;
	}
	goto tr1598;
st562:
	if ( ++p == pe )
		goto _test_eof562;
case 562:
	switch( (*p) ) {
		case 80: goto st563;
		case 112: goto st563;
	}
	goto tr1598;
st563:
	if ( ++p == pe )
		goto _test_eof563;
case 563:
	switch( (*p) ) {
		case 75: goto st564;
		case 107: goto st564;
	}
	goto tr1598;
st564:
	if ( ++p == pe )
		goto _test_eof564;
case 564:
	switch( (*p) ) {
		case 73: goto st565;
		case 105: goto st565;
	}
	goto tr1598;
st565:
	if ( ++p == pe )
		goto _test_eof565;
case 565:
	switch( (*p) ) {
		case 88: goto st566;
		case 120: goto st566;
	}
	goto tr1598;
st566:
	if ( ++p == pe )
		goto _test_eof566;
case 566:
	switch( (*p) ) {
		case 32: goto tr1622;
		case 59: goto tr1622;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1622;
	} else if ( (*p) >= 9 )
		goto tr1622;
	goto tr1598;
st567:
	if ( ++p == pe )
		goto _test_eof567;
case 567:
	switch( (*p) ) {
		case 71: goto st568;
		case 75: goto st570;
		case 103: goto st568;
		case 107: goto st570;
	}
	goto tr1598;
st568:
	if ( ++p == pe )
		goto _test_eof568;
case 568:
	switch( (*p) ) {
		case 80: goto st569;
		case 112: goto st569;
	}
	goto tr1598;
st569:
	if ( ++p == pe )
		goto _test_eof569;
case 569:
	switch( (*p) ) {
		case 32: goto tr1626;
		case 59: goto tr1626;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1626;
	} else if ( (*p) >= 9 )
		goto tr1626;
	goto tr1598;
st570:
	if ( ++p == pe )
		goto _test_eof570;
case 570:
	switch( (*p) ) {
		case 73: goto st571;
		case 105: goto st571;
	}
	goto tr1598;
st571:
	if ( ++p == pe )
		goto _test_eof571;
case 571:
	switch( (*p) ) {
		case 88: goto st572;
		case 120: goto st572;
	}
	goto tr1598;
st572:
	if ( ++p == pe )
		goto _test_eof572;
case 572:
	switch( (*p) ) {
		case 32: goto tr1629;
		case 59: goto tr1629;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1629;
	} else if ( (*p) >= 9 )
		goto tr1629;
	goto tr1598;
st573:
	if ( ++p == pe )
		goto _test_eof573;
case 573:
	switch( (*p) ) {
		case 80: goto st574;
		case 112: goto st574;
	}
	goto tr1598;
st574:
	if ( ++p == pe )
		goto _test_eof574;
case 574:
	switch( (*p) ) {
		case 75: goto st575;
		case 107: goto st575;
	}
	goto tr1598;
st575:
	if ( ++p == pe )
		goto _test_eof575;
case 575:
	switch( (*p) ) {
		case 73: goto st576;
		case 105: goto st576;
	}
	goto tr1598;
st576:
	if ( ++p == pe )
		goto _test_eof576;
case 576:
	switch( (*p) ) {
		case 32: goto tr1633;
		case 59: goto tr1633;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1633;
	} else if ( (*p) >= 9 )
		goto tr1633;
	goto tr1598;
st577:
	if ( ++p == pe )
		goto _test_eof577;
case 577:
	switch( (*p) ) {
		case 73: goto st578;
		case 105: goto st578;
	}
	goto tr1598;
st578:
	if ( ++p == pe )
		goto _test_eof578;
case 578:
	switch( (*p) ) {
		case 68: goto st579;
		case 100: goto st579;
	}
	goto tr1598;
st579:
	if ( ++p == pe )
		goto _test_eof579;
case 579:
	switch( (*p) ) {
		case 32: goto tr1636;
		case 59: goto tr1636;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1636;
	} else if ( (*p) >= 9 )
		goto tr1636;
	goto tr1598;
st580:
	if ( ++p == pe )
		goto _test_eof580;
case 580:
	switch( (*p) ) {
		case 71: goto st581;
		case 75: goto st583;
		case 103: goto st581;
		case 107: goto st583;
	}
	goto tr1598;
st581:
	if ( ++p == pe )
		goto _test_eof581;
case 581:
	switch( (*p) ) {
		case 80: goto st582;
		case 112: goto st582;
	}
	goto tr1598;
st582:
	if ( ++p == pe )
		goto _test_eof582;
case 582:
	switch( (*p) ) {
		case 32: goto tr1640;
		case 59: goto tr1640;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1640;
	} else if ( (*p) >= 9 )
		goto tr1640;
	goto tr1598;
st583:
	if ( ++p == pe )
		goto _test_eof583;
case 583:
	switch( (*p) ) {
		case 73: goto st584;
		case 105: goto st584;
	}
	goto tr1598;
st584:
	if ( ++p == pe )
		goto _test_eof584;
case 584:
	switch( (*p) ) {
		case 88: goto st585;
		case 120: goto st585;
	}
	goto tr1598;
st585:
	if ( ++p == pe )
		goto _test_eof585;
case 585:
	switch( (*p) ) {
		case 32: goto tr1643;
		case 59: goto tr1643;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1643;
	} else if ( (*p) >= 9 )
		goto tr1643;
	goto tr1598;
st586:
	if ( ++p == pe )
		goto _test_eof586;
case 586:
	switch( (*p) ) {
		case 80: goto st587;
		case 112: goto st587;
	}
	goto tr1598;
st587:
	if ( ++p == pe )
		goto _test_eof587;
case 587:
	switch( (*p) ) {
		case 75: goto st588;
		case 107: goto st588;
	}
	goto tr1598;
st588:
	if ( ++p == pe )
		goto _test_eof588;
case 588:
	switch( (*p) ) {
		case 73: goto st589;
		case 105: goto st589;
	}
	goto tr1598;
st589:
	if ( ++p == pe )
		goto _test_eof589;
case 589:
	switch( (*p) ) {
		case 32: goto tr1647;
		case 59: goto tr1647;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1647;
	} else if ( (*p) >= 9 )
		goto tr1647;
	goto tr1598;
st590:
	if ( ++p == pe )
		goto _test_eof590;
case 590:
	switch( (*p) ) {
		case 82: goto st591;
		case 114: goto st591;
	}
	goto tr1598;
st591:
	if ( ++p == pe )
		goto _test_eof591;
case 591:
	switch( (*p) ) {
		case 73: goto st592;
		case 105: goto st592;
	}
	goto tr1598;
st592:
	if ( ++p == pe )
		goto _test_eof592;
case 592:
	switch( (*p) ) {
		case 32: goto tr1650;
		case 59: goto tr1650;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1650;
	} else if ( (*p) >= 9 )
		goto tr1650;
	goto tr1598;
st593:
	if ( ++p == pe )
		goto _test_eof593;
case 593:
	if ( (*p) == 46 )
		goto tr1652;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1652;
	goto tr1651;
tr1652:
#line 763 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st594;
tr1654:
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st594;
st594:
	if ( ++p == pe )
		goto _test_eof594;
case 594:
#line 47394 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1653;
		case 46: goto tr1654;
		case 59: goto tr1653;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1653;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1654;
	} else
		goto tr1653;
	goto tr1651;
tr1653:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1094;
st1094:
	if ( ++p == pe )
		goto _test_eof1094;
case 1094:
#line 47433 "zscanner/scanner.c"
	goto st0;
st595:
	if ( ++p == pe )
		goto _test_eof595;
case 595:
	switch( (*p) ) {
		case 42: goto tr1655;
		case 92: goto tr1655;
		case 95: goto tr1655;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1655;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1655;
	} else
		goto tr1655;
	goto tr69;
tr1655:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 596; goto st249;} }
	goto st596;
st596:
	if ( ++p == pe )
		goto _test_eof596;
case 596:
#line 47465 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1656;
		case 59: goto tr1656;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1656;
	} else if ( (*p) >= 9 )
		goto tr1656;
	goto tr69;
tr1656:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1095;
st1095:
	if ( ++p == pe )
		goto _test_eof1095;
case 1095:
#line 47490 "zscanner/scanner.c"
	goto st0;
st597:
	if ( ++p == pe )
		goto _test_eof597;
case 597:
	switch( (*p) ) {
		case 42: goto tr1657;
		case 92: goto tr1657;
		case 95: goto tr1657;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1657;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1657;
	} else
		goto tr1657;
	goto tr69;
tr1657:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 598; goto st249;} }
	goto st598;
st598:
	if ( ++p == pe )
		goto _test_eof598;
case 598:
#line 47522 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1658;
		case 32: goto tr1658;
		case 40: goto tr1659;
		case 41: goto tr1660;
		case 1034: goto tr1661;
		case 1083: goto tr1662;
	}
	goto tr69;
tr1664:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st599;
tr1665:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st599;
tr1667:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st599;
tr1658:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st599;
tr1659:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st599;
tr1660:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st599;
tr1661:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st599;
st599:
	if ( ++p == pe )
		goto _test_eof599;
case 599:
#line 47640 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st599;
		case 32: goto st599;
		case 40: goto tr1664;
		case 41: goto tr1665;
		case 42: goto tr1666;
		case 92: goto tr1666;
		case 95: goto tr1666;
		case 1034: goto tr1667;
		case 1083: goto st628;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1666;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1666;
	} else
		goto tr1666;
	goto tr69;
tr1666:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 600; goto st249;} }
	goto st600;
st600:
	if ( ++p == pe )
		goto _test_eof600;
case 600:
#line 47687 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1669;
		case 32: goto tr1669;
		case 40: goto tr1670;
		case 41: goto tr1671;
		case 1034: goto tr1672;
		case 1083: goto tr1673;
	}
	goto tr69;
tr1676:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st601;
tr1677:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st601;
tr1679:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st601;
tr1669:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st601;
tr1670:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st601;
tr1671:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st601;
tr1672:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st601;
st601:
	if ( ++p == pe )
		goto _test_eof601;
case 601:
#line 47805 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st601;
		case 32: goto st601;
		case 40: goto tr1676;
		case 41: goto tr1677;
		case 1034: goto tr1679;
		case 1083: goto st627;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1678;
	goto tr1674;
tr1678:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st602;
tr1684:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st602;
st602:
	if ( ++p == pe )
		goto _test_eof602;
case 602:
#line 47873 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1681;
		case 32: goto tr1681;
		case 40: goto tr1682;
		case 41: goto tr1683;
		case 1034: goto tr1685;
		case 1083: goto tr1686;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1684;
	goto tr1674;
tr1688:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st603;
tr1689:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st603;
tr1691:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st603;
tr1681:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st603;
tr1682:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st603;
tr1683:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st603;
tr1685:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st603;
st603:
	if ( ++p == pe )
		goto _test_eof603;
case 603:
#line 47997 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st603;
		case 32: goto st603;
		case 40: goto tr1688;
		case 41: goto tr1689;
		case 1034: goto tr1691;
		case 1083: goto st626;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1690;
	goto tr1674;
tr1690:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st604;
tr1697:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st604;
st604:
	if ( ++p == pe )
		goto _test_eof604;
case 604:
#line 48065 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1694;
		case 32: goto tr1694;
		case 40: goto tr1695;
		case 41: goto tr1696;
		case 68: goto tr1698;
		case 72: goto tr1699;
		case 77: goto tr1700;
		case 83: goto st623;
		case 87: goto tr1702;
		case 100: goto tr1698;
		case 104: goto tr1699;
		case 109: goto tr1700;
		case 115: goto st623;
		case 119: goto tr1702;
		case 1034: goto tr1703;
		case 1083: goto tr1704;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1697;
	goto tr1693;
tr1706:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st605;
tr1707:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st605;
tr1709:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st605;
tr1694:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st605;
tr1695:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st605;
tr1696:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st605;
tr1703:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st605;
tr1794:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st605;
tr1795:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st605;
tr1796:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st605;
tr1798:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st605;
st605:
	if ( ++p == pe )
		goto _test_eof605;
case 605:
#line 48303 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st605;
		case 32: goto st605;
		case 40: goto tr1706;
		case 41: goto tr1707;
		case 1034: goto tr1709;
		case 1083: goto st622;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1708;
	goto tr1674;
tr1708:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st606;
tr1714:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st606;
st606:
	if ( ++p == pe )
		goto _test_eof606;
case 606:
#line 48371 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1711;
		case 32: goto tr1711;
		case 40: goto tr1712;
		case 41: goto tr1713;
		case 68: goto tr1715;
		case 72: goto tr1716;
		case 77: goto tr1717;
		case 83: goto st619;
		case 87: goto tr1719;
		case 100: goto tr1715;
		case 104: goto tr1716;
		case 109: goto tr1717;
		case 115: goto st619;
		case 119: goto tr1719;
		case 1034: goto tr1720;
		case 1083: goto tr1721;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1714;
	goto tr1693;
tr1723:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st607;
tr1724:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st607;
tr1726:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st607;
tr1711:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st607;
tr1712:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st607;
tr1713:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st607;
tr1720:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st607;
tr1781:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st607;
tr1782:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st607;
tr1783:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st607;
tr1785:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st607;
st607:
	if ( ++p == pe )
		goto _test_eof607;
case 607:
#line 48609 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st607;
		case 32: goto st607;
		case 40: goto tr1723;
		case 41: goto tr1724;
		case 1034: goto tr1726;
		case 1083: goto st618;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1725;
	goto tr1674;
tr1725:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st608;
tr1731:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st608;
st608:
	if ( ++p == pe )
		goto _test_eof608;
case 608:
#line 48677 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1728;
		case 32: goto tr1728;
		case 40: goto tr1729;
		case 41: goto tr1730;
		case 68: goto tr1732;
		case 72: goto tr1733;
		case 77: goto tr1734;
		case 83: goto st615;
		case 87: goto tr1736;
		case 100: goto tr1732;
		case 104: goto tr1733;
		case 109: goto tr1734;
		case 115: goto st615;
		case 119: goto tr1736;
		case 1034: goto tr1737;
		case 1083: goto tr1738;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1731;
	goto tr1693;
tr1740:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st609;
tr1741:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st609;
tr1743:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st609;
tr1728:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st609;
tr1729:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st609;
tr1730:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st609;
tr1737:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st609;
tr1768:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st609;
tr1769:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st609;
tr1770:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st609;
tr1772:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st609;
st609:
	if ( ++p == pe )
		goto _test_eof609;
case 609:
#line 48915 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st609;
		case 32: goto st609;
		case 40: goto tr1740;
		case 41: goto tr1741;
		case 1034: goto tr1743;
		case 1083: goto st614;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1742;
	goto tr1674;
tr1742:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st610;
tr1746:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st610;
st610:
	if ( ++p == pe )
		goto _test_eof610;
case 610:
#line 48983 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1745;
		case 59: goto tr1745;
		case 68: goto tr1747;
		case 72: goto tr1748;
		case 77: goto tr1749;
		case 83: goto st611;
		case 87: goto tr1751;
		case 100: goto tr1747;
		case 104: goto tr1748;
		case 109: goto tr1749;
		case 115: goto st611;
		case 119: goto tr1751;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1745;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1746;
	} else
		goto tr1745;
	goto tr1693;
tr1745:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1096;
tr1759:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1096;
st1096:
	if ( ++p == pe )
		goto _test_eof1096;
case 1096:
#line 49052 "zscanner/scanner.c"
	goto st0;
tr1747:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st611;
tr1748:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st611;
tr1749:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st611;
tr1751:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st611;
st611:
	if ( ++p == pe )
		goto _test_eof611;
case 611:
#line 49098 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1745;
		case 59: goto tr1745;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1745;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1752;
	} else
		goto tr1745;
	goto tr1693;
tr1753:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st612;
tr1752:
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st612;
tr1760:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st612;
st612:
	if ( ++p == pe )
		goto _test_eof612;
case 612:
#line 49192 "zscanner/scanner.c"
	switch( (*p) ) {
		case 68: goto tr1754;
		case 72: goto tr1755;
		case 77: goto tr1756;
		case 83: goto st613;
		case 87: goto tr1758;
		case 100: goto tr1754;
		case 104: goto tr1755;
		case 109: goto tr1756;
		case 115: goto st613;
		case 119: goto tr1758;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1753;
	goto tr1693;
tr1754:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st613;
tr1755:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st613;
tr1756:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st613;
tr1758:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st613;
st613:
	if ( ++p == pe )
		goto _test_eof613;
case 613:
#line 49252 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1759;
		case 59: goto tr1759;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1759;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1760;
	} else
		goto tr1759;
	goto tr1693;
tr1738:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st614;
tr1773:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st614;
st614:
	if ( ++p == pe )
		goto _test_eof614;
case 614:
#line 49303 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1743;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st614;
	goto tr69;
tr1732:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st615;
tr1733:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st615;
tr1734:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st615;
tr1736:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st615;
st615:
	if ( ++p == pe )
		goto _test_eof615;
case 615:
#line 49374 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1728;
		case 32: goto tr1728;
		case 40: goto tr1729;
		case 41: goto tr1730;
		case 1034: goto tr1737;
		case 1083: goto tr1738;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1761;
	goto tr1693;
tr1762:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st616;
tr1761:
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st616;
tr1771:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st616;
st616:
	if ( ++p == pe )
		goto _test_eof616;
case 616:
#line 49480 "zscanner/scanner.c"
	switch( (*p) ) {
		case 68: goto tr1763;
		case 72: goto tr1764;
		case 77: goto tr1765;
		case 83: goto st617;
		case 87: goto tr1767;
		case 100: goto tr1763;
		case 104: goto tr1764;
		case 109: goto tr1765;
		case 115: goto st617;
		case 119: goto tr1767;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1762;
	goto tr1693;
tr1763:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st617;
tr1764:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st617;
tr1765:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st617;
tr1767:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st617;
st617:
	if ( ++p == pe )
		goto _test_eof617;
case 617:
#line 49540 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1768;
		case 32: goto tr1768;
		case 40: goto tr1769;
		case 41: goto tr1770;
		case 1034: goto tr1772;
		case 1083: goto tr1773;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1771;
	goto tr1693;
tr1721:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st618;
tr1786:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st618;
st618:
	if ( ++p == pe )
		goto _test_eof618;
case 618:
#line 49603 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1726;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st618;
	goto tr69;
tr1715:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st619;
tr1716:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st619;
tr1717:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st619;
tr1719:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st619;
st619:
	if ( ++p == pe )
		goto _test_eof619;
case 619:
#line 49674 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1711;
		case 32: goto tr1711;
		case 40: goto tr1712;
		case 41: goto tr1713;
		case 1034: goto tr1720;
		case 1083: goto tr1721;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1774;
	goto tr1693;
tr1775:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st620;
tr1774:
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st620;
tr1784:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st620;
st620:
	if ( ++p == pe )
		goto _test_eof620;
case 620:
#line 49780 "zscanner/scanner.c"
	switch( (*p) ) {
		case 68: goto tr1776;
		case 72: goto tr1777;
		case 77: goto tr1778;
		case 83: goto st621;
		case 87: goto tr1780;
		case 100: goto tr1776;
		case 104: goto tr1777;
		case 109: goto tr1778;
		case 115: goto st621;
		case 119: goto tr1780;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1775;
	goto tr1693;
tr1776:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st621;
tr1777:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st621;
tr1778:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st621;
tr1780:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st621;
st621:
	if ( ++p == pe )
		goto _test_eof621;
case 621:
#line 49840 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1781;
		case 32: goto tr1781;
		case 40: goto tr1782;
		case 41: goto tr1783;
		case 1034: goto tr1785;
		case 1083: goto tr1786;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1784;
	goto tr1693;
tr1704:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st622;
tr1799:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st622;
st622:
	if ( ++p == pe )
		goto _test_eof622;
case 622:
#line 49903 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1709;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st622;
	goto tr69;
tr1698:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st623;
tr1699:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st623;
tr1700:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st623;
tr1702:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st623;
st623:
	if ( ++p == pe )
		goto _test_eof623;
case 623:
#line 49974 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1694;
		case 32: goto tr1694;
		case 40: goto tr1695;
		case 41: goto tr1696;
		case 1034: goto tr1703;
		case 1083: goto tr1704;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1787;
	goto tr1693;
tr1788:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st624;
tr1787:
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st624;
tr1797:
#line 427 "./zscanner/scanner_body.rl"
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 424 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st624;
st624:
	if ( ++p == pe )
		goto _test_eof624;
case 624:
#line 50080 "zscanner/scanner.c"
	switch( (*p) ) {
		case 68: goto tr1789;
		case 72: goto tr1790;
		case 77: goto tr1791;
		case 83: goto st625;
		case 87: goto tr1793;
		case 100: goto tr1789;
		case 104: goto tr1790;
		case 109: goto tr1791;
		case 115: goto st625;
		case 119: goto tr1793;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1788;
	goto tr1693;
tr1789:
#line 407 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st625;
tr1790:
#line 400 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st625;
tr1791:
#line 393 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st625;
tr1793:
#line 414 "./zscanner/scanner_body.rl"
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
	                  p--; {goto st247;}
	              }
	            }
	goto st625;
st625:
	if ( ++p == pe )
		goto _test_eof625;
case 625:
#line 50140 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1794;
		case 32: goto tr1794;
		case 40: goto tr1795;
		case 41: goto tr1796;
		case 1034: goto tr1798;
		case 1083: goto tr1799;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1797;
	goto tr1693;
tr1686:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st626;
st626:
	if ( ++p == pe )
		goto _test_eof626;
case 626:
#line 50182 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1691;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st626;
	goto tr69;
tr1673:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st627;
st627:
	if ( ++p == pe )
		goto _test_eof627;
case 627:
#line 50224 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1679;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st627;
	goto tr69;
tr1662:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st628;
st628:
	if ( ++p == pe )
		goto _test_eof628;
case 628:
#line 50266 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1667;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st628;
	goto tr69;
st629:
	if ( ++p == pe )
		goto _test_eof629;
case 629:
	switch( (*p) ) {
		case 32: goto tr69;
		case 59: goto tr69;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr69;
	} else if ( (*p) >= 9 )
		goto tr69;
	goto tr1800;
tr1800:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 565 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 630; goto st259;} }
	goto st630;
st630:
	if ( ++p == pe )
		goto _test_eof630;
case 630:
#line 50319 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1801;
		case 32: goto tr1801;
		case 40: goto tr1802;
		case 41: goto tr1803;
		case 1034: goto tr1804;
		case 1083: goto tr1805;
	}
	goto tr69;
tr1808:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st631;
tr1809:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st631;
tr1810:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st631;
tr1801:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st631;
tr1802:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st631;
tr1803:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st631;
tr1804:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st631;
st631:
	if ( ++p == pe )
		goto _test_eof631;
case 631:
#line 50445 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st631;
		case 32: goto st631;
		case 40: goto tr1808;
		case 41: goto tr1809;
		case 1034: goto tr1810;
		case 1083: goto st633;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1806;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1806;
	} else
		goto tr1806;
	goto tr69;
tr1806:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 565 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 632; goto st259;} }
	goto st632;
st632:
	if ( ++p == pe )
		goto _test_eof632;
case 632:
#line 50489 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1812;
		case 59: goto tr1812;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1812;
	} else if ( (*p) >= 9 )
		goto tr1812;
	goto tr69;
tr1812:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1097;
st1097:
	if ( ++p == pe )
		goto _test_eof1097;
case 1097:
#line 50521 "zscanner/scanner.c"
	goto st0;
tr1805:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st633;
st633:
	if ( ++p == pe )
		goto _test_eof633;
case 633:
#line 50540 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1810;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st633;
	goto tr69;
st634:
	if ( ++p == pe )
		goto _test_eof634;
case 634:
	switch( (*p) ) {
		case 42: goto tr1813;
		case 92: goto tr1813;
		case 95: goto tr1813;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1813;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1813;
	} else
		goto tr1813;
	goto tr69;
tr1813:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 635; goto st249;} }
	goto st635;
st635:
	if ( ++p == pe )
		goto _test_eof635;
case 635:
#line 50597 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1814;
		case 32: goto tr1814;
		case 40: goto tr1815;
		case 41: goto tr1816;
		case 1034: goto tr1817;
		case 1083: goto tr1818;
	}
	goto tr69;
tr1820:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st636;
tr1821:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st636;
tr1823:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st636;
tr1814:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st636;
tr1815:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st636;
tr1816:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st636;
tr1817:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st636;
st636:
	if ( ++p == pe )
		goto _test_eof636;
case 636:
#line 50715 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st636;
		case 32: goto st636;
		case 40: goto tr1820;
		case 41: goto tr1821;
		case 42: goto tr1822;
		case 92: goto tr1822;
		case 95: goto tr1822;
		case 1034: goto tr1823;
		case 1083: goto st638;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1822;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1822;
	} else
		goto tr1822;
	goto tr69;
tr1822:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 637; goto st249;} }
	goto st637;
st637:
	if ( ++p == pe )
		goto _test_eof637;
case 637:
#line 50762 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1825;
		case 59: goto tr1825;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1825;
	} else if ( (*p) >= 9 )
		goto tr1825;
	goto tr69;
tr1825:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
st1098:
	if ( ++p == pe )
		goto _test_eof1098;
case 1098:
#line 50787 "zscanner/scanner.c"
	goto st0;
tr1818:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st638;
st638:
	if ( ++p == pe )
		goto _test_eof638;
case 638:
#line 50804 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1823;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st638;
	goto tr69;
st639:
	if ( ++p == pe )
		goto _test_eof639;
case 639:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1826;
	goto tr1674;
tr1826:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st640;
tr1830:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st640;
st640:
	if ( ++p == pe )
		goto _test_eof640;
case 640:
#line 50880 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1827;
		case 32: goto tr1827;
		case 40: goto tr1828;
		case 41: goto tr1829;
		case 1034: goto tr1831;
		case 1083: goto tr1832;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1830;
	goto tr1674;
tr1834:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st641;
tr1835:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st641;
tr1837:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st641;
tr1827:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st641;
tr1828:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st641;
tr1829:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st641;
tr1831:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st641;
st641:
	if ( ++p == pe )
		goto _test_eof641;
case 641:
#line 51024 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st641;
		case 32: goto st641;
		case 40: goto tr1834;
		case 41: goto tr1835;
		case 42: goto tr1836;
		case 92: goto tr1836;
		case 95: goto tr1836;
		case 1034: goto tr1837;
		case 1083: goto st643;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1836;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1836;
	} else
		goto tr1836;
	goto tr69;
tr1836:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 642; goto st249;} }
	goto st642;
st642:
	if ( ++p == pe )
		goto _test_eof642;
case 642:
#line 51071 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1839;
		case 59: goto tr1839;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1839;
	} else if ( (*p) >= 9 )
		goto tr1839;
	goto tr69;
tr1839:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
st1099:
	if ( ++p == pe )
		goto _test_eof1099;
case 1099:
#line 51096 "zscanner/scanner.c"
	goto st0;
tr1832:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st643;
st643:
	if ( ++p == pe )
		goto _test_eof643;
case 643:
#line 51119 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1837;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st643;
	goto tr69;
st644:
	if ( ++p == pe )
		goto _test_eof644;
case 644:
	switch( (*p) ) {
		case 32: goto tr69;
		case 59: goto tr69;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr69;
	} else if ( (*p) >= 9 )
		goto tr69;
	goto tr1840;
tr1840:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 565 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 645; goto st259;} }
	goto st645;
st645:
	if ( ++p == pe )
		goto _test_eof645;
case 645:
#line 51172 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1841;
		case 32: goto tr1841;
		case 40: goto tr1842;
		case 41: goto tr1843;
		case 2058: goto tr1844;
		case 2107: goto tr1845;
		case 2314: goto tr1846;
		case 2363: goto tr1846;
		case 2570: goto tr1847;
		case 2619: goto tr1848;
	}
	goto tr69;
tr1850:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st646;
tr1851:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st646;
tr1852:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st646;
tr1841:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st646;
tr1842:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st646;
tr1843:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st646;
tr1844:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st646;
st646:
	if ( ++p == pe )
		goto _test_eof646;
case 646:
#line 51308 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st646;
		case 32: goto st646;
		case 40: goto tr1850;
		case 41: goto tr1851;
		case 2058: goto tr1852;
		case 2107: goto st647;
		case 2314: goto tr1854;
		case 2363: goto tr1854;
		case 2570: goto tr1855;
		case 2619: goto tr1856;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1840;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1840;
	} else
		goto tr1840;
	goto tr69;
tr1845:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st647;
st647:
	if ( ++p == pe )
		goto _test_eof647;
case 647:
#line 51367 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1852;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st647;
	goto tr69;
tr1854:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1846:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
st1100:
	if ( ++p == pe )
		goto _test_eof1100;
case 1100:
#line 51421 "zscanner/scanner.c"
	goto st0;
tr1855:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1101;
tr1847:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1101;
st1101:
	if ( ++p == pe )
		goto _test_eof1101;
case 1101:
#line 51458 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st646;
		case 32: goto st646;
		case 40: goto tr1850;
		case 41: goto tr1851;
		case 2058: goto tr1852;
		case 2107: goto st647;
		case 2314: goto tr1854;
		case 2363: goto tr1854;
		case 2570: goto tr1855;
		case 2619: goto tr1856;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1840;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1840;
	} else
		goto tr1840;
	goto tr69;
tr1856:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1102;
tr1848:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1102;
st1102:
	if ( ++p == pe )
		goto _test_eof1102;
case 1102:
#line 51527 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1852;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st647;
	goto tr69;
st648:
	if ( ++p == pe )
		goto _test_eof648;
case 648:
	if ( (*p) == 46 )
		goto tr1857;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr1857;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1857;
	} else
		goto tr1857;
	goto tr1651;
tr1857:
#line 763 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st649;
tr1859:
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st649;
st649:
	if ( ++p == pe )
		goto _test_eof649;
case 649:
#line 51601 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr1858;
		case 46: goto tr1859;
		case 59: goto tr1858;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr1858;
		} else if ( (*p) >= 9 )
			goto tr1858;
	} else if ( (*p) > 58 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr1859;
		} else if ( (*p) >= 65 )
			goto tr1859;
	} else
		goto tr1859;
	goto tr1651;
tr1858:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1103;
st1103:
	if ( ++p == pe )
		goto _test_eof1103;
case 1103:
#line 51646 "zscanner/scanner.c"
	goto st0;
st650:
	if ( ++p == pe )
		goto _test_eof650;
case 650:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1861;
	goto tr1860;
tr1865:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st651;
tr1861:
#line 1451 "./zscanner/scanner_body.rl"
	{
		memset(&(s->loc), 0, sizeof(s->loc));
		// Defaults.
		s->loc.siz = 100;
		s->loc.vp  = 1000;
		s->loc.hp  = 1000000;
		s->loc.lat_sign  = 1;
		s->loc.long_sign = 1;
		s->loc.alt_sign  = 1;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st651;
st651:
	if ( ++p == pe )
		goto _test_eof651;
case 651:
#line 51708 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1862;
		case 32: goto tr1862;
		case 40: goto tr1863;
		case 41: goto tr1864;
		case 1034: goto tr1866;
		case 1083: goto tr1867;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1865;
	goto tr1860;
tr1869:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st652;
tr1870:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st652;
tr1874:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st652;
tr1862:
#line 1346 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st652;
tr1863:
#line 1346 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st652;
tr1864:
#line 1346 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st652;
tr1866:
#line 1346 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st652;
st652:
	if ( ++p == pe )
		goto _test_eof652;
case 652:
#line 51828 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st652;
		case 32: goto st652;
		case 40: goto tr1869;
		case 41: goto tr1870;
		case 78: goto st657;
		case 83: goto st699;
		case 1034: goto tr1874;
		case 1083: goto st704;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1871;
	goto tr1860;
tr1871:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st653;
tr1879:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st653;
st653:
	if ( ++p == pe )
		goto _test_eof653;
case 653:
#line 51898 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1876;
		case 32: goto tr1876;
		case 40: goto tr1877;
		case 41: goto tr1878;
		case 1034: goto tr1880;
		case 1083: goto tr1881;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1879;
	goto tr1860;
tr1883:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st654;
tr1884:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st654;
tr1886:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st654;
tr1876:
#line 1362 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st654;
tr1877:
#line 1362 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st654;
tr1878:
#line 1362 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st654;
tr1880:
#line 1362 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st654;
st654:
	if ( ++p == pe )
		goto _test_eof654;
case 654:
#line 52018 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st654;
		case 32: goto st654;
		case 40: goto tr1883;
		case 41: goto tr1884;
		case 78: goto st657;
		case 83: goto st699;
		case 1034: goto tr1886;
		case 1083: goto st703;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1885;
	goto tr1860;
tr1892:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st655;
tr1885:
#line 323 "./zscanner/scanner_body.rl"
	{
		s->decimals = 3;
	}
#line 293 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter = 0;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st655;
st655:
	if ( ++p == pe )
		goto _test_eof655;
case 655:
#line 52096 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1888;
		case 32: goto tr1888;
		case 40: goto tr1889;
		case 41: goto tr1890;
		case 46: goto st701;
		case 1034: goto tr1893;
		case 1083: goto tr1894;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1892;
	goto tr1860;
tr1896:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st656;
tr1897:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st656;
tr1898:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st656;
tr1888:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st656;
tr1889:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st656;
tr1890:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st656;
tr1893:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st656;
tr2087:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st656;
tr2088:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st656;
tr2089:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st656;
tr2091:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st656;
st656:
	if ( ++p == pe )
		goto _test_eof656;
case 656:
#line 52401 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st656;
		case 32: goto st656;
		case 40: goto tr1896;
		case 41: goto tr1897;
		case 78: goto st657;
		case 83: goto st699;
		case 1034: goto tr1898;
		case 1083: goto st700;
	}
	goto tr1860;
st657:
	if ( ++p == pe )
		goto _test_eof657;
case 657:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st658;
		case 32: goto st658;
		case 40: goto tr1901;
		case 41: goto tr1902;
		case 1034: goto tr1903;
		case 1083: goto st698;
	}
	goto tr1860;
tr1901:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st658;
tr1902:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st658;
tr1903:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st658;
tr2082:
#line 1428 "./zscanner/scanner_body.rl"
	{
		s->loc.lat_sign = -1;
	}
	goto st658;
tr2083:
#line 1428 "./zscanner/scanner_body.rl"
	{
		s->loc.lat_sign = -1;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st658;
tr2084:
#line 1428 "./zscanner/scanner_body.rl"
	{
		s->loc.lat_sign = -1;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st658;
tr2085:
#line 1428 "./zscanner/scanner_body.rl"
	{
		s->loc.lat_sign = -1;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st658;
st658:
	if ( ++p == pe )
		goto _test_eof658;
case 658:
#line 52528 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st658;
		case 32: goto st658;
		case 40: goto tr1901;
		case 41: goto tr1902;
		case 1034: goto tr1903;
		case 1083: goto st698;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1905;
	goto tr1860;
tr1905:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st659;
tr1909:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st659;
st659:
	if ( ++p == pe )
		goto _test_eof659;
case 659:
#line 52596 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1906;
		case 32: goto tr1906;
		case 40: goto tr1907;
		case 41: goto tr1908;
		case 1034: goto tr1910;
		case 1083: goto tr1911;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1909;
	goto tr1860;
tr1913:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st660;
tr1914:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st660;
tr1918:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st660;
tr1906:
#line 1354 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st660;
tr1907:
#line 1354 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st660;
tr1908:
#line 1354 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st660;
tr1910:
#line 1354 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st660;
st660:
	if ( ++p == pe )
		goto _test_eof660;
case 660:
#line 52716 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st660;
		case 32: goto st660;
		case 40: goto tr1913;
		case 41: goto tr1914;
		case 69: goto st665;
		case 87: goto st692;
		case 1034: goto tr1918;
		case 1083: goto st697;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1915;
	goto tr1860;
tr1915:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st661;
tr1923:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st661;
st661:
	if ( ++p == pe )
		goto _test_eof661;
case 661:
#line 52786 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1920;
		case 32: goto tr1920;
		case 40: goto tr1921;
		case 41: goto tr1922;
		case 1034: goto tr1924;
		case 1083: goto tr1925;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1923;
	goto tr1860;
tr1927:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st662;
tr1928:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st662;
tr1930:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st662;
tr1920:
#line 1370 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st662;
tr1921:
#line 1370 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st662;
tr1922:
#line 1370 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st662;
tr1924:
#line 1370 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st662;
st662:
	if ( ++p == pe )
		goto _test_eof662;
case 662:
#line 52906 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st662;
		case 32: goto st662;
		case 40: goto tr1927;
		case 41: goto tr1928;
		case 69: goto st665;
		case 87: goto st692;
		case 1034: goto tr1930;
		case 1083: goto st696;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1929;
	goto tr1860;
tr1936:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st663;
tr1929:
#line 323 "./zscanner/scanner_body.rl"
	{
		s->decimals = 3;
	}
#line 293 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter = 0;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st663;
st663:
	if ( ++p == pe )
		goto _test_eof663;
case 663:
#line 52984 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1932;
		case 32: goto tr1932;
		case 40: goto tr1933;
		case 41: goto tr1934;
		case 46: goto st694;
		case 1034: goto tr1937;
		case 1083: goto tr1938;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1936;
	goto tr1860;
tr1940:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st664;
tr1941:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st664;
tr1942:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st664;
tr1932:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st664;
tr1933:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st664;
tr1934:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st664;
tr1937:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st664;
tr2075:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st664;
tr2076:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st664;
tr2077:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st664;
tr2079:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st664;
st664:
	if ( ++p == pe )
		goto _test_eof664;
case 664:
#line 53289 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st664;
		case 32: goto st664;
		case 40: goto tr1940;
		case 41: goto tr1941;
		case 69: goto st665;
		case 87: goto st692;
		case 1034: goto tr1942;
		case 1083: goto st693;
	}
	goto tr1860;
st665:
	if ( ++p == pe )
		goto _test_eof665;
case 665:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st666;
		case 32: goto st666;
		case 40: goto tr1945;
		case 41: goto tr1946;
		case 1034: goto tr1947;
		case 1083: goto st691;
	}
	goto tr1860;
tr1945:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st666;
tr1946:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st666;
tr1947:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st666;
tr2070:
#line 1431 "./zscanner/scanner_body.rl"
	{
		s->loc.long_sign = -1;
	}
	goto st666;
tr2071:
#line 1431 "./zscanner/scanner_body.rl"
	{
		s->loc.long_sign = -1;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st666;
tr2072:
#line 1431 "./zscanner/scanner_body.rl"
	{
		s->loc.long_sign = -1;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st666;
tr2073:
#line 1431 "./zscanner/scanner_body.rl"
	{
		s->loc.long_sign = -1;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st666;
st666:
	if ( ++p == pe )
		goto _test_eof666;
case 666:
#line 53416 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st666;
		case 32: goto st666;
		case 40: goto tr1945;
		case 41: goto tr1946;
		case 45: goto st667;
		case 1034: goto tr1947;
		case 1083: goto st691;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1950;
	goto tr1860;
st667:
	if ( ++p == pe )
		goto _test_eof667;
case 667:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1951;
	goto tr1860;
tr1956:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st668;
tr1950:
#line 320 "./zscanner/scanner_body.rl"
	{
		s->decimals = 2;
	}
#line 293 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter = 0;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st668;
tr1951:
#line 1434 "./zscanner/scanner_body.rl"
	{
		s->loc.alt_sign = -1;
	}
#line 320 "./zscanner/scanner_body.rl"
	{
		s->decimals = 2;
	}
#line 293 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter = 0;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st668;
st668:
	if ( ++p == pe )
		goto _test_eof668;
case 668:
#line 53533 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1952;
		case 32: goto tr1952;
		case 40: goto tr1953;
		case 41: goto tr1954;
		case 46: goto st688;
		case 109: goto tr1957;
		case 2058: goto tr1958;
		case 2107: goto tr1959;
		case 2314: goto tr1960;
		case 2363: goto tr1960;
		case 2570: goto tr1961;
		case 2619: goto tr1962;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1956;
	goto tr1860;
tr1964:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st669;
tr1965:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st669;
tr1967:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st669;
tr1952:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st669;
tr1953:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st669;
tr1954:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st669;
tr1958:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st669;
tr2059:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st669;
tr2060:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st669;
tr2061:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st669;
tr2064:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st669;
st669:
	if ( ++p == pe )
		goto _test_eof669;
case 669:
#line 53865 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st669;
		case 32: goto st669;
		case 40: goto tr1964;
		case 41: goto tr1965;
		case 2058: goto tr1967;
		case 2107: goto st687;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr1970;
		case 2619: goto tr1971;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1966;
	goto tr1860;
tr1976:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st670;
tr1966:
#line 320 "./zscanner/scanner_body.rl"
	{
		s->decimals = 2;
	}
#line 293 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter = 0;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st670;
st670:
	if ( ++p == pe )
		goto _test_eof670;
case 670:
#line 53951 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1972;
		case 32: goto tr1972;
		case 40: goto tr1973;
		case 41: goto tr1974;
		case 46: goto st684;
		case 109: goto tr1977;
		case 2058: goto tr1978;
		case 2107: goto tr1979;
		case 2314: goto tr1980;
		case 2363: goto tr1980;
		case 2570: goto tr1981;
		case 2619: goto tr1982;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1976;
	goto tr1860;
tr1984:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st671;
tr1985:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st671;
tr1987:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st671;
tr1972:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st671;
tr1973:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st671;
tr1974:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st671;
tr1978:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st671;
tr2048:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st671;
tr2049:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st671;
tr2050:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st671;
tr2053:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st671;
st671:
	if ( ++p == pe )
		goto _test_eof671;
case 671:
#line 54267 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st671;
		case 32: goto st671;
		case 40: goto tr1984;
		case 41: goto tr1985;
		case 2058: goto tr1987;
		case 2107: goto st683;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr1989;
		case 2619: goto tr1990;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1986;
	goto tr1860;
tr1995:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st672;
tr1986:
#line 320 "./zscanner/scanner_body.rl"
	{
		s->decimals = 2;
	}
#line 293 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter = 0;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st672;
st672:
	if ( ++p == pe )
		goto _test_eof672;
case 672:
#line 54353 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1991;
		case 32: goto tr1991;
		case 40: goto tr1992;
		case 41: goto tr1993;
		case 46: goto st680;
		case 109: goto tr1996;
		case 2058: goto tr1997;
		case 2107: goto tr1998;
		case 2314: goto tr1999;
		case 2363: goto tr1999;
		case 2570: goto tr2000;
		case 2619: goto tr2001;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1995;
	goto tr1860;
tr2003:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st673;
tr2004:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st673;
tr2006:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st673;
tr1991:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st673;
tr1992:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st673;
tr1993:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st673;
tr1997:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st673;
tr2037:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st673;
tr2038:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st673;
tr2039:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st673;
tr2042:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st673;
st673:
	if ( ++p == pe )
		goto _test_eof673;
case 673:
#line 54669 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st673;
		case 32: goto st673;
		case 40: goto tr2003;
		case 41: goto tr2004;
		case 2058: goto tr2006;
		case 2107: goto st679;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr2008;
		case 2619: goto tr2009;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2005;
	goto tr1860;
tr2014:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st674;
tr2005:
#line 320 "./zscanner/scanner_body.rl"
	{
		s->decimals = 2;
	}
#line 293 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter = 0;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st674;
st674:
	if ( ++p == pe )
		goto _test_eof674;
case 674:
#line 54755 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2010;
		case 32: goto tr2010;
		case 40: goto tr2011;
		case 41: goto tr2012;
		case 46: goto st677;
		case 109: goto tr2010;
		case 2058: goto tr2015;
		case 2107: goto tr2016;
		case 2314: goto tr2017;
		case 2363: goto tr2017;
		case 2570: goto tr2018;
		case 2619: goto tr2019;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2014;
	goto tr1860;
tr2021:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st675;
tr2022:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st675;
tr2023:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st675;
tr2010:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st675;
tr2011:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st675;
tr2012:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st675;
tr2015:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st675;
tr2027:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st675;
tr2028:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st675;
tr2029:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st675;
tr2031:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st675;
st675:
	if ( ++p == pe )
		goto _test_eof675;
case 675:
#line 55071 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st675;
		case 32: goto st675;
		case 40: goto tr2021;
		case 41: goto tr2022;
		case 2058: goto tr2023;
		case 2107: goto st676;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr2025;
		case 2619: goto tr2026;
	}
	goto tr1860;
tr2016:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st676;
tr2032:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st676;
st676:
	if ( ++p == pe )
		goto _test_eof676;
case 676:
#line 55161 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2023;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st676;
	goto tr1860;
tr1960:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr1969:
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr1980:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr1999:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr2017:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr2033:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr2044:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr2055:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
tr2066:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1104;
st1104:
	if ( ++p == pe )
		goto _test_eof1104;
case 1104:
#line 55685 "zscanner/scanner.c"
	goto st0;
tr2025:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1105;
tr2018:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1105;
tr2034:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1105;
st1105:
	if ( ++p == pe )
		goto _test_eof1105;
case 1105:
#line 55850 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st675;
		case 32: goto st675;
		case 40: goto tr2021;
		case 41: goto tr2022;
		case 2058: goto tr2023;
		case 2107: goto st676;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr2025;
		case 2619: goto tr2026;
	}
	goto tr1860;
tr2026:
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1106;
tr2019:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1106;
tr2035:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1420 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1106;
st1106:
	if ( ++p == pe )
		goto _test_eof1106;
case 1106:
#line 56035 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2023;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st676;
	goto tr1860;
st677:
	if ( ++p == pe )
		goto _test_eof677;
case 677:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2027;
		case 32: goto tr2027;
		case 40: goto tr2028;
		case 41: goto tr2029;
		case 109: goto tr2027;
		case 2058: goto tr2031;
		case 2107: goto tr2032;
		case 2314: goto tr2033;
		case 2363: goto tr2033;
		case 2570: goto tr2034;
		case 2619: goto tr2035;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2030;
	goto tr1860;
tr2030:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st678;
tr2036:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st678;
st678:
	if ( ++p == pe )
		goto _test_eof678;
case 678:
#line 56156 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2010;
		case 32: goto tr2010;
		case 40: goto tr2011;
		case 41: goto tr2012;
		case 109: goto tr2010;
		case 2058: goto tr2015;
		case 2107: goto tr2016;
		case 2314: goto tr2017;
		case 2363: goto tr2017;
		case 2570: goto tr2018;
		case 2619: goto tr2019;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2036;
	goto tr1860;
tr1998:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st679;
tr2043:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st679;
st679:
	if ( ++p == pe )
		goto _test_eof679;
case 679:
#line 56249 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2006;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st679;
	goto tr1860;
tr2008:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1107;
tr2000:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1107;
tr2045:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1107;
st1107:
	if ( ++p == pe )
		goto _test_eof1107;
case 1107:
#line 56439 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st673;
		case 32: goto st673;
		case 40: goto tr2003;
		case 41: goto tr2004;
		case 2058: goto tr2006;
		case 2107: goto st679;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr2008;
		case 2619: goto tr2009;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2005;
	goto tr1860;
tr2009:
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1108;
tr2001:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1108;
tr2046:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1108;
st1108:
	if ( ++p == pe )
		goto _test_eof1108;
case 1108:
#line 56626 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2006;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st679;
	goto tr1860;
st680:
	if ( ++p == pe )
		goto _test_eof680;
case 680:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2037;
		case 32: goto tr2037;
		case 40: goto tr2038;
		case 41: goto tr2039;
		case 109: goto tr2041;
		case 2058: goto tr2042;
		case 2107: goto tr2043;
		case 2314: goto tr2044;
		case 2363: goto tr2044;
		case 2570: goto tr2045;
		case 2619: goto tr2046;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2040;
	goto tr1860;
tr2040:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st681;
tr2047:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st681;
st681:
	if ( ++p == pe )
		goto _test_eof681;
case 681:
#line 56747 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1991;
		case 32: goto tr1991;
		case 40: goto tr1992;
		case 41: goto tr1993;
		case 109: goto tr1996;
		case 2058: goto tr1997;
		case 2107: goto tr1998;
		case 2314: goto tr1999;
		case 2363: goto tr1999;
		case 2570: goto tr2000;
		case 2619: goto tr2001;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2047;
	goto tr1860;
tr1996:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st682;
tr2041:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1412 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st682;
st682:
	if ( ++p == pe )
		goto _test_eof682;
case 682:
#line 56840 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st673;
		case 32: goto st673;
		case 40: goto tr2003;
		case 41: goto tr2004;
		case 2058: goto tr2006;
		case 2107: goto st679;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr2008;
		case 2619: goto tr2009;
	}
	goto tr1860;
tr1979:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st683;
tr2054:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st683;
st683:
	if ( ++p == pe )
		goto _test_eof683;
case 683:
#line 56930 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1987;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st683;
	goto tr1860;
tr1989:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1981:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr2056:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
st1109:
	if ( ++p == pe )
		goto _test_eof1109;
case 1109:
#line 57120 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st671;
		case 32: goto st671;
		case 40: goto tr1984;
		case 41: goto tr1985;
		case 2058: goto tr1987;
		case 2107: goto st683;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr1989;
		case 2619: goto tr1990;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1986;
	goto tr1860;
tr1990:
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1982:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr2057:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
st1110:
	if ( ++p == pe )
		goto _test_eof1110;
case 1110:
#line 57307 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1987;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st683;
	goto tr1860;
st684:
	if ( ++p == pe )
		goto _test_eof684;
case 684:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2048;
		case 32: goto tr2048;
		case 40: goto tr2049;
		case 41: goto tr2050;
		case 109: goto tr2052;
		case 2058: goto tr2053;
		case 2107: goto tr2054;
		case 2314: goto tr2055;
		case 2363: goto tr2055;
		case 2570: goto tr2056;
		case 2619: goto tr2057;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2051;
	goto tr1860;
tr2051:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st685;
tr2058:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st685;
st685:
	if ( ++p == pe )
		goto _test_eof685;
case 685:
#line 57428 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1972;
		case 32: goto tr1972;
		case 40: goto tr1973;
		case 41: goto tr1974;
		case 109: goto tr1977;
		case 2058: goto tr1978;
		case 2107: goto tr1979;
		case 2314: goto tr1980;
		case 2363: goto tr1980;
		case 2570: goto tr1981;
		case 2619: goto tr1982;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2058;
	goto tr1860;
tr1977:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st686;
tr2052:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1404 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st686;
st686:
	if ( ++p == pe )
		goto _test_eof686;
case 686:
#line 57521 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st671;
		case 32: goto st671;
		case 40: goto tr1984;
		case 41: goto tr1985;
		case 2058: goto tr1987;
		case 2107: goto st683;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr1989;
		case 2619: goto tr1990;
	}
	goto tr1860;
tr1959:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st687;
tr2065:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st687;
st687:
	if ( ++p == pe )
		goto _test_eof687;
case 687:
#line 57615 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1967;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st687;
	goto tr1860;
tr1961:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1111;
tr1970:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1111;
tr2067:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1111;
st1111:
	if ( ++p == pe )
		goto _test_eof1111;
case 1111:
#line 57809 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st669;
		case 32: goto st669;
		case 40: goto tr1964;
		case 41: goto tr1965;
		case 2058: goto tr1967;
		case 2107: goto st687;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr1970;
		case 2619: goto tr1971;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1966;
	goto tr1860;
tr1962:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1112;
tr1971:
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1112;
tr2068:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
#line 1461 "./zscanner/scanner_body.rl"
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1112;
st1112:
	if ( ++p == pe )
		goto _test_eof1112;
case 1112:
#line 58000 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1967;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st687;
	goto tr1860;
st688:
	if ( ++p == pe )
		goto _test_eof688;
case 688:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2059;
		case 32: goto tr2059;
		case 40: goto tr2060;
		case 41: goto tr2061;
		case 109: goto tr2063;
		case 2058: goto tr2064;
		case 2107: goto tr2065;
		case 2314: goto tr2066;
		case 2363: goto tr2066;
		case 2570: goto tr2067;
		case 2619: goto tr2068;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2062;
	goto tr1860;
tr2062:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st689;
tr2069:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st689;
st689:
	if ( ++p == pe )
		goto _test_eof689;
case 689:
#line 58121 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1952;
		case 32: goto tr1952;
		case 40: goto tr1953;
		case 41: goto tr1954;
		case 109: goto tr1957;
		case 2058: goto tr1958;
		case 2107: goto tr1959;
		case 2314: goto tr1960;
		case 2363: goto tr1960;
		case 2570: goto tr1961;
		case 2619: goto tr1962;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2069;
	goto tr1860;
tr1957:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st690;
tr2063:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1394 "./zscanner/scanner_body.rl"
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st690;
st690:
	if ( ++p == pe )
		goto _test_eof690;
case 690:
#line 58218 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st669;
		case 32: goto st669;
		case 40: goto tr1964;
		case 41: goto tr1965;
		case 2058: goto tr1967;
		case 2107: goto st687;
		case 2314: goto tr1969;
		case 2363: goto tr1969;
		case 2570: goto tr1970;
		case 2619: goto tr1971;
	}
	goto tr1860;
tr2074:
#line 1431 "./zscanner/scanner_body.rl"
	{
		s->loc.long_sign = -1;
	}
	goto st691;
st691:
	if ( ++p == pe )
		goto _test_eof691;
case 691:
#line 58262 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1947;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st691;
	goto tr1860;
st692:
	if ( ++p == pe )
		goto _test_eof692;
case 692:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2070;
		case 32: goto tr2070;
		case 40: goto tr2071;
		case 41: goto tr2072;
		case 1034: goto tr2073;
		case 1083: goto tr2074;
	}
	goto tr1860;
tr1938:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st693;
tr2080:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1386 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st693;
st693:
	if ( ++p == pe )
		goto _test_eof693;
case 693:
#line 58372 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1942;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st693;
	goto tr1860;
st694:
	if ( ++p == pe )
		goto _test_eof694;
case 694:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2075;
		case 32: goto tr2075;
		case 40: goto tr2076;
		case 41: goto tr2077;
		case 1034: goto tr2079;
		case 1083: goto tr2080;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2078;
	goto tr1860;
tr2078:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st695;
tr2081:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st695;
st695:
	if ( ++p == pe )
		goto _test_eof695;
case 695:
#line 58482 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1932;
		case 32: goto tr1932;
		case 40: goto tr1933;
		case 41: goto tr1934;
		case 1034: goto tr1937;
		case 1083: goto tr1938;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2081;
	goto tr1860;
tr1925:
#line 1370 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st696;
st696:
	if ( ++p == pe )
		goto _test_eof696;
case 696:
#line 58523 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1930;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st696;
	goto tr1860;
tr1911:
#line 1354 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st697;
st697:
	if ( ++p == pe )
		goto _test_eof697;
case 697:
#line 58565 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1918;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st697;
	goto tr1860;
tr2086:
#line 1428 "./zscanner/scanner_body.rl"
	{
		s->loc.lat_sign = -1;
	}
	goto st698;
st698:
	if ( ++p == pe )
		goto _test_eof698;
case 698:
#line 58602 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1903;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st698;
	goto tr1860;
st699:
	if ( ++p == pe )
		goto _test_eof699;
case 699:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2082;
		case 32: goto tr2082;
		case 40: goto tr2083;
		case 41: goto tr2084;
		case 1034: goto tr2085;
		case 1083: goto tr2086;
	}
	goto tr1860;
tr1894:
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st700;
tr2092:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 303 "./zscanner/scanner_body.rl"
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			SCANNER_WARNING(ZSCANNER_EFLOAT_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1378 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st700;
st700:
	if ( ++p == pe )
		goto _test_eof700;
case 700:
#line 58712 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1898;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st700;
	goto tr1860;
st701:
	if ( ++p == pe )
		goto _test_eof701;
case 701:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2087;
		case 32: goto tr2087;
		case 40: goto tr2088;
		case 41: goto tr2089;
		case 1034: goto tr2091;
		case 1083: goto tr2092;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2090;
	goto tr1860;
tr2090:
#line 296 "./zscanner/scanner_body.rl"
	{
		s->number64_tmp = s->number64;
	}
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st702;
tr2093:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 299 "./zscanner/scanner_body.rl"
	{
		s->decimal_counter++;
	}
	goto st702;
st702:
	if ( ++p == pe )
		goto _test_eof702;
case 702:
#line 58822 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1888;
		case 32: goto tr1888;
		case 40: goto tr1889;
		case 41: goto tr1890;
		case 1034: goto tr1893;
		case 1083: goto tr1894;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2093;
	goto tr1860;
tr1881:
#line 1362 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st703;
st703:
	if ( ++p == pe )
		goto _test_eof703;
case 703:
#line 58863 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1886;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st703;
	goto tr1860;
tr1867:
#line 1346 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
			p--; {goto st247;}
		}
	}
	goto st704;
st704:
	if ( ++p == pe )
		goto _test_eof704;
case 704:
#line 58905 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1874;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st704;
	goto tr1860;
st705:
	if ( ++p == pe )
		goto _test_eof705;
case 705:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2094;
	goto tr1674;
tr2094:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st706;
tr2098:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st706;
st706:
	if ( ++p == pe )
		goto _test_eof706;
case 706:
#line 58981 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2095;
		case 32: goto tr2095;
		case 40: goto tr2096;
		case 41: goto tr2097;
		case 1034: goto tr2099;
		case 1083: goto tr2100;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2098;
	goto tr1674;
tr2102:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st707;
tr2103:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st707;
tr2105:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st707;
tr2095:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st707;
tr2096:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st707;
tr2097:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st707;
tr2099:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st707;
st707:
	if ( ++p == pe )
		goto _test_eof707;
case 707:
#line 59105 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st707;
		case 32: goto st707;
		case 40: goto tr2102;
		case 41: goto tr2103;
		case 1034: goto tr2105;
		case 1083: goto st715;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2104;
	goto tr1674;
tr2104:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st708;
tr2110:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st708;
st708:
	if ( ++p == pe )
		goto _test_eof708;
case 708:
#line 59173 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2107;
		case 32: goto tr2107;
		case 40: goto tr2108;
		case 41: goto tr2109;
		case 1034: goto tr2111;
		case 1083: goto tr2112;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2110;
	goto tr1674;
tr2114:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st709;
tr2115:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st709;
tr2117:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st709;
tr2107:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st709;
tr2108:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st709;
tr2109:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st709;
tr2111:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st709;
st709:
	if ( ++p == pe )
		goto _test_eof709;
case 709:
#line 59297 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st709;
		case 32: goto st709;
		case 40: goto tr2114;
		case 41: goto tr2115;
		case 1034: goto tr2117;
		case 1083: goto st714;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2116;
	goto tr1674;
tr2116:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st710;
tr2122:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st710;
st710:
	if ( ++p == pe )
		goto _test_eof710;
case 710:
#line 59365 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2119;
		case 32: goto tr2119;
		case 40: goto tr2120;
		case 41: goto tr2121;
		case 1034: goto tr2123;
		case 1083: goto tr2124;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2122;
	goto tr1674;
tr2126:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st711;
tr2127:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st711;
tr2129:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st711;
tr2119:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st711;
tr2120:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st711;
tr2121:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st711;
tr2123:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st711;
st711:
	if ( ++p == pe )
		goto _test_eof711;
case 711:
#line 59509 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st711;
		case 32: goto st711;
		case 40: goto tr2126;
		case 41: goto tr2127;
		case 42: goto tr2128;
		case 92: goto tr2128;
		case 95: goto tr2128;
		case 1034: goto tr2129;
		case 1083: goto st713;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2128;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2128;
	} else
		goto tr2128;
	goto tr69;
tr2128:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 712; goto st249;} }
	goto st712;
st712:
	if ( ++p == pe )
		goto _test_eof712;
case 712:
#line 59556 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr2131;
		case 59: goto tr2131;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2131;
	} else if ( (*p) >= 9 )
		goto tr2131;
	goto tr69;
tr2131:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1113;
st1113:
	if ( ++p == pe )
		goto _test_eof1113;
case 1113:
#line 59581 "zscanner/scanner.c"
	goto st0;
tr2124:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st713;
st713:
	if ( ++p == pe )
		goto _test_eof713;
case 713:
#line 59604 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2129;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st713;
	goto tr69;
tr2112:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st714;
st714:
	if ( ++p == pe )
		goto _test_eof714;
case 714:
#line 59647 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2117;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st714;
	goto tr69;
tr2100:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st715;
st715:
	if ( ++p == pe )
		goto _test_eof715;
case 715:
#line 59690 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2105;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st715;
	goto tr69;
st716:
	if ( ++p == pe )
		goto _test_eof716;
case 716:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2132;
	goto tr1674;
tr2132:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st717;
tr2136:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st717;
st717:
	if ( ++p == pe )
		goto _test_eof717;
case 717:
#line 59766 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2133;
		case 32: goto tr2133;
		case 40: goto tr2134;
		case 41: goto tr2135;
		case 1034: goto tr2137;
		case 1083: goto tr2138;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2136;
	goto tr1674;
tr2140:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st718;
tr2141:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st718;
tr2143:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st718;
tr2133:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st718;
tr2134:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st718;
tr2135:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st718;
tr2137:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st718;
st718:
	if ( ++p == pe )
		goto _test_eof718;
case 718:
#line 59890 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st718;
		case 32: goto st718;
		case 40: goto tr2140;
		case 41: goto tr2141;
		case 1034: goto tr2143;
		case 1083: goto st732;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2142;
	goto tr1674;
tr2142:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st719;
tr2148:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st719;
st719:
	if ( ++p == pe )
		goto _test_eof719;
case 719:
#line 59958 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2145;
		case 32: goto tr2145;
		case 40: goto tr2146;
		case 41: goto tr2147;
		case 1034: goto tr2149;
		case 1083: goto tr2150;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2148;
	goto tr1674;
tr2153:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st720;
tr2154:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st720;
tr2155:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st720;
tr2145:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st720;
tr2146:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st720;
tr2147:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st720;
tr2149:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st720;
st720:
	if ( ++p == pe )
		goto _test_eof720;
case 720:
#line 60082 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st720;
		case 32: goto st720;
		case 40: goto tr2153;
		case 41: goto tr2154;
		case 1034: goto tr2155;
		case 1083: goto st731;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2151;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2151;
	} else
		goto tr2151;
	goto tr69;
tr2151:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 565 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 721; goto st259;} }
	goto st721;
st721:
	if ( ++p == pe )
		goto _test_eof721;
case 721:
#line 60126 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2157;
		case 32: goto tr2157;
		case 40: goto tr2158;
		case 41: goto tr2159;
		case 1034: goto tr2160;
		case 1083: goto tr2161;
	}
	goto tr69;
tr2164:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st722;
tr2165:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st722;
tr2166:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st722;
tr2157:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st722;
tr2158:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st722;
tr2159:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st722;
tr2160:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st722;
st722:
	if ( ++p == pe )
		goto _test_eof722;
case 722:
#line 60252 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st722;
		case 32: goto st722;
		case 40: goto tr2164;
		case 41: goto tr2165;
		case 1034: goto tr2166;
		case 1083: goto st730;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2162;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2162;
	} else
		goto tr2162;
	goto tr69;
tr2162:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 565 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 723; goto st259;} }
	goto st723;
st723:
	if ( ++p == pe )
		goto _test_eof723;
case 723:
#line 60296 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2168;
		case 32: goto tr2168;
		case 40: goto tr2169;
		case 41: goto tr2170;
		case 1034: goto tr2171;
		case 1083: goto tr2172;
	}
	goto tr69;
tr2175:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st724;
tr2176:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st724;
tr2177:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st724;
tr2168:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st724;
tr2169:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st724;
tr2170:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st724;
tr2171:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st724;
st724:
	if ( ++p == pe )
		goto _test_eof724;
case 724:
#line 60422 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st724;
		case 32: goto st724;
		case 40: goto tr2175;
		case 41: goto tr2176;
		case 1034: goto tr2177;
		case 1083: goto st729;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2173;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2173;
	} else
		goto tr2173;
	goto tr69;
tr2173:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 565 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 725; goto st259;} }
	goto st725;
st725:
	if ( ++p == pe )
		goto _test_eof725;
case 725:
#line 60466 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2179;
		case 32: goto tr2179;
		case 40: goto tr2180;
		case 41: goto tr2181;
		case 1034: goto tr2182;
		case 1083: goto tr2183;
	}
	goto tr69;
tr2185:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st726;
tr2186:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st726;
tr2188:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st726;
tr2179:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st726;
tr2180:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st726;
tr2181:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st726;
tr2182:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st726;
st726:
	if ( ++p == pe )
		goto _test_eof726;
case 726:
#line 60612 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st726;
		case 32: goto st726;
		case 40: goto tr2185;
		case 41: goto tr2186;
		case 42: goto tr2187;
		case 92: goto tr2187;
		case 95: goto tr2187;
		case 1034: goto tr2188;
		case 1083: goto st728;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2187;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2187;
	} else
		goto tr2187;
	goto tr69;
tr2187:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 727; goto st249;} }
	goto st727;
st727:
	if ( ++p == pe )
		goto _test_eof727;
case 727:
#line 60659 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr2190;
		case 59: goto tr2190;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2190;
	} else if ( (*p) >= 9 )
		goto tr2190;
	goto tr69;
tr2190:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1114;
st1114:
	if ( ++p == pe )
		goto _test_eof1114;
case 1114:
#line 60684 "zscanner/scanner.c"
	goto st0;
tr2183:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st728;
st728:
	if ( ++p == pe )
		goto _test_eof728;
case 728:
#line 60708 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2188;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st728;
	goto tr69;
tr2172:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st729;
st729:
	if ( ++p == pe )
		goto _test_eof729;
case 729:
#line 60752 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2177;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st729;
	goto tr69;
tr2161:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st730;
st730:
	if ( ++p == pe )
		goto _test_eof730;
case 730:
#line 60796 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2166;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st730;
	goto tr69;
tr2150:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st731;
st731:
	if ( ++p == pe )
		goto _test_eof731;
case 731:
#line 60839 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2155;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st731;
	goto tr69;
tr2138:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st732;
st732:
	if ( ++p == pe )
		goto _test_eof732;
case 732:
#line 60882 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2143;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st732;
	goto tr69;
st733:
	if ( ++p == pe )
		goto _test_eof733;
case 733:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2191;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2191;
	} else
		goto tr2191;
	goto tr69;
tr2191:
#line 1617 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 734; goto st552;} }
	goto st734;
st734:
	if ( ++p == pe )
		goto _test_eof734;
case 734:
#line 60930 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st735;
		case 32: goto st735;
		case 40: goto tr2193;
		case 41: goto tr2194;
		case 1034: goto tr2195;
		case 1083: goto st743;
	}
	goto tr69;
tr2193:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st735;
tr2194:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st735;
tr2195:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st735;
st735:
	if ( ++p == pe )
		goto _test_eof735;
case 735:
#line 60984 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st735;
		case 32: goto st735;
		case 40: goto tr2193;
		case 41: goto tr2194;
		case 1034: goto tr2195;
		case 1083: goto st743;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2197;
	goto tr1674;
tr2197:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st736;
tr2201:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st736;
st736:
	if ( ++p == pe )
		goto _test_eof736;
case 736:
#line 61052 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2198;
		case 32: goto tr2198;
		case 40: goto tr2199;
		case 41: goto tr2200;
		case 1034: goto tr2202;
		case 1083: goto tr2203;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2201;
	goto tr1674;
tr2205:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st737;
tr2206:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st737;
tr2208:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st737;
tr2198:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st737;
tr2199:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st737;
tr2200:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st737;
tr2202:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st737;
st737:
	if ( ++p == pe )
		goto _test_eof737;
case 737:
#line 61176 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st737;
		case 32: goto st737;
		case 40: goto tr2205;
		case 41: goto tr2206;
		case 1034: goto tr2208;
		case 1083: goto st742;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2207;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2207;
	} else
		goto tr2207;
	goto tr69;
tr2207:
#line 1602 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 738; goto st457;} }
	goto st738;
st738:
	if ( ++p == pe )
		goto _test_eof738;
case 738:
#line 61216 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st739;
		case 32: goto st739;
		case 40: goto tr2211;
		case 41: goto tr2212;
		case 1034: goto tr2213;
		case 1083: goto st741;
	}
	goto tr69;
tr2211:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st739;
tr2212:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st739;
tr2213:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st739;
st739:
	if ( ++p == pe )
		goto _test_eof739;
case 739:
#line 61270 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st739;
		case 32: goto st739;
		case 40: goto tr2211;
		case 41: goto tr2212;
		case 43: goto tr2215;
		case 1034: goto tr2213;
		case 1083: goto st741;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2215;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2215;
	} else
		goto tr2215;
	goto tr69;
tr2215:
#line 996 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 740; goto st311;} }
	goto st740;
st740:
	if ( ++p == pe )
		goto _test_eof740;
case 740:
#line 61311 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1280 + ((*p) - 0));
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1280 + ((*p) - 0));
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2216;
		case 1595: goto tr2216;
	}
	goto tr69;
tr2216:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1115;
st1115:
	if ( ++p == pe )
		goto _test_eof1115;
case 1115:
#line 61341 "zscanner/scanner.c"
	goto st0;
st741:
	if ( ++p == pe )
		goto _test_eof741;
case 741:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2213;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st741;
	goto tr69;
tr2203:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st742;
st742:
	if ( ++p == pe )
		goto _test_eof742;
case 742:
#line 61389 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2208;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st742;
	goto tr69;
st743:
	if ( ++p == pe )
		goto _test_eof743;
case 743:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2195;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st743;
	goto tr69;
tr2220:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st744;
tr2221:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st744;
tr2224:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st744;
tr2237:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	goto st744;
tr2238:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st744;
tr2239:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st744;
tr2241:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st744;
st744:
	if ( ++p == pe )
		goto _test_eof744;
case 744:
#line 61692 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st744;
		case 32: goto st744;
		case 33: goto tr2219;
		case 40: goto tr2220;
		case 41: goto tr2221;
		case 49: goto tr2222;
		case 50: goto tr2223;
		case 2058: goto tr2224;
		case 2107: goto st751;
		case 2314: goto tr2226;
		case 2363: goto tr2226;
		case 2570: goto tr2227;
		case 2619: goto tr2228;
	}
	goto tr2217;
tr2219:
#line 818 "./zscanner/scanner_body.rl"
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
#line 821 "./zscanner/scanner_body.rl"
	{
		s->apl.excl_flag = 128; // dec 128  = bin 10000000.
	}
	goto st745;
st745:
	if ( ++p == pe )
		goto _test_eof745;
case 745:
#line 61743 "zscanner/scanner.c"
	switch( (*p) ) {
		case 49: goto tr2229;
		case 50: goto tr2230;
	}
	goto tr2217;
tr2222:
#line 818 "./zscanner/scanner_body.rl"
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
#line 824 "./zscanner/scanner_body.rl"
	{
		s->apl.addr_family = 1;
	}
	goto st746;
tr2229:
#line 824 "./zscanner/scanner_body.rl"
	{
		s->apl.addr_family = 1;
	}
	goto st746;
st746:
	if ( ++p == pe )
		goto _test_eof746;
case 746:
#line 61769 "zscanner/scanner.c"
	if ( (*p) == 58 )
		goto st747;
	goto tr2217;
st747:
	if ( ++p == pe )
		goto _test_eof747;
case 747:
	if ( (*p) == 46 )
		goto tr2233;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2233;
	goto tr2232;
tr2233:
#line 763 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st748;
tr2234:
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st748;
st748:
	if ( ++p == pe )
		goto _test_eof748;
case 748:
#line 61814 "zscanner/scanner.c"
	if ( (*p) == 47 )
		goto tr2235;
	if ( 46 <= (*p) && (*p) <= 57 )
		goto tr2234;
	goto tr2232;
tr2235:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
	goto st749;
tr2249:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
	goto st749;
st749:
	if ( ++p == pe )
		goto _test_eof749;
case 749:
#line 61846 "zscanner/scanner.c"
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2236;
	goto tr2217;
tr2236:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st750;
tr2240:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st750;
st750:
	if ( ++p == pe )
		goto _test_eof750;
case 750:
#line 61892 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2237;
		case 32: goto tr2237;
		case 40: goto tr2238;
		case 41: goto tr2239;
		case 2058: goto tr2241;
		case 2107: goto tr2242;
		case 2314: goto tr2243;
		case 2363: goto tr2243;
		case 2570: goto tr2244;
		case 2619: goto tr2245;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2240;
	goto tr2217;
tr2242:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	goto st751;
st751:
	if ( ++p == pe )
		goto _test_eof751;
case 751:
#line 61981 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2224;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st751;
	goto tr69;
tr2226:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1116;
tr2243:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1116;
st1116:
	if ( ++p == pe )
		goto _test_eof1116;
case 1116:
#line 62071 "zscanner/scanner.c"
	goto st0;
tr2227:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1117;
tr2244:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1117;
st1117:
	if ( ++p == pe )
		goto _test_eof1117;
case 1117:
#line 62144 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st744;
		case 32: goto st744;
		case 33: goto tr2219;
		case 40: goto tr2220;
		case 41: goto tr2221;
		case 49: goto tr2222;
		case 50: goto tr2223;
		case 2058: goto tr2224;
		case 2107: goto st751;
		case 2314: goto tr2226;
		case 2363: goto tr2226;
		case 2570: goto tr2227;
		case 2619: goto tr2228;
	}
	goto tr2217;
tr2223:
#line 818 "./zscanner/scanner_body.rl"
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
#line 827 "./zscanner/scanner_body.rl"
	{
		s->apl.addr_family = 2;
	}
	goto st752;
tr2230:
#line 827 "./zscanner/scanner_body.rl"
	{
		s->apl.addr_family = 2;
	}
	goto st752;
st752:
	if ( ++p == pe )
		goto _test_eof752;
case 752:
#line 62201 "zscanner/scanner.c"
	if ( (*p) == 58 )
		goto st753;
	goto tr2217;
st753:
	if ( ++p == pe )
		goto _test_eof753;
case 753:
	if ( (*p) == 46 )
		goto tr2247;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr2247;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2247;
	} else
		goto tr2247;
	goto tr2232;
tr2247:
#line 763 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st754;
tr2248:
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st754;
st754:
	if ( ++p == pe )
		goto _test_eof754;
case 754:
#line 62252 "zscanner/scanner.c"
	if ( (*p) == 47 )
		goto tr2249;
	if ( (*p) < 65 ) {
		if ( 46 <= (*p) && (*p) <= 58 )
			goto tr2248;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2248;
	} else
		goto tr2248;
	goto tr2232;
tr2228:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1118;
tr2245:
#line 830 "./zscanner/scanner_body.rl"
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
	}
#line 839 "./zscanner/scanner_body.rl"
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			SCANNER_WARNING(ZSCANNER_EBAD_APL);
			p--; {goto st247;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1118;
st1118:
	if ( ++p == pe )
		goto _test_eof1118;
case 1118:
#line 62327 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2224;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st751;
	goto tr69;
st755:
	if ( ++p == pe )
		goto _test_eof755;
case 755:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2250;
	goto tr1674;
tr2250:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st756;
tr2254:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st756;
st756:
	if ( ++p == pe )
		goto _test_eof756;
case 756:
#line 62403 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2251;
		case 32: goto tr2251;
		case 40: goto tr2252;
		case 41: goto tr2253;
		case 1034: goto tr2255;
		case 1083: goto tr2256;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2254;
	goto tr1674;
tr2258:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st757;
tr2259:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st757;
tr2261:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st757;
tr2251:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st757;
tr2252:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st757;
tr2253:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st757;
tr2255:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st757;
st757:
	if ( ++p == pe )
		goto _test_eof757;
case 757:
#line 62527 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st757;
		case 32: goto st757;
		case 40: goto tr2258;
		case 41: goto tr2259;
		case 1034: goto tr2261;
		case 1083: goto st767;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2260;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2260;
	} else
		goto tr2260;
	goto tr69;
tr2260:
#line 1602 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 758; goto st457;} }
	goto st758;
st758:
	if ( ++p == pe )
		goto _test_eof758;
case 758:
#line 62567 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st759;
		case 32: goto st759;
		case 40: goto tr2264;
		case 41: goto tr2265;
		case 1034: goto tr2266;
		case 1083: goto st766;
	}
	goto tr69;
tr2264:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st759;
tr2265:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st759;
tr2266:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st759;
st759:
	if ( ++p == pe )
		goto _test_eof759;
case 759:
#line 62621 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st759;
		case 32: goto st759;
		case 40: goto tr2264;
		case 41: goto tr2265;
		case 1034: goto tr2266;
		case 1083: goto st766;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2268;
	goto tr1674;
tr2268:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st760;
tr2272:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st760;
st760:
	if ( ++p == pe )
		goto _test_eof760;
case 760:
#line 62689 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2269;
		case 32: goto tr2269;
		case 40: goto tr2270;
		case 41: goto tr2271;
		case 1034: goto tr2273;
		case 1083: goto tr2274;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2272;
	goto tr1674;
tr2277:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st761;
tr2278:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st761;
tr2280:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st761;
tr2269:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st761;
tr2270:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st761;
tr2271:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st761;
tr2273:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st761;
st761:
	if ( ++p == pe )
		goto _test_eof761;
case 761:
#line 62813 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st761;
		case 32: goto st761;
		case 40: goto tr2277;
		case 41: goto tr2278;
		case 1034: goto tr2280;
		case 1083: goto st765;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2279;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2279;
	} else
		goto tr2279;
	goto tr2275;
tr2279:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st762;
st762:
	if ( ++p == pe )
		goto _test_eof762;
case 762:
#line 62860 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2282;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2282;
	} else
		goto tr2282;
	goto tr2275;
tr2284:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st763;
tr2285:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st763;
tr2286:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st763;
tr2282:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st763;
st763:
	if ( ++p == pe )
		goto _test_eof763;
case 763:
#line 62907 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st763;
		case 32: goto st763;
		case 40: goto tr2284;
		case 41: goto tr2285;
		case 2058: goto tr2286;
		case 2107: goto st764;
		case 2314: goto tr2288;
		case 2363: goto tr2288;
		case 2570: goto tr2289;
		case 2619: goto tr2290;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2279;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2279;
	} else
		goto tr2279;
	goto tr2275;
st764:
	if ( ++p == pe )
		goto _test_eof764;
case 764:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2286;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st764;
	goto tr2275;
tr2288:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1119;
st1119:
	if ( ++p == pe )
		goto _test_eof1119;
case 1119:
#line 62989 "zscanner/scanner.c"
	goto st0;
tr2289:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1120;
st1120:
	if ( ++p == pe )
		goto _test_eof1120;
case 1120:
#line 63005 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st763;
		case 32: goto st763;
		case 40: goto tr2284;
		case 41: goto tr2285;
		case 2058: goto tr2286;
		case 2107: goto st764;
		case 2314: goto tr2288;
		case 2363: goto tr2288;
		case 2570: goto tr2289;
		case 2619: goto tr2290;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2279;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2279;
	} else
		goto tr2279;
	goto tr2275;
tr2290:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
st1121:
	if ( ++p == pe )
		goto _test_eof1121;
case 1121:
#line 63057 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2286;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st764;
	goto tr2275;
tr2274:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st765;
st765:
	if ( ++p == pe )
		goto _test_eof765;
case 765:
#line 63100 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2280;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st765;
	goto tr69;
st766:
	if ( ++p == pe )
		goto _test_eof766;
case 766:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2266;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st766;
	goto tr69;
tr2256:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st767;
st767:
	if ( ++p == pe )
		goto _test_eof767;
case 767:
#line 63173 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2261;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st767;
	goto tr69;
st768:
	if ( ++p == pe )
		goto _test_eof768;
case 768:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2291;
	goto tr1674;
tr2291:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st769;
tr2295:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st769;
st769:
	if ( ++p == pe )
		goto _test_eof769;
case 769:
#line 63249 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2292;
		case 32: goto tr2292;
		case 40: goto tr2293;
		case 41: goto tr2294;
		case 1034: goto tr2296;
		case 1083: goto tr2297;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2295;
	goto tr1674;
tr2299:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st770;
tr2300:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st770;
tr2302:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st770;
tr2292:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st770;
tr2293:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st770;
tr2294:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st770;
tr2296:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st770;
st770:
	if ( ++p == pe )
		goto _test_eof770;
case 770:
#line 63373 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st770;
		case 32: goto st770;
		case 40: goto tr2299;
		case 41: goto tr2300;
		case 1034: goto tr2302;
		case 1083: goto st777;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2301;
	goto tr1674;
tr2301:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st771;
tr2307:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st771;
st771:
	if ( ++p == pe )
		goto _test_eof771;
case 771:
#line 63441 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2304;
		case 32: goto tr2304;
		case 40: goto tr2305;
		case 41: goto tr2306;
		case 1034: goto tr2308;
		case 1083: goto tr2309;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2307;
	goto tr1674;
tr2311:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st772;
tr2312:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st772;
tr2314:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st772;
tr2304:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st772;
tr2305:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st772;
tr2306:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st772;
tr2308:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st772;
st772:
	if ( ++p == pe )
		goto _test_eof772;
case 772:
#line 63565 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st772;
		case 32: goto st772;
		case 40: goto tr2311;
		case 41: goto tr2312;
		case 1034: goto tr2314;
		case 1083: goto st776;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2313;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2313;
	} else
		goto tr2313;
	goto tr2275;
tr2313:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st773;
st773:
	if ( ++p == pe )
		goto _test_eof773;
case 773:
#line 63612 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2316;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2316;
	} else
		goto tr2316;
	goto tr2275;
tr2318:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st774;
tr2319:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st774;
tr2320:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st774;
tr2316:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st774;
st774:
	if ( ++p == pe )
		goto _test_eof774;
case 774:
#line 63659 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st774;
		case 32: goto st774;
		case 40: goto tr2318;
		case 41: goto tr2319;
		case 2058: goto tr2320;
		case 2107: goto st775;
		case 2314: goto tr2322;
		case 2363: goto tr2322;
		case 2570: goto tr2323;
		case 2619: goto tr2324;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2313;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2313;
	} else
		goto tr2313;
	goto tr2275;
st775:
	if ( ++p == pe )
		goto _test_eof775;
case 775:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2320;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st775;
	goto tr2275;
tr2322:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1122;
st1122:
	if ( ++p == pe )
		goto _test_eof1122;
case 1122:
#line 63741 "zscanner/scanner.c"
	goto st0;
tr2323:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1123;
st1123:
	if ( ++p == pe )
		goto _test_eof1123;
case 1123:
#line 63757 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st774;
		case 32: goto st774;
		case 40: goto tr2318;
		case 41: goto tr2319;
		case 2058: goto tr2320;
		case 2107: goto st775;
		case 2314: goto tr2322;
		case 2363: goto tr2322;
		case 2570: goto tr2323;
		case 2619: goto tr2324;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2313;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2313;
	} else
		goto tr2313;
	goto tr2275;
tr2324:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1124;
st1124:
	if ( ++p == pe )
		goto _test_eof1124;
case 1124:
#line 63809 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2320;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st775;
	goto tr2275;
tr2309:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st776;
st776:
	if ( ++p == pe )
		goto _test_eof776;
case 776:
#line 63852 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2314;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st776;
	goto tr69;
tr2297:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st777;
st777:
	if ( ++p == pe )
		goto _test_eof777;
case 777:
#line 63895 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2302;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st777;
	goto tr69;
st778:
	if ( ++p == pe )
		goto _test_eof778;
case 778:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2325;
	goto tr1674;
tr2325:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st779;
tr2329:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st779;
st779:
	if ( ++p == pe )
		goto _test_eof779;
case 779:
#line 63971 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2326;
		case 32: goto tr2326;
		case 40: goto tr2327;
		case 41: goto tr2328;
		case 1034: goto tr2330;
		case 1083: goto tr2331;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2329;
	goto tr1674;
tr2334:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st780;
tr2335:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st780;
tr2340:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st780;
tr2326:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st780;
tr2327:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st780;
tr2328:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st780;
tr2330:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st780;
st780:
	if ( ++p == pe )
		goto _test_eof780;
case 780:
#line 64095 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st780;
		case 32: goto st780;
		case 40: goto tr2334;
		case 41: goto tr2335;
		case 48: goto tr2336;
		case 49: goto tr2337;
		case 50: goto tr2338;
		case 51: goto tr2339;
		case 1034: goto tr2340;
		case 1083: goto st816;
	}
	goto tr2332;
tr2336:
#line 1093 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 0;
	}
	goto st781;
st781:
	if ( ++p == pe )
		goto _test_eof781;
case 781:
#line 64133 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st782;
		case 32: goto st782;
		case 40: goto tr2343;
		case 41: goto tr2344;
		case 1034: goto tr2345;
		case 1083: goto st794;
	}
	goto tr2332;
tr2343:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st782;
tr2344:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st782;
tr2345:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st782;
st782:
	if ( ++p == pe )
		goto _test_eof782;
case 782:
#line 64187 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st782;
		case 32: goto st782;
		case 40: goto tr2343;
		case 41: goto tr2344;
		case 1034: goto tr2345;
		case 1083: goto st794;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2348;
	goto tr2347;
tr2348:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st783;
tr2352:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st783;
st783:
	if ( ++p == pe )
		goto _test_eof783;
case 783:
#line 64255 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2349;
		case 32: goto tr2349;
		case 40: goto tr2350;
		case 41: goto tr2351;
		case 1034: goto tr2353;
		case 1083: goto tr2354;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2352;
	goto tr2347;
tr2356:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st784;
tr2357:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st784;
tr2359:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st784;
tr2349:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st784;
tr2350:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st784;
tr2351:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st784;
tr2353:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st784;
st784:
	if ( ++p == pe )
		goto _test_eof784;
case 784:
#line 64379 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st784;
		case 32: goto st784;
		case 40: goto tr2356;
		case 41: goto tr2357;
		case 46: goto st785;
		case 1034: goto tr2359;
		case 1083: goto st793;
	}
	goto tr2332;
st785:
	if ( ++p == pe )
		goto _test_eof785;
case 785:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5888 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto st786;
		case 6176: goto st786;
		case 6184: goto tr2363;
		case 6185: goto tr2364;
		case 6409: goto st789;
		case 6432: goto st789;
		case 6440: goto tr2366;
		case 6441: goto tr2367;
		case 6665: goto st791;
		case 6688: goto st791;
		case 6696: goto tr2369;
		case 6697: goto tr2370;
		case 9482: goto tr2371;
		case 9531: goto tr2371;
		case 9738: goto tr2371;
		case 9787: goto tr2371;
		case 10250: goto tr2372;
		case 10299: goto st788;
		case 10506: goto tr2371;
		case 10555: goto tr2371;
		case 10762: goto tr2374;
		case 10811: goto tr2375;
		case 11274: goto tr2376;
		case 11323: goto st790;
		case 11530: goto tr2371;
		case 11579: goto tr2371;
		case 11786: goto tr2378;
		case 11835: goto tr2379;
		case 12298: goto tr2380;
		case 12347: goto st792;
		case 12554: goto tr2371;
		case 12603: goto tr2371;
		case 12810: goto tr2382;
		case 12859: goto tr2383;
	}
	goto tr2361;
tr2363:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st786;
tr2364:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st786;
tr2372:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st786;
tr2407:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st786;
tr2408:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st786;
tr2409:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st786;
tr2417:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st786;
tr2448:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st786;
tr2449:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st786;
tr2450:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st786;
tr2458:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st786;
tr2488:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st786;
tr2489:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st786;
tr2490:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st786;
tr2498:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st786;
st786:
	if ( ++p == pe )
		goto _test_eof786;
case 786:
#line 64759 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3328 + ((*p) - 0));
					if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2816 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3328 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2816 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st786;
		case 3104: goto st786;
		case 3112: goto tr2363;
		case 3113: goto tr2364;
		case 3115: goto tr2385;
		case 4106: goto tr2372;
		case 4155: goto st788;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2385;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2385;
	} else
		goto tr2385;
	goto tr2384;
tr2385:
#line 996 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 787; goto st311;} }
	goto st787;
st787:
	if ( ++p == pe )
		goto _test_eof787;
case 787:
#line 64869 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1280 + ((*p) - 0));
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1280 + ((*p) - 0));
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2371;
		case 1595: goto tr2371;
	}
	goto tr2384;
tr2371:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1125;
tr2416:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1125;
tr2497:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1125;
tr2457:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1125;
st1125:
	if ( ++p == pe )
		goto _test_eof1125;
case 1125:
#line 64949 "zscanner/scanner.c"
	goto st0;
tr2418:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st788;
tr2459:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st788;
tr2499:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st788;
st788:
	if ( ++p == pe )
		goto _test_eof788;
case 788:
#line 64993 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3328 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3328 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3328 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2372;
	if ( 4096 <= _widec && _widec <= 4351 )
		goto st788;
	goto tr2384;
tr2366:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st789;
tr2367:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st789;
tr2376:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st789;
tr2410:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st789;
tr2411:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st789;
tr2412:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st789;
tr2421:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st789;
tr2451:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st789;
tr2452:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st789;
tr2453:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st789;
tr2462:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st789;
tr2491:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st789;
tr2492:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st789;
tr2493:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st789;
tr2502:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st789;
st789:
	if ( ++p == pe )
		goto _test_eof789;
case 789:
#line 65271 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(13056 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4352 + ((*p) - 0));
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4352 + ((*p) - 0));
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(13056 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4352 + ((*p) - 0));
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4352 + ((*p) - 0));
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st789;
		case 4640: goto st789;
		case 4648: goto tr2366;
		case 4649: goto tr2367;
		case 13578: goto tr2371;
		case 13627: goto tr2371;
		case 13834: goto tr2371;
		case 13883: goto tr2371;
		case 14346: goto tr2376;
		case 14395: goto st790;
		case 14602: goto tr2371;
		case 14651: goto tr2371;
		case 14858: goto tr2378;
		case 14907: goto tr2379;
	}
	goto tr2384;
tr2422:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st790;
tr2463:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st790;
tr2503:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st790;
st790:
	if ( ++p == pe )
		goto _test_eof790;
case 790:
#line 65385 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4864 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4864 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4864 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2376;
	if ( 5632 <= _widec && _widec <= 5887 )
		goto st790;
	goto tr2384;
tr2378:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1126;
tr2423:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1126;
tr2464:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1126;
tr2504:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1126;
st1126:
	if ( ++p == pe )
		goto _test_eof1126;
case 1126:
#line 65497 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(13056 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4352 + ((*p) - 0));
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4352 + ((*p) - 0));
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(13056 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4352 + ((*p) - 0));
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4352 + ((*p) - 0));
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st789;
		case 4640: goto st789;
		case 4648: goto tr2366;
		case 4649: goto tr2367;
		case 13578: goto tr2371;
		case 13627: goto tr2371;
		case 13834: goto tr2371;
		case 13883: goto tr2371;
		case 14346: goto tr2376;
		case 14395: goto st790;
		case 14602: goto tr2371;
		case 14651: goto tr2371;
		case 14858: goto tr2378;
		case 14907: goto tr2379;
	}
	goto tr2384;
tr2379:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1127;
tr2424:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1127;
tr2505:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1127;
tr2465:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1127;
st1127:
	if ( ++p == pe )
		goto _test_eof1127;
case 1127:
#line 65629 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4864 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4864 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4864 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2376;
	if ( 5632 <= _widec && _widec <= 5887 )
		goto st790;
	goto tr2384;
tr2369:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st791;
tr2370:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st791;
tr2380:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st791;
tr2413:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st791;
tr2414:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st791;
tr2415:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st791;
tr2425:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st791;
tr2454:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st791;
tr2455:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st791;
tr2456:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st791;
tr2466:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st791;
tr2494:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st791;
tr2495:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st791;
tr2496:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st791;
tr2506:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st791;
st791:
	if ( ++p == pe )
		goto _test_eof791;
case 791:
#line 65907 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8960 + ((*p) - 0));
					if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
					if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
					if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5888 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
					if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8960 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2816 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2385;
		case 6153: goto st786;
		case 6176: goto st786;
		case 6184: goto tr2363;
		case 6185: goto tr2364;
		case 6409: goto st789;
		case 6432: goto st789;
		case 6440: goto tr2366;
		case 6441: goto tr2367;
		case 6665: goto st791;
		case 6688: goto st791;
		case 6696: goto tr2369;
		case 6697: goto tr2370;
		case 9482: goto tr2371;
		case 9531: goto tr2371;
		case 9738: goto tr2371;
		case 9787: goto tr2371;
		case 10250: goto tr2372;
		case 10299: goto st788;
		case 10506: goto tr2371;
		case 10555: goto tr2371;
		case 10762: goto tr2374;
		case 10811: goto tr2375;
		case 11274: goto tr2376;
		case 11323: goto st790;
		case 11530: goto tr2371;
		case 11579: goto tr2371;
		case 11786: goto tr2378;
		case 11835: goto tr2379;
		case 12298: goto tr2380;
		case 12347: goto st792;
		case 12554: goto tr2371;
		case 12603: goto tr2371;
		case 12810: goto tr2382;
		case 12859: goto tr2383;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2385;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2385;
	} else
		goto tr2385;
	goto tr2384;
tr2374:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1128;
tr2419:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1128;
tr2460:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1128;
tr2500:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1128;
st1128:
	if ( ++p == pe )
		goto _test_eof1128;
case 1128:
#line 66137 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3328 + ((*p) - 0));
					if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2816 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3328 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2816 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st786;
		case 3104: goto st786;
		case 3112: goto tr2363;
		case 3113: goto tr2364;
		case 3115: goto tr2385;
		case 4106: goto tr2372;
		case 4155: goto st788;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2385;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2385;
	} else
		goto tr2385;
	goto tr2384;
tr2375:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1129;
tr2420:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1129;
tr2501:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1129;
tr2461:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1129;
st1129:
	if ( ++p == pe )
		goto _test_eof1129;
case 1129:
#line 66299 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3328 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3328 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3328 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2372;
	if ( 4096 <= _widec && _widec <= 4351 )
		goto st788;
	goto tr2384;
tr2426:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st792;
tr2467:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st792;
tr2507:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st792;
st792:
	if ( ++p == pe )
		goto _test_eof792;
case 792:
#line 66377 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6912 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6912 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6912 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2372;
		case 8202: goto tr2376;
		case 8714: goto tr2380;
	}
	if ( _widec < 8192 ) {
		if ( 7680 <= _widec && _widec <= 7935 )
			goto st788;
	} else if ( _widec > 8447 ) {
		if ( 8704 <= _widec && _widec <= 8959 )
			goto st792;
	} else
		goto st790;
	goto tr2384;
tr2382:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1130;
tr2427:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1130;
tr2468:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1130;
tr2508:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1130;
st1130:
	if ( ++p == pe )
		goto _test_eof1130;
case 1130:
#line 66507 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8960 + ((*p) - 0));
					if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
					if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
					if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5888 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
					if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2816 + ((*p) - 0));
					if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2816 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8960 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2816 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2385;
		case 6153: goto st786;
		case 6176: goto st786;
		case 6184: goto tr2363;
		case 6185: goto tr2364;
		case 6409: goto st789;
		case 6432: goto st789;
		case 6440: goto tr2366;
		case 6441: goto tr2367;
		case 6665: goto st791;
		case 6688: goto st791;
		case 6696: goto tr2369;
		case 6697: goto tr2370;
		case 9482: goto tr2371;
		case 9531: goto tr2371;
		case 9738: goto tr2371;
		case 9787: goto tr2371;
		case 10250: goto tr2372;
		case 10299: goto st788;
		case 10506: goto tr2371;
		case 10555: goto tr2371;
		case 10762: goto tr2374;
		case 10811: goto tr2375;
		case 11274: goto tr2376;
		case 11323: goto st790;
		case 11530: goto tr2371;
		case 11579: goto tr2371;
		case 11786: goto tr2378;
		case 11835: goto tr2379;
		case 12298: goto tr2380;
		case 12347: goto st792;
		case 12554: goto tr2371;
		case 12603: goto tr2371;
		case 12810: goto tr2382;
		case 12859: goto tr2383;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2385;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2385;
	} else
		goto tr2385;
	goto tr2384;
tr2383:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1131;
tr2428:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1131;
tr2509:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1131;
tr2469:
#line 793 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV6);
			p--; {goto st247;}
		}
	}
#line 801 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1131;
st1131:
	if ( ++p == pe )
		goto _test_eof1131;
case 1131:
#line 66721 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6912 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6912 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6912 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 512;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2372;
		case 8202: goto tr2376;
		case 8714: goto tr2380;
	}
	if ( _widec < 8192 ) {
		if ( 7680 <= _widec && _widec <= 7935 )
			goto st788;
	} else if ( _widec > 8447 ) {
		if ( 8704 <= _widec && _widec <= 8959 )
			goto st792;
	} else
		goto st790;
	goto tr2384;
tr2354:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st793;
st793:
	if ( ++p == pe )
		goto _test_eof793;
case 793:
#line 66791 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2359;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st793;
	goto tr2332;
st794:
	if ( ++p == pe )
		goto _test_eof794;
case 794:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2345;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st794;
	goto tr2332;
tr2337:
#line 1096 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 1;
	}
	goto st795;
st795:
	if ( ++p == pe )
		goto _test_eof795;
case 795:
#line 66858 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st796;
		case 32: goto st796;
		case 40: goto tr2387;
		case 41: goto tr2388;
		case 1034: goto tr2389;
		case 1083: goto st801;
	}
	goto tr2332;
tr2387:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st796;
tr2388:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st796;
tr2389:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st796;
st796:
	if ( ++p == pe )
		goto _test_eof796;
case 796:
#line 66912 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st796;
		case 32: goto st796;
		case 40: goto tr2387;
		case 41: goto tr2388;
		case 1034: goto tr2389;
		case 1083: goto st801;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2391;
	goto tr2347;
tr2391:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st797;
tr2395:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st797;
st797:
	if ( ++p == pe )
		goto _test_eof797;
case 797:
#line 66980 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2392;
		case 32: goto tr2392;
		case 40: goto tr2393;
		case 41: goto tr2394;
		case 1034: goto tr2396;
		case 1083: goto tr2397;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2395;
	goto tr2347;
tr2400:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st798;
tr2401:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st798;
tr2403:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st798;
tr2392:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st798;
tr2393:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st798;
tr2394:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st798;
tr2396:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st798;
st798:
	if ( ++p == pe )
		goto _test_eof798;
case 798:
#line 67104 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st798;
		case 32: goto st798;
		case 40: goto tr2400;
		case 41: goto tr2401;
		case 46: goto tr2402;
		case 1034: goto tr2403;
		case 1083: goto st800;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2402;
	goto tr2398;
tr2402:
#line 763 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st799;
tr2406:
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st799;
st799:
	if ( ++p == pe )
		goto _test_eof799;
case 799:
#line 67163 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5888 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2406;
		case 6153: goto tr2407;
		case 6176: goto tr2407;
		case 6184: goto tr2408;
		case 6185: goto tr2409;
		case 6409: goto tr2410;
		case 6432: goto tr2410;
		case 6440: goto tr2411;
		case 6441: goto tr2412;
		case 6665: goto tr2413;
		case 6688: goto tr2413;
		case 6696: goto tr2414;
		case 6697: goto tr2415;
		case 9482: goto tr2416;
		case 9531: goto tr2416;
		case 9738: goto tr2416;
		case 9787: goto tr2416;
		case 10250: goto tr2417;
		case 10299: goto tr2418;
		case 10506: goto tr2416;
		case 10555: goto tr2416;
		case 10762: goto tr2419;
		case 10811: goto tr2420;
		case 11274: goto tr2421;
		case 11323: goto tr2422;
		case 11530: goto tr2416;
		case 11579: goto tr2416;
		case 11786: goto tr2423;
		case 11835: goto tr2424;
		case 12298: goto tr2425;
		case 12347: goto tr2426;
		case 12554: goto tr2416;
		case 12603: goto tr2416;
		case 12810: goto tr2427;
		case 12859: goto tr2428;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2406;
	goto tr2405;
tr2397:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st800;
st800:
	if ( ++p == pe )
		goto _test_eof800;
case 800:
#line 67292 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2403;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st800;
	goto tr2332;
st801:
	if ( ++p == pe )
		goto _test_eof801;
case 801:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2389;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st801;
	goto tr2332;
tr2338:
#line 1099 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 2;
	}
	goto st802;
st802:
	if ( ++p == pe )
		goto _test_eof802;
case 802:
#line 67359 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st803;
		case 32: goto st803;
		case 40: goto tr2430;
		case 41: goto tr2431;
		case 1034: goto tr2432;
		case 1083: goto st808;
	}
	goto tr2332;
tr2430:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st803;
tr2431:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st803;
tr2432:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st803;
st803:
	if ( ++p == pe )
		goto _test_eof803;
case 803:
#line 67413 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st803;
		case 32: goto st803;
		case 40: goto tr2430;
		case 41: goto tr2431;
		case 1034: goto tr2432;
		case 1083: goto st808;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2434;
	goto tr2347;
tr2434:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st804;
tr2438:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st804;
st804:
	if ( ++p == pe )
		goto _test_eof804;
case 804:
#line 67481 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2435;
		case 32: goto tr2435;
		case 40: goto tr2436;
		case 41: goto tr2437;
		case 1034: goto tr2439;
		case 1083: goto tr2440;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2438;
	goto tr2347;
tr2442:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st805;
tr2443:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st805;
tr2445:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st805;
tr2435:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st805;
tr2436:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st805;
tr2437:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st805;
tr2439:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st805;
st805:
	if ( ++p == pe )
		goto _test_eof805;
case 805:
#line 67605 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st805;
		case 32: goto st805;
		case 40: goto tr2442;
		case 41: goto tr2443;
		case 46: goto tr2444;
		case 1034: goto tr2445;
		case 1083: goto st807;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2444;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2444;
	} else
		goto tr2444;
	goto tr2398;
tr2444:
#line 763 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st806;
tr2447:
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st806;
st806:
	if ( ++p == pe )
		goto _test_eof806;
case 806:
#line 67670 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5888 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2447;
		case 6153: goto tr2448;
		case 6176: goto tr2448;
		case 6184: goto tr2449;
		case 6185: goto tr2450;
		case 6409: goto tr2451;
		case 6432: goto tr2451;
		case 6440: goto tr2452;
		case 6441: goto tr2453;
		case 6665: goto tr2454;
		case 6688: goto tr2454;
		case 6696: goto tr2455;
		case 6697: goto tr2456;
		case 9482: goto tr2457;
		case 9531: goto tr2457;
		case 9738: goto tr2457;
		case 9787: goto tr2457;
		case 10250: goto tr2458;
		case 10299: goto tr2459;
		case 10506: goto tr2457;
		case 10555: goto tr2457;
		case 10762: goto tr2460;
		case 10811: goto tr2461;
		case 11274: goto tr2462;
		case 11323: goto tr2463;
		case 11530: goto tr2457;
		case 11579: goto tr2457;
		case 11786: goto tr2464;
		case 11835: goto tr2465;
		case 12298: goto tr2466;
		case 12347: goto tr2467;
		case 12554: goto tr2457;
		case 12603: goto tr2457;
		case 12810: goto tr2468;
		case 12859: goto tr2469;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2447;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2447;
	} else
		goto tr2447;
	goto tr2405;
tr2440:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st807;
st807:
	if ( ++p == pe )
		goto _test_eof807;
case 807:
#line 67805 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2445;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st807;
	goto tr2332;
st808:
	if ( ++p == pe )
		goto _test_eof808;
case 808:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2432;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st808;
	goto tr2332;
tr2339:
#line 1102 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) = 3;
	}
	goto st809;
st809:
	if ( ++p == pe )
		goto _test_eof809;
case 809:
#line 67872 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st810;
		case 32: goto st810;
		case 40: goto tr2471;
		case 41: goto tr2472;
		case 1034: goto tr2473;
		case 1083: goto st815;
	}
	goto tr2332;
tr2471:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st810;
tr2472:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st810;
tr2473:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st810;
st810:
	if ( ++p == pe )
		goto _test_eof810;
case 810:
#line 67926 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st810;
		case 32: goto st810;
		case 40: goto tr2471;
		case 41: goto tr2472;
		case 1034: goto tr2473;
		case 1083: goto st815;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2475;
	goto tr2347;
tr2475:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st811;
tr2479:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st811;
st811:
	if ( ++p == pe )
		goto _test_eof811;
case 811:
#line 67994 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2476;
		case 32: goto tr2476;
		case 40: goto tr2477;
		case 41: goto tr2478;
		case 1034: goto tr2480;
		case 1083: goto tr2481;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2479;
	goto tr2347;
tr2483:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st812;
tr2484:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st812;
tr2486:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st812;
tr2476:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st812;
tr2477:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st812;
tr2478:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st812;
tr2480:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st812;
st812:
	if ( ++p == pe )
		goto _test_eof812;
case 812:
#line 68118 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st812;
		case 32: goto st812;
		case 40: goto tr2483;
		case 41: goto tr2484;
		case 42: goto tr2485;
		case 92: goto tr2485;
		case 95: goto tr2485;
		case 1034: goto tr2486;
		case 1083: goto st814;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2485;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2485;
	} else
		goto tr2485;
	goto tr2332;
tr2485:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 813; goto st249;} }
	goto st813;
st813:
	if ( ++p == pe )
		goto _test_eof813;
case 813:
#line 68165 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5888 + ((*p) - 0));
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8960 + ((*p) - 0));
				if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
				if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
				if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 1024;
				if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5888 + ((*p) - 0));
			if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
			if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5888 + ((*p) - 0));
		if ( 
#line 1197 "./zscanner/scanner_body.rl"
 s->number64 != 0  ) _widec += 256;
		if ( 
#line 1198 "./zscanner/scanner_body.rl"
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto tr2488;
		case 6176: goto tr2488;
		case 6184: goto tr2489;
		case 6185: goto tr2490;
		case 6409: goto tr2491;
		case 6432: goto tr2491;
		case 6440: goto tr2492;
		case 6441: goto tr2493;
		case 6665: goto tr2494;
		case 6688: goto tr2494;
		case 6696: goto tr2495;
		case 6697: goto tr2496;
		case 9482: goto tr2497;
		case 9531: goto tr2497;
		case 9738: goto tr2497;
		case 9787: goto tr2497;
		case 10250: goto tr2498;
		case 10299: goto tr2499;
		case 10506: goto tr2497;
		case 10555: goto tr2497;
		case 10762: goto tr2500;
		case 10811: goto tr2501;
		case 11274: goto tr2502;
		case 11323: goto tr2503;
		case 11530: goto tr2497;
		case 11579: goto tr2497;
		case 11786: goto tr2504;
		case 11835: goto tr2505;
		case 12298: goto tr2506;
		case 12347: goto tr2507;
		case 12554: goto tr2497;
		case 12603: goto tr2497;
		case 12810: goto tr2508;
		case 12859: goto tr2509;
	}
	goto tr2361;
tr2481:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st814;
st814:
	if ( ++p == pe )
		goto _test_eof814;
case 814:
#line 68291 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2486;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st814;
	goto tr2332;
st815:
	if ( ++p == pe )
		goto _test_eof815;
case 815:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2473;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st815;
	goto tr2332;
tr2331:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st816;
st816:
	if ( ++p == pe )
		goto _test_eof816;
case 816:
#line 68364 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2340;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st816;
	goto tr69;
st817:
	if ( ++p == pe )
		goto _test_eof817;
case 817:
	switch( (*p) ) {
		case 65: goto st818;
		case 67: goto st852;
		case 68: goto st860;
		case 69: goto st874;
		case 72: goto st881;
		case 73: goto st886;
		case 75: goto st894;
		case 76: goto st898;
		case 77: goto st906;
		case 78: goto st912;
		case 80: goto st928;
		case 82: goto st931;
		case 83: goto st938;
		case 84: goto st949;
		case 97: goto st818;
		case 99: goto st852;
		case 100: goto st860;
		case 101: goto st874;
		case 104: goto st881;
		case 105: goto st886;
		case 107: goto st894;
		case 108: goto st898;
		case 109: goto st906;
		case 110: goto st912;
		case 112: goto st928;
		case 114: goto st931;
		case 115: goto st938;
		case 116: goto st949;
	}
	goto tr2510;
st818:
	if ( ++p == pe )
		goto _test_eof818;
case 818:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2525;
		case 32: goto tr2525;
		case 40: goto tr2526;
		case 41: goto tr2527;
		case 65: goto st843;
		case 70: goto st846;
		case 80: goto st850;
		case 97: goto st843;
		case 102: goto st846;
		case 112: goto st850;
		case 1034: goto tr2531;
		case 1083: goto tr2532;
	}
	goto tr2510;
tr2534:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2535:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2537:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2904:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st819;
tr2905:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2906:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2908:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2525:
#line 1210 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	goto st819;
tr2526:
#line 1210 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2527:
#line 1210 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2531:
#line 1210 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2620:
#line 1223 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	goto st819;
tr2621:
#line 1223 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2622:
#line 1223 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2623:
#line 1223 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2628:
#line 1220 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	goto st819;
tr2629:
#line 1220 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2630:
#line 1220 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2631:
#line 1220 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2634:
#line 1230 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	goto st819;
tr2635:
#line 1230 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2636:
#line 1230 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2637:
#line 1230 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2643:
#line 1228 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	goto st819;
tr2644:
#line 1228 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2645:
#line 1228 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2646:
#line 1228 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2651:
#line 1212 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	goto st819;
tr2652:
#line 1212 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2653:
#line 1212 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2654:
#line 1212 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2662:
#line 1237 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	goto st819;
tr2663:
#line 1237 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2664:
#line 1237 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2665:
#line 1237 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2671:
#line 1229 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	goto st819;
tr2672:
#line 1229 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2673:
#line 1229 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2674:
#line 1229 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2679:
#line 1236 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	goto st819;
tr2680:
#line 1236 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2681:
#line 1236 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2682:
#line 1236 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2684:
#line 1231 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	goto st819;
tr2685:
#line 1231 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2686:
#line 1231 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2687:
#line 1231 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2694:
#line 1246 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	goto st819;
tr2695:
#line 1246 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2696:
#line 1246 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2697:
#line 1246 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2700:
#line 1247 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	goto st819;
tr2701:
#line 1247 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2702:
#line 1247 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2703:
#line 1247 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2709:
#line 1215 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	goto st819;
tr2710:
#line 1215 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2711:
#line 1215 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2712:
#line 1215 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2721:
#line 1233 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	goto st819;
tr2722:
#line 1233 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2723:
#line 1233 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2724:
#line 1233 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2729:
#line 1222 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	goto st819;
tr2730:
#line 1222 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2731:
#line 1222 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2732:
#line 1222 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2734:
#line 1227 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	goto st819;
tr2735:
#line 1227 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2736:
#line 1227 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2737:
#line 1227 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2744:
#line 1243 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	goto st819;
tr2745:
#line 1243 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2746:
#line 1243 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2747:
#line 1243 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2750:
#line 1244 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	goto st819;
tr2751:
#line 1244 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2752:
#line 1244 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2753:
#line 1244 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2756:
#line 1224 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	goto st819;
tr2757:
#line 1224 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2758:
#line 1224 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2759:
#line 1224 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2761:
#line 1245 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	goto st819;
tr2762:
#line 1245 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2763:
#line 1245 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2764:
#line 1245 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2771:
#line 1216 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	goto st819;
tr2772:
#line 1216 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2773:
#line 1216 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2774:
#line 1216 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2776:
#line 1217 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	goto st819;
tr2777:
#line 1217 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2778:
#line 1217 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2779:
#line 1217 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2787:
#line 1226 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	goto st819;
tr2788:
#line 1226 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2789:
#line 1226 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2790:
#line 1226 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2793:
#line 1242 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	goto st819;
tr2794:
#line 1242 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2795:
#line 1242 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2796:
#line 1242 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2798:
#line 1211 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	goto st819;
tr2799:
#line 1211 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2800:
#line 1211 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2802:
#line 1211 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2805:
#line 1235 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	goto st819;
tr2806:
#line 1235 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2807:
#line 1235 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2809:
#line 1235 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2811:
#line 1238 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	goto st819;
tr2812:
#line 1238 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2813:
#line 1238 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2815:
#line 1238 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2821:
#line 1239 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	goto st819;
tr2822:
#line 1239 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2823:
#line 1239 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2824:
#line 1239 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2828:
#line 1214 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	goto st819;
tr2829:
#line 1214 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2830:
#line 1214 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2831:
#line 1214 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2836:
#line 1219 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	goto st819;
tr2837:
#line 1219 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2838:
#line 1219 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2839:
#line 1219 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2844:
#line 1234 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	goto st819;
tr2845:
#line 1234 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2846:
#line 1234 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2847:
#line 1234 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2849:
#line 1221 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	goto st819;
tr2850:
#line 1221 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2851:
#line 1221 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2852:
#line 1221 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2859:
#line 1213 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	goto st819;
tr2860:
#line 1213 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2861:
#line 1213 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2862:
#line 1213 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2865:
#line 1241 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	goto st819;
tr2866:
#line 1241 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2867:
#line 1241 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2868:
#line 1241 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2871:
#line 1225 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	goto st819;
tr2872:
#line 1225 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2873:
#line 1225 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2874:
#line 1225 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2879:
#line 1232 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	goto st819;
tr2880:
#line 1232 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2881:
#line 1232 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2882:
#line 1232 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2889:
#line 1240 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	goto st819;
tr2890:
#line 1240 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2891:
#line 1240 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2892:
#line 1240 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
tr2895:
#line 1218 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	goto st819;
tr2896:
#line 1218 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st819;
tr2897:
#line 1218 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st819;
tr2898:
#line 1218 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st819;
st819:
	if ( ++p == pe )
		goto _test_eof819;
case 819:
#line 69925 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st819;
		case 32: goto st819;
		case 40: goto tr2534;
		case 41: goto tr2535;
		case 1034: goto tr2537;
		case 1083: goto st842;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2536;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2536;
	} else
		goto tr2536;
	goto tr69;
tr2536:
#line 1602 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 820; goto st457;} }
	goto st820;
st820:
	if ( ++p == pe )
		goto _test_eof820;
case 820:
#line 69965 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st821;
		case 32: goto st821;
		case 40: goto tr2540;
		case 41: goto tr2541;
		case 1034: goto tr2542;
		case 1083: goto st841;
	}
	goto tr69;
tr2540:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st821;
tr2541:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st821;
tr2542:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st821;
st821:
	if ( ++p == pe )
		goto _test_eof821;
case 821:
#line 70019 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st821;
		case 32: goto st821;
		case 40: goto tr2540;
		case 41: goto tr2541;
		case 1034: goto tr2542;
		case 1083: goto st841;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2544;
	goto tr1674;
tr2544:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st822;
tr2548:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st822;
st822:
	if ( ++p == pe )
		goto _test_eof822;
case 822:
#line 70087 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2545;
		case 32: goto tr2545;
		case 40: goto tr2546;
		case 41: goto tr2547;
		case 1034: goto tr2549;
		case 1083: goto tr2550;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2548;
	goto tr1674;
tr2552:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st823;
tr2553:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st823;
tr2555:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st823;
tr2545:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st823;
tr2546:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st823;
tr2547:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st823;
tr2549:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st823;
st823:
	if ( ++p == pe )
		goto _test_eof823;
case 823:
#line 70211 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st823;
		case 32: goto st823;
		case 40: goto tr2552;
		case 41: goto tr2553;
		case 1034: goto tr2555;
		case 1083: goto st840;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2554;
	goto tr1674;
tr2554:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st824;
tr2560:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st824;
st824:
	if ( ++p == pe )
		goto _test_eof824;
case 824:
#line 70279 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2557;
		case 32: goto tr2557;
		case 40: goto tr2558;
		case 41: goto tr2559;
		case 1034: goto tr2561;
		case 1083: goto tr2562;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2560;
	goto tr1674;
tr2565:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st825;
tr2566:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st825;
tr2568:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st825;
tr2557:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st825;
tr2558:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st825;
tr2559:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st825;
tr2561:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st825;
st825:
	if ( ++p == pe )
		goto _test_eof825;
case 825:
#line 70403 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st825;
		case 32: goto st825;
		case 40: goto tr2565;
		case 41: goto tr2566;
		case 1034: goto tr2568;
		case 1083: goto st839;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2567;
	goto tr2563;
tr2567:
#line 445 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 448 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st826;
tr2573:
#line 448 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st826;
st826:
	if ( ++p == pe )
		goto _test_eof826;
case 826:
#line 70459 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2570;
		case 32: goto tr2570;
		case 40: goto tr2571;
		case 41: goto tr2572;
		case 1034: goto tr2574;
		case 1083: goto tr2575;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2573;
	goto tr2563;
tr2577:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st827;
tr2578:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st827;
tr2580:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st827;
tr2570:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
	goto st827;
tr2571:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st827;
tr2572:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st827;
tr2574:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st827;
st827:
	if ( ++p == pe )
		goto _test_eof827;
case 827:
#line 70687 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st827;
		case 32: goto st827;
		case 40: goto tr2577;
		case 41: goto tr2578;
		case 1034: goto tr2580;
		case 1083: goto st838;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2579;
	goto tr2563;
tr2579:
#line 445 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 448 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st828;
tr2585:
#line 448 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st828;
st828:
	if ( ++p == pe )
		goto _test_eof828;
case 828:
#line 70743 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2582;
		case 32: goto tr2582;
		case 40: goto tr2583;
		case 41: goto tr2584;
		case 1034: goto tr2586;
		case 1083: goto tr2587;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2585;
	goto tr2563;
tr2589:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st829;
tr2590:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st829;
tr2592:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st829;
tr2582:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
	goto st829;
tr2583:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st829;
tr2584:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st829;
tr2586:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st829;
st829:
	if ( ++p == pe )
		goto _test_eof829;
case 829:
#line 70971 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st829;
		case 32: goto st829;
		case 40: goto tr2589;
		case 41: goto tr2590;
		case 1034: goto tr2592;
		case 1083: goto st837;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2591;
	goto tr1674;
tr2591:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st830;
tr2597:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st830;
st830:
	if ( ++p == pe )
		goto _test_eof830;
case 830:
#line 71039 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2594;
		case 32: goto tr2594;
		case 40: goto tr2595;
		case 41: goto tr2596;
		case 1034: goto tr2598;
		case 1083: goto tr2599;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2597;
	goto tr1674;
tr2601:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st831;
tr2602:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st831;
tr2604:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st831;
tr2594:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st831;
tr2595:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st831;
tr2596:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st831;
tr2598:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st831;
st831:
	if ( ++p == pe )
		goto _test_eof831;
case 831:
#line 71183 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st831;
		case 32: goto st831;
		case 40: goto tr2601;
		case 41: goto tr2602;
		case 42: goto tr2603;
		case 92: goto tr2603;
		case 95: goto tr2603;
		case 1034: goto tr2604;
		case 1083: goto st836;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2603;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2603;
	} else
		goto tr2603;
	goto tr69;
tr2603:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 832; goto st249;} }
	goto st832;
st832:
	if ( ++p == pe )
		goto _test_eof832;
case 832:
#line 71230 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2606;
		case 32: goto tr2606;
		case 40: goto tr2607;
		case 41: goto tr2608;
		case 1034: goto tr2609;
		case 1083: goto tr2610;
	}
	goto tr69;
tr2612:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st833;
tr2613:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st833;
tr2615:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st833;
tr2606:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st833;
tr2607:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st833;
tr2608:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st833;
tr2609:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st833;
st833:
	if ( ++p == pe )
		goto _test_eof833;
case 833:
#line 71348 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st833;
		case 32: goto st833;
		case 40: goto tr2612;
		case 41: goto tr2613;
		case 43: goto tr2614;
		case 1034: goto tr2615;
		case 1083: goto st835;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2614;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2614;
	} else
		goto tr2614;
	goto tr69;
tr2614:
#line 996 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 834; goto st311;} }
	goto st834;
st834:
	if ( ++p == pe )
		goto _test_eof834;
case 834:
#line 71389 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1280 + ((*p) - 0));
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1280 + ((*p) - 0));
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2617;
		case 1595: goto tr2617;
	}
	goto tr69;
tr2617:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1132;
st1132:
	if ( ++p == pe )
		goto _test_eof1132;
case 1132:
#line 71419 "zscanner/scanner.c"
	goto st0;
tr2610:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st835;
st835:
	if ( ++p == pe )
		goto _test_eof835;
case 835:
#line 71436 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2615;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st835;
	goto tr69;
tr2599:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
	goto st836;
st836:
	if ( ++p == pe )
		goto _test_eof836;
case 836:
#line 71484 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2604;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st836;
	goto tr69;
tr2587:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
	goto st837;
st837:
	if ( ++p == pe )
		goto _test_eof837;
case 837:
#line 71553 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2592;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st837;
	goto tr69;
tr2575:
#line 456 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == KNOT_EOK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ret);
				p--; {goto st247;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP);
				p--; {goto st247;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
				p--; {goto st247;}
			}
		} else {
			SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_LENGTH);
			p--; {goto st247;}
		}
	}
	goto st838;
st838:
	if ( ++p == pe )
		goto _test_eof838;
case 838:
#line 71622 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2580;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st838;
	goto tr69;
tr2562:
#line 350 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER32_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st839;
st839:
	if ( ++p == pe )
		goto _test_eof839;
case 839:
#line 71665 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2568;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st839;
	goto tr69;
tr2550:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st840;
st840:
	if ( ++p == pe )
		goto _test_eof840;
case 840:
#line 71708 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2555;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st840;
	goto tr69;
st841:
	if ( ++p == pe )
		goto _test_eof841;
case 841:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2542;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st841;
	goto tr69;
tr2909:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st842;
tr2532:
#line 1210 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	goto st842;
tr2624:
#line 1223 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	goto st842;
tr2632:
#line 1220 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	goto st842;
tr2638:
#line 1230 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	goto st842;
tr2647:
#line 1228 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	goto st842;
tr2655:
#line 1212 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	goto st842;
tr2666:
#line 1237 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	goto st842;
tr2675:
#line 1229 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	goto st842;
tr2683:
#line 1236 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	goto st842;
tr2688:
#line 1231 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	goto st842;
tr2698:
#line 1246 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	goto st842;
tr2704:
#line 1247 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	goto st842;
tr2713:
#line 1215 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	goto st842;
tr2725:
#line 1233 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	goto st842;
tr2733:
#line 1222 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	goto st842;
tr2738:
#line 1227 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	goto st842;
tr2748:
#line 1243 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	goto st842;
tr2754:
#line 1244 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	goto st842;
tr2760:
#line 1224 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	goto st842;
tr2765:
#line 1245 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	goto st842;
tr2775:
#line 1216 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	goto st842;
tr2780:
#line 1217 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	goto st842;
tr2791:
#line 1226 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	goto st842;
tr2797:
#line 1242 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	goto st842;
tr2803:
#line 1211 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	goto st842;
tr2810:
#line 1235 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	goto st842;
tr2816:
#line 1238 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	goto st842;
tr2825:
#line 1239 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	goto st842;
tr2832:
#line 1214 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	goto st842;
tr2840:
#line 1219 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	goto st842;
tr2848:
#line 1234 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	goto st842;
tr2853:
#line 1221 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	goto st842;
tr2863:
#line 1213 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	goto st842;
tr2869:
#line 1241 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	goto st842;
tr2875:
#line 1225 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	goto st842;
tr2883:
#line 1232 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	goto st842;
tr2893:
#line 1240 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	goto st842;
tr2899:
#line 1218 "./zscanner/scanner_body.rl"
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	goto st842;
st842:
	if ( ++p == pe )
		goto _test_eof842;
case 842:
#line 71933 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2537;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st842;
	goto tr69;
st843:
	if ( ++p == pe )
		goto _test_eof843;
case 843:
	switch( (*p) ) {
		case 65: goto st844;
		case 97: goto st844;
	}
	goto tr2510;
st844:
	if ( ++p == pe )
		goto _test_eof844;
case 844:
	switch( (*p) ) {
		case 65: goto st845;
		case 97: goto st845;
	}
	goto tr2510;
st845:
	if ( ++p == pe )
		goto _test_eof845;
case 845:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2620;
		case 32: goto tr2620;
		case 40: goto tr2621;
		case 41: goto tr2622;
		case 1034: goto tr2623;
		case 1083: goto tr2624;
	}
	goto tr2510;
st846:
	if ( ++p == pe )
		goto _test_eof846;
case 846:
	switch( (*p) ) {
		case 83: goto st847;
		case 115: goto st847;
	}
	goto tr2510;
st847:
	if ( ++p == pe )
		goto _test_eof847;
case 847:
	switch( (*p) ) {
		case 68: goto st848;
		case 100: goto st848;
	}
	goto tr2510;
st848:
	if ( ++p == pe )
		goto _test_eof848;
case 848:
	switch( (*p) ) {
		case 66: goto st849;
		case 98: goto st849;
	}
	goto tr2510;
st849:
	if ( ++p == pe )
		goto _test_eof849;
case 849:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2628;
		case 32: goto tr2628;
		case 40: goto tr2629;
		case 41: goto tr2630;
		case 1034: goto tr2631;
		case 1083: goto tr2632;
	}
	goto tr2510;
st850:
	if ( ++p == pe )
		goto _test_eof850;
case 850:
	switch( (*p) ) {
		case 76: goto st851;
		case 108: goto st851;
	}
	goto tr2510;
st851:
	if ( ++p == pe )
		goto _test_eof851;
case 851:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2634;
		case 32: goto tr2634;
		case 40: goto tr2635;
		case 41: goto tr2636;
		case 1034: goto tr2637;
		case 1083: goto tr2638;
	}
	goto tr2510;
st852:
	if ( ++p == pe )
		goto _test_eof852;
case 852:
	switch( (*p) ) {
		case 69: goto st853;
		case 78: goto st856;
		case 101: goto st853;
		case 110: goto st856;
	}
	goto tr2510;
st853:
	if ( ++p == pe )
		goto _test_eof853;
case 853:
	switch( (*p) ) {
		case 82: goto st854;
		case 114: goto st854;
	}
	goto tr2510;
st854:
	if ( ++p == pe )
		goto _test_eof854;
case 854:
	switch( (*p) ) {
		case 84: goto st855;
		case 116: goto st855;
	}
	goto tr2510;
st855:
	if ( ++p == pe )
		goto _test_eof855;
case 855:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2643;
		case 32: goto tr2643;
		case 40: goto tr2644;
		case 41: goto tr2645;
		case 1034: goto tr2646;
		case 1083: goto tr2647;
	}
	goto tr2510;
st856:
	if ( ++p == pe )
		goto _test_eof856;
case 856:
	switch( (*p) ) {
		case 65: goto st857;
		case 97: goto st857;
	}
	goto tr2510;
st857:
	if ( ++p == pe )
		goto _test_eof857;
case 857:
	switch( (*p) ) {
		case 77: goto st858;
		case 109: goto st858;
	}
	goto tr2510;
st858:
	if ( ++p == pe )
		goto _test_eof858;
case 858:
	switch( (*p) ) {
		case 69: goto st859;
		case 101: goto st859;
	}
	goto tr2510;
st859:
	if ( ++p == pe )
		goto _test_eof859;
case 859:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2651;
		case 32: goto tr2651;
		case 40: goto tr2652;
		case 41: goto tr2653;
		case 1034: goto tr2654;
		case 1083: goto tr2655;
	}
	goto tr2510;
st860:
	if ( ++p == pe )
		goto _test_eof860;
case 860:
	switch( (*p) ) {
		case 72: goto st861;
		case 78: goto st865;
		case 83: goto st873;
		case 104: goto st861;
		case 110: goto st865;
		case 115: goto st873;
	}
	goto tr2510;
st861:
	if ( ++p == pe )
		goto _test_eof861;
case 861:
	switch( (*p) ) {
		case 67: goto st862;
		case 99: goto st862;
	}
	goto tr2510;
st862:
	if ( ++p == pe )
		goto _test_eof862;
case 862:
	switch( (*p) ) {
		case 73: goto st863;
		case 105: goto st863;
	}
	goto tr2510;
st863:
	if ( ++p == pe )
		goto _test_eof863;
case 863:
	switch( (*p) ) {
		case 68: goto st864;
		case 100: goto st864;
	}
	goto tr2510;
st864:
	if ( ++p == pe )
		goto _test_eof864;
case 864:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2662;
		case 32: goto tr2662;
		case 40: goto tr2663;
		case 41: goto tr2664;
		case 1034: goto tr2665;
		case 1083: goto tr2666;
	}
	goto tr2510;
st865:
	if ( ++p == pe )
		goto _test_eof865;
case 865:
	switch( (*p) ) {
		case 65: goto st866;
		case 83: goto st869;
		case 97: goto st866;
		case 115: goto st869;
	}
	goto tr2510;
st866:
	if ( ++p == pe )
		goto _test_eof866;
case 866:
	switch( (*p) ) {
		case 77: goto st867;
		case 109: goto st867;
	}
	goto tr2510;
st867:
	if ( ++p == pe )
		goto _test_eof867;
case 867:
	switch( (*p) ) {
		case 69: goto st868;
		case 101: goto st868;
	}
	goto tr2510;
st868:
	if ( ++p == pe )
		goto _test_eof868;
case 868:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2671;
		case 32: goto tr2671;
		case 40: goto tr2672;
		case 41: goto tr2673;
		case 1034: goto tr2674;
		case 1083: goto tr2675;
	}
	goto tr2510;
st869:
	if ( ++p == pe )
		goto _test_eof869;
case 869:
	switch( (*p) ) {
		case 75: goto st870;
		case 107: goto st870;
	}
	goto tr2510;
st870:
	if ( ++p == pe )
		goto _test_eof870;
case 870:
	switch( (*p) ) {
		case 69: goto st871;
		case 101: goto st871;
	}
	goto tr2510;
st871:
	if ( ++p == pe )
		goto _test_eof871;
case 871:
	switch( (*p) ) {
		case 89: goto st872;
		case 121: goto st872;
	}
	goto tr2510;
st872:
	if ( ++p == pe )
		goto _test_eof872;
case 872:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2679;
		case 32: goto tr2679;
		case 40: goto tr2680;
		case 41: goto tr2681;
		case 1034: goto tr2682;
		case 1083: goto tr2683;
	}
	goto tr2510;
st873:
	if ( ++p == pe )
		goto _test_eof873;
case 873:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2684;
		case 32: goto tr2684;
		case 40: goto tr2685;
		case 41: goto tr2686;
		case 1034: goto tr2687;
		case 1083: goto tr2688;
	}
	goto tr2510;
st874:
	if ( ++p == pe )
		goto _test_eof874;
case 874:
	switch( (*p) ) {
		case 85: goto st875;
		case 117: goto st875;
	}
	goto tr2510;
st875:
	if ( ++p == pe )
		goto _test_eof875;
case 875:
	switch( (*p) ) {
		case 73: goto st876;
		case 105: goto st876;
	}
	goto tr2510;
st876:
	if ( ++p == pe )
		goto _test_eof876;
case 876:
	switch( (*p) ) {
		case 52: goto st877;
		case 54: goto st879;
	}
	goto tr2510;
st877:
	if ( ++p == pe )
		goto _test_eof877;
case 877:
	if ( (*p) == 56 )
		goto st878;
	goto tr2510;
st878:
	if ( ++p == pe )
		goto _test_eof878;
case 878:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2694;
		case 32: goto tr2694;
		case 40: goto tr2695;
		case 41: goto tr2696;
		case 1034: goto tr2697;
		case 1083: goto tr2698;
	}
	goto tr2510;
st879:
	if ( ++p == pe )
		goto _test_eof879;
case 879:
	if ( (*p) == 52 )
		goto st880;
	goto tr2510;
st880:
	if ( ++p == pe )
		goto _test_eof880;
case 880:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2700;
		case 32: goto tr2700;
		case 40: goto tr2701;
		case 41: goto tr2702;
		case 1034: goto tr2703;
		case 1083: goto tr2704;
	}
	goto tr2510;
st881:
	if ( ++p == pe )
		goto _test_eof881;
case 881:
	switch( (*p) ) {
		case 73: goto st882;
		case 105: goto st882;
	}
	goto tr2510;
st882:
	if ( ++p == pe )
		goto _test_eof882;
case 882:
	switch( (*p) ) {
		case 78: goto st883;
		case 110: goto st883;
	}
	goto tr2510;
st883:
	if ( ++p == pe )
		goto _test_eof883;
case 883:
	switch( (*p) ) {
		case 70: goto st884;
		case 102: goto st884;
	}
	goto tr2510;
st884:
	if ( ++p == pe )
		goto _test_eof884;
case 884:
	switch( (*p) ) {
		case 79: goto st885;
		case 111: goto st885;
	}
	goto tr2510;
st885:
	if ( ++p == pe )
		goto _test_eof885;
case 885:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2709;
		case 32: goto tr2709;
		case 40: goto tr2710;
		case 41: goto tr2711;
		case 1034: goto tr2712;
		case 1083: goto tr2713;
	}
	goto tr2510;
st886:
	if ( ++p == pe )
		goto _test_eof886;
case 886:
	switch( (*p) ) {
		case 80: goto st887;
		case 112: goto st887;
	}
	goto tr2510;
st887:
	if ( ++p == pe )
		goto _test_eof887;
case 887:
	switch( (*p) ) {
		case 83: goto st888;
		case 115: goto st888;
	}
	goto tr2510;
st888:
	if ( ++p == pe )
		goto _test_eof888;
case 888:
	switch( (*p) ) {
		case 69: goto st889;
		case 101: goto st889;
	}
	goto tr2510;
st889:
	if ( ++p == pe )
		goto _test_eof889;
case 889:
	switch( (*p) ) {
		case 67: goto st890;
		case 99: goto st890;
	}
	goto tr2510;
st890:
	if ( ++p == pe )
		goto _test_eof890;
case 890:
	switch( (*p) ) {
		case 75: goto st891;
		case 107: goto st891;
	}
	goto tr2510;
st891:
	if ( ++p == pe )
		goto _test_eof891;
case 891:
	switch( (*p) ) {
		case 69: goto st892;
		case 101: goto st892;
	}
	goto tr2510;
st892:
	if ( ++p == pe )
		goto _test_eof892;
case 892:
	switch( (*p) ) {
		case 89: goto st893;
		case 121: goto st893;
	}
	goto tr2510;
st893:
	if ( ++p == pe )
		goto _test_eof893;
case 893:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2721;
		case 32: goto tr2721;
		case 40: goto tr2722;
		case 41: goto tr2723;
		case 1034: goto tr2724;
		case 1083: goto tr2725;
	}
	goto tr2510;
st894:
	if ( ++p == pe )
		goto _test_eof894;
case 894:
	switch( (*p) ) {
		case 69: goto st895;
		case 88: goto st897;
		case 101: goto st895;
		case 120: goto st897;
	}
	goto tr2510;
st895:
	if ( ++p == pe )
		goto _test_eof895;
case 895:
	switch( (*p) ) {
		case 89: goto st896;
		case 121: goto st896;
	}
	goto tr2510;
st896:
	if ( ++p == pe )
		goto _test_eof896;
case 896:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2729;
		case 32: goto tr2729;
		case 40: goto tr2730;
		case 41: goto tr2731;
		case 1034: goto tr2732;
		case 1083: goto tr2733;
	}
	goto tr2510;
st897:
	if ( ++p == pe )
		goto _test_eof897;
case 897:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2734;
		case 32: goto tr2734;
		case 40: goto tr2735;
		case 41: goto tr2736;
		case 1034: goto tr2737;
		case 1083: goto tr2738;
	}
	goto tr2510;
st898:
	if ( ++p == pe )
		goto _test_eof898;
case 898:
	switch( (*p) ) {
		case 51: goto st899;
		case 54: goto st901;
		case 79: goto st903;
		case 80: goto st905;
		case 111: goto st903;
		case 112: goto st905;
	}
	goto tr2510;
st899:
	if ( ++p == pe )
		goto _test_eof899;
case 899:
	if ( (*p) == 50 )
		goto st900;
	goto tr2510;
st900:
	if ( ++p == pe )
		goto _test_eof900;
case 900:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2744;
		case 32: goto tr2744;
		case 40: goto tr2745;
		case 41: goto tr2746;
		case 1034: goto tr2747;
		case 1083: goto tr2748;
	}
	goto tr2510;
st901:
	if ( ++p == pe )
		goto _test_eof901;
case 901:
	if ( (*p) == 52 )
		goto st902;
	goto tr2510;
st902:
	if ( ++p == pe )
		goto _test_eof902;
case 902:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2750;
		case 32: goto tr2750;
		case 40: goto tr2751;
		case 41: goto tr2752;
		case 1034: goto tr2753;
		case 1083: goto tr2754;
	}
	goto tr2510;
st903:
	if ( ++p == pe )
		goto _test_eof903;
case 903:
	switch( (*p) ) {
		case 67: goto st904;
		case 99: goto st904;
	}
	goto tr2510;
st904:
	if ( ++p == pe )
		goto _test_eof904;
case 904:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2756;
		case 32: goto tr2756;
		case 40: goto tr2757;
		case 41: goto tr2758;
		case 1034: goto tr2759;
		case 1083: goto tr2760;
	}
	goto tr2510;
st905:
	if ( ++p == pe )
		goto _test_eof905;
case 905:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2761;
		case 32: goto tr2761;
		case 40: goto tr2762;
		case 41: goto tr2763;
		case 1034: goto tr2764;
		case 1083: goto tr2765;
	}
	goto tr2510;
st906:
	if ( ++p == pe )
		goto _test_eof906;
case 906:
	switch( (*p) ) {
		case 73: goto st907;
		case 88: goto st911;
		case 105: goto st907;
		case 120: goto st911;
	}
	goto tr2510;
st907:
	if ( ++p == pe )
		goto _test_eof907;
case 907:
	switch( (*p) ) {
		case 78: goto st908;
		case 110: goto st908;
	}
	goto tr2510;
st908:
	if ( ++p == pe )
		goto _test_eof908;
case 908:
	switch( (*p) ) {
		case 70: goto st909;
		case 102: goto st909;
	}
	goto tr2510;
st909:
	if ( ++p == pe )
		goto _test_eof909;
case 909:
	switch( (*p) ) {
		case 79: goto st910;
		case 111: goto st910;
	}
	goto tr2510;
st910:
	if ( ++p == pe )
		goto _test_eof910;
case 910:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2771;
		case 32: goto tr2771;
		case 40: goto tr2772;
		case 41: goto tr2773;
		case 1034: goto tr2774;
		case 1083: goto tr2775;
	}
	goto tr2510;
st911:
	if ( ++p == pe )
		goto _test_eof911;
case 911:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2776;
		case 32: goto tr2776;
		case 40: goto tr2777;
		case 41: goto tr2778;
		case 1034: goto tr2779;
		case 1083: goto tr2780;
	}
	goto tr2510;
st912:
	if ( ++p == pe )
		goto _test_eof912;
case 912:
	switch( (*p) ) {
		case 65: goto st913;
		case 73: goto st917;
		case 83: goto st919;
		case 97: goto st913;
		case 105: goto st917;
		case 115: goto st919;
	}
	goto tr2510;
st913:
	if ( ++p == pe )
		goto _test_eof913;
case 913:
	switch( (*p) ) {
		case 80: goto st914;
		case 112: goto st914;
	}
	goto tr2510;
st914:
	if ( ++p == pe )
		goto _test_eof914;
case 914:
	switch( (*p) ) {
		case 84: goto st915;
		case 116: goto st915;
	}
	goto tr2510;
st915:
	if ( ++p == pe )
		goto _test_eof915;
case 915:
	switch( (*p) ) {
		case 82: goto st916;
		case 114: goto st916;
	}
	goto tr2510;
st916:
	if ( ++p == pe )
		goto _test_eof916;
case 916:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2787;
		case 32: goto tr2787;
		case 40: goto tr2788;
		case 41: goto tr2789;
		case 1034: goto tr2790;
		case 1083: goto tr2791;
	}
	goto tr2510;
st917:
	if ( ++p == pe )
		goto _test_eof917;
case 917:
	switch( (*p) ) {
		case 68: goto st918;
		case 100: goto st918;
	}
	goto tr2510;
st918:
	if ( ++p == pe )
		goto _test_eof918;
case 918:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2793;
		case 32: goto tr2793;
		case 40: goto tr2794;
		case 41: goto tr2795;
		case 1034: goto tr2796;
		case 1083: goto tr2797;
	}
	goto tr2510;
st919:
	if ( ++p == pe )
		goto _test_eof919;
case 919:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2798;
		case 32: goto tr2798;
		case 40: goto tr2799;
		case 41: goto tr2800;
		case 69: goto st920;
		case 101: goto st920;
		case 1034: goto tr2802;
		case 1083: goto tr2803;
	}
	goto tr2510;
st920:
	if ( ++p == pe )
		goto _test_eof920;
case 920:
	switch( (*p) ) {
		case 67: goto st921;
		case 99: goto st921;
	}
	goto tr2510;
st921:
	if ( ++p == pe )
		goto _test_eof921;
case 921:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2805;
		case 32: goto tr2805;
		case 40: goto tr2806;
		case 41: goto tr2807;
		case 51: goto st922;
		case 1034: goto tr2809;
		case 1083: goto tr2810;
	}
	goto tr2510;
st922:
	if ( ++p == pe )
		goto _test_eof922;
case 922:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2811;
		case 32: goto tr2811;
		case 40: goto tr2812;
		case 41: goto tr2813;
		case 80: goto st923;
		case 112: goto st923;
		case 1034: goto tr2815;
		case 1083: goto tr2816;
	}
	goto tr2510;
st923:
	if ( ++p == pe )
		goto _test_eof923;
case 923:
	switch( (*p) ) {
		case 65: goto st924;
		case 97: goto st924;
	}
	goto tr2510;
st924:
	if ( ++p == pe )
		goto _test_eof924;
case 924:
	switch( (*p) ) {
		case 82: goto st925;
		case 114: goto st925;
	}
	goto tr2510;
st925:
	if ( ++p == pe )
		goto _test_eof925;
case 925:
	switch( (*p) ) {
		case 65: goto st926;
		case 97: goto st926;
	}
	goto tr2510;
st926:
	if ( ++p == pe )
		goto _test_eof926;
case 926:
	switch( (*p) ) {
		case 77: goto st927;
		case 109: goto st927;
	}
	goto tr2510;
st927:
	if ( ++p == pe )
		goto _test_eof927;
case 927:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2821;
		case 32: goto tr2821;
		case 40: goto tr2822;
		case 41: goto tr2823;
		case 1034: goto tr2824;
		case 1083: goto tr2825;
	}
	goto tr2510;
st928:
	if ( ++p == pe )
		goto _test_eof928;
case 928:
	switch( (*p) ) {
		case 84: goto st929;
		case 116: goto st929;
	}
	goto tr2510;
st929:
	if ( ++p == pe )
		goto _test_eof929;
case 929:
	switch( (*p) ) {
		case 82: goto st930;
		case 114: goto st930;
	}
	goto tr2510;
st930:
	if ( ++p == pe )
		goto _test_eof930;
case 930:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2828;
		case 32: goto tr2828;
		case 40: goto tr2829;
		case 41: goto tr2830;
		case 1034: goto tr2831;
		case 1083: goto tr2832;
	}
	goto tr2510;
st931:
	if ( ++p == pe )
		goto _test_eof931;
case 931:
	switch( (*p) ) {
		case 80: goto st932;
		case 82: goto st933;
		case 84: goto st937;
		case 112: goto st932;
		case 114: goto st933;
		case 116: goto st937;
	}
	goto tr2510;
st932:
	if ( ++p == pe )
		goto _test_eof932;
case 932:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2836;
		case 32: goto tr2836;
		case 40: goto tr2837;
		case 41: goto tr2838;
		case 1034: goto tr2839;
		case 1083: goto tr2840;
	}
	goto tr2510;
st933:
	if ( ++p == pe )
		goto _test_eof933;
case 933:
	switch( (*p) ) {
		case 83: goto st934;
		case 115: goto st934;
	}
	goto tr2510;
st934:
	if ( ++p == pe )
		goto _test_eof934;
case 934:
	switch( (*p) ) {
		case 73: goto st935;
		case 105: goto st935;
	}
	goto tr2510;
st935:
	if ( ++p == pe )
		goto _test_eof935;
case 935:
	switch( (*p) ) {
		case 71: goto st936;
		case 103: goto st936;
	}
	goto tr2510;
st936:
	if ( ++p == pe )
		goto _test_eof936;
case 936:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2844;
		case 32: goto tr2844;
		case 40: goto tr2845;
		case 41: goto tr2846;
		case 1034: goto tr2847;
		case 1083: goto tr2848;
	}
	goto tr2510;
st937:
	if ( ++p == pe )
		goto _test_eof937;
case 937:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2849;
		case 32: goto tr2849;
		case 40: goto tr2850;
		case 41: goto tr2851;
		case 1034: goto tr2852;
		case 1083: goto tr2853;
	}
	goto tr2510;
st938:
	if ( ++p == pe )
		goto _test_eof938;
case 938:
	switch( (*p) ) {
		case 79: goto st939;
		case 80: goto st941;
		case 82: goto st943;
		case 83: goto st945;
		case 111: goto st939;
		case 112: goto st941;
		case 114: goto st943;
		case 115: goto st945;
	}
	goto tr2510;
st939:
	if ( ++p == pe )
		goto _test_eof939;
case 939:
	switch( (*p) ) {
		case 65: goto st940;
		case 97: goto st940;
	}
	goto tr2510;
st940:
	if ( ++p == pe )
		goto _test_eof940;
case 940:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2859;
		case 32: goto tr2859;
		case 40: goto tr2860;
		case 41: goto tr2861;
		case 1034: goto tr2862;
		case 1083: goto tr2863;
	}
	goto tr2510;
st941:
	if ( ++p == pe )
		goto _test_eof941;
case 941:
	switch( (*p) ) {
		case 70: goto st942;
		case 102: goto st942;
	}
	goto tr2510;
st942:
	if ( ++p == pe )
		goto _test_eof942;
case 942:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2865;
		case 32: goto tr2865;
		case 40: goto tr2866;
		case 41: goto tr2867;
		case 1034: goto tr2868;
		case 1083: goto tr2869;
	}
	goto tr2510;
st943:
	if ( ++p == pe )
		goto _test_eof943;
case 943:
	switch( (*p) ) {
		case 86: goto st944;
		case 118: goto st944;
	}
	goto tr2510;
st944:
	if ( ++p == pe )
		goto _test_eof944;
case 944:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2871;
		case 32: goto tr2871;
		case 40: goto tr2872;
		case 41: goto tr2873;
		case 1034: goto tr2874;
		case 1083: goto tr2875;
	}
	goto tr2510;
st945:
	if ( ++p == pe )
		goto _test_eof945;
case 945:
	switch( (*p) ) {
		case 72: goto st946;
		case 104: goto st946;
	}
	goto tr2510;
st946:
	if ( ++p == pe )
		goto _test_eof946;
case 946:
	switch( (*p) ) {
		case 70: goto st947;
		case 102: goto st947;
	}
	goto tr2510;
st947:
	if ( ++p == pe )
		goto _test_eof947;
case 947:
	switch( (*p) ) {
		case 80: goto st948;
		case 112: goto st948;
	}
	goto tr2510;
st948:
	if ( ++p == pe )
		goto _test_eof948;
case 948:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2879;
		case 32: goto tr2879;
		case 40: goto tr2880;
		case 41: goto tr2881;
		case 1034: goto tr2882;
		case 1083: goto tr2883;
	}
	goto tr2510;
st949:
	if ( ++p == pe )
		goto _test_eof949;
case 949:
	switch( (*p) ) {
		case 76: goto st950;
		case 88: goto st953;
		case 89: goto st955;
		case 108: goto st950;
		case 120: goto st953;
		case 121: goto st955;
	}
	goto tr2510;
st950:
	if ( ++p == pe )
		goto _test_eof950;
case 950:
	switch( (*p) ) {
		case 83: goto st951;
		case 115: goto st951;
	}
	goto tr2510;
st951:
	if ( ++p == pe )
		goto _test_eof951;
case 951:
	switch( (*p) ) {
		case 65: goto st952;
		case 97: goto st952;
	}
	goto tr2510;
st952:
	if ( ++p == pe )
		goto _test_eof952;
case 952:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2889;
		case 32: goto tr2889;
		case 40: goto tr2890;
		case 41: goto tr2891;
		case 1034: goto tr2892;
		case 1083: goto tr2893;
	}
	goto tr2510;
st953:
	if ( ++p == pe )
		goto _test_eof953;
case 953:
	switch( (*p) ) {
		case 84: goto st954;
		case 116: goto st954;
	}
	goto tr2510;
st954:
	if ( ++p == pe )
		goto _test_eof954;
case 954:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2895;
		case 32: goto tr2895;
		case 40: goto tr2896;
		case 41: goto tr2897;
		case 1034: goto tr2898;
		case 1083: goto tr2899;
	}
	goto tr2510;
st955:
	if ( ++p == pe )
		goto _test_eof955;
case 955:
	switch( (*p) ) {
		case 80: goto st956;
		case 112: goto st956;
	}
	goto tr2510;
st956:
	if ( ++p == pe )
		goto _test_eof956;
case 956:
	switch( (*p) ) {
		case 69: goto st957;
		case 101: goto st957;
	}
	goto tr2510;
st957:
	if ( ++p == pe )
		goto _test_eof957;
case 957:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2903;
	goto tr2902;
tr2903:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st958;
tr2907:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st958;
st958:
	if ( ++p == pe )
		goto _test_eof958;
case 958:
#line 73732 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2904;
		case 32: goto tr2904;
		case 40: goto tr2905;
		case 41: goto tr2906;
		case 1034: goto tr2908;
		case 1083: goto tr2909;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2907;
	goto tr2902;
st959:
	if ( ++p == pe )
		goto _test_eof959;
case 959:
	switch( (*p) ) {
		case 42: goto tr2910;
		case 92: goto tr2910;
		case 95: goto tr2910;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr2910;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2910;
	} else
		goto tr2910;
	goto tr69;
tr2910:
#line 254 "./zscanner/scanner_body.rl"
	{
		s->dname = rdata_tail;
	}
#line 201 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 960; goto st249;} }
	goto st960;
st960:
	if ( ++p == pe )
		goto _test_eof960;
case 960:
#line 73788 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr2911;
		case 59: goto tr2911;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2911;
	} else if ( (*p) >= 9 )
		goto tr2911;
	goto tr69;
tr2911:
#line 257 "./zscanner/scanner_body.rl"
	{
		rdata_tail += s->dname_tmp_length;
	}
#line 219 "./zscanner/scanner_body.rl"
	{
		s->r_data_blocks[++(s->r_data_blocks_count)] =
			rdata_tail - s->r_data;
	}
#line 1342 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 961; goto st318;} }
	goto st961;
st961:
	if ( ++p == pe )
		goto _test_eof961;
case 961:
#line 73816 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr2912;
		case 59: goto tr2912;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2912;
	} else if ( (*p) >= 9 )
		goto tr2912;
	goto tr69;
tr2912:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1133;
st1133:
	if ( ++p == pe )
		goto _test_eof1133;
case 1133:
#line 73837 "zscanner/scanner.c"
	goto st0;
st962:
	if ( ++p == pe )
		goto _test_eof962;
case 962:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2913;
	goto tr1674;
tr2913:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st963;
tr2917:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st963;
st963:
	if ( ++p == pe )
		goto _test_eof963;
case 963:
#line 73888 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2914;
		case 32: goto tr2914;
		case 40: goto tr2915;
		case 41: goto tr2916;
		case 1034: goto tr2918;
		case 1083: goto tr2919;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2917;
	goto tr1674;
tr2921:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st964;
tr2922:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st964;
tr2924:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st964;
tr2914:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st964;
tr2915:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st964;
tr2916:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st964;
tr2918:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st964;
st964:
	if ( ++p == pe )
		goto _test_eof964;
case 964:
#line 74012 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st964;
		case 32: goto st964;
		case 40: goto tr2921;
		case 41: goto tr2922;
		case 1034: goto tr2924;
		case 1083: goto st972;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2923;
	goto tr1674;
tr2923:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st965;
tr2929:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st965;
st965:
	if ( ++p == pe )
		goto _test_eof965;
case 965:
#line 74080 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2926;
		case 32: goto tr2926;
		case 40: goto tr2927;
		case 41: goto tr2928;
		case 1034: goto tr2930;
		case 1083: goto tr2931;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2929;
	goto tr1674;
tr2933:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st966;
tr2934:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st966;
tr2936:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st966;
tr2926:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st966;
tr2927:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st966;
tr2928:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st966;
tr2930:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st966;
st966:
	if ( ++p == pe )
		goto _test_eof966;
case 966:
#line 74204 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st966;
		case 32: goto st966;
		case 40: goto tr2933;
		case 41: goto tr2934;
		case 1034: goto tr2936;
		case 1083: goto st971;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2935;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2935;
	} else
		goto tr2935;
	goto tr69;
tr2935:
#line 1602 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 967; goto st457;} }
	goto st967;
st967:
	if ( ++p == pe )
		goto _test_eof967;
case 967:
#line 74244 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st968;
		case 32: goto st968;
		case 40: goto tr2939;
		case 41: goto tr2940;
		case 1034: goto tr2941;
		case 1083: goto st970;
	}
	goto tr69;
tr2939:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st968;
tr2940:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st968;
tr2941:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st968;
st968:
	if ( ++p == pe )
		goto _test_eof968;
case 968:
#line 74298 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st968;
		case 32: goto st968;
		case 40: goto tr2939;
		case 41: goto tr2940;
		case 43: goto tr2943;
		case 1034: goto tr2941;
		case 1083: goto st970;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2943;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2943;
	} else
		goto tr2943;
	goto tr69;
tr2943:
#line 996 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 969; goto st311;} }
	goto st969;
st969:
	if ( ++p == pe )
		goto _test_eof969;
case 969:
#line 74339 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1280 + ((*p) - 0));
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1280 + ((*p) - 0));
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2944;
		case 1595: goto tr2944;
	}
	goto tr69;
tr2944:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1134;
st1134:
	if ( ++p == pe )
		goto _test_eof1134;
case 1134:
#line 74369 "zscanner/scanner.c"
	goto st0;
st970:
	if ( ++p == pe )
		goto _test_eof970;
case 970:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2941;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st970;
	goto tr69;
tr2931:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st971;
st971:
	if ( ++p == pe )
		goto _test_eof971;
case 971:
#line 74417 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2936;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st971;
	goto tr69;
tr2919:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st972;
st972:
	if ( ++p == pe )
		goto _test_eof972;
case 972:
#line 74460 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2924;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st972;
	goto tr69;
st973:
	if ( ++p == pe )
		goto _test_eof973;
case 973:
	if ( (*p) == 43 )
		goto tr2945;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr2945;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2945;
	} else
		goto tr2945;
	goto tr69;
tr2945:
#line 996 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 974; goto st311;} }
	goto st974;
st974:
	if ( ++p == pe )
		goto _test_eof974;
case 974:
#line 74510 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1280 + ((*p) - 0));
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1280 + ((*p) - 0));
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2946;
		case 1595: goto tr2946;
	}
	goto tr69;
tr2946:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1135;
st1135:
	if ( ++p == pe )
		goto _test_eof1135;
case 1135:
#line 74540 "zscanner/scanner.c"
	goto st0;
st975:
	if ( ++p == pe )
		goto _test_eof975;
case 975:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2947;
	goto tr1674;
tr2947:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st976;
tr2951:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st976;
st976:
	if ( ++p == pe )
		goto _test_eof976;
case 976:
#line 74591 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2948;
		case 32: goto tr2948;
		case 40: goto tr2949;
		case 41: goto tr2950;
		case 1034: goto tr2952;
		case 1083: goto tr2953;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2951;
	goto tr1674;
tr2955:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st977;
tr2956:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st977;
tr2958:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st977;
tr2948:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st977;
tr2949:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st977;
tr2950:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st977;
tr2952:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st977;
st977:
	if ( ++p == pe )
		goto _test_eof977;
case 977:
#line 74715 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st977;
		case 32: goto st977;
		case 40: goto tr2955;
		case 41: goto tr2956;
		case 1034: goto tr2958;
		case 1083: goto st1003;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2957;
	goto tr1674;
tr2957:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st978;
tr2963:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st978;
st978:
	if ( ++p == pe )
		goto _test_eof978;
case 978:
#line 74783 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2960;
		case 32: goto tr2960;
		case 40: goto tr2961;
		case 41: goto tr2962;
		case 1034: goto tr2964;
		case 1083: goto tr2965;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2963;
	goto tr1674;
tr2967:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st979;
tr2968:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st979;
tr2970:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st979;
tr2960:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st979;
tr2961:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st979;
tr2962:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st979;
tr2964:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st979;
st979:
	if ( ++p == pe )
		goto _test_eof979;
case 979:
#line 74907 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st979;
		case 32: goto st979;
		case 40: goto tr2967;
		case 41: goto tr2968;
		case 1034: goto tr2970;
		case 1083: goto st1002;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2969;
	goto tr1674;
tr2969:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st980;
tr2975:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st980;
st980:
	if ( ++p == pe )
		goto _test_eof980;
case 980:
#line 74975 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2972;
		case 32: goto tr2972;
		case 40: goto tr2973;
		case 41: goto tr2974;
		case 1034: goto tr2976;
		case 1083: goto tr2977;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2975;
	goto tr1674;
tr2979:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st981;
tr2980:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st981;
tr2983:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st981;
tr2972:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st981;
tr2973:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st981;
tr2974:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st981;
tr2976:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st981;
st981:
	if ( ++p == pe )
		goto _test_eof981;
case 981:
#line 75099 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st981;
		case 32: goto st981;
		case 40: goto tr2979;
		case 41: goto tr2980;
		case 45: goto tr2981;
		case 1034: goto tr2983;
		case 1083: goto st1001;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2982;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2982;
	} else
		goto tr2982;
	goto tr2275;
tr2981:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
	goto st982;
st982:
	if ( ++p == pe )
		goto _test_eof982;
case 982:
#line 75142 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2985;
		case 32: goto tr2985;
		case 40: goto tr2986;
		case 41: goto tr2987;
		case 1034: goto tr2988;
		case 1083: goto tr2989;
	}
	goto tr2275;
tr2992:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st983;
tr2993:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st983;
tr2995:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st983;
tr2985:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st983;
tr2986:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st983;
tr2987:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st983;
tr2988:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st983;
st983:
	if ( ++p == pe )
		goto _test_eof983;
case 983:
#line 75268 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st983;
		case 32: goto st983;
		case 40: goto tr2992;
		case 41: goto tr2993;
		case 1034: goto tr2995;
		case 1083: goto st998;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2994;
	} else if ( _widec > 86 ) {
		if ( 97 <= _widec && _widec <= 118 )
			goto tr2994;
	} else
		goto tr2994;
	goto tr2990;
tr2994:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 1000 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st984;
tr3009:
#line 1000 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st984;
st984:
	if ( ++p == pe )
		goto _test_eof984;
case 984:
#line 75330 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2997;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr2997;
	} else
		goto tr2997;
	goto tr2990;
tr2997:
#line 1008 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += second_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st985;
st985:
	if ( ++p == pe )
		goto _test_eof985;
case 985:
#line 75357 "zscanner/scanner.c"
	if ( (*p) == 61 )
		goto st996;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2998;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr2998;
	} else
		goto tr2998;
	goto tr2990;
tr2998:
#line 1018 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += third_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st986;
st986:
	if ( ++p == pe )
		goto _test_eof986;
case 986:
#line 75379 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3000;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3000;
	} else
		goto tr3000;
	goto tr2990;
tr3000:
#line 1021 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += fourth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fourth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st987;
st987:
	if ( ++p == pe )
		goto _test_eof987;
case 987:
#line 75406 "zscanner/scanner.c"
	if ( (*p) == 61 )
		goto st995;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3001;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3001;
	} else
		goto tr3001;
	goto tr2990;
tr3001:
#line 1031 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += fifth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fifth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st988;
st988:
	if ( ++p == pe )
		goto _test_eof988;
case 988:
#line 75435 "zscanner/scanner.c"
	if ( (*p) == 61 )
		goto st993;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3003;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3003;
	} else
		goto tr3003;
	goto tr2990;
tr3003:
#line 1041 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += sixth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st989;
st989:
	if ( ++p == pe )
		goto _test_eof989;
case 989:
#line 75457 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3005;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3005;
	} else
		goto tr3005;
	goto tr2990;
tr3005:
#line 1044 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += seventh_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = seventh_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st990;
st990:
	if ( ++p == pe )
		goto _test_eof990;
case 990:
#line 75484 "zscanner/scanner.c"
	if ( (*p) == 61 )
		goto st991;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3006;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3006;
	} else
		goto tr3006;
	goto tr2990;
tr3006:
#line 1054 "./zscanner/scanner_body.rl"
	{
		*(rdata_tail++) += eighth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st991;
st991:
	if ( ++p == pe )
		goto _test_eof991;
case 991:
#line 75506 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3008;
		case 59: goto tr3008;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3008;
		} else if ( (*p) >= 9 )
			goto tr3008;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 86 ) {
			if ( 97 <= (*p) && (*p) <= 118 )
				goto tr3009;
		} else if ( (*p) >= 65 )
			goto tr3009;
	} else
		goto tr3009;
	goto tr2990;
tr3008:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 1342 "./zscanner/scanner_body.rl"
	{ p--; {stack[top++] = 992; goto st318;} }
	goto st992;
st992:
	if ( ++p == pe )
		goto _test_eof992;
case 992:
#line 75545 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3010;
		case 59: goto tr3010;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3010;
	} else if ( (*p) >= 9 )
		goto tr3010;
	goto tr69;
tr3010:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1136;
st1136:
	if ( ++p == pe )
		goto _test_eof1136;
case 1136:
#line 75566 "zscanner/scanner.c"
	goto st0;
st993:
	if ( ++p == pe )
		goto _test_eof993;
case 993:
	if ( (*p) == 61 )
		goto st994;
	goto tr2990;
st994:
	if ( ++p == pe )
		goto _test_eof994;
case 994:
	if ( (*p) == 61 )
		goto st991;
	goto tr2990;
st995:
	if ( ++p == pe )
		goto _test_eof995;
case 995:
	if ( (*p) == 61 )
		goto st993;
	goto tr2990;
st996:
	if ( ++p == pe )
		goto _test_eof996;
case 996:
	if ( (*p) == 61 )
		goto st997;
	goto tr2990;
st997:
	if ( ++p == pe )
		goto _test_eof997;
case 997:
	if ( (*p) == 61 )
		goto st995;
	goto tr2990;
tr2989:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st998;
st998:
	if ( ++p == pe )
		goto _test_eof998;
case 998:
#line 75620 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2995;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st998;
	goto tr69;
tr3014:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st999;
tr2982:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st999;
st999:
	if ( ++p == pe )
		goto _test_eof999;
case 999:
#line 75677 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3013;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3013;
	} else
		goto tr3013;
	goto tr2275;
tr3013:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1000;
st1000:
	if ( ++p == pe )
		goto _test_eof1000;
case 1000:
#line 75698 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2985;
		case 32: goto tr2985;
		case 40: goto tr2986;
		case 41: goto tr2987;
		case 1034: goto tr2988;
		case 1083: goto tr2989;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3014;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3014;
	} else
		goto tr3014;
	goto tr2275;
tr2977:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1001;
st1001:
	if ( ++p == pe )
		goto _test_eof1001;
case 1001:
#line 75746 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2983;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1001;
	goto tr69;
tr2965:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1002;
st1002:
	if ( ++p == pe )
		goto _test_eof1002;
case 1002:
#line 75789 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2970;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1002;
	goto tr69;
tr2953:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1003;
st1003:
	if ( ++p == pe )
		goto _test_eof1003;
case 1003:
#line 75832 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2958;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1003;
	goto tr69;
st1004:
	if ( ++p == pe )
		goto _test_eof1004;
case 1004:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3015;
	goto tr1674;
tr3015:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1005;
tr3019:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1005;
st1005:
	if ( ++p == pe )
		goto _test_eof1005;
case 1005:
#line 75908 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3016;
		case 32: goto tr3016;
		case 40: goto tr3017;
		case 41: goto tr3018;
		case 1034: goto tr3020;
		case 1083: goto tr3021;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3019;
	goto tr1674;
tr3023:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1006;
tr3024:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1006;
tr3026:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1006;
tr3016:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1006;
tr3017:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1006;
tr3018:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1006;
tr3020:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1006;
st1006:
	if ( ++p == pe )
		goto _test_eof1006;
case 1006:
#line 76032 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1006;
		case 32: goto st1006;
		case 40: goto tr3023;
		case 41: goto tr3024;
		case 1034: goto tr3026;
		case 1083: goto st1016;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3025;
	goto tr1674;
tr3025:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1007;
tr3031:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1007;
st1007:
	if ( ++p == pe )
		goto _test_eof1007;
case 1007:
#line 76100 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3028;
		case 32: goto tr3028;
		case 40: goto tr3029;
		case 41: goto tr3030;
		case 1034: goto tr3032;
		case 1083: goto tr3033;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3031;
	goto tr1674;
tr3035:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1008;
tr3036:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1008;
tr3038:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1008;
tr3028:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1008;
tr3029:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1008;
tr3030:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1008;
tr3032:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1008;
st1008:
	if ( ++p == pe )
		goto _test_eof1008;
case 1008:
#line 76224 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1008;
		case 32: goto st1008;
		case 40: goto tr3035;
		case 41: goto tr3036;
		case 1034: goto tr3038;
		case 1083: goto st1015;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3037;
	goto tr1674;
tr3037:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1009;
tr3043:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1009;
st1009:
	if ( ++p == pe )
		goto _test_eof1009;
case 1009:
#line 76292 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3040;
		case 32: goto tr3040;
		case 40: goto tr3041;
		case 41: goto tr3042;
		case 1034: goto tr3044;
		case 1083: goto tr3045;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3043;
	goto tr1674;
tr3047:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1010;
tr3048:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1010;
tr3051:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1010;
tr3040:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1010;
tr3041:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1010;
tr3042:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1010;
tr3044:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1010;
st1010:
	if ( ++p == pe )
		goto _test_eof1010;
case 1010:
#line 76416 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1010;
		case 32: goto st1010;
		case 40: goto tr3047;
		case 41: goto tr3048;
		case 45: goto tr3049;
		case 1034: goto tr3051;
		case 1083: goto st1014;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3050;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3050;
	} else
		goto tr3050;
	goto tr2275;
tr3049:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
	goto st1011;
st1011:
	if ( ++p == pe )
		goto _test_eof1011;
case 1011:
#line 76459 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3053;
		case 59: goto tr3053;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3053;
	} else if ( (*p) >= 9 )
		goto tr3053;
	goto tr2275;
tr3053:
#line 208 "./zscanner/scanner_body.rl"
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			SCANNER_WARNING(ZSCANNER_EITEM_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1137;
st1137:
	if ( ++p == pe )
		goto _test_eof1137;
case 1137:
#line 76491 "zscanner/scanner.c"
	goto st0;
tr3055:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1012;
tr3050:
#line 205 "./zscanner/scanner_body.rl"
	{
		s->item_length_location = rdata_tail++;
	}
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1012;
st1012:
	if ( ++p == pe )
		goto _test_eof1012;
case 1012:
#line 76523 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3054;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3054;
	} else
		goto tr3054;
	goto tr2275;
tr3054:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1013;
st1013:
	if ( ++p == pe )
		goto _test_eof1013;
case 1013:
#line 76544 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3053;
		case 59: goto tr3053;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3053;
		} else if ( (*p) >= 9 )
			goto tr3053;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr3055;
		} else if ( (*p) >= 65 )
			goto tr3055;
	} else
		goto tr3055;
	goto tr2275;
tr3045:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1014;
st1014:
	if ( ++p == pe )
		goto _test_eof1014;
case 1014:
#line 76580 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3051;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1014;
	goto tr69;
tr3033:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1015;
st1015:
	if ( ++p == pe )
		goto _test_eof1015;
case 1015:
#line 76623 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3038;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1015;
	goto tr69;
tr3021:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1016;
st1016:
	if ( ++p == pe )
		goto _test_eof1016;
case 1016:
#line 76666 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3026;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1016;
	goto tr69;
st1017:
	if ( ++p == pe )
		goto _test_eof1017;
case 1017:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3056;
	goto tr1674;
tr3056:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1018;
tr3060:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1018;
st1018:
	if ( ++p == pe )
		goto _test_eof1018;
case 1018:
#line 76742 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3057;
		case 32: goto tr3057;
		case 40: goto tr3058;
		case 41: goto tr3059;
		case 1034: goto tr3061;
		case 1083: goto tr3062;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3060;
	goto tr1674;
tr3064:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1019;
tr3065:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1019;
tr3067:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1019;
tr3057:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1019;
tr3058:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1019;
tr3059:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1019;
tr3061:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1019;
st1019:
	if ( ++p == pe )
		goto _test_eof1019;
case 1019:
#line 76866 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1019;
		case 32: goto st1019;
		case 40: goto tr3064;
		case 41: goto tr3065;
		case 1034: goto tr3067;
		case 1083: goto st1029;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3066;
	goto tr1674;
tr3066:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1020;
tr3072:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1020;
st1020:
	if ( ++p == pe )
		goto _test_eof1020;
case 1020:
#line 76934 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3069;
		case 32: goto tr3069;
		case 40: goto tr3070;
		case 41: goto tr3071;
		case 1034: goto tr3073;
		case 1083: goto tr3074;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3072;
	goto tr1674;
tr3076:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1021;
tr3077:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1021;
tr3079:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1021;
tr3069:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1021;
tr3070:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1021;
tr3071:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1021;
tr3073:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1021;
st1021:
	if ( ++p == pe )
		goto _test_eof1021;
case 1021:
#line 77058 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1021;
		case 32: goto st1021;
		case 40: goto tr3076;
		case 41: goto tr3077;
		case 1034: goto tr3079;
		case 1083: goto st1028;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3078;
	goto tr1674;
tr3078:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1022;
tr3084:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1022;
st1022:
	if ( ++p == pe )
		goto _test_eof1022;
case 1022:
#line 77126 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3081;
		case 32: goto tr3081;
		case 40: goto tr3082;
		case 41: goto tr3083;
		case 1034: goto tr3085;
		case 1083: goto tr3086;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3084;
	goto tr1674;
tr3088:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1023;
tr3089:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1023;
tr3091:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1023;
tr3081:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1023;
tr3082:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1023;
tr3083:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1023;
tr3085:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1023;
st1023:
	if ( ++p == pe )
		goto _test_eof1023;
case 1023:
#line 77250 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1023;
		case 32: goto st1023;
		case 40: goto tr3088;
		case 41: goto tr3089;
		case 1034: goto tr3091;
		case 1083: goto st1027;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3090;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3090;
	} else
		goto tr3090;
	goto tr2275;
tr3090:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1024;
st1024:
	if ( ++p == pe )
		goto _test_eof1024;
case 1024:
#line 77297 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3093;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3093;
	} else
		goto tr3093;
	goto tr2275;
tr3095:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1025;
tr3096:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1025;
tr3097:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1025;
tr3093:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1025;
st1025:
	if ( ++p == pe )
		goto _test_eof1025;
case 1025:
#line 77344 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1025;
		case 32: goto st1025;
		case 40: goto tr3095;
		case 41: goto tr3096;
		case 2058: goto tr3097;
		case 2107: goto st1026;
		case 2314: goto tr3099;
		case 2363: goto tr3099;
		case 2570: goto tr3100;
		case 2619: goto tr3101;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3090;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3090;
	} else
		goto tr3090;
	goto tr2275;
st1026:
	if ( ++p == pe )
		goto _test_eof1026;
case 1026:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3097;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1026;
	goto tr2275;
tr3099:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1138;
st1138:
	if ( ++p == pe )
		goto _test_eof1138;
case 1138:
#line 77426 "zscanner/scanner.c"
	goto st0;
tr3100:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1139;
st1139:
	if ( ++p == pe )
		goto _test_eof1139;
case 1139:
#line 77442 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1792 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
			if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1792 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		if ( 
#line 63 "./zscanner/scanner_body.rl"
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1025;
		case 32: goto st1025;
		case 40: goto tr3095;
		case 41: goto tr3096;
		case 2058: goto tr3097;
		case 2107: goto st1026;
		case 2314: goto tr3099;
		case 2363: goto tr3099;
		case 2570: goto tr3100;
		case 2619: goto tr3101;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3090;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3090;
	} else
		goto tr3090;
	goto tr2275;
tr3101:
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1140;
st1140:
	if ( ++p == pe )
		goto _test_eof1140;
case 1140:
#line 77494 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3097;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1026;
	goto tr2275;
tr3086:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1027;
st1027:
	if ( ++p == pe )
		goto _test_eof1027;
case 1027:
#line 77537 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3091;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1027;
	goto tr69;
tr3074:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1028;
st1028:
	if ( ++p == pe )
		goto _test_eof1028;
case 1028:
#line 77580 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3079;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1028;
	goto tr69;
tr3062:
#line 332 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER8_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1029;
st1029:
	if ( ++p == pe )
		goto _test_eof1029;
case 1029:
#line 77623 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3067;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1029;
	goto tr69;
st1030:
	if ( ++p == pe )
		goto _test_eof1030;
case 1030:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3102;
	goto tr1674;
tr3102:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1031;
tr3106:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1031;
st1031:
	if ( ++p == pe )
		goto _test_eof1031;
case 1031:
#line 77699 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3103;
		case 32: goto tr3103;
		case 40: goto tr3104;
		case 41: goto tr3105;
		case 1034: goto tr3107;
		case 1083: goto tr3108;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3106;
	goto tr1674;
tr3110:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1032;
tr3111:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1032;
tr3113:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1032;
tr3103:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1032;
tr3104:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1032;
tr3105:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1032;
tr3107:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1032;
st1032:
	if ( ++p == pe )
		goto _test_eof1032;
case 1032:
#line 77823 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1032;
		case 32: goto st1032;
		case 40: goto tr3110;
		case 41: goto tr3111;
		case 46: goto tr3112;
		case 1034: goto tr3113;
		case 1083: goto st1034;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3112;
	goto tr1651;
tr3112:
#line 763 "./zscanner/scanner_body.rl"
	{
		s->buffer_length = 0;
	}
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1033;
tr3116:
#line 766 "./zscanner/scanner_body.rl"
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1033;
st1033:
	if ( ++p == pe )
		goto _test_eof1033;
case 1033:
#line 77882 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3115;
		case 46: goto tr3116;
		case 59: goto tr3115;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr3115;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3116;
	} else
		goto tr3115;
	goto tr1651;
tr3115:
#line 780 "./zscanner/scanner_body.rl"
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			SCANNER_WARNING(ZSCANNER_EBAD_IPV4);
			p--; {goto st247;}
		}
	}
#line 788 "./zscanner/scanner_body.rl"
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1141;
st1141:
	if ( ++p == pe )
		goto _test_eof1141;
case 1141:
#line 77921 "zscanner/scanner.c"
	goto st0;
tr3108:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1034;
st1034:
	if ( ++p == pe )
		goto _test_eof1034;
case 1034:
#line 77939 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3113;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1034;
	goto tr69;
st1035:
	if ( ++p == pe )
		goto _test_eof1035;
case 1035:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3117;
	goto tr1674;
tr3117:
#line 282 "./zscanner/scanner_body.rl"
	{
		s->number64 = 0;
	}
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1036;
tr3121:
#line 265 "./zscanner/scanner_body.rl"
	{
		// Overflow check: 10*(s->number64) + fc - ASCII_0 <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + ASCII_0)
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER64_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1036;
st1036:
	if ( ++p == pe )
		goto _test_eof1036;
case 1036:
#line 78015 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3118;
		case 32: goto tr3118;
		case 40: goto tr3119;
		case 41: goto tr3120;
		case 1034: goto tr3122;
		case 1083: goto tr3123;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3121;
	goto tr1674;
tr3125:
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1037;
tr3126:
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1037;
tr3128:
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1037;
tr3118:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1037;
tr3119:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 29 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == true) {
			SCANNER_ERROR(ZSCANNER_ELEFT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = true;
	}
	goto st1037;
tr3120:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 36 "./zscanner/scanner_body.rl"
	{
		if (s->multiline == false) {
			SCANNER_ERROR(ZSCANNER_ERIGHT_PARENTHESIS);
			p--; {goto st247;}
		}
		s->multiline = false;
	}
	goto st1037;
tr3122:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
#line 25 "./zscanner/scanner_body.rl"
	{
		s->line_counter++;
	}
	goto st1037;
st1037:
	if ( ++p == pe )
		goto _test_eof1037;
case 1037:
#line 78139 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1037;
		case 32: goto st1037;
		case 40: goto tr3125;
		case 41: goto tr3126;
		case 1034: goto tr3128;
		case 1083: goto st1047;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3127;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3127;
	} else
		goto tr3127;
	goto tr2275;
tr3127:
#line 1551 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
	}
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1038;
st1038:
	if ( ++p == pe )
		goto _test_eof1038;
case 1038:
#line 78190 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3130;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3130;
	} else
		goto tr3130;
	goto tr2275;
tr3130:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1039;
st1039:
	if ( ++p == pe )
		goto _test_eof1039;
case 1039:
#line 78211 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3131;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3131;
	} else
		goto tr3131;
	goto tr2275;
tr3131:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1040;
st1040:
	if ( ++p == pe )
		goto _test_eof1040;
case 1040:
#line 78236 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3132;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3132;
	} else
		goto tr3132;
	goto tr2275;
tr3132:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1041;
st1041:
	if ( ++p == pe )
		goto _test_eof1041;
case 1041:
#line 78257 "zscanner/scanner.c"
	if ( (*p) == 58 )
		goto tr3134;
	goto tr3133;
tr3134:
#line 1554 "./zscanner/scanner_body.rl"
	{
		s->item_length++;
	}
	goto st1042;
st1042:
	if ( ++p == pe )
		goto _test_eof1042;
case 1042:
#line 78271 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3135;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3135;
	} else
		goto tr3135;
	goto tr2275;
tr3135:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1043;
st1043:
	if ( ++p == pe )
		goto _test_eof1043;
case 1043:
#line 78296 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3136;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3136;
	} else
		goto tr3136;
	goto tr2275;
tr3136:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1044;
st1044:
	if ( ++p == pe )
		goto _test_eof1044;
case 1044:
#line 78317 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3137;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3137;
	} else
		goto tr3137;
	goto tr2275;
tr3137:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1045;
st1045:
	if ( ++p == pe )
		goto _test_eof1045;
case 1045:
#line 78342 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3138;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3138;
	} else
		goto tr3138;
	goto tr2275;
tr3138:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1046;
st1046:
	if ( ++p == pe )
		goto _test_eof1046;
case 1046:
#line 78363 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3140;
		case 58: goto tr3134;
		case 59: goto tr3140;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3140;
	} else if ( (*p) >= 9 )
		goto tr3140;
	goto tr3139;
tr3140:
#line 1554 "./zscanner/scanner_body.rl"
	{
		s->item_length++;
	}
#line 1557 "./zscanner/scanner_body.rl"
	{
		if (s->item_length != 4) {
			SCANNER_WARNING(ZSCANNER_EBAD_L64_LENGTH);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1142;
st1142:
	if ( ++p == pe )
		goto _test_eof1142;
case 1142:
#line 78396 "zscanner/scanner.c"
	goto st0;
tr3123:
#line 341 "./zscanner/scanner_body.rl"
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			SCANNER_WARNING(ZSCANNER_ENUMBER16_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1047;
st1047:
	if ( ++p == pe )
		goto _test_eof1047;
case 1047:
#line 78414 "zscanner/scanner.c"
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(768 + ((*p) - 0));
			if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(768 + ((*p) - 0));
		if ( 
#line 54 "./zscanner/scanner_body.rl"
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3128;
	if ( 1024 <= _widec && _widec <= 1279 )
		goto st1047;
	goto tr69;
st1048:
	if ( ++p == pe )
		goto _test_eof1048;
case 1048:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3141;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3141;
	} else
		goto tr3141;
	goto tr2275;
tr3141:
#line 1518 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
	}
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1049;
st1049:
	if ( ++p == pe )
		goto _test_eof1049;
case 1049:
#line 78473 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3142;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3142;
	} else
		goto tr3142;
	goto tr2275;
tr3142:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1050;
st1050:
	if ( ++p == pe )
		goto _test_eof1050;
case 1050:
#line 78494 "zscanner/scanner.c"
	if ( (*p) == 45 )
		goto tr3144;
	goto tr3143;
tr3144:
#line 1521 "./zscanner/scanner_body.rl"
	{
		s->item_length++;
	}
	goto st1051;
st1051:
	if ( ++p == pe )
		goto _test_eof1051;
case 1051:
#line 78508 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3145;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3145;
	} else
		goto tr3145;
	goto tr2275;
tr3145:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1052;
st1052:
	if ( ++p == pe )
		goto _test_eof1052;
case 1052:
#line 78533 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3146;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3146;
	} else
		goto tr3146;
	goto tr2275;
tr3146:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1053;
st1053:
	if ( ++p == pe )
		goto _test_eof1053;
case 1053:
#line 78554 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3147;
		case 45: goto tr3144;
		case 59: goto tr3147;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3147;
	} else if ( (*p) >= 9 )
		goto tr3147;
	goto tr3143;
tr3147:
#line 1521 "./zscanner/scanner_body.rl"
	{
		s->item_length++;
	}
#line 1524 "./zscanner/scanner_body.rl"
	{
		if (s->item_length != 6) {
			SCANNER_WARNING(ZSCANNER_EBAD_EUI_LENGTH);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1143;
st1143:
	if ( ++p == pe )
		goto _test_eof1143;
case 1143:
#line 78587 "zscanner/scanner.c"
	goto st0;
st1054:
	if ( ++p == pe )
		goto _test_eof1054;
case 1054:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3148;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3148;
	} else
		goto tr3148;
	goto tr2275;
tr3148:
#line 1518 "./zscanner/scanner_body.rl"
	{
		s->item_length = 0;
	}
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1055;
st1055:
	if ( ++p == pe )
		goto _test_eof1055;
case 1055:
#line 78621 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3149;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3149;
	} else
		goto tr3149;
	goto tr2275;
tr3149:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1056;
st1056:
	if ( ++p == pe )
		goto _test_eof1056;
case 1056:
#line 78642 "zscanner/scanner.c"
	if ( (*p) == 45 )
		goto tr3150;
	goto tr3143;
tr3150:
#line 1521 "./zscanner/scanner_body.rl"
	{
		s->item_length++;
	}
	goto st1057;
st1057:
	if ( ++p == pe )
		goto _test_eof1057;
case 1057:
#line 78656 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3151;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3151;
	} else
		goto tr3151;
	goto tr2275;
tr3151:
#line 893 "./zscanner/scanner_body.rl"
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			SCANNER_WARNING(ZSCANNER_ERDATA_OVERFLOW);
			p--; {goto st247;}
		}
	}
	goto st1058;
st1058:
	if ( ++p == pe )
		goto _test_eof1058;
case 1058:
#line 78681 "zscanner/scanner.c"
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3152;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3152;
	} else
		goto tr3152;
	goto tr2275;
tr3152:
#line 901 "./zscanner/scanner_body.rl"
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1059;
st1059:
	if ( ++p == pe )
		goto _test_eof1059;
case 1059:
#line 78702 "zscanner/scanner.c"
	switch( (*p) ) {
		case 32: goto tr3153;
		case 45: goto tr3150;
		case 59: goto tr3153;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3153;
	} else if ( (*p) >= 9 )
		goto tr3153;
	goto tr3143;
tr3153:
#line 1521 "./zscanner/scanner_body.rl"
	{
		s->item_length++;
	}
#line 1530 "./zscanner/scanner_body.rl"
	{
		if (s->item_length != 8) {
			SCANNER_WARNING(ZSCANNER_EBAD_EUI_LENGTH);
			p--; {goto st247;}
		}
	}
#line 20 "./zscanner/scanner_body.rl"
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1144;
st1144:
	if ( ++p == pe )
		goto _test_eof1144;
case 1144:
#line 78735 "zscanner/scanner.c"
	goto st0;
	}
	_test_eof1060: cs = 1060; goto _test_eof; 
	_test_eof1: cs = 1; goto _test_eof; 
	_test_eof2: cs = 2; goto _test_eof; 
	_test_eof3: cs = 3; goto _test_eof; 
	_test_eof4: cs = 4; goto _test_eof; 
	_test_eof5: cs = 5; goto _test_eof; 
	_test_eof6: cs = 6; goto _test_eof; 
	_test_eof7: cs = 7; goto _test_eof; 
	_test_eof8: cs = 8; goto _test_eof; 
	_test_eof9: cs = 9; goto _test_eof; 
	_test_eof10: cs = 10; goto _test_eof; 
	_test_eof11: cs = 11; goto _test_eof; 
	_test_eof12: cs = 12; goto _test_eof; 
	_test_eof13: cs = 13; goto _test_eof; 
	_test_eof1061: cs = 1061; goto _test_eof; 
	_test_eof14: cs = 14; goto _test_eof; 
	_test_eof15: cs = 15; goto _test_eof; 
	_test_eof16: cs = 16; goto _test_eof; 
	_test_eof17: cs = 17; goto _test_eof; 
	_test_eof18: cs = 18; goto _test_eof; 
	_test_eof19: cs = 19; goto _test_eof; 
	_test_eof20: cs = 20; goto _test_eof; 
	_test_eof21: cs = 21; goto _test_eof; 
	_test_eof22: cs = 22; goto _test_eof; 
	_test_eof23: cs = 23; goto _test_eof; 
	_test_eof24: cs = 24; goto _test_eof; 
	_test_eof25: cs = 25; goto _test_eof; 
	_test_eof26: cs = 26; goto _test_eof; 
	_test_eof27: cs = 27; goto _test_eof; 
	_test_eof28: cs = 28; goto _test_eof; 
	_test_eof29: cs = 29; goto _test_eof; 
	_test_eof30: cs = 30; goto _test_eof; 
	_test_eof31: cs = 31; goto _test_eof; 
	_test_eof32: cs = 32; goto _test_eof; 
	_test_eof33: cs = 33; goto _test_eof; 
	_test_eof34: cs = 34; goto _test_eof; 
	_test_eof35: cs = 35; goto _test_eof; 
	_test_eof36: cs = 36; goto _test_eof; 
	_test_eof37: cs = 37; goto _test_eof; 
	_test_eof38: cs = 38; goto _test_eof; 
	_test_eof39: cs = 39; goto _test_eof; 
	_test_eof40: cs = 40; goto _test_eof; 
	_test_eof41: cs = 41; goto _test_eof; 
	_test_eof42: cs = 42; goto _test_eof; 
	_test_eof43: cs = 43; goto _test_eof; 
	_test_eof44: cs = 44; goto _test_eof; 
	_test_eof45: cs = 45; goto _test_eof; 
	_test_eof46: cs = 46; goto _test_eof; 
	_test_eof47: cs = 47; goto _test_eof; 
	_test_eof48: cs = 48; goto _test_eof; 
	_test_eof49: cs = 49; goto _test_eof; 
	_test_eof50: cs = 50; goto _test_eof; 
	_test_eof51: cs = 51; goto _test_eof; 
	_test_eof52: cs = 52; goto _test_eof; 
	_test_eof53: cs = 53; goto _test_eof; 
	_test_eof54: cs = 54; goto _test_eof; 
	_test_eof55: cs = 55; goto _test_eof; 
	_test_eof56: cs = 56; goto _test_eof; 
	_test_eof57: cs = 57; goto _test_eof; 
	_test_eof58: cs = 58; goto _test_eof; 
	_test_eof59: cs = 59; goto _test_eof; 
	_test_eof60: cs = 60; goto _test_eof; 
	_test_eof61: cs = 61; goto _test_eof; 
	_test_eof62: cs = 62; goto _test_eof; 
	_test_eof63: cs = 63; goto _test_eof; 
	_test_eof64: cs = 64; goto _test_eof; 
	_test_eof65: cs = 65; goto _test_eof; 
	_test_eof66: cs = 66; goto _test_eof; 
	_test_eof67: cs = 67; goto _test_eof; 
	_test_eof68: cs = 68; goto _test_eof; 
	_test_eof69: cs = 69; goto _test_eof; 
	_test_eof70: cs = 70; goto _test_eof; 
	_test_eof71: cs = 71; goto _test_eof; 
	_test_eof72: cs = 72; goto _test_eof; 
	_test_eof73: cs = 73; goto _test_eof; 
	_test_eof74: cs = 74; goto _test_eof; 
	_test_eof75: cs = 75; goto _test_eof; 
	_test_eof76: cs = 76; goto _test_eof; 
	_test_eof77: cs = 77; goto _test_eof; 
	_test_eof78: cs = 78; goto _test_eof; 
	_test_eof79: cs = 79; goto _test_eof; 
	_test_eof80: cs = 80; goto _test_eof; 
	_test_eof81: cs = 81; goto _test_eof; 
	_test_eof82: cs = 82; goto _test_eof; 
	_test_eof83: cs = 83; goto _test_eof; 
	_test_eof84: cs = 84; goto _test_eof; 
	_test_eof85: cs = 85; goto _test_eof; 
	_test_eof86: cs = 86; goto _test_eof; 
	_test_eof87: cs = 87; goto _test_eof; 
	_test_eof88: cs = 88; goto _test_eof; 
	_test_eof89: cs = 89; goto _test_eof; 
	_test_eof90: cs = 90; goto _test_eof; 
	_test_eof91: cs = 91; goto _test_eof; 
	_test_eof92: cs = 92; goto _test_eof; 
	_test_eof93: cs = 93; goto _test_eof; 
	_test_eof94: cs = 94; goto _test_eof; 
	_test_eof95: cs = 95; goto _test_eof; 
	_test_eof96: cs = 96; goto _test_eof; 
	_test_eof97: cs = 97; goto _test_eof; 
	_test_eof98: cs = 98; goto _test_eof; 
	_test_eof99: cs = 99; goto _test_eof; 
	_test_eof100: cs = 100; goto _test_eof; 
	_test_eof101: cs = 101; goto _test_eof; 
	_test_eof102: cs = 102; goto _test_eof; 
	_test_eof103: cs = 103; goto _test_eof; 
	_test_eof104: cs = 104; goto _test_eof; 
	_test_eof105: cs = 105; goto _test_eof; 
	_test_eof106: cs = 106; goto _test_eof; 
	_test_eof107: cs = 107; goto _test_eof; 
	_test_eof108: cs = 108; goto _test_eof; 
	_test_eof109: cs = 109; goto _test_eof; 
	_test_eof110: cs = 110; goto _test_eof; 
	_test_eof111: cs = 111; goto _test_eof; 
	_test_eof112: cs = 112; goto _test_eof; 
	_test_eof113: cs = 113; goto _test_eof; 
	_test_eof114: cs = 114; goto _test_eof; 
	_test_eof115: cs = 115; goto _test_eof; 
	_test_eof116: cs = 116; goto _test_eof; 
	_test_eof117: cs = 117; goto _test_eof; 
	_test_eof118: cs = 118; goto _test_eof; 
	_test_eof119: cs = 119; goto _test_eof; 
	_test_eof120: cs = 120; goto _test_eof; 
	_test_eof121: cs = 121; goto _test_eof; 
	_test_eof122: cs = 122; goto _test_eof; 
	_test_eof123: cs = 123; goto _test_eof; 
	_test_eof124: cs = 124; goto _test_eof; 
	_test_eof125: cs = 125; goto _test_eof; 
	_test_eof126: cs = 126; goto _test_eof; 
	_test_eof127: cs = 127; goto _test_eof; 
	_test_eof128: cs = 128; goto _test_eof; 
	_test_eof1062: cs = 1062; goto _test_eof; 
	_test_eof129: cs = 129; goto _test_eof; 
	_test_eof130: cs = 130; goto _test_eof; 
	_test_eof131: cs = 131; goto _test_eof; 
	_test_eof132: cs = 132; goto _test_eof; 
	_test_eof133: cs = 133; goto _test_eof; 
	_test_eof134: cs = 134; goto _test_eof; 
	_test_eof135: cs = 135; goto _test_eof; 
	_test_eof136: cs = 136; goto _test_eof; 
	_test_eof137: cs = 137; goto _test_eof; 
	_test_eof138: cs = 138; goto _test_eof; 
	_test_eof1063: cs = 1063; goto _test_eof; 
	_test_eof139: cs = 139; goto _test_eof; 
	_test_eof140: cs = 140; goto _test_eof; 
	_test_eof141: cs = 141; goto _test_eof; 
	_test_eof142: cs = 142; goto _test_eof; 
	_test_eof143: cs = 143; goto _test_eof; 
	_test_eof144: cs = 144; goto _test_eof; 
	_test_eof145: cs = 145; goto _test_eof; 
	_test_eof146: cs = 146; goto _test_eof; 
	_test_eof1064: cs = 1064; goto _test_eof; 
	_test_eof147: cs = 147; goto _test_eof; 
	_test_eof148: cs = 148; goto _test_eof; 
	_test_eof149: cs = 149; goto _test_eof; 
	_test_eof1065: cs = 1065; goto _test_eof; 
	_test_eof150: cs = 150; goto _test_eof; 
	_test_eof151: cs = 151; goto _test_eof; 
	_test_eof152: cs = 152; goto _test_eof; 
	_test_eof153: cs = 153; goto _test_eof; 
	_test_eof154: cs = 154; goto _test_eof; 
	_test_eof155: cs = 155; goto _test_eof; 
	_test_eof156: cs = 156; goto _test_eof; 
	_test_eof157: cs = 157; goto _test_eof; 
	_test_eof158: cs = 158; goto _test_eof; 
	_test_eof159: cs = 159; goto _test_eof; 
	_test_eof160: cs = 160; goto _test_eof; 
	_test_eof1066: cs = 1066; goto _test_eof; 
	_test_eof161: cs = 161; goto _test_eof; 
	_test_eof162: cs = 162; goto _test_eof; 
	_test_eof163: cs = 163; goto _test_eof; 
	_test_eof164: cs = 164; goto _test_eof; 
	_test_eof1067: cs = 1067; goto _test_eof; 
	_test_eof165: cs = 165; goto _test_eof; 
	_test_eof166: cs = 166; goto _test_eof; 
	_test_eof167: cs = 167; goto _test_eof; 
	_test_eof168: cs = 168; goto _test_eof; 
	_test_eof169: cs = 169; goto _test_eof; 
	_test_eof170: cs = 170; goto _test_eof; 
	_test_eof171: cs = 171; goto _test_eof; 
	_test_eof172: cs = 172; goto _test_eof; 
	_test_eof173: cs = 173; goto _test_eof; 
	_test_eof174: cs = 174; goto _test_eof; 
	_test_eof175: cs = 175; goto _test_eof; 
	_test_eof176: cs = 176; goto _test_eof; 
	_test_eof177: cs = 177; goto _test_eof; 
	_test_eof178: cs = 178; goto _test_eof; 
	_test_eof179: cs = 179; goto _test_eof; 
	_test_eof1068: cs = 1068; goto _test_eof; 
	_test_eof180: cs = 180; goto _test_eof; 
	_test_eof181: cs = 181; goto _test_eof; 
	_test_eof182: cs = 182; goto _test_eof; 
	_test_eof183: cs = 183; goto _test_eof; 
	_test_eof184: cs = 184; goto _test_eof; 
	_test_eof185: cs = 185; goto _test_eof; 
	_test_eof186: cs = 186; goto _test_eof; 
	_test_eof187: cs = 187; goto _test_eof; 
	_test_eof188: cs = 188; goto _test_eof; 
	_test_eof189: cs = 189; goto _test_eof; 
	_test_eof190: cs = 190; goto _test_eof; 
	_test_eof191: cs = 191; goto _test_eof; 
	_test_eof192: cs = 192; goto _test_eof; 
	_test_eof193: cs = 193; goto _test_eof; 
	_test_eof194: cs = 194; goto _test_eof; 
	_test_eof195: cs = 195; goto _test_eof; 
	_test_eof1069: cs = 1069; goto _test_eof; 
	_test_eof196: cs = 196; goto _test_eof; 
	_test_eof197: cs = 197; goto _test_eof; 
	_test_eof198: cs = 198; goto _test_eof; 
	_test_eof199: cs = 199; goto _test_eof; 
	_test_eof200: cs = 200; goto _test_eof; 
	_test_eof201: cs = 201; goto _test_eof; 
	_test_eof202: cs = 202; goto _test_eof; 
	_test_eof203: cs = 203; goto _test_eof; 
	_test_eof204: cs = 204; goto _test_eof; 
	_test_eof205: cs = 205; goto _test_eof; 
	_test_eof206: cs = 206; goto _test_eof; 
	_test_eof207: cs = 207; goto _test_eof; 
	_test_eof208: cs = 208; goto _test_eof; 
	_test_eof209: cs = 209; goto _test_eof; 
	_test_eof210: cs = 210; goto _test_eof; 
	_test_eof211: cs = 211; goto _test_eof; 
	_test_eof212: cs = 212; goto _test_eof; 
	_test_eof213: cs = 213; goto _test_eof; 
	_test_eof214: cs = 214; goto _test_eof; 
	_test_eof215: cs = 215; goto _test_eof; 
	_test_eof216: cs = 216; goto _test_eof; 
	_test_eof217: cs = 217; goto _test_eof; 
	_test_eof218: cs = 218; goto _test_eof; 
	_test_eof219: cs = 219; goto _test_eof; 
	_test_eof220: cs = 220; goto _test_eof; 
	_test_eof221: cs = 221; goto _test_eof; 
	_test_eof222: cs = 222; goto _test_eof; 
	_test_eof223: cs = 223; goto _test_eof; 
	_test_eof224: cs = 224; goto _test_eof; 
	_test_eof225: cs = 225; goto _test_eof; 
	_test_eof226: cs = 226; goto _test_eof; 
	_test_eof227: cs = 227; goto _test_eof; 
	_test_eof228: cs = 228; goto _test_eof; 
	_test_eof229: cs = 229; goto _test_eof; 
	_test_eof230: cs = 230; goto _test_eof; 
	_test_eof231: cs = 231; goto _test_eof; 
	_test_eof232: cs = 232; goto _test_eof; 
	_test_eof233: cs = 233; goto _test_eof; 
	_test_eof234: cs = 234; goto _test_eof; 
	_test_eof235: cs = 235; goto _test_eof; 
	_test_eof236: cs = 236; goto _test_eof; 
	_test_eof237: cs = 237; goto _test_eof; 
	_test_eof238: cs = 238; goto _test_eof; 
	_test_eof239: cs = 239; goto _test_eof; 
	_test_eof240: cs = 240; goto _test_eof; 
	_test_eof241: cs = 241; goto _test_eof; 
	_test_eof242: cs = 242; goto _test_eof; 
	_test_eof243: cs = 243; goto _test_eof; 
	_test_eof244: cs = 244; goto _test_eof; 
	_test_eof245: cs = 245; goto _test_eof; 
	_test_eof246: cs = 246; goto _test_eof; 
	_test_eof247: cs = 247; goto _test_eof; 
	_test_eof248: cs = 248; goto _test_eof; 
	_test_eof1070: cs = 1070; goto _test_eof; 
	_test_eof249: cs = 249; goto _test_eof; 
	_test_eof250: cs = 250; goto _test_eof; 
	_test_eof1071: cs = 1071; goto _test_eof; 
	_test_eof251: cs = 251; goto _test_eof; 
	_test_eof252: cs = 252; goto _test_eof; 
	_test_eof253: cs = 253; goto _test_eof; 
	_test_eof254: cs = 254; goto _test_eof; 
	_test_eof255: cs = 255; goto _test_eof; 
	_test_eof256: cs = 256; goto _test_eof; 
	_test_eof257: cs = 257; goto _test_eof; 
	_test_eof258: cs = 258; goto _test_eof; 
	_test_eof259: cs = 259; goto _test_eof; 
	_test_eof260: cs = 260; goto _test_eof; 
	_test_eof1072: cs = 1072; goto _test_eof; 
	_test_eof1073: cs = 1073; goto _test_eof; 
	_test_eof261: cs = 261; goto _test_eof; 
	_test_eof262: cs = 262; goto _test_eof; 
	_test_eof263: cs = 263; goto _test_eof; 
	_test_eof264: cs = 264; goto _test_eof; 
	_test_eof265: cs = 265; goto _test_eof; 
	_test_eof266: cs = 266; goto _test_eof; 
	_test_eof267: cs = 267; goto _test_eof; 
	_test_eof268: cs = 268; goto _test_eof; 
	_test_eof269: cs = 269; goto _test_eof; 
	_test_eof270: cs = 270; goto _test_eof; 
	_test_eof271: cs = 271; goto _test_eof; 
	_test_eof272: cs = 272; goto _test_eof; 
	_test_eof273: cs = 273; goto _test_eof; 
	_test_eof274: cs = 274; goto _test_eof; 
	_test_eof1074: cs = 1074; goto _test_eof; 
	_test_eof275: cs = 275; goto _test_eof; 
	_test_eof276: cs = 276; goto _test_eof; 
	_test_eof277: cs = 277; goto _test_eof; 
	_test_eof278: cs = 278; goto _test_eof; 
	_test_eof279: cs = 279; goto _test_eof; 
	_test_eof280: cs = 280; goto _test_eof; 
	_test_eof281: cs = 281; goto _test_eof; 
	_test_eof282: cs = 282; goto _test_eof; 
	_test_eof283: cs = 283; goto _test_eof; 
	_test_eof284: cs = 284; goto _test_eof; 
	_test_eof1075: cs = 1075; goto _test_eof; 
	_test_eof285: cs = 285; goto _test_eof; 
	_test_eof286: cs = 286; goto _test_eof; 
	_test_eof287: cs = 287; goto _test_eof; 
	_test_eof288: cs = 288; goto _test_eof; 
	_test_eof289: cs = 289; goto _test_eof; 
	_test_eof290: cs = 290; goto _test_eof; 
	_test_eof291: cs = 291; goto _test_eof; 
	_test_eof292: cs = 292; goto _test_eof; 
	_test_eof293: cs = 293; goto _test_eof; 
	_test_eof294: cs = 294; goto _test_eof; 
	_test_eof295: cs = 295; goto _test_eof; 
	_test_eof296: cs = 296; goto _test_eof; 
	_test_eof297: cs = 297; goto _test_eof; 
	_test_eof298: cs = 298; goto _test_eof; 
	_test_eof299: cs = 299; goto _test_eof; 
	_test_eof1076: cs = 1076; goto _test_eof; 
	_test_eof300: cs = 300; goto _test_eof; 
	_test_eof301: cs = 301; goto _test_eof; 
	_test_eof302: cs = 302; goto _test_eof; 
	_test_eof303: cs = 303; goto _test_eof; 
	_test_eof304: cs = 304; goto _test_eof; 
	_test_eof305: cs = 305; goto _test_eof; 
	_test_eof306: cs = 306; goto _test_eof; 
	_test_eof307: cs = 307; goto _test_eof; 
	_test_eof1077: cs = 1077; goto _test_eof; 
	_test_eof308: cs = 308; goto _test_eof; 
	_test_eof309: cs = 309; goto _test_eof; 
	_test_eof310: cs = 310; goto _test_eof; 
	_test_eof311: cs = 311; goto _test_eof; 
	_test_eof312: cs = 312; goto _test_eof; 
	_test_eof313: cs = 313; goto _test_eof; 
	_test_eof314: cs = 314; goto _test_eof; 
	_test_eof315: cs = 315; goto _test_eof; 
	_test_eof316: cs = 316; goto _test_eof; 
	_test_eof1078: cs = 1078; goto _test_eof; 
	_test_eof1079: cs = 1079; goto _test_eof; 
	_test_eof1080: cs = 1080; goto _test_eof; 
	_test_eof317: cs = 317; goto _test_eof; 
	_test_eof318: cs = 318; goto _test_eof; 
	_test_eof319: cs = 319; goto _test_eof; 
	_test_eof320: cs = 320; goto _test_eof; 
	_test_eof321: cs = 321; goto _test_eof; 
	_test_eof322: cs = 322; goto _test_eof; 
	_test_eof323: cs = 323; goto _test_eof; 
	_test_eof324: cs = 324; goto _test_eof; 
	_test_eof1081: cs = 1081; goto _test_eof; 
	_test_eof1082: cs = 1082; goto _test_eof; 
	_test_eof325: cs = 325; goto _test_eof; 
	_test_eof326: cs = 326; goto _test_eof; 
	_test_eof327: cs = 327; goto _test_eof; 
	_test_eof328: cs = 328; goto _test_eof; 
	_test_eof1083: cs = 1083; goto _test_eof; 
	_test_eof329: cs = 329; goto _test_eof; 
	_test_eof330: cs = 330; goto _test_eof; 
	_test_eof331: cs = 331; goto _test_eof; 
	_test_eof332: cs = 332; goto _test_eof; 
	_test_eof333: cs = 333; goto _test_eof; 
	_test_eof334: cs = 334; goto _test_eof; 
	_test_eof335: cs = 335; goto _test_eof; 
	_test_eof336: cs = 336; goto _test_eof; 
	_test_eof337: cs = 337; goto _test_eof; 
	_test_eof338: cs = 338; goto _test_eof; 
	_test_eof339: cs = 339; goto _test_eof; 
	_test_eof340: cs = 340; goto _test_eof; 
	_test_eof341: cs = 341; goto _test_eof; 
	_test_eof342: cs = 342; goto _test_eof; 
	_test_eof343: cs = 343; goto _test_eof; 
	_test_eof344: cs = 344; goto _test_eof; 
	_test_eof345: cs = 345; goto _test_eof; 
	_test_eof346: cs = 346; goto _test_eof; 
	_test_eof347: cs = 347; goto _test_eof; 
	_test_eof348: cs = 348; goto _test_eof; 
	_test_eof349: cs = 349; goto _test_eof; 
	_test_eof350: cs = 350; goto _test_eof; 
	_test_eof351: cs = 351; goto _test_eof; 
	_test_eof352: cs = 352; goto _test_eof; 
	_test_eof353: cs = 353; goto _test_eof; 
	_test_eof354: cs = 354; goto _test_eof; 
	_test_eof355: cs = 355; goto _test_eof; 
	_test_eof356: cs = 356; goto _test_eof; 
	_test_eof357: cs = 357; goto _test_eof; 
	_test_eof358: cs = 358; goto _test_eof; 
	_test_eof359: cs = 359; goto _test_eof; 
	_test_eof360: cs = 360; goto _test_eof; 
	_test_eof361: cs = 361; goto _test_eof; 
	_test_eof362: cs = 362; goto _test_eof; 
	_test_eof363: cs = 363; goto _test_eof; 
	_test_eof364: cs = 364; goto _test_eof; 
	_test_eof365: cs = 365; goto _test_eof; 
	_test_eof366: cs = 366; goto _test_eof; 
	_test_eof367: cs = 367; goto _test_eof; 
	_test_eof368: cs = 368; goto _test_eof; 
	_test_eof369: cs = 369; goto _test_eof; 
	_test_eof370: cs = 370; goto _test_eof; 
	_test_eof371: cs = 371; goto _test_eof; 
	_test_eof372: cs = 372; goto _test_eof; 
	_test_eof373: cs = 373; goto _test_eof; 
	_test_eof374: cs = 374; goto _test_eof; 
	_test_eof375: cs = 375; goto _test_eof; 
	_test_eof376: cs = 376; goto _test_eof; 
	_test_eof377: cs = 377; goto _test_eof; 
	_test_eof378: cs = 378; goto _test_eof; 
	_test_eof379: cs = 379; goto _test_eof; 
	_test_eof380: cs = 380; goto _test_eof; 
	_test_eof381: cs = 381; goto _test_eof; 
	_test_eof382: cs = 382; goto _test_eof; 
	_test_eof383: cs = 383; goto _test_eof; 
	_test_eof384: cs = 384; goto _test_eof; 
	_test_eof385: cs = 385; goto _test_eof; 
	_test_eof386: cs = 386; goto _test_eof; 
	_test_eof387: cs = 387; goto _test_eof; 
	_test_eof388: cs = 388; goto _test_eof; 
	_test_eof389: cs = 389; goto _test_eof; 
	_test_eof390: cs = 390; goto _test_eof; 
	_test_eof391: cs = 391; goto _test_eof; 
	_test_eof392: cs = 392; goto _test_eof; 
	_test_eof393: cs = 393; goto _test_eof; 
	_test_eof394: cs = 394; goto _test_eof; 
	_test_eof395: cs = 395; goto _test_eof; 
	_test_eof396: cs = 396; goto _test_eof; 
	_test_eof397: cs = 397; goto _test_eof; 
	_test_eof398: cs = 398; goto _test_eof; 
	_test_eof399: cs = 399; goto _test_eof; 
	_test_eof400: cs = 400; goto _test_eof; 
	_test_eof401: cs = 401; goto _test_eof; 
	_test_eof402: cs = 402; goto _test_eof; 
	_test_eof403: cs = 403; goto _test_eof; 
	_test_eof404: cs = 404; goto _test_eof; 
	_test_eof405: cs = 405; goto _test_eof; 
	_test_eof406: cs = 406; goto _test_eof; 
	_test_eof407: cs = 407; goto _test_eof; 
	_test_eof408: cs = 408; goto _test_eof; 
	_test_eof409: cs = 409; goto _test_eof; 
	_test_eof410: cs = 410; goto _test_eof; 
	_test_eof411: cs = 411; goto _test_eof; 
	_test_eof412: cs = 412; goto _test_eof; 
	_test_eof413: cs = 413; goto _test_eof; 
	_test_eof414: cs = 414; goto _test_eof; 
	_test_eof415: cs = 415; goto _test_eof; 
	_test_eof416: cs = 416; goto _test_eof; 
	_test_eof417: cs = 417; goto _test_eof; 
	_test_eof418: cs = 418; goto _test_eof; 
	_test_eof419: cs = 419; goto _test_eof; 
	_test_eof420: cs = 420; goto _test_eof; 
	_test_eof421: cs = 421; goto _test_eof; 
	_test_eof422: cs = 422; goto _test_eof; 
	_test_eof423: cs = 423; goto _test_eof; 
	_test_eof424: cs = 424; goto _test_eof; 
	_test_eof425: cs = 425; goto _test_eof; 
	_test_eof426: cs = 426; goto _test_eof; 
	_test_eof427: cs = 427; goto _test_eof; 
	_test_eof428: cs = 428; goto _test_eof; 
	_test_eof429: cs = 429; goto _test_eof; 
	_test_eof430: cs = 430; goto _test_eof; 
	_test_eof431: cs = 431; goto _test_eof; 
	_test_eof432: cs = 432; goto _test_eof; 
	_test_eof433: cs = 433; goto _test_eof; 
	_test_eof434: cs = 434; goto _test_eof; 
	_test_eof435: cs = 435; goto _test_eof; 
	_test_eof436: cs = 436; goto _test_eof; 
	_test_eof437: cs = 437; goto _test_eof; 
	_test_eof438: cs = 438; goto _test_eof; 
	_test_eof439: cs = 439; goto _test_eof; 
	_test_eof440: cs = 440; goto _test_eof; 
	_test_eof441: cs = 441; goto _test_eof; 
	_test_eof442: cs = 442; goto _test_eof; 
	_test_eof443: cs = 443; goto _test_eof; 
	_test_eof444: cs = 444; goto _test_eof; 
	_test_eof1084: cs = 1084; goto _test_eof; 
	_test_eof1085: cs = 1085; goto _test_eof; 
	_test_eof1086: cs = 1086; goto _test_eof; 
	_test_eof445: cs = 445; goto _test_eof; 
	_test_eof446: cs = 446; goto _test_eof; 
	_test_eof447: cs = 447; goto _test_eof; 
	_test_eof448: cs = 448; goto _test_eof; 
	_test_eof449: cs = 449; goto _test_eof; 
	_test_eof1087: cs = 1087; goto _test_eof; 
	_test_eof450: cs = 450; goto _test_eof; 
	_test_eof451: cs = 451; goto _test_eof; 
	_test_eof452: cs = 452; goto _test_eof; 
	_test_eof453: cs = 453; goto _test_eof; 
	_test_eof1088: cs = 1088; goto _test_eof; 
	_test_eof1089: cs = 1089; goto _test_eof; 
	_test_eof1090: cs = 1090; goto _test_eof; 
	_test_eof454: cs = 454; goto _test_eof; 
	_test_eof455: cs = 455; goto _test_eof; 
	_test_eof1091: cs = 1091; goto _test_eof; 
	_test_eof456: cs = 456; goto _test_eof; 
	_test_eof457: cs = 457; goto _test_eof; 
	_test_eof458: cs = 458; goto _test_eof; 
	_test_eof1092: cs = 1092; goto _test_eof; 
	_test_eof459: cs = 459; goto _test_eof; 
	_test_eof460: cs = 460; goto _test_eof; 
	_test_eof461: cs = 461; goto _test_eof; 
	_test_eof462: cs = 462; goto _test_eof; 
	_test_eof463: cs = 463; goto _test_eof; 
	_test_eof464: cs = 464; goto _test_eof; 
	_test_eof465: cs = 465; goto _test_eof; 
	_test_eof466: cs = 466; goto _test_eof; 
	_test_eof467: cs = 467; goto _test_eof; 
	_test_eof468: cs = 468; goto _test_eof; 
	_test_eof469: cs = 469; goto _test_eof; 
	_test_eof470: cs = 470; goto _test_eof; 
	_test_eof471: cs = 471; goto _test_eof; 
	_test_eof472: cs = 472; goto _test_eof; 
	_test_eof473: cs = 473; goto _test_eof; 
	_test_eof474: cs = 474; goto _test_eof; 
	_test_eof475: cs = 475; goto _test_eof; 
	_test_eof476: cs = 476; goto _test_eof; 
	_test_eof477: cs = 477; goto _test_eof; 
	_test_eof478: cs = 478; goto _test_eof; 
	_test_eof479: cs = 479; goto _test_eof; 
	_test_eof480: cs = 480; goto _test_eof; 
	_test_eof481: cs = 481; goto _test_eof; 
	_test_eof482: cs = 482; goto _test_eof; 
	_test_eof483: cs = 483; goto _test_eof; 
	_test_eof484: cs = 484; goto _test_eof; 
	_test_eof485: cs = 485; goto _test_eof; 
	_test_eof486: cs = 486; goto _test_eof; 
	_test_eof487: cs = 487; goto _test_eof; 
	_test_eof488: cs = 488; goto _test_eof; 
	_test_eof489: cs = 489; goto _test_eof; 
	_test_eof490: cs = 490; goto _test_eof; 
	_test_eof491: cs = 491; goto _test_eof; 
	_test_eof492: cs = 492; goto _test_eof; 
	_test_eof493: cs = 493; goto _test_eof; 
	_test_eof494: cs = 494; goto _test_eof; 
	_test_eof495: cs = 495; goto _test_eof; 
	_test_eof496: cs = 496; goto _test_eof; 
	_test_eof497: cs = 497; goto _test_eof; 
	_test_eof498: cs = 498; goto _test_eof; 
	_test_eof499: cs = 499; goto _test_eof; 
	_test_eof500: cs = 500; goto _test_eof; 
	_test_eof501: cs = 501; goto _test_eof; 
	_test_eof502: cs = 502; goto _test_eof; 
	_test_eof503: cs = 503; goto _test_eof; 
	_test_eof504: cs = 504; goto _test_eof; 
	_test_eof505: cs = 505; goto _test_eof; 
	_test_eof506: cs = 506; goto _test_eof; 
	_test_eof507: cs = 507; goto _test_eof; 
	_test_eof508: cs = 508; goto _test_eof; 
	_test_eof509: cs = 509; goto _test_eof; 
	_test_eof510: cs = 510; goto _test_eof; 
	_test_eof511: cs = 511; goto _test_eof; 
	_test_eof512: cs = 512; goto _test_eof; 
	_test_eof513: cs = 513; goto _test_eof; 
	_test_eof514: cs = 514; goto _test_eof; 
	_test_eof515: cs = 515; goto _test_eof; 
	_test_eof516: cs = 516; goto _test_eof; 
	_test_eof517: cs = 517; goto _test_eof; 
	_test_eof518: cs = 518; goto _test_eof; 
	_test_eof519: cs = 519; goto _test_eof; 
	_test_eof520: cs = 520; goto _test_eof; 
	_test_eof521: cs = 521; goto _test_eof; 
	_test_eof522: cs = 522; goto _test_eof; 
	_test_eof523: cs = 523; goto _test_eof; 
	_test_eof524: cs = 524; goto _test_eof; 
	_test_eof525: cs = 525; goto _test_eof; 
	_test_eof526: cs = 526; goto _test_eof; 
	_test_eof527: cs = 527; goto _test_eof; 
	_test_eof528: cs = 528; goto _test_eof; 
	_test_eof529: cs = 529; goto _test_eof; 
	_test_eof530: cs = 530; goto _test_eof; 
	_test_eof531: cs = 531; goto _test_eof; 
	_test_eof532: cs = 532; goto _test_eof; 
	_test_eof533: cs = 533; goto _test_eof; 
	_test_eof534: cs = 534; goto _test_eof; 
	_test_eof535: cs = 535; goto _test_eof; 
	_test_eof536: cs = 536; goto _test_eof; 
	_test_eof537: cs = 537; goto _test_eof; 
	_test_eof538: cs = 538; goto _test_eof; 
	_test_eof539: cs = 539; goto _test_eof; 
	_test_eof540: cs = 540; goto _test_eof; 
	_test_eof541: cs = 541; goto _test_eof; 
	_test_eof542: cs = 542; goto _test_eof; 
	_test_eof543: cs = 543; goto _test_eof; 
	_test_eof544: cs = 544; goto _test_eof; 
	_test_eof545: cs = 545; goto _test_eof; 
	_test_eof546: cs = 546; goto _test_eof; 
	_test_eof547: cs = 547; goto _test_eof; 
	_test_eof548: cs = 548; goto _test_eof; 
	_test_eof549: cs = 549; goto _test_eof; 
	_test_eof550: cs = 550; goto _test_eof; 
	_test_eof551: cs = 551; goto _test_eof; 
	_test_eof552: cs = 552; goto _test_eof; 
	_test_eof553: cs = 553; goto _test_eof; 
	_test_eof1093: cs = 1093; goto _test_eof; 
	_test_eof554: cs = 554; goto _test_eof; 
	_test_eof555: cs = 555; goto _test_eof; 
	_test_eof556: cs = 556; goto _test_eof; 
	_test_eof557: cs = 557; goto _test_eof; 
	_test_eof558: cs = 558; goto _test_eof; 
	_test_eof559: cs = 559; goto _test_eof; 
	_test_eof560: cs = 560; goto _test_eof; 
	_test_eof561: cs = 561; goto _test_eof; 
	_test_eof562: cs = 562; goto _test_eof; 
	_test_eof563: cs = 563; goto _test_eof; 
	_test_eof564: cs = 564; goto _test_eof; 
	_test_eof565: cs = 565; goto _test_eof; 
	_test_eof566: cs = 566; goto _test_eof; 
	_test_eof567: cs = 567; goto _test_eof; 
	_test_eof568: cs = 568; goto _test_eof; 
	_test_eof569: cs = 569; goto _test_eof; 
	_test_eof570: cs = 570; goto _test_eof; 
	_test_eof571: cs = 571; goto _test_eof; 
	_test_eof572: cs = 572; goto _test_eof; 
	_test_eof573: cs = 573; goto _test_eof; 
	_test_eof574: cs = 574; goto _test_eof; 
	_test_eof575: cs = 575; goto _test_eof; 
	_test_eof576: cs = 576; goto _test_eof; 
	_test_eof577: cs = 577; goto _test_eof; 
	_test_eof578: cs = 578; goto _test_eof; 
	_test_eof579: cs = 579; goto _test_eof; 
	_test_eof580: cs = 580; goto _test_eof; 
	_test_eof581: cs = 581; goto _test_eof; 
	_test_eof582: cs = 582; goto _test_eof; 
	_test_eof583: cs = 583; goto _test_eof; 
	_test_eof584: cs = 584; goto _test_eof; 
	_test_eof585: cs = 585; goto _test_eof; 
	_test_eof586: cs = 586; goto _test_eof; 
	_test_eof587: cs = 587; goto _test_eof; 
	_test_eof588: cs = 588; goto _test_eof; 
	_test_eof589: cs = 589; goto _test_eof; 
	_test_eof590: cs = 590; goto _test_eof; 
	_test_eof591: cs = 591; goto _test_eof; 
	_test_eof592: cs = 592; goto _test_eof; 
	_test_eof593: cs = 593; goto _test_eof; 
	_test_eof594: cs = 594; goto _test_eof; 
	_test_eof1094: cs = 1094; goto _test_eof; 
	_test_eof595: cs = 595; goto _test_eof; 
	_test_eof596: cs = 596; goto _test_eof; 
	_test_eof1095: cs = 1095; goto _test_eof; 
	_test_eof597: cs = 597; goto _test_eof; 
	_test_eof598: cs = 598; goto _test_eof; 
	_test_eof599: cs = 599; goto _test_eof; 
	_test_eof600: cs = 600; goto _test_eof; 
	_test_eof601: cs = 601; goto _test_eof; 
	_test_eof602: cs = 602; goto _test_eof; 
	_test_eof603: cs = 603; goto _test_eof; 
	_test_eof604: cs = 604; goto _test_eof; 
	_test_eof605: cs = 605; goto _test_eof; 
	_test_eof606: cs = 606; goto _test_eof; 
	_test_eof607: cs = 607; goto _test_eof; 
	_test_eof608: cs = 608; goto _test_eof; 
	_test_eof609: cs = 609; goto _test_eof; 
	_test_eof610: cs = 610; goto _test_eof; 
	_test_eof1096: cs = 1096; goto _test_eof; 
	_test_eof611: cs = 611; goto _test_eof; 
	_test_eof612: cs = 612; goto _test_eof; 
	_test_eof613: cs = 613; goto _test_eof; 
	_test_eof614: cs = 614; goto _test_eof; 
	_test_eof615: cs = 615; goto _test_eof; 
	_test_eof616: cs = 616; goto _test_eof; 
	_test_eof617: cs = 617; goto _test_eof; 
	_test_eof618: cs = 618; goto _test_eof; 
	_test_eof619: cs = 619; goto _test_eof; 
	_test_eof620: cs = 620; goto _test_eof; 
	_test_eof621: cs = 621; goto _test_eof; 
	_test_eof622: cs = 622; goto _test_eof; 
	_test_eof623: cs = 623; goto _test_eof; 
	_test_eof624: cs = 624; goto _test_eof; 
	_test_eof625: cs = 625; goto _test_eof; 
	_test_eof626: cs = 626; goto _test_eof; 
	_test_eof627: cs = 627; goto _test_eof; 
	_test_eof628: cs = 628; goto _test_eof; 
	_test_eof629: cs = 629; goto _test_eof; 
	_test_eof630: cs = 630; goto _test_eof; 
	_test_eof631: cs = 631; goto _test_eof; 
	_test_eof632: cs = 632; goto _test_eof; 
	_test_eof1097: cs = 1097; goto _test_eof; 
	_test_eof633: cs = 633; goto _test_eof; 
	_test_eof634: cs = 634; goto _test_eof; 
	_test_eof635: cs = 635; goto _test_eof; 
	_test_eof636: cs = 636; goto _test_eof; 
	_test_eof637: cs = 637; goto _test_eof; 
	_test_eof1098: cs = 1098; goto _test_eof; 
	_test_eof638: cs = 638; goto _test_eof; 
	_test_eof639: cs = 639; goto _test_eof; 
	_test_eof640: cs = 640; goto _test_eof; 
	_test_eof641: cs = 641; goto _test_eof; 
	_test_eof642: cs = 642; goto _test_eof; 
	_test_eof1099: cs = 1099; goto _test_eof; 
	_test_eof643: cs = 643; goto _test_eof; 
	_test_eof644: cs = 644; goto _test_eof; 
	_test_eof645: cs = 645; goto _test_eof; 
	_test_eof646: cs = 646; goto _test_eof; 
	_test_eof647: cs = 647; goto _test_eof; 
	_test_eof1100: cs = 1100; goto _test_eof; 
	_test_eof1101: cs = 1101; goto _test_eof; 
	_test_eof1102: cs = 1102; goto _test_eof; 
	_test_eof648: cs = 648; goto _test_eof; 
	_test_eof649: cs = 649; goto _test_eof; 
	_test_eof1103: cs = 1103; goto _test_eof; 
	_test_eof650: cs = 650; goto _test_eof; 
	_test_eof651: cs = 651; goto _test_eof; 
	_test_eof652: cs = 652; goto _test_eof; 
	_test_eof653: cs = 653; goto _test_eof; 
	_test_eof654: cs = 654; goto _test_eof; 
	_test_eof655: cs = 655; goto _test_eof; 
	_test_eof656: cs = 656; goto _test_eof; 
	_test_eof657: cs = 657; goto _test_eof; 
	_test_eof658: cs = 658; goto _test_eof; 
	_test_eof659: cs = 659; goto _test_eof; 
	_test_eof660: cs = 660; goto _test_eof; 
	_test_eof661: cs = 661; goto _test_eof; 
	_test_eof662: cs = 662; goto _test_eof; 
	_test_eof663: cs = 663; goto _test_eof; 
	_test_eof664: cs = 664; goto _test_eof; 
	_test_eof665: cs = 665; goto _test_eof; 
	_test_eof666: cs = 666; goto _test_eof; 
	_test_eof667: cs = 667; goto _test_eof; 
	_test_eof668: cs = 668; goto _test_eof; 
	_test_eof669: cs = 669; goto _test_eof; 
	_test_eof670: cs = 670; goto _test_eof; 
	_test_eof671: cs = 671; goto _test_eof; 
	_test_eof672: cs = 672; goto _test_eof; 
	_test_eof673: cs = 673; goto _test_eof; 
	_test_eof674: cs = 674; goto _test_eof; 
	_test_eof675: cs = 675; goto _test_eof; 
	_test_eof676: cs = 676; goto _test_eof; 
	_test_eof1104: cs = 1104; goto _test_eof; 
	_test_eof1105: cs = 1105; goto _test_eof; 
	_test_eof1106: cs = 1106; goto _test_eof; 
	_test_eof677: cs = 677; goto _test_eof; 
	_test_eof678: cs = 678; goto _test_eof; 
	_test_eof679: cs = 679; goto _test_eof; 
	_test_eof1107: cs = 1107; goto _test_eof; 
	_test_eof1108: cs = 1108; goto _test_eof; 
	_test_eof680: cs = 680; goto _test_eof; 
	_test_eof681: cs = 681; goto _test_eof; 
	_test_eof682: cs = 682; goto _test_eof; 
	_test_eof683: cs = 683; goto _test_eof; 
	_test_eof1109: cs = 1109; goto _test_eof; 
	_test_eof1110: cs = 1110; goto _test_eof; 
	_test_eof684: cs = 684; goto _test_eof; 
	_test_eof685: cs = 685; goto _test_eof; 
	_test_eof686: cs = 686; goto _test_eof; 
	_test_eof687: cs = 687; goto _test_eof; 
	_test_eof1111: cs = 1111; goto _test_eof; 
	_test_eof1112: cs = 1112; goto _test_eof; 
	_test_eof688: cs = 688; goto _test_eof; 
	_test_eof689: cs = 689; goto _test_eof; 
	_test_eof690: cs = 690; goto _test_eof; 
	_test_eof691: cs = 691; goto _test_eof; 
	_test_eof692: cs = 692; goto _test_eof; 
	_test_eof693: cs = 693; goto _test_eof; 
	_test_eof694: cs = 694; goto _test_eof; 
	_test_eof695: cs = 695; goto _test_eof; 
	_test_eof696: cs = 696; goto _test_eof; 
	_test_eof697: cs = 697; goto _test_eof; 
	_test_eof698: cs = 698; goto _test_eof; 
	_test_eof699: cs = 699; goto _test_eof; 
	_test_eof700: cs = 700; goto _test_eof; 
	_test_eof701: cs = 701; goto _test_eof; 
	_test_eof702: cs = 702; goto _test_eof; 
	_test_eof703: cs = 703; goto _test_eof; 
	_test_eof704: cs = 704; goto _test_eof; 
	_test_eof705: cs = 705; goto _test_eof; 
	_test_eof706: cs = 706; goto _test_eof; 
	_test_eof707: cs = 707; goto _test_eof; 
	_test_eof708: cs = 708; goto _test_eof; 
	_test_eof709: cs = 709; goto _test_eof; 
	_test_eof710: cs = 710; goto _test_eof; 
	_test_eof711: cs = 711; goto _test_eof; 
	_test_eof712: cs = 712; goto _test_eof; 
	_test_eof1113: cs = 1113; goto _test_eof; 
	_test_eof713: cs = 713; goto _test_eof; 
	_test_eof714: cs = 714; goto _test_eof; 
	_test_eof715: cs = 715; goto _test_eof; 
	_test_eof716: cs = 716; goto _test_eof; 
	_test_eof717: cs = 717; goto _test_eof; 
	_test_eof718: cs = 718; goto _test_eof; 
	_test_eof719: cs = 719; goto _test_eof; 
	_test_eof720: cs = 720; goto _test_eof; 
	_test_eof721: cs = 721; goto _test_eof; 
	_test_eof722: cs = 722; goto _test_eof; 
	_test_eof723: cs = 723; goto _test_eof; 
	_test_eof724: cs = 724; goto _test_eof; 
	_test_eof725: cs = 725; goto _test_eof; 
	_test_eof726: cs = 726; goto _test_eof; 
	_test_eof727: cs = 727; goto _test_eof; 
	_test_eof1114: cs = 1114; goto _test_eof; 
	_test_eof728: cs = 728; goto _test_eof; 
	_test_eof729: cs = 729; goto _test_eof; 
	_test_eof730: cs = 730; goto _test_eof; 
	_test_eof731: cs = 731; goto _test_eof; 
	_test_eof732: cs = 732; goto _test_eof; 
	_test_eof733: cs = 733; goto _test_eof; 
	_test_eof734: cs = 734; goto _test_eof; 
	_test_eof735: cs = 735; goto _test_eof; 
	_test_eof736: cs = 736; goto _test_eof; 
	_test_eof737: cs = 737; goto _test_eof; 
	_test_eof738: cs = 738; goto _test_eof; 
	_test_eof739: cs = 739; goto _test_eof; 
	_test_eof740: cs = 740; goto _test_eof; 
	_test_eof1115: cs = 1115; goto _test_eof; 
	_test_eof741: cs = 741; goto _test_eof; 
	_test_eof742: cs = 742; goto _test_eof; 
	_test_eof743: cs = 743; goto _test_eof; 
	_test_eof744: cs = 744; goto _test_eof; 
	_test_eof745: cs = 745; goto _test_eof; 
	_test_eof746: cs = 746; goto _test_eof; 
	_test_eof747: cs = 747; goto _test_eof; 
	_test_eof748: cs = 748; goto _test_eof; 
	_test_eof749: cs = 749; goto _test_eof; 
	_test_eof750: cs = 750; goto _test_eof; 
	_test_eof751: cs = 751; goto _test_eof; 
	_test_eof1116: cs = 1116; goto _test_eof; 
	_test_eof1117: cs = 1117; goto _test_eof; 
	_test_eof752: cs = 752; goto _test_eof; 
	_test_eof753: cs = 753; goto _test_eof; 
	_test_eof754: cs = 754; goto _test_eof; 
	_test_eof1118: cs = 1118; goto _test_eof; 
	_test_eof755: cs = 755; goto _test_eof; 
	_test_eof756: cs = 756; goto _test_eof; 
	_test_eof757: cs = 757; goto _test_eof; 
	_test_eof758: cs = 758; goto _test_eof; 
	_test_eof759: cs = 759; goto _test_eof; 
	_test_eof760: cs = 760; goto _test_eof; 
	_test_eof761: cs = 761; goto _test_eof; 
	_test_eof762: cs = 762; goto _test_eof; 
	_test_eof763: cs = 763; goto _test_eof; 
	_test_eof764: cs = 764; goto _test_eof; 
	_test_eof1119: cs = 1119; goto _test_eof; 
	_test_eof1120: cs = 1120; goto _test_eof; 
	_test_eof1121: cs = 1121; goto _test_eof; 
	_test_eof765: cs = 765; goto _test_eof; 
	_test_eof766: cs = 766; goto _test_eof; 
	_test_eof767: cs = 767; goto _test_eof; 
	_test_eof768: cs = 768; goto _test_eof; 
	_test_eof769: cs = 769; goto _test_eof; 
	_test_eof770: cs = 770; goto _test_eof; 
	_test_eof771: cs = 771; goto _test_eof; 
	_test_eof772: cs = 772; goto _test_eof; 
	_test_eof773: cs = 773; goto _test_eof; 
	_test_eof774: cs = 774; goto _test_eof; 
	_test_eof775: cs = 775; goto _test_eof; 
	_test_eof1122: cs = 1122; goto _test_eof; 
	_test_eof1123: cs = 1123; goto _test_eof; 
	_test_eof1124: cs = 1124; goto _test_eof; 
	_test_eof776: cs = 776; goto _test_eof; 
	_test_eof777: cs = 777; goto _test_eof; 
	_test_eof778: cs = 778; goto _test_eof; 
	_test_eof779: cs = 779; goto _test_eof; 
	_test_eof780: cs = 780; goto _test_eof; 
	_test_eof781: cs = 781; goto _test_eof; 
	_test_eof782: cs = 782; goto _test_eof; 
	_test_eof783: cs = 783; goto _test_eof; 
	_test_eof784: cs = 784; goto _test_eof; 
	_test_eof785: cs = 785; goto _test_eof; 
	_test_eof786: cs = 786; goto _test_eof; 
	_test_eof787: cs = 787; goto _test_eof; 
	_test_eof1125: cs = 1125; goto _test_eof; 
	_test_eof788: cs = 788; goto _test_eof; 
	_test_eof789: cs = 789; goto _test_eof; 
	_test_eof790: cs = 790; goto _test_eof; 
	_test_eof1126: cs = 1126; goto _test_eof; 
	_test_eof1127: cs = 1127; goto _test_eof; 
	_test_eof791: cs = 791; goto _test_eof; 
	_test_eof1128: cs = 1128; goto _test_eof; 
	_test_eof1129: cs = 1129; goto _test_eof; 
	_test_eof792: cs = 792; goto _test_eof; 
	_test_eof1130: cs = 1130; goto _test_eof; 
	_test_eof1131: cs = 1131; goto _test_eof; 
	_test_eof793: cs = 793; goto _test_eof; 
	_test_eof794: cs = 794; goto _test_eof; 
	_test_eof795: cs = 795; goto _test_eof; 
	_test_eof796: cs = 796; goto _test_eof; 
	_test_eof797: cs = 797; goto _test_eof; 
	_test_eof798: cs = 798; goto _test_eof; 
	_test_eof799: cs = 799; goto _test_eof; 
	_test_eof800: cs = 800; goto _test_eof; 
	_test_eof801: cs = 801; goto _test_eof; 
	_test_eof802: cs = 802; goto _test_eof; 
	_test_eof803: cs = 803; goto _test_eof; 
	_test_eof804: cs = 804; goto _test_eof; 
	_test_eof805: cs = 805; goto _test_eof; 
	_test_eof806: cs = 806; goto _test_eof; 
	_test_eof807: cs = 807; goto _test_eof; 
	_test_eof808: cs = 808; goto _test_eof; 
	_test_eof809: cs = 809; goto _test_eof; 
	_test_eof810: cs = 810; goto _test_eof; 
	_test_eof811: cs = 811; goto _test_eof; 
	_test_eof812: cs = 812; goto _test_eof; 
	_test_eof813: cs = 813; goto _test_eof; 
	_test_eof814: cs = 814; goto _test_eof; 
	_test_eof815: cs = 815; goto _test_eof; 
	_test_eof816: cs = 816; goto _test_eof; 
	_test_eof817: cs = 817; goto _test_eof; 
	_test_eof818: cs = 818; goto _test_eof; 
	_test_eof819: cs = 819; goto _test_eof; 
	_test_eof820: cs = 820; goto _test_eof; 
	_test_eof821: cs = 821; goto _test_eof; 
	_test_eof822: cs = 822; goto _test_eof; 
	_test_eof823: cs = 823; goto _test_eof; 
	_test_eof824: cs = 824; goto _test_eof; 
	_test_eof825: cs = 825; goto _test_eof; 
	_test_eof826: cs = 826; goto _test_eof; 
	_test_eof827: cs = 827; goto _test_eof; 
	_test_eof828: cs = 828; goto _test_eof; 
	_test_eof829: cs = 829; goto _test_eof; 
	_test_eof830: cs = 830; goto _test_eof; 
	_test_eof831: cs = 831; goto _test_eof; 
	_test_eof832: cs = 832; goto _test_eof; 
	_test_eof833: cs = 833; goto _test_eof; 
	_test_eof834: cs = 834; goto _test_eof; 
	_test_eof1132: cs = 1132; goto _test_eof; 
	_test_eof835: cs = 835; goto _test_eof; 
	_test_eof836: cs = 836; goto _test_eof; 
	_test_eof837: cs = 837; goto _test_eof; 
	_test_eof838: cs = 838; goto _test_eof; 
	_test_eof839: cs = 839; goto _test_eof; 
	_test_eof840: cs = 840; goto _test_eof; 
	_test_eof841: cs = 841; goto _test_eof; 
	_test_eof842: cs = 842; goto _test_eof; 
	_test_eof843: cs = 843; goto _test_eof; 
	_test_eof844: cs = 844; goto _test_eof; 
	_test_eof845: cs = 845; goto _test_eof; 
	_test_eof846: cs = 846; goto _test_eof; 
	_test_eof847: cs = 847; goto _test_eof; 
	_test_eof848: cs = 848; goto _test_eof; 
	_test_eof849: cs = 849; goto _test_eof; 
	_test_eof850: cs = 850; goto _test_eof; 
	_test_eof851: cs = 851; goto _test_eof; 
	_test_eof852: cs = 852; goto _test_eof; 
	_test_eof853: cs = 853; goto _test_eof; 
	_test_eof854: cs = 854; goto _test_eof; 
	_test_eof855: cs = 855; goto _test_eof; 
	_test_eof856: cs = 856; goto _test_eof; 
	_test_eof857: cs = 857; goto _test_eof; 
	_test_eof858: cs = 858; goto _test_eof; 
	_test_eof859: cs = 859; goto _test_eof; 
	_test_eof860: cs = 860; goto _test_eof; 
	_test_eof861: cs = 861; goto _test_eof; 
	_test_eof862: cs = 862; goto _test_eof; 
	_test_eof863: cs = 863; goto _test_eof; 
	_test_eof864: cs = 864; goto _test_eof; 
	_test_eof865: cs = 865; goto _test_eof; 
	_test_eof866: cs = 866; goto _test_eof; 
	_test_eof867: cs = 867; goto _test_eof; 
	_test_eof868: cs = 868; goto _test_eof; 
	_test_eof869: cs = 869; goto _test_eof; 
	_test_eof870: cs = 870; goto _test_eof; 
	_test_eof871: cs = 871; goto _test_eof; 
	_test_eof872: cs = 872; goto _test_eof; 
	_test_eof873: cs = 873; goto _test_eof; 
	_test_eof874: cs = 874; goto _test_eof; 
	_test_eof875: cs = 875; goto _test_eof; 
	_test_eof876: cs = 876; goto _test_eof; 
	_test_eof877: cs = 877; goto _test_eof; 
	_test_eof878: cs = 878; goto _test_eof; 
	_test_eof879: cs = 879; goto _test_eof; 
	_test_eof880: cs = 880; goto _test_eof; 
	_test_eof881: cs = 881; goto _test_eof; 
	_test_eof882: cs = 882; goto _test_eof; 
	_test_eof883: cs = 883; goto _test_eof; 
	_test_eof884: cs = 884; goto _test_eof; 
	_test_eof885: cs = 885; goto _test_eof; 
	_test_eof886: cs = 886; goto _test_eof; 
	_test_eof887: cs = 887; goto _test_eof; 
	_test_eof888: cs = 888; goto _test_eof; 
	_test_eof889: cs = 889; goto _test_eof; 
	_test_eof890: cs = 890; goto _test_eof; 
	_test_eof891: cs = 891; goto _test_eof; 
	_test_eof892: cs = 892; goto _test_eof; 
	_test_eof893: cs = 893; goto _test_eof; 
	_test_eof894: cs = 894; goto _test_eof; 
	_test_eof895: cs = 895; goto _test_eof; 
	_test_eof896: cs = 896; goto _test_eof; 
	_test_eof897: cs = 897; goto _test_eof; 
	_test_eof898: cs = 898; goto _test_eof; 
	_test_eof899: cs = 899; goto _test_eof; 
	_test_eof900: cs = 900; goto _test_eof; 
	_test_eof901: cs = 901; goto _test_eof; 
	_test_eof902: cs = 902; goto _test_eof; 
	_test_eof903: cs = 903; goto _test_eof; 
	_test_eof904: cs = 904; goto _test_eof; 
	_test_eof905: cs = 905; goto _test_eof; 
	_test_eof906: cs = 906; goto _test_eof; 
	_test_eof907: cs = 907; goto _test_eof; 
	_test_eof908: cs = 908; goto _test_eof; 
	_test_eof909: cs = 909; goto _test_eof; 
	_test_eof910: cs = 910; goto _test_eof; 
	_test_eof911: cs = 911; goto _test_eof; 
	_test_eof912: cs = 912; goto _test_eof; 
	_test_eof913: cs = 913; goto _test_eof; 
	_test_eof914: cs = 914; goto _test_eof; 
	_test_eof915: cs = 915; goto _test_eof; 
	_test_eof916: cs = 916; goto _test_eof; 
	_test_eof917: cs = 917; goto _test_eof; 
	_test_eof918: cs = 918; goto _test_eof; 
	_test_eof919: cs = 919; goto _test_eof; 
	_test_eof920: cs = 920; goto _test_eof; 
	_test_eof921: cs = 921; goto _test_eof; 
	_test_eof922: cs = 922; goto _test_eof; 
	_test_eof923: cs = 923; goto _test_eof; 
	_test_eof924: cs = 924; goto _test_eof; 
	_test_eof925: cs = 925; goto _test_eof; 
	_test_eof926: cs = 926; goto _test_eof; 
	_test_eof927: cs = 927; goto _test_eof; 
	_test_eof928: cs = 928; goto _test_eof; 
	_test_eof929: cs = 929; goto _test_eof; 
	_test_eof930: cs = 930; goto _test_eof; 
	_test_eof931: cs = 931; goto _test_eof; 
	_test_eof932: cs = 932; goto _test_eof; 
	_test_eof933: cs = 933; goto _test_eof; 
	_test_eof934: cs = 934; goto _test_eof; 
	_test_eof935: cs = 935; goto _test_eof; 
	_test_eof936: cs = 936; goto _test_eof; 
	_test_eof937: cs = 937; goto _test_eof; 
	_test_eof938: cs = 938; goto _test_eof; 
	_test_eof939: cs = 939; goto _test_eof; 
	_test_eof940: cs = 940; goto _test_eof; 
	_test_eof941: cs = 941; goto _test_eof; 
	_test_eof942: cs = 942; goto _test_eof; 
	_test_eof943: cs = 943; goto _test_eof; 
	_test_eof944: cs = 944; goto _test_eof; 
	_test_eof945: cs = 945; goto _test_eof; 
	_test_eof946: cs = 946; goto _test_eof; 
	_test_eof947: cs = 947; goto _test_eof; 
	_test_eof948: cs = 948; goto _test_eof; 
	_test_eof949: cs = 949; goto _test_eof; 
	_test_eof950: cs = 950; goto _test_eof; 
	_test_eof951: cs = 951; goto _test_eof; 
	_test_eof952: cs = 952; goto _test_eof; 
	_test_eof953: cs = 953; goto _test_eof; 
	_test_eof954: cs = 954; goto _test_eof; 
	_test_eof955: cs = 955; goto _test_eof; 
	_test_eof956: cs = 956; goto _test_eof; 
	_test_eof957: cs = 957; goto _test_eof; 
	_test_eof958: cs = 958; goto _test_eof; 
	_test_eof959: cs = 959; goto _test_eof; 
	_test_eof960: cs = 960; goto _test_eof; 
	_test_eof961: cs = 961; goto _test_eof; 
	_test_eof1133: cs = 1133; goto _test_eof; 
	_test_eof962: cs = 962; goto _test_eof; 
	_test_eof963: cs = 963; goto _test_eof; 
	_test_eof964: cs = 964; goto _test_eof; 
	_test_eof965: cs = 965; goto _test_eof; 
	_test_eof966: cs = 966; goto _test_eof; 
	_test_eof967: cs = 967; goto _test_eof; 
	_test_eof968: cs = 968; goto _test_eof; 
	_test_eof969: cs = 969; goto _test_eof; 
	_test_eof1134: cs = 1134; goto _test_eof; 
	_test_eof970: cs = 970; goto _test_eof; 
	_test_eof971: cs = 971; goto _test_eof; 
	_test_eof972: cs = 972; goto _test_eof; 
	_test_eof973: cs = 973; goto _test_eof; 
	_test_eof974: cs = 974; goto _test_eof; 
	_test_eof1135: cs = 1135; goto _test_eof; 
	_test_eof975: cs = 975; goto _test_eof; 
	_test_eof976: cs = 976; goto _test_eof; 
	_test_eof977: cs = 977; goto _test_eof; 
	_test_eof978: cs = 978; goto _test_eof; 
	_test_eof979: cs = 979; goto _test_eof; 
	_test_eof980: cs = 980; goto _test_eof; 
	_test_eof981: cs = 981; goto _test_eof; 
	_test_eof982: cs = 982; goto _test_eof; 
	_test_eof983: cs = 983; goto _test_eof; 
	_test_eof984: cs = 984; goto _test_eof; 
	_test_eof985: cs = 985; goto _test_eof; 
	_test_eof986: cs = 986; goto _test_eof; 
	_test_eof987: cs = 987; goto _test_eof; 
	_test_eof988: cs = 988; goto _test_eof; 
	_test_eof989: cs = 989; goto _test_eof; 
	_test_eof990: cs = 990; goto _test_eof; 
	_test_eof991: cs = 991; goto _test_eof; 
	_test_eof992: cs = 992; goto _test_eof; 
	_test_eof1136: cs = 1136; goto _test_eof; 
	_test_eof993: cs = 993; goto _test_eof; 
	_test_eof994: cs = 994; goto _test_eof; 
	_test_eof995: cs = 995; goto _test_eof; 
	_test_eof996: cs = 996; goto _test_eof; 
	_test_eof997: cs = 997; goto _test_eof; 
	_test_eof998: cs = 998; goto _test_eof; 
	_test_eof999: cs = 999; goto _test_eof; 
	_test_eof1000: cs = 1000; goto _test_eof; 
	_test_eof1001: cs = 1001; goto _test_eof; 
	_test_eof1002: cs = 1002; goto _test_eof; 
	_test_eof1003: cs = 1003; goto _test_eof; 
	_test_eof1004: cs = 1004; goto _test_eof; 
	_test_eof1005: cs = 1005; goto _test_eof; 
	_test_eof1006: cs = 1006; goto _test_eof; 
	_test_eof1007: cs = 1007; goto _test_eof; 
	_test_eof1008: cs = 1008; goto _test_eof; 
	_test_eof1009: cs = 1009; goto _test_eof; 
	_test_eof1010: cs = 1010; goto _test_eof; 
	_test_eof1011: cs = 1011; goto _test_eof; 
	_test_eof1137: cs = 1137; goto _test_eof; 
	_test_eof1012: cs = 1012; goto _test_eof; 
	_test_eof1013: cs = 1013; goto _test_eof; 
	_test_eof1014: cs = 1014; goto _test_eof; 
	_test_eof1015: cs = 1015; goto _test_eof; 
	_test_eof1016: cs = 1016; goto _test_eof; 
	_test_eof1017: cs = 1017; goto _test_eof; 
	_test_eof1018: cs = 1018; goto _test_eof; 
	_test_eof1019: cs = 1019; goto _test_eof; 
	_test_eof1020: cs = 1020; goto _test_eof; 
	_test_eof1021: cs = 1021; goto _test_eof; 
	_test_eof1022: cs = 1022; goto _test_eof; 
	_test_eof1023: cs = 1023; goto _test_eof; 
	_test_eof1024: cs = 1024; goto _test_eof; 
	_test_eof1025: cs = 1025; goto _test_eof; 
	_test_eof1026: cs = 1026; goto _test_eof; 
	_test_eof1138: cs = 1138; goto _test_eof; 
	_test_eof1139: cs = 1139; goto _test_eof; 
	_test_eof1140: cs = 1140; goto _test_eof; 
	_test_eof1027: cs = 1027; goto _test_eof; 
	_test_eof1028: cs = 1028; goto _test_eof; 
	_test_eof1029: cs = 1029; goto _test_eof; 
	_test_eof1030: cs = 1030; goto _test_eof; 
	_test_eof1031: cs = 1031; goto _test_eof; 
	_test_eof1032: cs = 1032; goto _test_eof; 
	_test_eof1033: cs = 1033; goto _test_eof; 
	_test_eof1141: cs = 1141; goto _test_eof; 
	_test_eof1034: cs = 1034; goto _test_eof; 
	_test_eof1035: cs = 1035; goto _test_eof; 
	_test_eof1036: cs = 1036; goto _test_eof; 
	_test_eof1037: cs = 1037; goto _test_eof; 
	_test_eof1038: cs = 1038; goto _test_eof; 
	_test_eof1039: cs = 1039; goto _test_eof; 
	_test_eof1040: cs = 1040; goto _test_eof; 
	_test_eof1041: cs = 1041; goto _test_eof; 
	_test_eof1042: cs = 1042; goto _test_eof; 
	_test_eof1043: cs = 1043; goto _test_eof; 
	_test_eof1044: cs = 1044; goto _test_eof; 
	_test_eof1045: cs = 1045; goto _test_eof; 
	_test_eof1046: cs = 1046; goto _test_eof; 
	_test_eof1142: cs = 1142; goto _test_eof; 
	_test_eof1047: cs = 1047; goto _test_eof; 
	_test_eof1048: cs = 1048; goto _test_eof; 
	_test_eof1049: cs = 1049; goto _test_eof; 
	_test_eof1050: cs = 1050; goto _test_eof; 
	_test_eof1051: cs = 1051; goto _test_eof; 
	_test_eof1052: cs = 1052; goto _test_eof; 
	_test_eof1053: cs = 1053; goto _test_eof; 
	_test_eof1143: cs = 1143; goto _test_eof; 
	_test_eof1054: cs = 1054; goto _test_eof; 
	_test_eof1055: cs = 1055; goto _test_eof; 
	_test_eof1056: cs = 1056; goto _test_eof; 
	_test_eof1057: cs = 1057; goto _test_eof; 
	_test_eof1058: cs = 1058; goto _test_eof; 
	_test_eof1059: cs = 1059; goto _test_eof; 
	_test_eof1144: cs = 1144; goto _test_eof; 

	_test_eof: {}
	if ( p == eof )
	{
	switch ( cs ) {
	case 7: 
	case 8: 
	case 138: 
	case 162: 
	case 164: 
	case 212: 
	case 225: 
	case 243: 
	case 299: 
	case 300: 
	case 307: 
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 249: 
	case 250: 
	case 251: 
	case 252: 
	case 257: 
	case 258: 
#line 189 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_DNAME_CHAR);
		p--; {goto st247;}
	}
	break;
	case 147: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
	break;
	case 266: 
#line 513 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT);
		p--; {goto st247;}
	}
	break;
	case 271: 
	case 279: 
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	break;
	case 280: 
	case 281: 
	case 282: 
	case 286: 
	case 292: 
#line 600 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_ORIGIN);
		p--; {goto st247;}
	}
	break;
	case 294: 
#line 629 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
		p--; {goto st247;}
	}
	break;
	case 297: 
	case 301: 
	case 308: 
#line 640 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_ORIGIN);
		p--; {goto st247;}
	}
	break;
	case 1064: 
#line 721 "./zscanner/scanner_body.rl"
	{
		s->stop = false;
	}
	break;
	case 139: 
	case 140: 
	case 141: 
	case 142: 
	case 143: 
	case 144: 
	case 145: 
	case 146: 
	case 181: 
	case 182: 
	case 183: 
	case 184: 
	case 185: 
	case 186: 
	case 187: 
	case 188: 
	case 189: 
#line 724 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_DIRECTIVE);
		p--; {goto st247;}
	}
	break;
	case 311: 
	case 312: 
	case 313: 
	case 314: 
	case 315: 
	case 316: 
	case 317: 
#line 974 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_BASE64_CHAR);
		p--; {goto st247;}
	}
	break;
	case 318: 
	case 319: 
	case 320: 
	case 321: 
	case 322: 
	case 323: 
	case 324: 
	case 325: 
	case 326: 
	case 327: 
	case 328: 
	case 329: 
	case 330: 
	case 331: 
	case 332: 
	case 333: 
	case 334: 
	case 335: 
	case 336: 
	case 337: 
	case 338: 
	case 339: 
	case 340: 
	case 341: 
	case 342: 
	case 343: 
	case 344: 
	case 345: 
	case 346: 
	case 347: 
	case 348: 
	case 349: 
	case 350: 
	case 351: 
	case 352: 
	case 353: 
	case 354: 
	case 355: 
	case 356: 
	case 357: 
	case 358: 
	case 359: 
	case 360: 
	case 361: 
	case 362: 
	case 363: 
	case 364: 
	case 365: 
	case 366: 
	case 367: 
	case 368: 
	case 369: 
	case 370: 
	case 371: 
	case 372: 
	case 373: 
	case 374: 
	case 375: 
	case 376: 
	case 377: 
	case 378: 
	case 379: 
	case 380: 
	case 381: 
	case 382: 
	case 383: 
	case 384: 
	case 385: 
	case 386: 
	case 387: 
	case 388: 
	case 389: 
	case 390: 
	case 391: 
	case 392: 
	case 393: 
	case 394: 
	case 395: 
	case 396: 
	case 397: 
	case 398: 
	case 399: 
	case 400: 
	case 401: 
	case 402: 
	case 403: 
	case 404: 
	case 405: 
	case 406: 
	case 407: 
	case 408: 
	case 409: 
	case 410: 
	case 411: 
	case 412: 
	case 413: 
	case 414: 
	case 415: 
	case 416: 
	case 417: 
	case 418: 
	case 419: 
	case 420: 
	case 421: 
	case 422: 
	case 423: 
	case 424: 
	case 425: 
	case 426: 
	case 427: 
	case 428: 
	case 429: 
	case 430: 
	case 431: 
	case 432: 
	case 433: 
	case 434: 
	case 435: 
	case 436: 
	case 437: 
#line 1334 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_BITMAP);
		p--; {goto st247;}
	}
	break;
	case 438: 
	case 445: 
	case 446: 
	case 447: 
	case 454: 
	case 456: 
#line 1500 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
	break;
	case 457: 
	case 458: 
	case 459: 
	case 460: 
	case 461: 
	case 462: 
	case 463: 
	case 464: 
	case 465: 
	case 466: 
	case 467: 
	case 468: 
	case 469: 
	case 470: 
	case 471: 
	case 472: 
	case 473: 
	case 474: 
	case 475: 
	case 476: 
	case 477: 
	case 478: 
	case 479: 
	case 480: 
	case 481: 
	case 482: 
	case 483: 
	case 484: 
	case 485: 
	case 486: 
	case 487: 
	case 488: 
	case 489: 
	case 490: 
	case 491: 
	case 492: 
	case 493: 
	case 494: 
	case 495: 
	case 496: 
	case 497: 
	case 498: 
	case 499: 
	case 500: 
	case 501: 
	case 502: 
	case 503: 
	case 504: 
	case 505: 
	case 506: 
	case 507: 
	case 508: 
	case 509: 
	case 510: 
	case 511: 
	case 512: 
	case 513: 
	case 514: 
	case 515: 
	case 516: 
	case 517: 
	case 518: 
	case 519: 
	case 520: 
	case 521: 
	case 522: 
	case 523: 
	case 524: 
	case 525: 
	case 526: 
	case 527: 
	case 528: 
	case 529: 
	case 530: 
	case 531: 
	case 532: 
	case 533: 
	case 534: 
	case 535: 
	case 536: 
	case 537: 
	case 538: 
	case 539: 
	case 540: 
	case 541: 
	case 542: 
	case 543: 
	case 544: 
	case 545: 
	case 546: 
	case 547: 
	case 548: 
	case 549: 
	case 550: 
	case 551: 
#line 1576 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ALGORITHM);
		p--; {goto st247;}
	}
	break;
	case 552: 
	case 553: 
	case 554: 
	case 555: 
	case 556: 
	case 557: 
	case 558: 
	case 559: 
	case 560: 
	case 561: 
	case 562: 
	case 563: 
	case 564: 
	case 565: 
	case 566: 
	case 567: 
	case 568: 
	case 569: 
	case 570: 
	case 571: 
	case 572: 
	case 573: 
	case 574: 
	case 575: 
	case 576: 
	case 577: 
	case 578: 
	case 579: 
	case 580: 
	case 581: 
	case 582: 
	case 583: 
	case 584: 
	case 585: 
	case 586: 
	case 587: 
	case 588: 
	case 589: 
	case 590: 
	case 591: 
	case 592: 
#line 1580 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CERT_TYPE);
		p--; {goto st247;}
	}
	break;
	case 5: 
	case 9: 
	case 11: 
	case 228: 
	case 595: 
	case 596: 
	case 597: 
	case 598: 
	case 599: 
	case 600: 
	case 614: 
	case 618: 
	case 622: 
	case 626: 
	case 627: 
	case 628: 
	case 629: 
	case 630: 
	case 631: 
	case 632: 
	case 633: 
	case 634: 
	case 635: 
	case 636: 
	case 637: 
	case 638: 
	case 641: 
	case 642: 
	case 643: 
	case 644: 
	case 645: 
	case 646: 
	case 647: 
	case 711: 
	case 712: 
	case 713: 
	case 714: 
	case 715: 
	case 720: 
	case 721: 
	case 722: 
	case 723: 
	case 724: 
	case 725: 
	case 726: 
	case 727: 
	case 728: 
	case 729: 
	case 730: 
	case 731: 
	case 732: 
	case 733: 
	case 734: 
	case 737: 
	case 738: 
	case 739: 
	case 740: 
	case 741: 
	case 742: 
	case 743: 
	case 751: 
	case 757: 
	case 758: 
	case 765: 
	case 766: 
	case 767: 
	case 776: 
	case 777: 
	case 816: 
	case 819: 
	case 820: 
	case 831: 
	case 832: 
	case 833: 
	case 834: 
	case 835: 
	case 836: 
	case 837: 
	case 838: 
	case 839: 
	case 840: 
	case 841: 
	case 842: 
	case 959: 
	case 960: 
	case 961: 
	case 966: 
	case 967: 
	case 968: 
	case 969: 
	case 970: 
	case 971: 
	case 972: 
	case 973: 
	case 974: 
	case 992: 
	case 998: 
	case 1001: 
	case 1002: 
	case 1003: 
	case 1014: 
	case 1015: 
	case 1016: 
	case 1027: 
	case 1028: 
	case 1029: 
	case 1034: 
	case 1047: 
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 3: 
	case 17: 
	case 18: 
	case 19: 
	case 22: 
	case 23: 
	case 24: 
	case 26: 
	case 27: 
	case 28: 
	case 29: 
	case 31: 
	case 32: 
	case 33: 
	case 35: 
	case 36: 
	case 37: 
	case 40: 
	case 41: 
	case 42: 
	case 43: 
	case 45: 
	case 47: 
	case 48: 
	case 49: 
	case 50: 
	case 52: 
	case 53: 
	case 54: 
	case 55: 
	case 56: 
	case 57: 
	case 58: 
	case 59: 
	case 60: 
	case 61: 
	case 63: 
	case 64: 
	case 67: 
	case 68: 
	case 70: 
	case 72: 
	case 75: 
	case 76: 
	case 77: 
	case 78: 
	case 81: 
	case 82: 
	case 83: 
	case 84: 
	case 86: 
	case 89: 
	case 92: 
	case 93: 
	case 94: 
	case 95: 
	case 97: 
	case 98: 
	case 100: 
	case 102: 
	case 103: 
	case 104: 
	case 107: 
	case 108: 
	case 110: 
	case 112: 
	case 114: 
	case 115: 
	case 116: 
	case 118: 
	case 119: 
	case 120: 
	case 122: 
	case 124: 
	case 125: 
	case 128: 
	case 130: 
	case 131: 
	case 137: 
	case 152: 
	case 156: 
	case 214: 
	case 215: 
	case 217: 
	case 218: 
	case 219: 
	case 221: 
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	break;
	case 274: 
	case 275: 
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	break;
	case 283: 
	case 284: 
	case 285: 
	case 291: 
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 600 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_ORIGIN);
		p--; {goto st247;}
	}
	break;
	case 253: 
	case 254: 
	case 255: 
	case 256: 
#line 143 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 189 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_DNAME_CHAR);
		p--; {goto st247;}
	}
	break;
	case 287: 
	case 288: 
	case 289: 
	case 290: 
#line 143 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 600 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_ORIGIN);
		p--; {goto st247;}
	}
	break;
	case 302: 
	case 303: 
	case 304: 
	case 305: 
#line 143 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 640 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_ORIGIN);
		p--; {goto st247;}
	}
	break;
	case 209: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 230: 
	case 231: 
	case 232: 
	case 233: 
	case 234: 
	case 235: 
	case 236: 
	case 237: 
	case 238: 
	case 239: 
	case 240: 
	case 241: 
	case 242: 
	case 245: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	break;
	case 272: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	break;
	case 439: 
	case 440: 
	case 448: 
	case 449: 
	case 455: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1500 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
	break;
	case 601: 
	case 602: 
	case 603: 
	case 605: 
	case 607: 
	case 609: 
	case 639: 
	case 640: 
	case 705: 
	case 706: 
	case 707: 
	case 708: 
	case 709: 
	case 710: 
	case 716: 
	case 717: 
	case 718: 
	case 719: 
	case 735: 
	case 736: 
	case 755: 
	case 756: 
	case 759: 
	case 760: 
	case 768: 
	case 769: 
	case 770: 
	case 771: 
	case 778: 
	case 779: 
	case 821: 
	case 822: 
	case 823: 
	case 824: 
	case 829: 
	case 830: 
	case 962: 
	case 963: 
	case 964: 
	case 965: 
	case 975: 
	case 976: 
	case 977: 
	case 978: 
	case 979: 
	case 980: 
	case 1004: 
	case 1005: 
	case 1006: 
	case 1007: 
	case 1008: 
	case 1009: 
	case 1017: 
	case 1018: 
	case 1019: 
	case 1020: 
	case 1021: 
	case 1022: 
	case 1030: 
	case 1031: 
	case 1035: 
	case 1036: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 126: 
	case 132: 
	case 148: 
	case 151: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	break;
	case 825: 
	case 826: 
	case 827: 
	case 828: 
#line 491 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIMESTAMP_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 259: 
	case 260: 
	case 265: 
#line 509 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT_CHAR);
		p--; {goto st247;}
	}
#line 513 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT);
		p--; {goto st247;}
	}
	break;
	case 295: 
#line 629 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_FILENAME);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 296: 
	case 298: 
	case 306: 
#line 640 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_INCLUDE_ORIGIN);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 593: 
	case 594: 
	case 648: 
	case 649: 
	case 1032: 
	case 1033: 
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 744: 
	case 745: 
	case 746: 
	case 749: 
	case 750: 
	case 752: 
#line 875 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_APL);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 761: 
	case 762: 
	case 763: 
	case 764: 
	case 772: 
	case 773: 
	case 774: 
	case 775: 
	case 981: 
	case 982: 
	case 999: 
	case 1000: 
	case 1010: 
	case 1011: 
	case 1012: 
	case 1013: 
	case 1023: 
	case 1024: 
	case 1025: 
	case 1026: 
	case 1037: 
	case 1038: 
	case 1039: 
	case 1040: 
	case 1042: 
	case 1043: 
	case 1044: 
	case 1045: 
	case 1048: 
	case 1049: 
	case 1051: 
	case 1052: 
	case 1054: 
	case 1055: 
	case 1057: 
	case 1058: 
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 983: 
	case 984: 
	case 985: 
	case 986: 
	case 987: 
	case 988: 
	case 989: 
	case 990: 
	case 991: 
	case 993: 
	case 994: 
	case 995: 
	case 996: 
	case 997: 
#line 1058 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_BASE32HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 780: 
	case 781: 
	case 784: 
	case 793: 
	case 794: 
	case 795: 
	case 800: 
	case 801: 
	case 802: 
	case 807: 
	case 808: 
	case 809: 
	case 812: 
	case 814: 
	case 815: 
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 786: 
	case 787: 
	case 788: 
	case 789: 
	case 790: 
	case 791: 
	case 792: 
#line 1186 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY_KEY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 817: 
	case 818: 
	case 843: 
	case 844: 
	case 845: 
	case 846: 
	case 847: 
	case 848: 
	case 849: 
	case 850: 
	case 851: 
	case 852: 
	case 853: 
	case 854: 
	case 855: 
	case 856: 
	case 857: 
	case 858: 
	case 859: 
	case 860: 
	case 861: 
	case 862: 
	case 863: 
	case 864: 
	case 865: 
	case 866: 
	case 867: 
	case 868: 
	case 869: 
	case 870: 
	case 871: 
	case 872: 
	case 873: 
	case 874: 
	case 875: 
	case 876: 
	case 877: 
	case 878: 
	case 879: 
	case 880: 
	case 881: 
	case 882: 
	case 883: 
	case 884: 
	case 885: 
	case 886: 
	case 887: 
	case 888: 
	case 889: 
	case 890: 
	case 891: 
	case 892: 
	case 893: 
	case 894: 
	case 895: 
	case 896: 
	case 897: 
	case 898: 
	case 899: 
	case 900: 
	case 901: 
	case 902: 
	case 903: 
	case 904: 
	case 905: 
	case 906: 
	case 907: 
	case 908: 
	case 909: 
	case 910: 
	case 911: 
	case 912: 
	case 913: 
	case 914: 
	case 915: 
	case 916: 
	case 917: 
	case 918: 
	case 919: 
	case 920: 
	case 921: 
	case 922: 
	case 923: 
	case 924: 
	case 925: 
	case 926: 
	case 927: 
	case 928: 
	case 929: 
	case 930: 
	case 931: 
	case 932: 
	case 933: 
	case 934: 
	case 935: 
	case 936: 
	case 937: 
	case 938: 
	case 939: 
	case 940: 
	case 941: 
	case 942: 
	case 943: 
	case 944: 
	case 945: 
	case 946: 
	case 947: 
	case 948: 
	case 949: 
	case 950: 
	case 951: 
	case 952: 
	case 953: 
	case 954: 
	case 955: 
	case 956: 
#line 1204 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 650: 
	case 651: 
	case 652: 
	case 653: 
	case 654: 
	case 655: 
	case 656: 
	case 657: 
	case 658: 
	case 659: 
	case 660: 
	case 661: 
	case 662: 
	case 663: 
	case 664: 
	case 665: 
	case 666: 
	case 667: 
	case 668: 
	case 669: 
	case 670: 
	case 671: 
	case 672: 
	case 673: 
	case 674: 
	case 675: 
	case 676: 
	case 677: 
	case 678: 
	case 679: 
	case 680: 
	case 681: 
	case 682: 
	case 683: 
	case 684: 
	case 685: 
	case 686: 
	case 687: 
	case 688: 
	case 689: 
	case 690: 
	case 691: 
	case 692: 
	case 693: 
	case 694: 
	case 695: 
	case 696: 
	case 697: 
	case 698: 
	case 699: 
	case 700: 
	case 701: 
	case 702: 
	case 703: 
	case 704: 
#line 1487 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_LOC_DATA);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 6: 
	case 10: 
	case 12: 
	case 13: 
	case 21: 
	case 160: 
	case 163: 
	case 179: 
	case 211: 
	case 213: 
	case 223: 
	case 224: 
	case 226: 
	case 227: 
	case 229: 
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 16: 
	case 193: 
	case 244: 
	case 246: 
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 4: 
	case 20: 
	case 25: 
	case 30: 
	case 34: 
	case 38: 
	case 39: 
	case 44: 
	case 46: 
	case 51: 
	case 62: 
	case 65: 
	case 66: 
	case 69: 
	case 71: 
	case 73: 
	case 74: 
	case 79: 
	case 80: 
	case 85: 
	case 87: 
	case 88: 
	case 90: 
	case 91: 
	case 96: 
	case 99: 
	case 101: 
	case 105: 
	case 106: 
	case 109: 
	case 111: 
	case 113: 
	case 117: 
	case 121: 
	case 123: 
	case 216: 
	case 220: 
	case 222: 
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 190: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 157: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 782: 
	case 783: 
	case 796: 
	case 797: 
	case 803: 
	case 804: 
	case 810: 
	case 811: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 957: 
	case 958: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1204 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 1: 
	case 129: 
	case 161: 
	case 192: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 127: 
	case 158: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 277: 
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	break;
	case 604: 
	case 606: 
	case 608: 
	case 610: 
	case 611: 
	case 612: 
	case 613: 
	case 615: 
	case 616: 
	case 617: 
	case 619: 
	case 620: 
	case 621: 
	case 623: 
	case 624: 
	case 625: 
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 2: 
	case 133: 
	case 134: 
	case 135: 
	case 136: 
	case 153: 
	case 154: 
	case 155: 
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	break;
	case 261: 
	case 262: 
	case 263: 
	case 264: 
	case 267: 
	case 268: 
	case 269: 
	case 270: 
#line 543 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 509 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT_CHAR);
		p--; {goto st247;}
	}
#line 513 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TEXT);
		p--; {goto st247;}
	}
	break;
	case 747: 
	case 748: 
	case 753: 
	case 754: 
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 875 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_APL);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 798: 
	case 805: 
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 441: 
	case 442: 
	case 443: 
	case 444: 
	case 450: 
	case 451: 
	case 452: 
	case 453: 
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 932 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
#line 1500 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_RDATA);
		p--; {goto st247;}
	}
	break;
	case 1041: 
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1563 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CHAR_COLON);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 785: 
	case 813: 
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1186 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY_KEY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 1050: 
	case 1053: 
	case 1056: 
	case 1059: 
#line 1536 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CHAR_DASH);                       
		p--; {goto st247;}
	}
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 1046: 
#line 1563 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_CHAR_COLON);
		p--; {goto st247;}
	}
#line 905 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_HEX_CHAR);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 180: 
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 724 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_DIRECTIVE);
		p--; {goto st247;}
	}
	break;
	case 159: 
	case 166: 
	case 167: 
	case 168: 
	case 169: 
	case 170: 
	case 171: 
	case 172: 
	case 173: 
	case 174: 
	case 175: 
	case 176: 
	case 177: 
	case 178: 
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 150: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
	break;
	case 194: 
	case 196: 
	case 197: 
	case 198: 
	case 199: 
	case 200: 
	case 201: 
	case 202: 
	case 203: 
	case 204: 
	case 205: 
	case 206: 
	case 207: 
	case 208: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 14: 
	case 165: 
	case 195: 
	case 210: 
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 273: 
	case 276: 
	case 278: 
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
#line 583 "./zscanner/scanner_body.rl"
	{
		SCANNER_ERROR(ZSCANNER_EBAD_TTL);
		p--; {goto st247;}
	}
	break;
	case 799: 
	case 806: 
#line 775 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_ADDRESS_CHAR);
		p--; {goto st247;}
	}
#line 1182 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY);
		p--; {goto st247;}
	}
#line 1186 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_GATEWAY_KEY);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
	break;
	case 15: 
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
	case 191: 
#line 242 "./zscanner/scanner_body.rl"
	{
		s->r_owner_length = 0;
		SCANNER_WARNING(ZSCANNER_EBAD_OWNER);
		p--; {goto st247;}
	}
#line 386 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_TIME_UNIT);
		p--; {goto st247;}
	}
#line 285 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_NUMBER);
		p--; {goto st247;}
	}
#line 1880 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EUNSUPPORTED_TYPE);
		p--; {goto st247;}
	}
#line 1626 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_RDATA);
		p--; {goto st247;}
	}
#line 44 "./zscanner/scanner_body.rl"
	{
		SCANNER_WARNING(ZSCANNER_EBAD_REST);
		p--; {goto st247;}
	}
	break;
#line 81802 "zscanner/scanner.c"
	}
	}

	_out: {}
	}

#line 144 "./zscanner/scanner.rl"

	// Check if scanner state machine is in uncovered state.
	if (cs == zone_scanner_error) {
		SCANNER_ERROR(ZSCANNER_UNCOVERED_STATE);
		s->error_counter++;

		// Fill error context data.
		for (s->buffer_length = 0;
		     ((p + s->buffer_length) < pe) &&
		     (s->buffer_length < sizeof(s->buffer) - 1);
		     s->buffer_length++)
		{
			// Only rest of the current line.
			if (*(p + s->buffer_length) == '\n') {
				break;
			}
			s->buffer[s->buffer_length] = *(p + s->buffer_length);
		}

		// Ending string in buffer.
		s->buffer[s->buffer_length++] = 0;

		// Processing error.
		s->process_error(s);

		return -1;
	}

	// Check unclosed multiline record.
	if (is_last_block && s->multiline) {
		SCANNER_ERROR(ZSCANNER_UNCLOSED_MULTILINE);
		s->error_counter++;
		s->process_error(s);
	}

	// Storing scanner states.
	s->cs  = cs;
	s->top = top;
	memcpy(s->stack, stack, sizeof(stack));

	// Storing r_data pointer.
	s->r_data_tail = rdata_tail - s->r_data;

	// Check if any errors has occured.
	if (s->error_counter > 0) {
		return -1;
	}

	return 0;
}
